/*
   dialog.c

   Dialogroutinen fuer User-Ein-/Ausgaben

   Historie:
   13.7.95   Till  erste Version
   5.6.96    Till  toll 3d und sound ...
   6.6.96    Till  scales eingebaut
   10.11.96  Chip  einige korrekturen
   14.11.96  Chip  put_bitmap nur durchsichtig
   17.11.96  Chip  farbvariablen, korrektes Schwarz und Weiss
   26.11.96  Chip  keine multicolor-fonts mehr
   28.11.96  Chip  Typen auf tm_* umgestellt
   01.12.96  Chip  16-Farbmodus korrigiert
   04.12.96  Till  Mausfehler korrigiert
   25.01.97  Chip  texteingabe
   23.08.97  Till  File-Selektor
   06.01.98  Till  neuer File-Selektor
   25.01.98  Till  Warnings unter Irix entfernt
*/

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "machine.h"
#ifdef MACHINE_X11
#define FS_OK
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/param.h>
#endif

#include "dialog.h"
#include "gr_prim.h"
#include "grafik.h"
#include "joystick.h"
#include "misc.h"
#include "text.h"
#include "timer.h"
#include "maze.h"
#include "machine.h"
#include "audio.h"
#include "net.h"
#include "console.h"
#include "drawmaze.h"

#define ABSOLUTE_SENSE 200    /* 50 .. 200 */

#ifdef MACHINE_HAS_FILESYSTEM

#endif

#define SEL_DARK

static tm_uword dialog_bg_col,dialog_header_col;
static tm_uword dialog_entry_on_col,dialog_entry_off_col;
static tm_uword border_light_col,border_dark_col;

tm_uword language;

#define MAX_NAMELEN  (40)
tm_char player_name[MAX_NAMELEN+1]="noname";

static tm_char text_buffer[512];

tm_void
init_dialog(tm_void) {
  /* get players name (from env, passwd or whatever */
  get_player_name(player_name,MAX_NAMELEN);

  if(colors<16) {
    dialog_bg_col         = BW_WHITE;
    dialog_header_col     = BW_BLACK;
    dialog_entry_off_col  = BW_BLACK;
    dialog_entry_on_col   = BW_BLACK;
/*    border_light_col      = BW_WHITE; */
    border_light_col      = BW_MGREY;
    border_dark_col       = BW_BLACK;
  } else {
    dialog_bg_col         = COL_LGREY;
    dialog_header_col     = COL_GREEN;
    dialog_entry_off_col  = COL_DGREEN;
    dialog_entry_on_col   = COL_GREEN;

    border_light_col      = COL_WHITE;
    border_dark_col       = COL_DGREY;
  }
}

#define MENU_BORDER (font_height/3)
#define BD          (MENU_BORDER/3)

static
tm_void draw_menu_text(tm_uword x, tm_uword y, tm_char *string,
                       tm_uword col, tm_bool act)
{
  if(colors<16) {
    draw_text(x  ,y  ,string,col);
    if (act)
      draw_text(x+1,y  ,string,col);
  } else {
    draw_text(x-1,y-1,string,border_light_col);
    draw_text(x+1,y+1,string,border_dark_col);
    draw_text(x  ,y  ,string,col);
  }
}

/* Menu mit vorgegebener Breite und Hoehe malen, mit farbigem Titel */
/* die sich ergebenden Koordinaten werden zurueckgegeben */
static
tm_void draw_dialog_border(tm_uword *x, tm_uword *y, tm_uword w, tm_uword h,
                           tm_char *titel, tm_uword fg_col, tm_uword bg_col)
{
  tm_uword cnt;
  tm_uword topm = font_height + 1 + font_height/7;

  /* oben kommt noch eine Textzeile drueber */
  h = h + topm;

  /* X- und Y-Position der Box */
  *x=(width-w)/2;
  *y=(height-h)/2;

  draw_rect(*x-MENU_BORDER,*y-MENU_BORDER-1,
            w+2*MENU_BORDER+1,h+2*MENU_BORDER+1,bg_col);

  for(cnt=0;cnt<MAX(BD,1);cnt++) {
    /* Box um die Box */
    draw_hline(*x-MENU_BORDER+cnt-1,*y-MENU_BORDER+cnt-1,w+2*MENU_BORDER-2*cnt+3,border_light_col);
    draw_hline(*x-MENU_BORDER+cnt-1,*y+h+MENU_BORDER-cnt-1,w+2*MENU_BORDER-2*cnt+3,border_dark_col);
    draw_vline(*x-MENU_BORDER+cnt-1,*y-MENU_BORDER+cnt-1,h+2*MENU_BORDER-2*cnt,border_light_col);
    draw_vline(*x+w+MENU_BORDER-cnt+1,*y-MENU_BORDER+cnt-1,h+2*MENU_BORDER-2*cnt,border_dark_col);

    /* Box um die Entries */
    draw_vline(*x-cnt-1,*y-cnt+topm,h+2*cnt-topm,border_dark_col);
    draw_vline(*x+w+cnt+1,*y-cnt+topm,h+2*cnt-topm,border_light_col);
    draw_hline(*x-cnt-1,*y-cnt+topm-1,w+2*cnt+3,border_dark_col);
    draw_hline(*x-cnt-1,*y+h+cnt,w+2*cnt+3,border_light_col);
  }

  /* noch zwei schmucke Linien */
  draw_hline(*x,*y-3,w,border_light_col);
  draw_hline(*x,*y-1,w,border_dark_col);
  draw_hline(*x,*y+font_height-3,w,border_light_col);
  draw_hline(*x,*y+font_height-1,w,border_dark_col);

  /* Titel-Text */
  draw_menu_text(*x+(w-get_text_width(titel))/2,*y,titel,fg_col,FALSE);

  *y += topm;
}

#define MENU_IOFF    (font_height/7)
#define CHECKSIZE    (font_height/2)

static
tm_void draw_check(tm_uword x, tm_uword y, tm_bool on, tm_uword col)
{
  tm_uword  i,col1,col2;

  col1=on?border_dark_col:border_light_col;
  col2=on?border_light_col:border_dark_col;

  draw_rect(x+(font_height-CHECKSIZE)/2,y+(font_height-CHECKSIZE)/2,CHECKSIZE,CHECKSIZE,col);

  for(i=0;i<MAX(font_height/10,1);i++) {
    draw_hline(x+(font_height-CHECKSIZE)/2+i,y+(font_height-CHECKSIZE)/2+i,CHECKSIZE+1-2*i, col1);
    draw_vline(x+(font_height-CHECKSIZE)/2+i,y+(font_height-CHECKSIZE)/2+i,CHECKSIZE+1-2*i, col1);
    draw_hline(x+(font_height-CHECKSIZE)/2+i,y+(font_height+CHECKSIZE)/2-i,CHECKSIZE+1-2*i, col2);
    draw_vline(x+(font_height+CHECKSIZE)/2-i,y+(font_height-CHECKSIZE)/2+i,CHECKSIZE+1-2*i, col2);
  }
}

#define SCALE_WIDTH  (font_height*5)
#define MENU_IOFF    (font_height/7)
#define SLIDEW       (font_height/3)
#define SLIDEH       (font_height/2)

static
tm_void draw_scale(tm_uword x, tm_uword y, tm_uword slider, tm_uword col)
{
  draw_rect(x,y,SCALE_WIDTH,font_height,dialog_bg_col);

  draw_hline(x+font_height/2+MENU_IOFF,y+font_height/2-1,SCALE_WIDTH-1-font_height,border_light_col);
  draw_hline(x+font_height/2+MENU_IOFF,y+font_height/2+1,SCALE_WIDTH-1-font_height,border_dark_col);

  /* draw the slider */
  slider = x+MENU_IOFF+font_height/2+slider*(SCALE_WIDTH-SLIDEW-font_height)/100;
  draw_rect( slider,y+font_height/2-SLIDEH/2,SLIDEW,SLIDEH,col);
  draw_hline(slider,y+font_height/2-SLIDEH/2,SLIDEW,border_light_col);
  draw_hline(slider,y+font_height-1-SLIDEH/2,SLIDEW,border_dark_col);
  draw_vline(slider,y+font_height/2-SLIDEH/2,SLIDEH,border_light_col);
  draw_vline(slider+SLIDEW-1,y+font_height/2-SLIDEH/2,SLIDEH,border_dark_col);
}

#define  X_BUTTONS    (11)
#define  Y_BUTTONS    (4)

static const tm_char button[][X_BUTTONS*Y_BUTTONS+1]={
"!@#$%^&*()+"   "QWERTYUIOP~"   "ASDFGHJKL;'"    "ZXCVBNM[]? ",
"1234567890="   "qwertyuiop-"   "asdfghjkl:\""   "zxcvbnm,./ " };

#define  BUTTON_DELETE  (0)
#define  BUTTON_SHIFT   (1)
#define  BUTTON_OK      (2)
#define  S_BUTTONS      (3)

static tm_uword str_x,str_y,str_w,str_wsum,str_hsum;

#define MENU_IOFF    (font_height/7)
#define BUTTON_GAP   (font_height/10)
#define MENU_SAFETY  (10)

#define SUNKEN       FALSE
#define RAISED       TRUE

static
tm_void draw_one_button(tm_uword bx,tm_uword by,
                        tm_bool shift,tm_bool raised,tm_bool act)
{
  tm_char   c[2]=" ",*s;
  tm_uword  x,y,w,xs;

  if (by<Y_BUTTONS) {
    c[0]=button[shift?0:1][by*X_BUTTONS+bx];
    s=c;

    x=str_x+MENU_IOFF + bx*(str_w+3) +
      (bx*(str_wsum-2*MENU_IOFF-X_BUTTONS*(str_w+3)))/(X_BUTTONS-1);
    w=str_w;
    xs=x+2+w/2-get_text_width(s)/2;
  } else {
    switch (bx) {
    case BUTTON_DELETE:
      s=get_string_delete[LANGUAGE];
      x=str_x+MENU_IOFF;
      break;
    case BUTTON_SHIFT:
      s=get_string_shift[LANGUAGE];
      x=str_x+str_wsum/2-(get_text_width(s)+3)/2;
      break;
    default:
      s=get_string_ok[LANGUAGE];
      x=str_x+str_wsum-MENU_IOFF-get_text_width(s)-3;
      break;
    }
    w=get_text_width(s);
    xs=x+2;
  }
  y=str_y+MENU_IOFF+by*(font_height+4+BUTTON_GAP);

  draw_hline(x,y,w+3,raised?border_light_col:border_dark_col);
  draw_vline(x,y,font_height+4,raised?border_light_col:border_dark_col);
  draw_hline(x,y+font_height+3,w+3,raised?border_dark_col:border_light_col);
  draw_vline(x+w+2,y,font_height+4,raised?border_dark_col:border_light_col);
  if (colors<16) {
    draw_text(xs,y+2,s,dialog_entry_on_col);
    if (act)
      draw_text(xs+1,y+2,s,dialog_entry_on_col);
  } else
    draw_text(xs,y+2,s,act?dialog_entry_on_col:dialog_entry_off_col);
}

tm_void get_string(tm_char *title,tm_char *string, tm_uword max)
{
  tm_char   c[2]=" ";
  tm_bool   shift=FALSE;
  tm_uword  joy;
  tm_word   joy_x,joy_y;
  tm_word   pos_x,pos_y,old_x,old_y;
  tm_uword  x,y,w;

  init_joystick_delay();
  save_screen();

  str_w=0;
  for (x=0; x<Y_BUTTONS*X_BUTTONS; x++) {
    c[0]=button[0][x];
    w=get_text_width(c);
    if (str_w<w)  str_w=w;
  }
  for (x=0; x<Y_BUTTONS*X_BUTTONS; x++) {
    c[0]=button[1][x];
    w=get_text_width(c);
    if (str_w<w)  str_w=w;
  }

  str_hsum=2*MENU_IOFF+Y_BUTTONS*BUTTON_GAP+(Y_BUTTONS+1)*(font_height+4);

  if(kind_of_joystick()&ABSOLUTE_DEVICE) {
    get_joystick_analog(&joy_x,&joy_y);
    pos_y=((Y_BUTTONS+1) * (joy_y+ABSOLUTE_SENSE/2)) / ABSOLUTE_SENSE;
    if (pos_y<0)               pos_y=0;
    else if (pos_y>Y_BUTTONS)  pos_y=Y_BUTTONS;
    if (pos_y!=Y_BUTTONS) {
      pos_x=(X_BUTTONS * (joy_x+ABSOLUTE_SENSE/2)) / ABSOLUTE_SENSE;
      if (pos_x<0)                pos_x=0;
      else if (pos_x>=X_BUTTONS)  pos_x=X_BUTTONS-1;
    } else {
      pos_x=(S_BUTTONS * (joy_x+ABSOLUTE_SENSE/2)) / ABSOLUTE_SENSE;
      if (pos_x<0)                pos_x=0;
      else if (pos_x>=S_BUTTONS)  pos_x=S_BUTTONS-1;
    }
  } else {
    pos_x=0;
    pos_y=0;
  }

  for (;;) {
    strcpy(text_buffer,title);
    strcat(text_buffer,string);

    str_wsum=get_text_width(text_buffer);
    w=2*MENU_IOFF+(X_BUTTONS-1)*BUTTON_GAP+X_BUTTONS*(str_w+3);
    if (str_wsum<w)  str_wsum=w;
    w=2*MENU_IOFF+get_text_width(get_string_shift[LANGUAGE])+3
      + 2 * (BUTTON_GAP+get_text_width(get_string_delete[LANGUAGE])+3);
    if (str_wsum<w)  str_wsum=w;
    w=2*MENU_IOFF+get_text_width(get_string_shift[LANGUAGE])+3
      + 2 * (BUTTON_GAP+get_text_width(get_string_ok[LANGUAGE])+3);
    if (str_wsum<w)  str_wsum=w;

    restore_screen();
    draw_dialog_border(&str_x,&str_y,str_wsum,str_hsum,
                       text_buffer,dialog_header_col,dialog_bg_col);
    for (y=0; y<Y_BUTTONS; y++)
      for (x=0; x<X_BUTTONS; x++)
        draw_one_button(x,y,shift,RAISED,x==pos_x && y==pos_y);
    draw_one_button(BUTTON_DELETE,Y_BUTTONS,shift,RAISED,
                    pos_x==BUTTON_DELETE && pos_y==Y_BUTTONS);
    draw_one_button(BUTTON_SHIFT,Y_BUTTONS,shift,shift?SUNKEN:RAISED,
                    pos_x==BUTTON_SHIFT && pos_y==Y_BUTTONS);
    draw_one_button(BUTTON_OK,Y_BUTTONS,shift,RAISED,
                    pos_x==BUTTON_OK && pos_y==Y_BUTTONS);
    flip_screen();

    do {
      net_handle_messages();
      joy=get_joystick_delay();
      if (joy&JOY_FIRE0) {
        restore_screen();
        return;
      }

      if(kind_of_joystick()&ABSOLUTE_DEVICE) {
        old_x=pos_x;
        old_y=pos_y;
        get_joystick_analog(&joy_x,&joy_y);
        pos_y=((Y_BUTTONS+1) * (joy_y+ABSOLUTE_SENSE/2)) / ABSOLUTE_SENSE;
        if (pos_y<0)               pos_y=0;
        else if (pos_y>Y_BUTTONS)  pos_y=Y_BUTTONS;
        if (pos_y!=Y_BUTTONS) {
          pos_x=(X_BUTTONS * (joy_x+ABSOLUTE_SENSE/2)) / ABSOLUTE_SENSE;
          if (pos_x<0)                pos_x=0;
          else if (pos_x>=X_BUTTONS)  pos_x=X_BUTTONS-1;
        } else {
          pos_x=(S_BUTTONS * (joy_x+ABSOLUTE_SENSE/2)) / ABSOLUTE_SENSE;
          if (pos_x<0)                pos_x=0;
          else if (pos_x>=S_BUTTONS)  pos_x=S_BUTTONS-1;
        }
        if (pos_x!=old_x || pos_y!=old_y) {
          do_audio(SOUND_CLICK);
          break;
        }
      }

      if((joy&JOY_ALL)!=0) {

        /* Joystick, Tastatur etc. */
        if( ((joy&JOY_UP)||(joy&JOY_DOWN)||(joy&JOY_LEFT)||(joy&JOY_RIGHT)) &&
            (kind_of_joystick()&RELATIVE_DEVICE)) {
          if (joy&JOY_UP) {
            if (pos_y==Y_BUTTONS)    pos_x=(X_BUTTONS*pos_x)/S_BUTTONS;
            else if (pos_y==0)       pos_x=(S_BUTTONS*pos_x)/X_BUTTONS;
            if (--pos_y<0)           pos_y=Y_BUTTONS;
          } else if (joy&JOY_DOWN) {
            if (++pos_y>Y_BUTTONS)   pos_y=0;
            if (pos_y==Y_BUTTONS)    pos_x=(S_BUTTONS*pos_x)/X_BUTTONS;
            else if (pos_y==0)       pos_x=(X_BUTTONS*pos_x)/S_BUTTONS;
          }

          if (joy&JOY_LEFT)         pos_x--;
          else if (joy&JOY_RIGHT)   pos_x++;
          if (pos_y!=Y_BUTTONS) {
            if (pos_x<0)               pos_x=X_BUTTONS-1;
            else if (pos_x>=X_BUTTONS) pos_x=0;
          } else {
            if (pos_x<0)               pos_x=S_BUTTONS-1;
            else if (pos_x>=S_BUTTONS) pos_x=0;
          }
          do_audio(SOUND_CLICK);
          break;
        }

        if(joy&JOY_FIRE1) {
          do_audio(SOUND_CLICK);
          if (pos_y!=Y_BUTTONS) {
            w=strlen(string);
            if (w<max) {
              string[w++]=button[shift?0:1][pos_y*X_BUTTONS+pos_x];
              string[w]='\0';
            }
            w=2*MENU_BORDER+3+MENU_SAFETY;
            w+=get_text_width(title);
            w+=get_text_width(string);
            if (w>width) {
              w=strlen(string);
              string[w-1]='\0';
            }
          } else {
            switch (pos_x) {
            case BUTTON_DELETE:
              w=strlen(string);
              if (w>0)
                string[w-1]='\0';
              break;
            case BUTTON_SHIFT:
              shift=!shift;
              break;
            case BUTTON_OK:
              restore_screen();
              return;
            default:
              break;
            }
          }
        }
      } /* joy&JOY_ALL */
    }
    while( !(joy&JOY_FIRE1));
  }
}

#ifdef MACHINE_HAS_FILESYSTEM
#define MAX_FILENAMES   128
#define MAX_FILENAMELEN 128
#define FSEL_HEIGHT     8

int 
fname_compare(tm_char *s0, tm_char *s1) {
  int a=2,b=2;
  tm_char s2[128],s3[128];

  strcpy(s2,s0);
  if( s2[strlen(s2)-1] == '/') {
    s2[strlen(s2)-1]=0;
    a=1;
  }

  strcpy(s3,s1);
  if( s3[strlen(s3)-1] == '/') {
    s3[strlen(s3)-1]=0;
    b=1;
  }

  if( s2[strlen(s2)-1] == ':') a=0;
  if( s1[strlen(s1)-1] == ':') b=0;

  if(a<b) return -1;
  if(a>b) return 1;
  return(strcmp(s2,s3));
}

#ifdef MACHINE_X11

tm_word
scan_directory( tm_char filenames[MAX_FILENAMES][MAX_FILENAMELEN]) {
  tm_word n=0;
  struct dirent *dent;
  DIR *dir;
  struct stat fstat;

  dir=opendir(".");
  if(dir==NULL) {
    perror("open dir");
    return 0;
  } else {
    dent = readdir(dir);
    while((dent != NULL)&&(n<MAX_FILENAMES)) {
      strncpy(filenames[n],dent->d_name,MAX_FILENAMELEN-1);
      stat(dent->d_name,&fstat);
      if(S_ISDIR(fstat.st_mode))
	strcat(filenames[n],"/");

      n++;
      dent = readdir(dir);
    }
  }
  closedir(dir);

  qsort(filenames,n,MAX_FILENAMELEN, 
	(int(*)(const void *, const void *))fname_compare);

  return(n);
}

tm_void 
get_directory(tm_char *dir, tm_word size) {
  getcwd(dir,size-1);
  if(dir[strlen(dir)-1]!='/')
    strcat(dir,"/");
}

tm_void
set_directory(tm_char *dir, tm_bool drive) {
  if(chdir(dir)<0)
    perror("chdir");
}
#endif /* UNIX/X11 scan_directory */

#ifdef MACHINE_PC_DOS
#define FS_OK
#include <dir.h>
#include <ctype.h>
#include <dos.h>

unsigned int drivemap=0;

tm_word
scan_directory( tm_char filenames[MAX_FILENAMES][MAX_FILENAMELEN]) {
  tm_word done,n=0,i;
  struct ffblk ffblock;
  unsigned int drives;
  int disk;
  tm_char diskname[]="A:";

  /* drives abscannen (gibts anscheinend kein dos/bios fuer) */
  if(drivemap==0) {
    disk=getdisk();
    for(i=0;i<32;i++) {
      setdisk(i);
      if(getdisk()==i)
	drivemap |= (1<<i);
    }
    setdisk(disk);
  }

  for(i=0;i<32;i++)
    if(drivemap & (1<<i)) {
      filenames[n][0]='A'+i;
      filenames[n][1]=':';
      filenames[n][2]=0;
      n++;
    }

  done=findfirst("*.*",&ffblock,FA_RDONLY|FA_DIREC|FA_ARCH);

  while((!done)&&(n<MAX_FILENAMES)) {
    strncpy(filenames[n],ffblock.ff_name,MAX_FILENAMELEN-1);
    if(ffblock.ff_attrib & FA_DIREC)
      strcat(filenames[n],"/");

    n++;
    done=findnext(&ffblock);
  }

  qsort(filenames,n,MAX_FILENAMELEN,
	(int(*)(const void *, const void *))fname_compare);

  return(n);
}

tm_void 
get_directory(tm_char *dir, tm_word size) {
  char *p=dir;

  getcwd(dir,size);
  while(*p!=0) {
    if(isalpha(*p))
      *p=toupper(*p);
    p++;
  }

  /* djgpp/dos benutzt unix-slashes!! */
  if(dir[strlen(dir)-1]!='/')
    strcat(dir,"/");
}

tm_void
set_directory(tm_char *dir, tm_bool drive) {
  if(drive)
    setdisk(dir[0]-'A');
  else
    chdir(dir);
}
#endif /* PC/DOS scan_directory */

#ifdef MACHINE_ATARI_ST
#define FS_OK
#include <tos.h>

tm_word
scan_directory( tm_char filenames[MAX_FILENAMES][MAX_FILENAMELEN]) {
  tm_word done,n=0,i;
  long drivemap;
  struct DTA *dta;
  tm_char diskname[]="A:";

  drivemap=Dsetdrv(Dgetdrv());
  for(i=0;i<32;i++)
    if(drivemap & (1<<i)) {
      filenames[n][0]='A'+i;
      filenames[n][1]=':';
      filenames[n][2]=0;
      n++;
    }

  dta=Fgetdta();
  done=Fsfirst("*.*",FA_RDONLY|FA_DIREC|FA_ARCH);

  while((!done)&&(n<MAX_FILENAMES)) {
    strncpy(filenames[n],dta->d_fname,MAX_FILENAMELEN-1);
    if(dta->d_attrib & FA_DIREC)
      strcat(filename[n],"/");

    n++;
    done=Fsnext();
  }

  qsort(filenames,n,MAX_FILENAMELEN,
	(int(*)(const void *, const void *))fname_compare);

  return(n);
}

tm_void 
get_directory(tm_char *dir, tm_word size) {
  dir[0]='A'+Dgetdrv();
  dir[1]=':';
  Dgetpath(&drv[2],0);
  if(dir[strlen(dir)-1]!='\\')  
    strcat(drv,"\\");
}

tm_void
set_directory(tm_char *dir, tm_bool drive) {
  if(drive)
    Dsetdrv(dir[0]-'A');
  else
    Dsetpath(dir);
}
#endif /* Atari-ST scan_directory */

#ifndef FS_OK
#error Please add the file system functions for your machine or undef MACHINE_HAS_FILESYSTEM!
#endif

tm_void 
draw_fsel_button(tm_word x, tm_word y, tm_word w, tm_bool raised, tm_bool act, tm_char *text) {
  tm_word xs= x+ (w-get_text_width(text))/2;

  draw_hline(x,y,w+3,raised?border_light_col:border_dark_col);
  draw_vline(x,y,font_height+4,raised?border_light_col:border_dark_col);
  draw_hline(x,y+font_height+3,w+3,raised?border_dark_col:border_light_col);
  draw_vline(x+w+2,y,font_height+4,raised?border_dark_col:border_light_col);
  if (colors<16) {
    draw_text(xs,y+2,text,dialog_entry_on_col);
    if (act)
      draw_text(xs+1,y+2,text,dialog_entry_on_col);
  } else
    draw_text(xs,y+2,text,act?dialog_entry_on_col:dialog_entry_off_col);
}

tm_void
select_draw_box(tm_word x, tm_word y, tm_word w, tm_word h, tm_bool inner, tm_bool up) {
  tm_word cnt, col1, col2;

  if(up) {
    col1 = border_dark_col;
    col2 = border_light_col;
  } else {
    col1 = border_light_col;
    col2 = border_dark_col;
  }

  if(inner) {
    for(cnt=0;cnt<MAX(BD,1);cnt++) {
      /* Box um die Box */
      draw_hline(x+cnt,       y+cnt,       w-2*cnt, col1);
      draw_hline(x+cnt,       y+h-cnt-1,   w-2*cnt, col2);
      draw_vline(x+cnt,       y+cnt,       h-2*cnt, col1);
      draw_vline(x+w-cnt-1,   y+cnt,       h-2*cnt, col2);
    }
  } else {
    for(cnt=0;cnt<MAX(BD,1);cnt++) {
      /* Box um die Box */
      draw_hline(x-cnt-1,     y-cnt-1,     w+2*cnt+2, col1);  /* oberkante  */
      draw_hline(x-cnt-1,     y+h+cnt,     w+2*cnt+2, col2);  /* unterkante */
      draw_vline(x-cnt-1,     y-cnt-1,     h+2*cnt+2, col1);  /* links */
      draw_vline(x+w+cnt,     y-cnt-1,     h+2*cnt+2, col2);  /* rechts */
    }
  }
}

tm_void
sel_draw_entry(tm_word x, tm_word y, tm_word width, tm_char *str, tm_bool sel) {
  tm_char buffer[128];
  tm_word col;

  strcpy(buffer,str);

  if(colors<16) {
    col = sel?COL_WHITE:COL_BLACK;
#ifdef SEL_DARK
    if(sel) draw_rect(x,y,width,font_height,COL_BLACK);
#endif  
  } else {
    col = sel?dialog_entry_on_col:dialog_entry_off_col;
#ifdef SEL_DARK
    if(sel) draw_rect(x,y,width,font_height,border_dark_col);
#endif  
  }

  /* string fits into box? */
  if(get_text_width(buffer)<=width) {
    draw_text(x,y,str,col);
  } else {
    while(get_text_width(buffer)>(width-get_text_width("..."))) {
      buffer[strlen(buffer)-1]=0;
    }  
    strcat(buffer,"...");
    draw_text(x,y,buffer,col);
  }
}

tm_void
select_logo(tm_word x, tm_word y, tm_word w, tm_word h, tm_word type, tm_bool sel) {
  tm_word n,m;
  tm_word col_gr, col_rd;

  if(colors<16) {
    if(sel) draw_rect(x,y,w,h,COL_BLACK);
    col_gr = col_rd = sel?COL_WHITE:COL_BLACK;
  } else {
    col_gr = sel?COL_GREEN:COL_DGREEN;
    col_rd = sel?COL_RED:COL_DRED;
  }
  
  /* resize box */
  n=w/10;
  x+=n; y+=n;
  w-=(2*n); h-=(2*n);

  switch(type) {
    /* arrow up */
    case 0:
      for(n=0;n<h;n++)
	draw_hline(x + w/2 - ((n*w)/h/2), y+n, (n*w)/h, col_gr);
      break;

    /* arrow down */
    case 1:
      for(n=0;n<h;n++)
	draw_hline(x + w/2 - ((n*w)/h/2), y+h-n-1, (n*w)/h, col_gr);
      break;

    /* ok */
    case 2:
      m=MAX(BD,1);
      for(n=-m;n<m;n++) {
#if 0
	draw_line(x+w/4, y+w/2+n, x+w/2, y+(3*w)/4+n, col_gr);
	draw_line(x+w/2, y+(3*w)/4+n, x+w, y+n, col_gr);
#else
	draw_line(x, y+w/2+n, x+w/2, y+w+n, col_gr);
	draw_line(x+w/2, y+w+n, x+w, y+n, col_gr);
#endif
      }
      break;

    /* exit */
    case 3:
      m=MAX(BD,1);
      for(n=-m;n<m;n++) {
	draw_line(x, y+n,   x+w, y+w+n, col_rd);
	draw_line(x, y+w+n, x+w, y+n, col_rd);
      }
      break;

  }
}

tm_void
draw_select_box(tm_word x, tm_word y, tm_word w, tm_word h, tm_word entries, 
		tm_char entry_str[MAX_FILENAMES][MAX_FILENAMELEN], tm_word act, tm_word but) {
  tm_word vis_ent, start_ent, ent;
  tm_word sl_x,sl_h,sl_y, b_x;
  tm_bool pressed=get_joystick()&JOY_FIRE1;

  /* number of visible entries */
  vis_ent = h/font_height;

  /* real height of selector */
  h = vis_ent * font_height;

  /* some space for the slider */
  b_x = x + w - font_height;
  sl_x = x + w - (font_height*5)/2;
  w -= 3*font_height;

  /* border around selection area */
  select_draw_box(x, y, w, h, FALSE, TRUE);

  /* and now draw a slider */
  sl_h = (h * vis_ent) / entries;
  sl_h = (sl_h>h)?h:sl_h;    /* slider im verhaeltnis zu allem */
  sl_y = y+((h-sl_h)*act)/(entries-1);
  select_draw_box(sl_x,    y, font_height,    h, FALSE, TRUE);
  select_draw_box(sl_x, sl_y, font_height, sl_h, TRUE, FALSE);

  /* now draw the entries */
  start_ent = act - vis_ent/2;
  if(start_ent<0) start_ent = 0;
  if((entries>=vis_ent)&&(start_ent>entries-vis_ent)) start_ent = entries-vis_ent;
  if(vis_ent>entries) start_ent=0;
  
  for(ent=start_ent; ent< start_ent+vis_ent; ent++) {
    if(ent<entries) {
      sel_draw_entry(x,y+(ent-start_ent)*font_height,w,entry_str[ent],ent==act);
    }
  }

  select_draw_box(b_x, y, font_height, font_height, FALSE, ((but==0)&&(pressed))?TRUE:FALSE);
  select_logo(b_x, y, font_height, font_height, 0, but==0);
  for(ent=0;ent<3;ent++) {
    select_draw_box(b_x, y + ent*(h-font_height)/3 + (h-font_height)/3, 
		    font_height, font_height, FALSE, ((but==(ent+1))&&(pressed))?TRUE:FALSE );
    select_logo(b_x, y + ent*(h-font_height)/3 + (h-font_height)/3, 
		font_height, font_height, ent+1, but==(ent+1));
  }

}	

static tm_void
do_fileselector( tm_void* function) {
  tm_word fs_x, fs_y, fs_w, fs_h, joy, n, files, i, joy_x, joy_y;
  tm_word but;
  tm_char text_buffer[192];
  tm_char filenames[MAX_FILENAMES][MAX_FILENAMELEN];
  tm_word quit=0;

  /* Fileselektor malen */
  init_joystick_delay();
  save_screen();
  files = scan_directory(filenames);
  i=0;

  fs_h  = FSEL_HEIGHT * (font_height+1);
  fs_w  = (4*fs_h)/3;

  /* erstmal Taste loslassen */
  while(joy&JOY_FIRE1) {
    net_handle_messages();
    joy=get_joystick_delay();
  }

  but=0;

  do {
    n=but;
    if(kind_of_joystick()&ABSOLUTE_DEVICE) {
      joy = get_joystick_analog(&joy_x,&joy_y);
      joy=get_joystick_delay();
      but = (joy_y+100)/51;  
    } else {
      joy = get_joystick_delay();
      if(joy&JOY_UP)   but-=1;
      if(joy&JOY_DOWN) but+=1;
      if(but<0) but=0;
      if(but>3) but=3;
    }

    if(n!=but) do_audio(SOUND_CLICK);

    restore_screen();

    /* aktuellen Pfad auslesen */
    get_directory(&text_buffer[3],128);

    /* und Slashes rumdrehen */
    n=3; while(text_buffer[n]!=0) {
      if(text_buffer[n]=='\\')
	text_buffer[n]='/';
      n++;
    }

    /* draw empty dialog box */
    if(get_text_width(&text_buffer[3])<=(fs_w-2*font_height))
      draw_dialog_border(&fs_x,&fs_y,fs_w,fs_h,
			 &text_buffer[3],dialog_header_col,dialog_bg_col);
    else {
      /* if path is too long, cut it */
      n=3;
      while(get_text_width("...")+get_text_width(&text_buffer[n])>(fs_w-2*font_height)) n++;
      text_buffer[n-3]=text_buffer[n-2]=text_buffer[n-1]='.';
      draw_dialog_border(&fs_x,&fs_y,fs_w,fs_h,
			 &text_buffer[n-3],dialog_header_col,dialog_bg_col);      
    }

    draw_select_box(fs_x + font_height/2, fs_y + font_height/2, 
		    fs_w - font_height, fs_h - font_height, 
		    files, filenames, i, but);

    flip_screen();
    
    if(joy & JOY_FIRE1) {
      n=i;
      if(but==0) i--;
      if(but==1) i++;

      if(i<0)      i=0;
      if(i>=files) i=files-1;
      
      if(n!=i) do_audio(SOUND_CLICK);

      if(but==2) {
	if(filenames[i][strlen(filenames[i])-1]=='/') { /* directory/drive */
	  strcpy(text_buffer,filenames[i]);
	  filenames[i][strlen(filenames[i])-1]=0;
	  set_directory(filenames[i],FALSE);
	  i=0;
	} else {
	  if(filenames[i][strlen(filenames[i])-1]==':') { /* directory/drive */
	    set_directory(filenames[i],TRUE);
	    i=0;
	  } else {
	    quit = 1;
	  }
	}
	  
	files=scan_directory(filenames);
	while(joy&JOY_FIRE1) {
	  net_handle_messages();
	  joy=get_joystick_delay();
	}
      }
      if(but==3) quit = 2;
    }

    if(joy & JOY_FIRE0) quit = 2;

    net_handle_messages();
  } while( !quit);

  /* Sprung in Funktion */
  if(quit!=2) {
    /* Datei zusammenbasteln */
    get_directory(text_buffer,128);
    strcat(text_buffer,filenames[i]);
    
    (*(tm_bool(*)(tm_char*))function)(text_buffer);
  }
  do_audio(SOUND_BONUS);
}
#endif

/* globaler Menustatus */
tm_uword g_top, g_left, g_entries,g_max_len, g_max_height;

static tm_void
draw_entries( tm_uword pos, MENU_ENTRY *menu)
{
  tm_uword cnt,value;
  tm_uword text_left;
  tm_bool  bval;

  for(cnt=1;cnt<g_entries;cnt++) {
    text_left = g_left+(g_max_len-get_text_width(menu[cnt].str[language]))/2;

    if((menu[cnt].action == MENU_SCALE)||
       (menu[cnt].action == MENU_CHECK)||
       (menu[cnt].action == MENU_STRING))
      text_left = g_left + MENU_IOFF;

    if(menu[cnt].action == MENU_SCALE) {
      (*(tm_bool(*)(tm_uword*, tm_bool))menu[cnt].value)(&value, PARM_GET);
      draw_scale(g_left+g_max_len-SCALE_WIDTH,g_top+(cnt-1)*(font_height+1),
                 value,(cnt==pos)?dialog_entry_on_col:dialog_entry_off_col);
    }

    if(menu[cnt].action == MENU_CHECK) {
      (*(tm_bool(*)(tm_bool*, tm_bool))menu[cnt].value)(&bval, PARM_GET);
      draw_check(g_left+g_max_len-font_height,g_top+(cnt-1)*(font_height+1),
                 bval,(cnt==pos)?dialog_entry_on_col:dialog_entry_off_col);
    }

    if(menu[cnt].action == MENU_STRING) {
      draw_text(g_left+g_max_len-get_text_width((tm_char *)(menu[cnt].value)),
                g_top+(cnt-1)*(font_height+1),menu[cnt].value,
                (cnt==pos)?dialog_entry_on_col:dialog_entry_off_col);
    }

    draw_menu_text(text_left,g_top+(cnt-1)*(font_height+1),menu[cnt].str[language],
                   (cnt==pos)?dialog_entry_on_col:dialog_entry_off_col,cnt==pos);
  }
}

static tm_void
draw_menu(tm_uword *pos,MENU_ENTRY *menu)
{
  tm_uword cnt,width;
  tm_word  joy_x,joy_y;

  for(g_entries=1;menu[g_entries-1].action!=MENU_END;g_entries++);
  g_max_height=(g_entries-1)*(font_height+1);
  for(cnt=0,g_max_len=0;cnt<g_entries;cnt++) {
    width = get_text_width(menu[cnt].str[language]);
    if (menu[cnt].action == MENU_SCALE)
      width += SCALE_WIDTH;
    else if (menu[cnt].action == MENU_STRING)
      width += get_text_width((tm_char *)(menu[cnt].value));
    if (width + MENU_IOFF > g_max_len)
      g_max_len = width + MENU_IOFF;
  }

  restore_screen();

  draw_dialog_border(&g_left,&g_top,g_max_len,g_max_height,menu[0].str[language],
                     dialog_header_col,dialog_bg_col);

  /* Maus vorpositionieren */
  if (*pos==0)
  {
    if(kind_of_joystick()&ABSOLUTE_DEVICE) {
      get_joystick_analog(&joy_x,&joy_y);
      *pos = (joy_y/((tm_word)(ABSOLUTE_SENSE/g_entries)))+(g_entries/2);
    } else
      *pos = 1;
  }

  if(*pos<1)               *pos=g_entries-1;
  else if(*pos>=g_entries) *pos=1;

  draw_entries(*pos,menu);

  flip_screen();
}

tm_bool
main_menu(tm_void)
{
  tm_uword joy,old_pos,pos;
  tm_word  item=0,joy_x,joy_y,value;
  tm_bool  quit=FALSE,bval;

  init_joystick_delay();
  save_screen();

  do {
    if(item<0)
      item=0;

    do_audio(SOUND_BONUS);
    pos=0;
    draw_menu(&pos,menu[item]);

    do {
      net_handle_messages();
      joy=get_joystick_delay();

      if(kind_of_joystick()&ABSOLUTE_DEVICE) {
        old_pos = pos;
        get_joystick_analog(&joy_x,&joy_y);
        pos = (joy_y/((tm_word)(ABSOLUTE_SENSE/g_entries)))+(g_entries/2);
        if (pos<1)               pos=1;
        else if (pos>=g_entries) pos=g_entries-1;
        if (pos!=old_pos) {
          do_audio(SOUND_CLICK);
          draw_menu(&pos,menu[item]);
        }
      }

      if((joy&JOY_ALL)!=0) {

        /* Joystick, Tastatur etc. */
        if( ((joy&JOY_UP)||(joy&JOY_DOWN)) &&
            (kind_of_joystick()&RELATIVE_DEVICE)) {
          if (joy&JOY_UP)           pos--;
          else if (joy&JOY_DOWN)    pos++;
          if (pos<1)                pos=g_entries-1;
          else if (pos>=g_entries)  pos=1;
          do_audio(SOUND_CLICK);
          draw_menu(&pos,menu[item]);
        }

        if(joy&JOY_FIRE1) {
          switch(menu[item][pos].action) {
          case MENU_SUB:
          case MENU_END:
            item=(tm_word)menu[item][pos].value;
            break;

          case MENU_QUIT:
            quit = TRUE;
            break;

          case MENU_EXEC:
            joy=0;
            do_audio(SOUND_BONUS);
            restore_screen();
            flip_screen();
            (*(tm_void(*)(tm_void))menu[item][pos].value)();
            draw_menu(&pos,menu[item]);
            break;

          case MENU_SCALE:
            if(kind_of_joystick()&RELATIVE_DEVICE)
              (*(tm_bool(*)(tm_word*, tm_bool))menu[item][pos].value)(&value, PARM_GET);

            do {
              if(kind_of_joystick()&ABSOLUTE_DEVICE) {
                joy = get_joystick_analog(&joy_x,&joy_y);
                value = (joy_x/2)+50;
              } else {
                joy = get_joystick();
                if (joy&JOY_LEFT)       value--;
                else if (joy&JOY_RIGHT) value++;
                if (value<0)            value=0;
                else if (value>100)     value=100;
              }
              (*(tm_bool(*)(tm_word*, tm_bool))menu[item][pos].value)(&value, PARM_SET);
              if(kind_of_joystick()&ABSOLUTE_DEVICE)
                (*(tm_bool(*)(tm_word*, tm_bool))menu[item][pos].value)(&value, PARM_GET);
              draw_menu(&pos,menu[item]);
              net_handle_messages();
            } while(joy & JOY_FIRE1) ;
            break;

          case MENU_CHECK:
            (*(tm_bool(*)(tm_bool*, tm_bool))menu[item][pos].value)( &bval, PARM_GET);
            bval = !bval;
            (*(tm_bool(*)(tm_bool*, tm_bool))menu[item][pos].value)( &bval, PARM_SET);
            do_audio(SOUND_CLICK);
            draw_menu(&pos,menu[item]);
            do {
              net_handle_messages();
              joy = get_joystick_delay();
            } while(joy & JOY_FIRE1);
            break;

          case MENU_STRING:
            joy=0;
            do_audio(SOUND_BONUS);
            restore_screen();
            flip_screen();
            get_string(menu[item][pos].str[language],
                       (tm_char *)(menu[item][pos].value),
                       MAX_NAMELEN);
            draw_menu(&pos,menu[item]);
            break;

#ifdef MACHINE_HAS_FILESYSTEM
	  case MENU_FILESEL:
	    joy=0;
	    do_audio(SOUND_BONUS);
	    restore_screen();
	    do_fileselector((tm_void *)(menu[item][pos].value));
            draw_menu(&pos,menu[item]);
            break;
#endif

          default:
            console_puts("unbekanntes Kommando");
            break;
          }
        }
      }
    }
    while( !(joy&JOY_FIRE1) && !(joy&JOY_FIRE0));
  }
  while(!( (joy&JOY_FIRE0)||  /* mit Fire0 (Schuss) immer Abbruch */
           (quit) ||
           ((joy&JOY_FIRE1)&&(item==MENU_TOP)&&(pos==g_entries-1)) ));

  /* Alle Tasten loslassen */
  while((get_joystick()&(JOY_FIRE0|JOY_FIRE1))!=0);

  if(quit)
    return(TRUE);

  return(FALSE);
}

tm_void
text_box(tm_char *titel, tm_char *text)
{
  tm_uword cnt,max_height,max_len,left,top,entries;
  tm_char  string[60],*p;
  tm_bool  split;

  save_screen();
  restore_screen();

  max_len=get_text_width(titel);
  entries=0; p=text; cnt=0;
  do {
    if((*p!='|')&&(*p!=0))
      string[cnt++]=*p;
    else {
      string[cnt++]=0;
      if(get_text_width(string)>max_len)
        max_len=get_text_width(string);
      entries++;
      cnt=0;
    }
  }
  while(*(p++)!=0);

  max_height=entries*(font_height+1);

  draw_dialog_border(&left,&top,max_len+font_height/5,max_height,titel,
                     dialog_header_col,dialog_bg_col);
  left+=font_height/10;

  p=text; cnt=0; entries=0; split=FALSE;
  do {
    /* geESCAPEtes & */
    if((*p=='&')&&(*(p-1)=='\\'))
      string[cnt-1]=*p;
    else
      if((*p!='|')&&(*p!=0)&&(*p!='&'))
        string[cnt++]=*p;
      else {
        string[cnt]=0;
        cnt=0;
        if(*p=='&') {
          draw_text(left,top+entries*(font_height+1),string,dialog_entry_off_col);
          split=TRUE;
        } else {
          if(!split)
            draw_text_width(left,top+entries*(font_height+1),string,
                            dialog_entry_off_col,max_len);
          else
            draw_text(left+max_len-get_text_width(string),
                      top+entries*(font_height+1),string,
                      dialog_entry_off_col);
          split=FALSE;
          entries++;
        }
      }
  }
  while(*(p++)!=0);

  flip_screen();
}

tm_void text_box_butf(tm_char *titel, tm_char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vsprintf(text_buffer,fmt,ap);
  va_end(ap);
  text_box(titel,text_buffer);
  wait_4_button();
}

tm_void dialog_program_info(tm_void)
{
  text_box_butf(dialog_program_title[language],dialog_program_msg[language],
                machine_name[MACHINE],video_driver_name,joystick_driver_name,
                timer_driver_name,audio_driver_name,net_driver_name);
}

tm_void dialog_general_info(tm_void)
{
  text_box_butf(dialog_general_title[language],dialog_general_msg[language]);
}

#define MEASURE_MSEC 1500  /* 1.5 Sekunden pro Messung */
tm_void dialog_system_info(tm_void)
{
  tm_ulong t,m,n;

  m=0; t=get_ticks(); while(ticks2msec(get_ticks()-t)<MEASURE_MSEC) { draw_maze();   m++; };
  n=0; t=get_ticks(); while(ticks2msec(get_ticks()-t)<MEASURE_MSEC) { flip_screen(); n++; };
  text_box_butf(dialog_system_title[language],dialog_system_msg[language],
                (MEASURE_MSEC*100l/m)/100l,  (MEASURE_MSEC*100l/m)%100l,
                (MEASURE_MSEC*100l/n)/100l,  (MEASURE_MSEC*100l/n)%100l,
                (m*100000l/MEASURE_MSEC)/100l,(m*100000l/MEASURE_MSEC)%100l );
}
