/*
  joystick.c

  Joystick-Treiber fuer PC

  12. 7.95  Till   Tastatur-Version
  18. 7.95  Till   Printer-Port-Adapter
  17.10.95  Till   PC-Joystick / PC-Mouse
  19.10.95  Till   Analog-Routinen eingebaut
   4. 2.97  Till   zu DJGPP gewechselt
*/


#include "ctypes.h"
#include "joystick.h"
#include "misc.h"
#include "dialog.h"
#include "gr_prim.h"
#include "console.h"
#include "text.h"
#include "grafik.h"
#include <dos.h>
#include <conio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/kbd.h>
#include <pc.h>
#include <go32.h>
#include <dpmi.h>

#define JOY_MOUSE  0
#define JOY_KBD    1
#define JOY_PCJOY  2
#define JOY_PRTJOY 3

/* PRINTER ADAPTER */
#define BIOSVARSEG    0x01
#define BIOSVAROFF    0x40
#define LPT1          0x08
unsigned short parportAddr;

/* MOUSE */
tm_ushort mouse_buttons;
#define MOUSEINT   0x33   /* mouse driver interrupt */
#define MOUSE_RES   201

/* all devices */
tm_word joy_inp;
char joystick_driver_name[]="DOS V1.2";

tm_void show_joystick_options(tm_void) {
  tm_word i;
  console_printf("%s input driver options:\n",joystick_driver_name);
  console_puts("id=n\tset input device to mouse (n=m), keyboard (n=k),\n"
	       "\tpc-joystick (n=j) or parallel joystick (n=p), default is mouse");
}

tm_void init_joystick_options(tm_void) {
  joy_inp=JOY_MOUSE;
}

tm_bool set_joystick_option(tm_char *cmd) {
  if(cmd[1]=='=') {
    if((cmd[0]=='d')&&((cmd[2]=='k')||(cmd[2]=='m')||
		       (cmd[2]=='j')||(cmd[2]=='p'))) {
      if(cmd[2]=='k') joy_inp=JOY_KBD;
      if(cmd[2]=='m') joy_inp=JOY_MOUSE;
      if(cmd[2]=='j') joy_inp=JOY_PCJOY;
      if(cmd[2]=='p') joy_inp=JOY_PRTJOY;
      return(OK);
    }
  }
  return(FALSE);
}

#define JOY_OVER (100000)

/* read pc analog joystick */
tm_uword xlow,xhigh,xmin,xmax;
tm_uword ylow,yhigh,ymin,ymax;
tm_uword read_joystick(tm_uword *x, tm_uword *y, tm_uword *buttons) {
  tm_uchar c;
  tm_uword xp,yp;

  xp=yp=0;

  disable();

  /* Flip-Flops triggern */
  outportb(0x201,0xff);

  do {
    c = inp(0x201);
    if(c&0x01) xp++;
    if(c&0x02) yp++;
  } while(((c&0x03)!=0x00)&&(xp<JOY_OVER));

  enable();
  *buttons=((c>>4)&0x03)^0x03;
  *x=xp; *y=yp;
  return(OK);
}

tm_bool 
init_joystick(void) {
  tm_uword port,xc,yc,b,col[3]={0,0,0};
  __dpmi_regs reg;
  tm_bool retval=RELATIVE_DEVICE;

  switch(joy_inp) {

  case JOY_PRTJOY:
    parportAddr = _farpeekw(_dos_ds, BIOSVARSEG * 16 + BIOSVAROFF + LPT1);
    /*    port=inp(parportAddr+1); */
    break;
	
  case JOY_KBD:
    break;
    
  case JOY_PCJOY:
    /* dialogs need this */
    init_dialog();
    init_maze_colors();

    read_joystick(&xmin,&ymin,&b);
    if(xmin == JOY_OVER) {
      draw_rect(0,0,width,height,COL_BLACK);
      text_box(error_msg[language],no_joy[language]);
      console_waitkey();
      return(NO_DEVICE);
    }
    
    draw_rect(0,0,width,height,COL_BLACK);
    text_box(joystick_kalibration[language],joystick_kal1[language]);
    wait_4_button();
    read_joystick(&xmin,&ymin,&b);
    flip_screen();

    draw_rect(0,0,width,height,COL_BLACK);
    text_box(joystick_kalibration[language],joystick_kal2[language]);
    wait_4_button();
    read_joystick(&xmax,&ymax,&b);
    flip_screen();
    
    draw_rect(0,0,width,height,COL_BLACK);
    text_box(joystick_kalibration[language],joystick_kal3[language]);
    wait_4_button();
    read_joystick(&xc,&yc,&b);
    flip_screen();
    
    xlow=(xc+xmin)/2;  ylow=(yc+ymin)/2;
    xhigh=(xmax+xc)/2; yhigh=(ymax+yc)/2;

    /* Joystick-Bereich einengen um */
    /* Maximalwerte zu sichern */
    xc=(xmax-xmin)/20;  yc=(ymax-ymin)/20;
    xmin+=xc; ymin+=yc; xmax-=xc; ymax-=yc;
    
    /* everything black again */
    for(b=0;b<colors;b++) set_color(b,col);
    break;
	
  case JOY_MOUSE:
    reg.x.ax = reg.x.bx = 0;
    __dpmi_int(MOUSEINT,&reg);
	
    if(reg.x.ax != 0xffff) {
      init_dialog();
      init_maze_colors();
      draw_rect(0,0,width,height,COL_BLACK);
      text_box(error_msg[language],no_mouse[language]);
      console_waitkey();        /* Taste, denn wir sind schon im Grafikmodus */
      return(NO_DEVICE);
    }
    mouse_buttons = reg.x.bx;
	
    /* set x range */
    reg.x.ax = 7;
    reg.x.cx = 0;
    reg.x.dx = MOUSE_RES;
    /* set y range */
    __dpmi_int(MOUSEINT,&reg);
    reg.x.ax = 8;
    reg.x.cx = 0;
    reg.x.dx = MOUSE_RES;
    __dpmi_int(MOUSEINT,&reg);
    /* set mouse cursor position */
    reg.x.ax = 4;
    reg.x.cx = MOUSE_RES/2;
    reg.x.dx = MOUSE_RES/2;
    __dpmi_int(MOUSEINT,&reg);
    
    retval=ABSOLUTE_DEVICE|JITTER_DEVICE;
  }
  
  kbd_passon = OFF;   /* dont pass to systems handler */
  kbd_install(0);

  atexit(exit_joystick);

  return(retval);
}

tm_void exit_joystick(tm_void)
{
  kbd_remove();
}

tm_uword get_joystick(tm_void)
{
  tm_uword port,joy=0,x,y,b;
  __dpmi_regs reg;

  switch(joy_inp) {
	
  case JOY_PRTJOY:
    port=inp(parportAddr+1);
    if(!(port&0x08)) joy|=JOY_FIRE0;
    if(!(port&0x10)) joy|=JOY_UP;
    if(!(port&0x20)) joy|=JOY_DOWN;
    if(!(port&0x40)) joy|=JOY_RIGHT;
    if(port&0x80)    joy|=JOY_LEFT;
    break;
	
  case JOY_KBD:
    if(kbd_pressed[SCAN_RIGHT])  joy|=JOY_RIGHT;
    if(kbd_pressed[SCAN_LEFT])   joy|=JOY_LEFT;
    if(kbd_pressed[SCAN_UP])     joy|=JOY_UP;
    if(kbd_pressed[SCAN_DOWN])   joy|=JOY_DOWN;
    if(kbd_pressed[SCAN_SPACE])  joy|=JOY_FIRE0;
    if(kbd_pressed[SCAN_RETURN]) joy|=JOY_FIRE1;
    break;

  case JOY_PCJOY:
    read_joystick(&x,&y,&b);
    if(x<xlow)  joy|=JOY_LEFT;
    if(x>xhigh) joy|=JOY_RIGHT;
    if(y<ylow)  joy|=JOY_UP;
    if(y>yhigh) joy|=JOY_DOWN;
    if(b&1)     joy|=JOY_FIRE0;
    if(b&2)     joy|=JOY_FIRE1;
    break;

  case JOY_MOUSE:
    reg.x.ax = 3;
    __dpmi_int(MOUSEINT,&reg);
    if(reg.x.cx<(MOUSE_RES/6))   joy|=JOY_LEFT;
    if(reg.x.cx>(MOUSE_RES*5/6)) joy|=JOY_RIGHT;
    if(reg.x.dx<(MOUSE_RES/6))   joy|=JOY_UP;
    if(reg.x.dx>(MOUSE_RES*5/6)) joy|=JOY_DOWN;
    if(reg.x.bx&1)               joy|=JOY_FIRE0;
    if(reg.x.bx&2)               joy|=JOY_FIRE1;
    if(reg.x.bx&4)               joy|=JOY_FIRE2;
    break;
  }

  if(kbd_pressed[SCAN_F1]) joy|=JOY_FIRE0;
  if(kbd_pressed[SCAN_F2]) joy|=JOY_FIRE1;
  if(kbd_pressed[SCAN_F3]) joy|=JOY_FIRE2;

  return(joy);
}

tm_uword get_joystick_analog(tm_word *x, tm_word *y)
{
  tm_uword joy=0,b,xt,yt;
  __dpmi_regs reg;

  switch(joy_inp) {
	
  case JOY_PRTJOY:
  case JOY_KBD:
    *x=*y=0;
    joy=get_joystick();
    if(joy&JOY_UP)    *y=-100;
    if(joy&JOY_DOWN)  *y=+100;
    if(joy&JOY_LEFT)  *x=-100;
    if(joy&JOY_RIGHT) *x=+100;
    break;

  case JOY_PCJOY:
    read_joystick(&xt,&yt,&b);
    if(xt<xmin) xt=xmin;
    if(xt>xmax) xt=xmax;
    if(yt<ymin) yt=ymin;
    if(yt>ymax) yt=ymax;
    /* beware of division by zero !! */
    *x=((xt-xmin)*201l)/(xmax-xmin)-101;
    *y=((yt-ymin)*201l)/(ymax-ymin)-101;
    if(b&1)  joy|=JOY_FIRE0;
    if(b&2)  joy|=JOY_FIRE1;
    break;
	
  case JOY_MOUSE:
    reg.x.ax = 3;
    __dpmi_int(MOUSEINT,&reg);
    *x=reg.x.cx-(MOUSE_RES/2);
    *y=reg.x.dx-(MOUSE_RES/2);
    if(reg.x.bx&1) joy|=JOY_FIRE0;
    if(reg.x.bx&2) joy|=JOY_FIRE1;
    if(reg.x.bx&4) joy|=JOY_FIRE2;
    break;
  }
  
  if(kbd_pressed[SCAN_F1]) joy|=JOY_FIRE0;
  if(kbd_pressed[SCAN_F2]) joy|=JOY_FIRE1;
  if(kbd_pressed[SCAN_F3]) joy|=JOY_FIRE2;

  return(joy);
}

tm_uword kind_of_joystick(void)
{
  if(joy_inp == JOY_MOUSE)
    return(ABSOLUTE_DEVICE|JITTER_DEVICE);
  
  return(RELATIVE_DEVICE);
}
