/*
   dos/sblaster.c

*/

#include <sb_lib.h>
#include "ctypes.h"
#include "audio.h"
#include "joystick.h"
#include "net.h"
#include "misc.h"
#include "dialog.h"
#include "maze.h"
#include "gr_prim.h"
#include "grafik.h"
#include "music/music.h"
#include "the_song.h"

/* dumm-di-dumm trallala */
#define MUSIC_ENDLESS
#define MUSIC_PREFILL         (4)     /* have 3 beats in queue */
#define MUSIC_OUTPUT_RATE     (FREQ)  /* */
#define MUSIC_BEATS_PER_MIN   (80)
#define MUSIC_OUTPUT_SIZE     (60*MUSIC_OUTPUT_RATE/MUSIC_BEATS_PER_MIN)
#define MUSIC_BUFFER_SIZE     (MUS_MAX_PRECALC+MUSIC_OUTPUT_SIZE)
#define MUSIC_BUFFERS         (MUSIC_PREFILL+4)

tm_char audio_driver_name[]="SBlaster V0.8";
tm_bool aon=TRUE;
sb_sample *snd[16],sam[MUSIC_BUFFERS];
sb_status stat;

tm_void show_audio_options(tm_void)
{
}

tm_void init_audio_options(tm_void)
{
}

tm_bool set_audio_option(tm_char *cmd)
{
  return(FALSE);
}

tm_bool 
audio_on(tm_bool *on, tm_bool get) {
  if(stat == SB_SUCCESS) {
    if(get) *on = aon;
    else    aon = *on;
  } else
    if(get) *on = OFF;

  return(OK);
}

/* volume ranges from 0..100 */
tm_bool 
vol_audio(tm_uword *volume, tm_bool get) 
{
  int err,i;
  static int vol=32;

  if(stat == SB_SUCCESS) {
    if(get) {
      *volume = vol * 100 / 63;
    } else {
      vol = *volume * 63 / 100;
      for(i=0;i<SOUND_LAST;i++)
	snd[i]->left_volume=snd[i]->right_volume=vol;
      /*
      for(i=0;i<MUSIC_BUFFERS;i++)
	sam[i]->left_volume=sam[i]->right_volume=vol;
	*/
    }
  } else
    if(get) *volume=0;

  return(OK);
}

extern tm_char *tm_sounds[];
extern tm_word tm_sound_len[];

unsigned int eof;
void increase_eof(void) { eof++; }

tm_bool init_audio(tm_void)
{
  tm_word i;
  sb_sample orig;

  /* open the blaster with 8kHz */
  stat = sb_install_driver(FREQ);
  if(stat != SB_SUCCESS) {  
    init_dialog();
    init_maze_colors();
    draw_rect(0,0,width,height,COL_BLACK);
    text_box_butf("Sound Blaster","Error opening the|Blaster hardware!");
    return(ERROR);
  }

  orig.stereo=orig.bits=0;
  orig.left_volume=orig.right_volume=32;
  orig.callback=increase_eof;

  for(i=0;i<SOUND_LAST;i++) {
    orig.data=tm_sounds[i];
    orig.length=tm_sound_len[i];
    snd[i]=sb_convert_frequency(&orig,8000,sb_sample_frequency);
  }

  atexit(exit_audio);

  return(OK);
}

tm_bool exit_audio(tm_void)
{
  if(stat == SB_SUCCESS)
    sb_uninstall_driver();

  return(OK);
}

tm_long   music_buffer[MUSIC_BUFFER_SIZE];
tm_uchar  music_output_buffer[MUSIC_BUFFERS][MUSIC_OUTPUT_SIZE+10];

tm_void
do_music(tm_void) {
  tm_uchar *pos=(tm_uchar*)mus_song,sample;
  tm_word  i,time,vol,len,joy,beatlen;
  unsigned int old_eof=-MUSIC_PREFILL,buf=0;
  tm_ulong val,factor;
  tm_long  *dest;
  tm_short *src;
  /*tm_void flush_audio(tm_void); */

  if((aon)&&(stat==SB_SUCCESS)) {
    /* clear buffer */
    eof=0;
    memset(music_buffer,0,sizeof(music_buffer));
    
    /* get length of beat in n/32 */
    beatlen=*pos++;
	
    do {
      time=0;  /* start a new beat */
      while((time<beatlen)&&(*pos!=MUS_END)) {   /* one beat completed ? */
	
	sample=*pos++;                      /* get sample no */
	
	if(sample==MUS_WAIT) {
	  time+=*pos++;     /* time in n/32 beats */
	} else {
	  /* get frequency factor  */
	  factor=*pos++*256*256*256; factor+=*pos++*256*256; factor+=*pos++*256; factor+=*pos++;
	  vol=*pos++; len=*pos++; /* get volume and length */
	  
	  /* compose samples in buffer */
	  dest=&music_buffer[time*(MUSIC_OUTPUT_SIZE/beatlen)]; src=(tm_short*)mus_sample[sample];
	  for(i=0;i<((tm_word)mus_sample_len[sample]<<16)/factor;i++)
	    *dest+++=(vol*src[(i*factor)>>16])>>8;
	}
#ifdef MUSIC_ENDLESS
	if(*pos==MUS_END)
	  pos=(tm_uchar*)mus_song+1;
#endif
      }
      
      /* play this beat */
      /* output buffer (4*long -> signed short -> ulaw) */
      for(i=0;i<MUSIC_OUTPUT_SIZE;i++)
	music_output_buffer[buf][i]=((music_buffer[i]/MUS_INSTRUMENTS/8)/256);
      /* shift music buffer down */
      memcpy(music_buffer,&music_buffer[MUSIC_OUTPUT_SIZE],(MUSIC_BUFFER_SIZE-MUSIC_OUTPUT_SIZE)*sizeof(tm_long));
      /* clear buffer end */
      memset(&music_buffer[MUSIC_BUFFER_SIZE-MUSIC_OUTPUT_SIZE],0,MUSIC_OUTPUT_SIZE*sizeof(tm_long));
      
      /* and down to the device */
      sam[buf].data=music_output_buffer[buf];
      sam[buf].length=MUSIC_OUTPUT_SIZE;
      sam[buf].stereo=sam[buf].bits=0;
      sam[buf].left_volume=sam[buf].right_volume=32;
      sam[buf].callback=increase_eof;
      sb_queue_sample(&sam[buf]);
      
      buf++;
      if(buf==MUSIC_BUFFERS) buf=0;
      
#if 0      
      write(Audio_fd,music_output_buffer,MUSIC_OUTPUT_SIZE);
      audio_play_eof(Audio_fd);
#endif
	  
      /* wait for end of beat */
      do {
	joy=get_joystick();
      } while((eof==old_eof)&&!(joy&JOY_FIRE));
      old_eof=eof;
      
      net_handle_messages();
      /* music running and nobody pressed fire and game not running? */
    } while((*pos!=MUS_END)&&!(joy&JOY_FIRE)&&!(game_running));
    
    /* wait for button release */
    if(joy&JOY_FIRE) {
#if 0
      flush_audio();
#endif
      wait_msec(10);
      do
	joy=get_joystick();
      while(joy&JOY_FIRE);
    } else {
      if(!game_running)
	/* wait two beats for music to stop */
	wait_msec(2000*MUSIC_BEATS_PER_MIN/60);
#if 0
      else
	flush_audio();
#endif
    }
  } else {
    do {
      joy=get_joystick();
      net_handle_messages();
    } while(!(joy&JOY_FIRE)&&(!game_running));
    while(joy&JOY_FIRE) joy=get_joystick();
  }
}

tm_void 
do_audio(tm_sound sound)
{
  if(aon)
    sb_mix_sample(snd[sound]);
}

tm_void 
test_audio(tm_void)
{
  do_audio(SOUND_TEST);
}

