/*
   DOS\VGA.C

   Hardware-Treiber fuer PC-VGA (256 Farben)

    8.7.95   Till  erste Version
   10.7.95   Till  256 Farben, kein BGI
    4.2.97   Till  BC -> DJGPP
    7.2.97   Till  VESA option
   14.2.98   Till  linear mapping with VBE 2.0
   15.2.98   Till  double buffering with VBE 2.0 and sufficient memory
   17.2.98   Till  one executable for std vga and vesa
*/

#include <dos.h>
#include "gr_prim.h"
#include "ctypes.h"
#include "text.h"
#include "grafik.h"
#include "console.h"
#include "gif.h"
#include "misc.h"
#include "machine.h"
#include "maze.h"
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pc.h>
#include <dpmi.h>
#include <go32.h>
#include <sys/movedata.h>
#include <sys/nearptr.h>
#include <sys/farptr.h>
#include <sys/segments.h>


#define WAIT_VSYNC

#define VGA_VER "0.91"

char video_driver_name[32]="VGA V" VGA_VER;

#define STEPS 50
#define SLEEP 2

#define USE_CLIB    /* Grafikprimitiva duerfen string.h benutzen */

#define STDVGA256
#define COPY_TO_SCREEN

#define WIDTH_A (width)
#define WIDTH   (width)
#define HEIGHT  (height)

#define COLORS  (256)
#define COL256

#define VIDEO     0x10  /* Interrupt-Nummer */

#define DAC_CPM  0x3c6  /* DAC-Mask        */
#define DAC_CPRI 0x3c7  /* DAC-Read-Index  */
#define DAC_CPWI 0x3c8  /* DAC-Write-Index */
#define DAC_CPD  0x3c9  /* DAC-Data        */

#define ST01     0x3da  /* input status register 1 */
#define VSYNC { while(!(inp(ST01)&0x08)); while((inp(ST01)&0x08)); }

char  *video;
extern tm_picture pic_tm;
#ifdef THREE_D
extern tm_picture pic_tm3d;
#endif


#define MASK_LINEAR(addr)     (addr & 0x000FFFFF)
#define RM_TO_LINEAR(addr)    (((addr & 0xFFFF0000) >> 12) + (addr & 0xFFFF))
#define RM_OFFSET(addr)       (MASK_LINEAR(addr) & 0xFFFF)
#define RM_SEGMENT(addr)      ((MASK_LINEAR(addr) & 0xFFFF0000) >> 4)

typedef struct VESA_INFO         /* VESA information block structure */
{ 
   unsigned char  VESASignature[4]     __attribute__ ((packed));
   unsigned short VESAVersion          __attribute__ ((packed));
   unsigned long  OEMStringPtr         __attribute__ ((packed));
   unsigned char  Capabilities[4]      __attribute__ ((packed));
   unsigned long  VideoModePtr         __attribute__ ((packed)); 
   unsigned short TotalMemory          __attribute__ ((packed)); 
   unsigned short OemSoftwareRev       __attribute__ ((packed)); 
   unsigned long  OemVendorNamePtr     __attribute__ ((packed)); 
   unsigned long  OemProductNamePtr    __attribute__ ((packed)); 
   unsigned long  OemProductRevPtr     __attribute__ ((packed)); 
   unsigned char  Reserved[222]        __attribute__ ((packed)); 
   unsigned char  OemData[256]         __attribute__ ((packed)); 
} VESA_INFO;

typedef struct MODE_INFO         /* VESA information for a specific mode */
{
   unsigned short ModeAttributes       __attribute__ ((packed)); 
   unsigned char  WinAAttributes       __attribute__ ((packed)); 
   unsigned char  WinBAttributes       __attribute__ ((packed)); 
   unsigned short WinGranularity       __attribute__ ((packed)); 
   unsigned short WinSize              __attribute__ ((packed)); 
   unsigned short WinASegment          __attribute__ ((packed)); 
   unsigned short WinBSegment          __attribute__ ((packed)); 
   unsigned long  WinFuncPtr           __attribute__ ((packed)); 
   unsigned short BytesPerScanLine     __attribute__ ((packed)); 
   unsigned short XResolution          __attribute__ ((packed)); 
   unsigned short YResolution          __attribute__ ((packed)); 
   unsigned char  XCharSize            __attribute__ ((packed)); 
   unsigned char  YCharSize            __attribute__ ((packed)); 
   unsigned char  NumberOfPlanes       __attribute__ ((packed)); 
   unsigned char  BitsPerPixel         __attribute__ ((packed)); 
   unsigned char  NumberOfBanks        __attribute__ ((packed)); 
   unsigned char  MemoryModel          __attribute__ ((packed)); 
   unsigned char  BankSize             __attribute__ ((packed)); 
   unsigned char  NumberOfImagePages   __attribute__ ((packed));
   unsigned char  Reserved_page        __attribute__ ((packed)); 
   unsigned char  RedMaskSize          __attribute__ ((packed)); 
   unsigned char  RedFieldPosition     __attribute__ ((packed)); 
   unsigned char  GreenMaskSize        __attribute__ ((packed)); 
   unsigned char  GreenFieldPosition   __attribute__ ((packed));
   unsigned char  BlueMaskSize         __attribute__ ((packed)); 
   unsigned char  BlueFieldPosition    __attribute__ ((packed)); 
   unsigned char  RsvdMaskSize         __attribute__ ((packed)); 
   unsigned char  RsvdFieldPosition    __attribute__ ((packed)); 
   unsigned char  DirectColorModeInfo  __attribute__ ((packed));
   unsigned long  PhysBasePtr          __attribute__ ((packed)); 
   unsigned long  OffScreenMemOffset   __attribute__ ((packed)); 
   unsigned short OffScreenMemSize     __attribute__ ((packed)); 
   unsigned char  Reserved[206]        __attribute__ ((packed)); 
} MODE_INFO;

typedef struct PM_INFO           /* VESA 2.0 protected mode interface */
{
   unsigned short setWindow            __attribute__ ((packed)); 
   unsigned short setDisplayStart      __attribute__ ((packed)); 
   unsigned short setPalette           __attribute__ ((packed)); 
   unsigned short IOPrivInfo           __attribute__ ((packed)); 
   /* pmode code is located in this space */
} PM_INFO;

static VESA_INFO vesa_info;            /* SVGA info block */
static MODE_INFO mode_info;            /* info for this video mode */
__dpmi_regs _dpmi_reg;                 /* for calling int 10 bank switch */
tm_bool linear;
static unsigned long lb_linear = 0;    /* linear address of framebuffer */
static int lb_segment = 0;             /* descriptor for the buffer */


/* get_vesa_info:
 *  Retrieves a VESA info block structure, returning 0 for success.
 */
static int get_vesa_info()
{
   int c;

   for (c=0; c<sizeof(VESA_INFO); c++)
      _farpokeb(_dos_ds, MASK_LINEAR(__tb)+c, 0);

   dosmemput("VBE2", 4, MASK_LINEAR(__tb));

   _dpmi_reg.x.ax = 0x4F00;
   _dpmi_reg.x.di = RM_OFFSET(__tb);
   _dpmi_reg.x.es = RM_SEGMENT(__tb);
   __dpmi_int(0x10, &_dpmi_reg);
   if (_dpmi_reg.h.ah)
      return -1;

   dosmemget(MASK_LINEAR(__tb), sizeof(VESA_INFO), &vesa_info);

   if (strncmp(vesa_info.VESASignature, "VESA", 4) != 0)
      return -1;

   return 0;
}



/* get_mode_info:
 *  Retrieves a mode info block structure, for a specific graphics mode.
 *  Returns 0 for success.
 */
static int get_mode_info(int mode)
{
   int c;

   for (c=0; c<sizeof(MODE_INFO); c++)
      _farpokeb(_dos_ds, MASK_LINEAR(__tb)+c, 0);

   _dpmi_reg.x.ax = 0x4F01;
   _dpmi_reg.x.di = RM_OFFSET(__tb);
   _dpmi_reg.x.es = RM_SEGMENT(__tb);
   _dpmi_reg.x.cx = mode;
   __dpmi_int(0x10, &_dpmi_reg);
   if (_dpmi_reg.h.ah)
      return -1;
   
   dosmemget(MASK_LINEAR(__tb), sizeof(MODE_INFO), &mode_info);
   return 0;
}



/* _vesa_vidmem_check:
 *  Trying to autodetect the available video memory in my hardware level 
 *  card drivers is a hopeless task. Fortunately that seems to be one of 
 *  the few things that VESA usually gets right, so if VESA is available 
 *  the non-VESA drivers can use it to confirm how much vidmem is present.
 */
long _vesa_vidmem_check(long mem)
{
   if (get_vesa_info() != 0)
      return mem;

   if (vesa_info.TotalMemory <= 0)
      return mem;

   return MIN(mem, (vesa_info.TotalMemory << 16));
}

/* find_vesa_mode:
 *  Tries to find a VESA mode number for the specified screen size.
 *  Searches the mode list from the VESA info block, and if that doesn't
 *  work, uses the standard VESA mode numbers.
 */
static int find_vesa_mode(int w, int h)
{
   #define MAX_VESA_MODES 1024

   unsigned short mode[MAX_VESA_MODES];
   int c, modes;
   long mode_ptr;

   if (get_vesa_info() != 0) {
      console_puts("VESA not found");
      return 0;
   }

   mode_ptr = RM_TO_LINEAR(vesa_info.VideoModePtr);
   modes = 0;

   while ((mode[modes] = _farpeekw(_dos_ds, mode_ptr)) != 0xFFFF) {
      modes++;
      mode_ptr += 2;
   }

   /* search the list of modes */
   for (c=0; c<modes; c++) {
      if (get_mode_info(mode[c]) == 0) {
	 if ((mode_info.ModeAttributes & 1) &&     /* mode is available */
	     (mode_info.ModeAttributes & 8) &&     /* and is a color mode */
	     (mode_info.ModeAttributes & 16) &&    /* and a graphics mode */
	     (mode_info.XResolution == w) &&       /* check width */
	     (mode_info.YResolution == h) &&       /* check height */
	     (mode_info.NumberOfPlanes == 1) &&    /* no planes please */
	     (mode_info.BitsPerPixel == 8) &&      /* want 256 colors */
	     (mode_info.MemoryModel == 4))         /* want packed pixels */
	    /* looks like this will do... */
	    return mode[c];
      } 
   }

   /* try the standard mode numbers */
   if ((w == 640) && (h == 400))        c = 0x100;
   else if ((w == 640) && (h == 480))   c = 0x101;
   else if ((w == 800) && (h == 600))   c = 0x103;
   else if ((w == 1024) && (h == 768))  c = 0x105;
   else if ((w == 1280) && (h == 1024)) c = 0x107;
   else {
      console_puts("VESA: Resolution not supported");
      return 0; 
   }

   if (get_mode_info(c) == 0)
      return c;

   console_puts("VESA: Resolution not supported");
   return 0;
}

tm_word granul;
tm_bool double_buffer;
tm_word double_page, double_cnt;
tm_bool vesa;


char *vtext;
long screen_base;
tm_uchar video_mode;
tm_uword font_height;
extern tm_uword font_height0, font_height1;
CHARACTER charTable[128];
extern CHARACTER charTable0[128], charTable1[128];

#ifdef THREE_D
tm_3d_dev use_3d;
#endif

#ifdef TEXTURE
tm_bool texon;
#endif

#define TEXTSIZE 4000

#define UPDATE_PAGE    0
#define UPDATE_LINEAR  1
#define UPDATE_DOUBLE  2
#define UPDATE_TRIPPLE 3
tm_word umode;
tm_uchar *offscreen_buffer;

tm_uchar orig_text[TEXTSIZE];
tm_uchar orig_cols[256][3];
tm_ushort orig_cursor;

tm_void show_video_options(tm_void)
{
  console_printf("%s video driver options:\n",video_driver_name);
  console_printf("vs=n*m\tset window size, default is %d*%d, (min %d*%d, max %d*%d)\n",
		 DEF_REZ_WIDTH, DEF_REZ_HEIGHT, MIN_REZ_WIDTH, 
		 MIN_REZ_HEIGHT, MAX_REZ_WIDTH, MAX_REZ_HEIGHT);
  console_puts("vm=n\tset update mode to copy page (n=p), copy linear (n=l),"
               "\n\tdouble buffer (n=d) or tripple buffer (n=t), default is best");
#ifdef THREE_D
  console_puts("v3=n\tswitch 3d effect off (n=n), use 3D-MAX glasses (n=3) or"
	       "\n\tuse CyberMaxx hmd (n=c), default is off");
#endif
#ifdef TEXTURE
  console_puts("vt=n\tuse texture mapping (n=on/off), default is on");
#endif
}

tm_void init_video_options(tm_void)
{
  width = DEF_REZ_WIDTH;
  height = DEF_REZ_HEIGHT;
  umode = UPDATE_TRIPPLE;

#ifdef TEXTURE
  texon=TRUE;
#endif
#ifdef THREE_D
  use_3d=DEV_NONE;
#endif
}

tm_bool set_video_option(tm_char *cmd)
{
  tm_word n;
  char str[10];

  if(cmd[1]=='=') {
    if((cmd[0]=='m')||(cmd[0]=='M')) {
      switch(cmd[2]) {
	case 'p':
	case 'P':
          umode = UPDATE_PAGE;
	  return(TRUE);
	  break;

	case 'l':
        case 'L':
          umode = UPDATE_LINEAR;
	  return(TRUE);
          break;

        case 'd':
        case 'D':
          umode = UPDATE_DOUBLE;
	  return(TRUE);
	  break;

        case 't':
        case 'T':
          umode = UPDATE_TRIPPLE;
	  return(TRUE);
	  break;
      }
    }

    if((cmd[0]=='s')||(cmd[0]=='S')) {
      n=0;
      while(isdigit(cmd[n+2])) { 
	str[n]=cmd[n+2]; n++; 
      }
      str[n]=0;
      
      if((cmd[n+2]!='*')&&(cmd[n+2]!='x'))
	return(FALSE);
      
      width = atoi(str);
      height = atoi(&cmd[n+3]);
      
      if((width<MIN_REZ_WIDTH)||(height<MIN_REZ_HEIGHT)||
	 (width>MAX_REZ_WIDTH)||(height>MAX_REZ_HEIGHT)) {
	width = DEF_REZ_WIDTH;
	height = DEF_REZ_HEIGHT;
	return(FALSE);
      }
      return(OK);
    }
#ifdef THREE_D
    if((cmd[0]=='3')&&((cmd[2]=='n')||(cmd[2]=='c')||(cmd[2]=='3'))) {
      if(cmd[2]=='c') use_3d=DEV_CBMAX;
      if(cmd[2]=='3') use_3d=DEV_3DMAX;
      return(OK);
    }
#endif
#ifdef TEXTURE
    if((cmd[0]=='t')
       &&((cmd[2]=='O')||(cmd[2]=='o'))
       &&((cmd[3]=='F')||(cmd[3]=='f'))
       &&((cmd[4]=='F')||(cmd[4]=='f'))) {
      texon=FALSE;
      return(OK);
    }
#endif
  }

  return(FALSE);
}

#define COPY_WIN(z,y)   regs.x.ax = 0x4f05;  regs.x.bx = 0;  regs.x.dx = z*granul; \
                        __dpmi_int(VIDEO, &regs); \
		        memcpy(video, (char*)screen_base+z*65536, y);

void flip_screen(void)
{
  __dpmi_regs regs;
  tm_long i=0, size = width*height;

  if(double_buffer) {
    regs.x.ax = 0x4f07;
    regs.h.bh = 0;
    regs.h.bl = 0x80;   /* set display start */
    regs.x.cx = 0;   /* first displayed pixel in scan line */
    regs.x.dx = double_page*height; /* first displayed scan line */    
    __dpmi_int(VIDEO, &regs);

    /* Zeichenspeicher ist immer eins vor dem aktuell dargestellten */
    /* das ist vor allem wichtig bei tripple buffering */
    if(++double_page > double_cnt)
      double_page=0;

    /* switch to other buffer */
    screen_base = (long)(video + double_page*size);

    return;
  }

#ifdef WAIT_VSYNC
  VSYNC;
#endif

  if(linear)
    memcpy(video, (char*)screen_base, size);
  else {

    /* copy video buffer in 64k chunks */
    while(size>65535) {
      COPY_WIN(i,65536);
      i++;
      size-=65536;
    }
    
    if(size>0) {
      COPY_WIN(i,size);
    }
  }
}

tm_bool open_screen( tm_uword argc, tm_char **argv, tm_uword *width, tm_uword *height, tm_uword *colors )
{
  __dpmi_regs regs;
  __dpmi_meminfo info;
  int i,m,vmode;

  linear = FALSE;
  double_buffer = FALSE;

  __djgpp_nearptr_enable();
  video = (char*)0xa0000 + __djgpp_conventional_base;
  vtext = (char*)0xb8000 + __djgpp_conventional_base;

  /* save text mode contents */
  memcpy(orig_text, vtext, TEXTSIZE);

  /* VGA feststellen */
  regs.x.ax = 0x1a00;
  __dpmi_int(VIDEO, &regs);
  if(regs.h.al!=0x1a) {
    puts(no_vga[language]);
    return(FALSE);
  }

  /* VESA feststellen */
  /* search for maching graphics mode */
  if((vmode = find_vesa_mode(*width, *height))==0) {
    vesa=FALSE;
    *width = 320;
    *height = 200;
    vmode = 0x13;
    linear = TRUE;
  } else {
    vesa=TRUE;
    granul=64/mode_info.WinSize;
  }

  screen_base=(long)malloc((size_t)*width * *height);
  if(screen_base==0l) {
    puts(no_video_buffer[language]);
    return(FALSE);
  }

  /* copy right character info */
  if((*width>=640)&&(*height>=400)) {
    font_height = font_height1;
    memcpy(charTable, charTable1, 128* sizeof(CHARACTER));
  } else {
    font_height = font_height0;
    memcpy(charTable, charTable0, 128* sizeof(CHARACTER));
  }

  /* originale Farben lesen */
  outp(DAC_CPRI,0);
  for(i=0;i<256;i++) {
    orig_cols[i][0]=inp(DAC_CPD);
    orig_cols[i][1]=inp(DAC_CPD);
    orig_cols[i][2]=inp(DAC_CPD);
  }

  /* cursor sichern */
  regs.h.ah = 0x03;
  regs.h.bh = 0x00;
  __dpmi_int(VIDEO, &regs);
  orig_cursor = regs.x.dx;

#ifdef THREE_D
  if(use_3d == DEV_3DMAX) {
    /* switch 3d glasses on */
    regs.x.ax = 0x4ed0;
    __dpmi_int(VIDEO, &regs);
    if(regs.x.dx!=0x3344) {         /* '3D' */
      use_3d = FALSE;
      console_puts("No 3d bios installed, switching to non 3d mode!");
      console_puts("Press any key to continue!");
      console_waitkey();
    }
  }
#endif

#ifdef TEXTURE
  if(texon)
    use_texture=TRUE;
#endif

  if ((vesa)&&(mode_info.ModeAttributes & 0x80)&&(umode>UPDATE_PAGE)) {   /* linear framebuffer? */
    linear = TRUE;

    info.size = vesa_info.TotalMemory<<16;
    info.address = mode_info.PhysBasePtr;
    if(__dpmi_physical_address_mapping(&info)==-1) {
      console_printf("vga.c: physical mapping of vesa address 0x%08x failed!\n", 
		     mode_info.PhysBasePtr);
      linear=FALSE;
    } else {
      video = (tm_char*)(info.address + __djgpp_conventional_base);
    }
  }

  if((vesa)&&(linear)&&(umode>UPDATE_LINEAR)) {
    if(vesa_info.TotalMemory<<16 >= (2* *width * *height)) {
      /* enable double buffering */
      double_buffer=TRUE;
      double_cnt=1;
      double_page = 0;
      free((void*)screen_base);  /* with double buffer we dont need offscreen mem */

      /* tripple buffering possible?? */
      if((vesa_info.TotalMemory<<16 >= (3* *width * *height))&&(umode>UPDATE_DOUBLE)) {
	double_cnt=2;
	double_page=1;
      }
      screen_base = (long)(video + double_page * (*width * *height));
    }
  }

  /* alten Videomodus sichern */
  regs.x.ax = 0x0f00;
  __dpmi_int(VIDEO, &regs);
  video_mode = regs.h.al;

  /* fade to black */
  outp(DAC_CPM,0xff);

  /* originale Farben lesen */
  outp(DAC_CPRI,0);
  for(i=0;i<256;i++) {
    orig_cols[i][0]=inp(DAC_CPD);
    orig_cols[i][1]=inp(DAC_CPD);
    orig_cols[i][2]=inp(DAC_CPD);
  }

  for(m=STEPS;m>=0;m--) {
    outp(DAC_CPWI,0);
    for(i=0;i<256;i++) {
      outp(DAC_CPD,orig_cols[i][0]*m/STEPS);
      outp(DAC_CPD,orig_cols[i][1]*m/STEPS);
      outp(DAC_CPD,orig_cols[i][2]*m/STEPS);
    }
    VSYNC;
  }

  offscreen_buffer = (tm_uchar*)malloc(MAX(*width,pic_tm.width) * MAX(*height,pic_tm.height));
  if(offscreen_buffer == NULL) {
    console_printf("vga.c: out of memory allocating %d bytes\n", *width * *height);
    return(FALSE);
  }

  if(vesa) {
    /* Video-Modus setzen */
    regs.x.ax = 0x4f02;
    regs.x.bx = vmode;
    if (linear) {
      regs.x.bx |= 0x4000;
      if(double_buffer) {
	if(double_cnt == 1)
	  strcpy(video_driver_name,"VESA DBL VGA V" VGA_VER);
	else
	  strcpy(video_driver_name,"VESA TRI VGA V" VGA_VER);
      } else
	strcpy(video_driver_name,"VESA LIN VGA V" VGA_VER);
    } else
      strcpy(video_driver_name,"VESA VGA V" VGA_VER);
  } else {
    /* Video-Modus setzen */
    regs.x.ax = vmode;
    regs.x.bx = 0x00;
    strcpy(video_driver_name,"STD VGA V" VGA_VER);
  }
  __dpmi_int(VIDEO, &regs);

  *colors = COLORS;

#ifdef THREE_D
  if(use_3d == DEV_3DMAX) {
    regs.x.ax = 0x4fd0;
    regs.x.bx = 0x3d01;   /* switch interlace on */
    __dpmi_int(VIDEO, &regs);
    regs.x.ax = 0x4fd0;
    regs.x.bx = 0x3d03;   /* switch glasses on */
    __dpmi_int(VIDEO, &regs);
  }
#endif
  /* give vga abd crt some seconds */
  sleep(SLEEP);

  return(OK);
}

tm_void fade_title_pic(tm_char *copyright) {
  int  i,j,c;
  tm_word col[3];
  int N,M;
  unsigned char *src,*p,*dst;

  /* all colors black */
  col[0]=col[1]=col[2]=0;
  for (i=0; i<COLORS; i++)
    set_color(i,col);
  
  /* show title screen */
  if((width==pic_tm.width)&&(height==pic_tm.height))
    show_gif("tm",(tm_char*)screen_base);
  else {

    show_gif("tm",offscreen_buffer);

    /* now scale title pic to screen size */
    for(dst=(tm_char*)screen_base,src=offscreen_buffer,M=0,i=0;i<height;i++)  
    {  /* do this for all lines */
      while(M>height) { M-=height; src+=pic_tm.width; }
      M+=pic_tm.height;
	
      for(p=src,N=0,j=0;j<width;j++) {
	while(N>width) { N-=width; p++; }
	N+=pic_tm.width;
	*dst++=*p;
      }
    }
  }

  /* display copyright message */
  draw_text((WIDTH-get_text_width(copyright))/2,HEIGHT/6,copyright,COL_RED);

  flip_screen();

  for (j=0; j<=STEPS; j++) {
    for (i=0; i<COLORS; i++) {
#ifdef THREE_D
      if(use_3d) {
	col[0] = ((pic_tm3d.color[i].r * 1000L) / 255L) * j / STEPS;
	col[1] = ((pic_tm3d.color[i].g * 1000L) / 255L) * j / STEPS;
	col[2] = ((pic_tm3d.color[i].b * 1000L) / 255L) * j / STEPS;
	set_color(i,col);
      } else
#endif
      {
	col[0] = ((pic_tm.color[i].r * 1000L) / 255L) * j / STEPS;
	col[1] = ((pic_tm.color[i].g * 1000L) / 255L) * j / STEPS;
	col[2] = ((pic_tm.color[i].b * 1000L) / 255L) * j / STEPS;
	set_color(i,col);
      }
    }
    VSYNC;
  }
}

/* kopiert die Statuszeile vom aktuellen in den Backup-Puffer */
tm_void
update_double_buffer(tm_void) {
  int i;
  
  /* nur wenn der Treiber wirklich doppelt puffert */
  if(double_buffer) {
    for(i=1;i<=double_cnt;i++)
      memcpy( video + (((double_page+i)%(double_cnt+1)) * height + pf_height) * width,
	      video + (  double_page                    * height + pf_height) * width,
	      width*(height-pf_height));
  }
}

void close_screen(void)
{
  __dpmi_regs regs;
  int   tmp_cols[256][3],i,m;

  free(offscreen_buffer);

  /* fade to black */
  outp(DAC_CPM,0xff);

  /* originale Farben lesen */
  outp(DAC_CPRI,0);
  for(i=0;i<256;i++) {
    tmp_cols[i][0]=inp(DAC_CPD);
    tmp_cols[i][1]=inp(DAC_CPD);
    tmp_cols[i][2]=inp(DAC_CPD);
  }
  
  for(m=STEPS;m>=0;m--) {
    outp(DAC_CPWI,0);
    for(i=0;i<256;i++) {
      outp(DAC_CPD,tmp_cols[i][0]*m/STEPS);
      outp(DAC_CPD,tmp_cols[i][1]*m/STEPS);
      outp(DAC_CPD,tmp_cols[i][2]*m/STEPS);
    }
    VSYNC;
  }

#ifdef THREE_D
  if(use_3d == DEV_3DMAX) {
    regs.x.ax = 0x4fd0;
    regs.x.bx = 0x3d02;   /* switch glasses off */
    __dpmi_int(VIDEO, &regs);
    regs.x.ax = 0x4fd0;
    regs.x.bx = 0x3d00;   /* switch interlace off */
    __dpmi_int(VIDEO, &regs);
  }
#endif

  if(!double_buffer)
    free((void*)screen_base);

  regs.h.ah = 0x00;
  regs.h.al = video_mode;      /* alter Videomodus */
  regs.x.bx = 0x00;
  __dpmi_int(VIDEO, &regs);

  /* all colors black */
  outp(DAC_CPWI,0);
  for(i=0;i<768;i++)
    outp(DAC_CPD,0);

  /* restore text mode contents */
  memcpy(vtext, orig_text, TEXTSIZE);

  /* restore cursor */
  regs.h.ah = 0x02;
  regs.h.bh = 0x00;
  regs.x.dx = orig_cursor;
  __dpmi_int(VIDEO, &regs);

  /* give vga abd crt some seconds */
  sleep(SLEEP);

  for(m=0;m<=STEPS;m++) {
    outp(DAC_CPWI,0);
    for(i=0;i<256;i++) {
      outp(DAC_CPD,orig_cols[i][0]*m/STEPS);
      outp(DAC_CPD,orig_cols[i][1]*m/STEPS);
      outp(DAC_CPD,orig_cols[i][2]*m/STEPS);
    }
    VSYNC;
  }
  __djgpp_nearptr_disable();
}

void set_color(tm_uword col, tm_uword prom[3])
{
  outp(DAC_CPWI,col);
  outp(DAC_CPD,prom[0]*16/253);
  outp(DAC_CPD,prom[1]*16/253);
  outp(DAC_CPD,prom[2]*16/253);
}

void  save_colors(void)
{
}

void restore_colors(void)
{
}

tm_void 
copy_back_screen(tm_void) {
}

#define RAW256
#include "gr_prim.c"

