/*
   drawmaze.c

   09.03.97  Chip   aus maze.c extrahiert
   23.11.97  Chip   Smileys erstmal primitiv
   10.09.97  Till   Shading
   21.12.97  Till   Texture-Mapping
*/

#include "ctypes.h"
#include "mazes.h"
#include "maze.h"
#include "drawmaze.h"
#include "gr_prim.h"
#include "grafik.h"
#include "move.h"
#include "smiley.h"
#include "ninja.h"
#include "texture.h"
#include "machine.h"
#include "net.h"

#include "console.h"  /* spaeter raus */
#include "timer.h"
#include "misc.h"
#include <stdio.h>

#ifdef THREE_D
#include <mem.h>
#endif

#define MAX_SMILEY_PER_FIELD  (9)

#define SCR_XOFFS  (width/2)
#define SCR_YOFFS  (pf_height/2)

/* Das Fadenkreuz */
#define CROSS      (4)

#define  X_VISIBLE  (OB_VISIBLE<<OB_X_SHIFT)
#define  Y_VISIBLE  (OB_VISIBLE<<OB_Y_SHIFT)


/* --- maze-koordinaten vorbereiten --- */

static tm_long  xx[MAX_MAZE],xz[MAX_MAZE],yx[MAX_MAZE],yz[MAX_MAZE];
static tm_long  sin,cos;

static tm_void trans_maze(tm_void)  /* x- und y-werte vorbereiten */
{
  tm_uword i;
  tm_word  d;

  sin=sintab[player[my_id].rot_z>>4];
  cos=costab[player[my_id].rot_z>>4];

  for (i=0, d=-player[my_id].x; i<akt_maze->x_size; i++, d+=FIELD)
  {
    xx[i]=+d*cos;
    xz[i]=+d*sin;
  }
  for (i=0, d=-player[my_id].y; i<akt_maze->y_size; i++, d+=FIELD)
  {
    yx[i]=-d*sin;
    yz[i]=+d*cos;
  }
}


/* --- zeichenstack fuer waende und figuren  --- */

#define STACK_END    (0)
#define STACK_WALL2  (1)
#define STACK_WALL4  (2)
#define STACK_WALL6  (3)
#define STACK_SMILEY (4)
#define STACK_WALLTX (5)

#define STACK_LEN    (MAX_REZ_WIDTH+MAX_PLAYERS)

#ifdef TEXTURE
#define STACK_ELEM_MAX (10)  /* maximale groesse eines eintrags */
#else
#define STACK_ELEM_MAX (8)
#endif

static tm_word stack_buffer[STACK_LEN*STACK_ELEM_MAX];
static tm_word *stack;

#ifdef TEXTURE
#define stack_wall4tex(x1,h1,x2,h2,wall,xa,za,xb,zb)  { \
  *stack++=zb;  *stack++=xb;                            \
  *stack++=za;  *stack++=xa;                            \
  *stack++=wall;                                        \
  *stack++=h2;  *stack++=x2;                            \
  *stack++=h1;  *stack++=x1;                            \
  *stack++=STACK_WALLTX;                                }
#endif

#define stack_wall2(x,h,col)  {   \
  *stack++=col;                   \
  *stack++=h;  *stack++=x;        \
  *stack++=STACK_WALL2;           }

#define stack_wall4(x1,h1,x2,h2,col)  {   \
  *stack++=col;                           \
  *stack++=h2;  *stack++=x2;              \
  *stack++=h1;  *stack++=x1;              \
  *stack++=STACK_WALL4;                   }

#define stack_wall6(x1,h1,x2,h2,x3,h3,col)  {   \
  *stack++=col;                                 \
  *stack++=h3;  *stack++=x3;                    \
  *stack++=h2;  *stack++=x2;                    \
  *stack++=h1;  *stack++=x1;                    \
  *stack++=STACK_WALL6;                         }

#define stack_smiley(x,h,rot,id) { \
  *stack++=id; *stack++=rot;       \
  *stack++=h;  *stack++=x;         \
  *stack++=STACK_SMILEY;           }

static
tm_void init_stack(tm_void)
{
  stack=stack_buffer;
  *stack++=STACK_END;
}

static
tm_void draw_stack(tm_void)
{
  tm_word x1,h1,x2,h2,x3,h3,col_id;
#ifdef TEXTURE
  tm_word xa,za,xb,zb;
#endif

  for (;;)
  {
    switch (*--stack)
    {
    case STACK_END:
      return;
#ifdef TEXTURE
    case STACK_WALLTX:
      x1=*--stack;
      h1=*--stack;
      x2=*--stack;
      h2=*--stack;
      col_id=*--stack;
      xa=*--stack;
      za=*--stack;
      xb=*--stack;
      zb=*--stack;
      wall_4tex(x1,h1,x2,h2,col_id,xa,za,xb,zb);
      break;
#endif
    case STACK_WALL2:
      x1=*--stack;
      h1=*--stack;
      col_id=*--stack;
      wall_2(x1,h1,col_id,wall_border);
      break;
    case STACK_WALL4:
      x1=*--stack;
      h1=*--stack;
      x2=*--stack;
      h2=*--stack;
      col_id=*--stack;
      wall_4(x1,h1,x2,h2,col_id,wall_border);
      break;
    case STACK_WALL6:
      x1=*--stack;
      h1=*--stack;
      x2=*--stack;
      h2=*--stack;
      x3=*--stack;
      h3=*--stack;
      col_id=*--stack;
      wall_6(x1,h1,x2,h2,x3,h3,col_id,wall_border);
      break;
    case STACK_SMILEY:
/* noch nicht fertig */
      x1=*--stack;
      h1=*--stack;
      x3=*--stack;
      col_id=*--stack;
      draw_smiley(x1,h1,x3,col_id);

#if 0
      x2=x1+h1;
      x1-=h1;
      if (x1<0)            x1=0;
      if (x1>=width)       x1=width-1;
      if (x2<0)            x2=0;
      if (x2>=width)       x2=width-1;
      if (h1>=pf_height/2) h1=pf_height/2-1;

      /* ist auf jeden fall mglich, weil wir nach auen verzerren */
      draw_rect(x1,SCR_YOFFS-h1,x2-x1+1,h1+h1+1,15);
      draw_vline(x1,SCR_YOFFS-h1,h1+h1+1,BW_WHITE);
      draw_vline(x2,SCR_YOFFS-h1,h1+h1+1,BW_WHITE);
      draw_hline(x1,SCR_YOFFS-h1,x2-x1+1,BW_WHITE);
      draw_hline(x1,SCR_YOFFS+h1,x2-x1+1,BW_WHITE);
#endif
      break;
    }
  }
}


/* --- clipping --- */
/* verkettete liste freier bereiche */

static tm_word  c_pix[MAX_REZ_WIDTH+1];
static tm_word  c_p0,c_p1,c_p2;
static tm_bool  c_finished;

static tm_word  c_fact_x,c_fact_y,c_height;

static tm_void c_init(tm_void)
{
  c_pix[0]=width-1;
  c_pix[width-1]=-1;
  c_pix[width]=0;

  c_finished=FALSE;

  c_fact_x=SCR_XOFFS;
  c_fact_y=SCR_YOFFS-1;
  c_height=(tm_word)(((tm_long)WALL_HEIGHT*c_fact_x)/c_fact_y/2);
}

static tm_word c_findfirst(tm_word x)
{
  if (x<0)
    x=0;
  else if (x>=width-1)
    return(-1);

  c_p0=width;
  for (;;) {
    c_p1=c_pix[c_p0];
    if (c_p1<0)
      return(-1);
    c_p2=c_pix[c_p1];
    if (c_p2>x)          /* c_p0 <= belegt < c_p1 <= frei < c_p2 */
      break;
    c_p0=c_p2;
    }
  if (c_p1<x)
    return(x);
  else
    return(c_p1);
}

static tm_word c_findnext(tm_word x)
{
  if (x!=c_p1)
    c_p0=x;
  c_p1=c_pix[c_p0];
  if (c_p1>=0)
    c_p2=c_pix[c_p1];
  return(c_p1);
}

static tm_word c_use(tm_word x1,tm_word x2)
{
  if (x2+1>=c_p2)                /* bis ans ende */
  {
    x2=c_p2-1;
    if (x1==c_p1)
      c_pix[c_p0]=c_pix[c_p2];   /* freien bereich loeschen */
    else
    {
      c_pix[c_p1]=x1;
      c_pix[x1]=c_pix[c_p2];     /* freien bereich von rechts kuerzen */
    }
  }
  else
  {
    if (x1==c_p1)
    {
      c_pix[c_p0]=x2+1;
      c_pix[x2+1]=c_p2;          /* freien bereich von links kuerzen */
    }
    else
    {
      c_pix[c_p1]=x1;
      c_pix[x1]=x2+1;
      c_pix[x2+1]=c_p2;          /* belegten bereich einfuegen */
    }
  }
  if (c_pix[width]<0)
    c_finished=TRUE;
  return(x2);
}


/* --- darstellung --- */

tm_uword wall_color[OB_MASK+1];
tm_uword wall_border,heaven_color,floor_color;

#ifdef TEXTURE

static tm_void draw_wall_4tex(tm_word x1, tm_word h1,
                           tm_word x2, tm_word h2,
                           tm_uword wall,
			   tm_word xa, tm_word za, tm_word xb, tm_word zb)
{
  tm_word  xl,xr,hl,hr;

  x1+=SCR_XOFFS;
  x2+=SCR_XOFFS;

  xl=c_findfirst(x1);

  while (xl>=0 && xl<x2) {
    if (xl==x1)
      hl=h1;
    else
      hl=(tm_word)(h1 + ((tm_long)(h2-h1)*(xl-x1))/(x2-x1));

    xr=c_use(xl,x2-1);
    if (xr==x2-1)
      hr=h2;
    else
      hr=(tm_word)(h1 + ((tm_long)(h2-h1)*(xr+1-x1))/(x2-x1));

    stack_wall4tex(xl,hl,xr+1,hr,wall,xa,za,xb,zb);

    if (xr==x2-1)
      return;

    xl=c_findnext(xl);
  }
}
#endif

static tm_void draw_wall_2(tm_word x,tm_word h,
                           tm_uword wall)
{
  x+=SCR_XOFFS;

  if (c_findfirst(x)!=x)
    return;

  (tm_void)c_use(x,x);
  stack_wall2(x,h,wall_color[wall]);
}

static tm_void draw_wall_4(tm_word x1,tm_word h1,
                           tm_word x2,tm_word h2,
                           tm_uword wall)
{
  tm_word  xl,xr,hl,hr;

  x1+=SCR_XOFFS;
  x2+=SCR_XOFFS;

  xl=c_findfirst(x1);

  while (xl>=0 && xl<x2) {
    if (xl==x1)
      hl=h1;
    else
      hl=(tm_word)(h1 + ((tm_long)(h2-h1)*(xl-x1))/(x2-x1));

    xr=c_use(xl,x2-1);
    if (xr==x2-1)
      hr=h2;
    else
      hr=(tm_word)(h1 + ((tm_long)(h2-h1)*(xr+1-x1))/(x2-x1));

    stack_wall4(xl,hl,xr+1,hr,wall_color[wall]);
    if (xr==x2-1)
      return;
    xl=c_findnext(xl);
  }
}

static tm_void draw_wall_6(tm_word x1,tm_word h1,
                           tm_word x2,tm_word h2,
                           tm_word x3,tm_word h3,
                           tm_uword wall)
{
  tm_word xl,xr,hl,hr;

  x1+=SCR_XOFFS;
  x2+=SCR_XOFFS;
  x3+=SCR_XOFFS;

  xl=c_findfirst(x1);

  while (xl>=0 && xl<x2) {
    if (xl==x1)
      hl=h1;
    else
      hl=(tm_word)(h1 + ((tm_long)(h2-h1)*(xl-x1))/(x2-x1));

    xr=c_use(xl,x3-1);
    if (xr<x2-1)
    {
      if (xr==x2-1)
        hr=h2;
      else
        hr=(tm_word)(h1 + ((tm_long)(h2-h1)*(xr+1-x1))/(x2-x1));
      stack_wall4(xl,hl,xr+1,hr,wall_color[wall]);
    }
    else
    {
      if (xr==x3-1)
        hr=h3;
      else
        hr=(tm_word)(h2 + ((tm_long)(h3-h2)*(xr+1-x2))/(x3-x2));
      stack_wall6(xl,hl,x2,h2,xr+1,hr,wall_color[wall]);
    }

    if (xr==x3-1)
      return;
    xl=c_findnext(xl);
  }

  while (xl>=0 && xl<x3) {
    if (xl==x2)
      hl=h2;
    else
      hl=(tm_word)(h2 + ((tm_long)(h3-h2)*(xl-x2))/(x3-x2));
    xr=c_use(xl,x3-1);
    if (xr==x3-1)
      hr=h3;
    else
      hr=(tm_word)(h2 + ((tm_long)(h3-h2)*(xr+1-x2))/(x3-x2));
    stack_wall4(xl,hl,xr+1,hr,wall_color[wall]);

    if (xr==x3-1)
      return;
    xl=c_findnext(xl);
  }
}

/* hier: DIE Haupt-Clipping-Funktion... */

tm_void draw_wall(tm_word x1,tm_word z1,tm_word x2,tm_word z2,
                  tm_uword wall)
{
/* nur von links nach rechts malen, wird durch aufruf sichergestellt */
  tm_word  sx1,sy1,sx2,sy2,sx3,sy3;

#ifdef TEXTURE
  /* orig. Werte merken */
  tm_word xa,za,xb,zb;
  xa=x1; za=z1; xb=x2; zb=z2; 
#endif

  /* links clippen */
  if (x1<-z1)
  {
    if (x2<-z2)
      return;
    x1-=(tm_word)((tm_long)(x2-x1)*(x1+z1)/(x2-x1+z2-z1));
    z1=-x1;
  }
  else if (x2<-z2)
  {
    x2-=(tm_word)((tm_long)(x1-x2)*(x2+z2)/(x1-x2+z1-z2));
    z2=-x2;
  }

  /* rechts clippen */
  if (x1>z1)
  {
    if (x2>z2)
      return;
    x1-=(tm_word)((tm_long)(x2-x1)*(x1-z1)/(x2-x1-z2+z1));
    z1=x1;
  }
  else if (x2>z2)
  {
    x2-=(tm_word)((tm_long)(x1-x2)*(x2-z2)/(x1-x2-z1+z2));
    z2=x2;
  }

#ifdef TEXTURE
  if(use_texture) {
    if((z1 !=0)&&(z2!=0)) {
      /* texture wird vertikal woanders geclippt */
      sx1=(tm_word)((tm_long)c_fact_x*x1/z1);
      sy1=(tm_word)((tm_long)c_fact_y*c_height/z1);
      sx2=(tm_word)((tm_long)c_fact_x*x2/z2);
      sy2=(tm_word)((tm_long)c_fact_y*c_height/z2);
      draw_wall_4tex(sx1,sy1,sx2,sy2,wall,xa,za,xb,zb);
    }
    return;
  }
#endif

  /* oben/unten clippen */
  if (z1<c_height)
  {
    if (z1<=0)
      {
      if (z2<=0)
        return;
      sx1=-c_fact_x;
      }
    else
      sx1=(tm_word)((tm_long)c_fact_x*x1/z1);
    sy1=c_fact_y;
    if (z2<c_height)
    {
      if (z2<=0)
        sx2=c_fact_x;
      else
        sx2=(tm_word)((tm_long)c_fact_x*x2/z2);
      sy2=c_fact_y;
      if (sx1<sx2)
        draw_wall_4(sx1,sy1,sx2,sy2,wall);
      else
        draw_wall_2(sx1,sy1,wall);
    }
    else  /* z2>=c_height */
    {
      sx3=x1-(tm_word)((tm_long)(x1-x2)*(c_height-z1)/(z2-z1));
      sx2=(tm_word)((tm_long)c_fact_x*sx3/c_height);
      sy2=c_fact_y;
      sx3=(tm_word)((tm_long)c_fact_x*x2/z2);
      sy3=(tm_word)((tm_long)c_fact_y*c_height/z2);
      if (sx1<sx2)
      {
        if (sx2<sx3)
          draw_wall_6(sx1,sy1,sx2,sy2,sx3,sy3,wall);
        else
          draw_wall_4(sx1,sy1,sx3,sy2,wall);
      }
      else if (sx1<sx3)
        draw_wall_4(sx1,sy1,sx3,sy3,wall);
      else
        draw_wall_2(sx1,sy1,wall);
    }
  }
  else  /* z1>=c_height */
  {
    sx1=(tm_word)((tm_long)c_fact_x*x1/z1);
    sy1=(tm_word)((tm_long)c_fact_y*c_height/z1);
    if (z2<c_height)
    {
      sx3=x2-(tm_word)((tm_long)(x2-x1)*(c_height-z2)/(z1-z2));
      sx2=(tm_word)((tm_long)c_fact_x*sx3/c_height);
      sy2=c_fact_y;
      if (z2<=0)
        sx3=c_fact_x;
      else
        sx3=(tm_word)((tm_long)c_fact_x*x2/z2);
      sy3=c_fact_y;
      if (sx1<sx2)
      {
        if (sx2<sx3)
          draw_wall_6(sx1,sy1,sx2,sy2,sx3,sy3,wall);
        else
          draw_wall_4(sx1,sy1,sx3,sy3,wall);
      }
      else if (sx1<sx3)
        draw_wall_4(sx1,sy2,sx3,sy3,wall);
      else
        draw_wall_2(sx1,sy3,wall);
    }
    else  /* z2>=c_height */
    {
      sx2=(tm_word)((tm_long)c_fact_x*x2/z2);
      sy2=(tm_word)((tm_long)c_fact_y*c_height/z2);
      if (sx1<sx2)
        draw_wall_4(sx1,sy1,sx2,sy2,wall);
      else if (sy1<sy2)
        draw_wall_2(sx1,sy2,wall);
      else
        draw_wall_2(sx1,sy1,wall);
    }
  }

  /* hinten clippen ist nicht noetig */
}


/* --- smileys merken --- */

typedef struct
{
  tm_word  fx1,fx2,fy1,fy2;
  tm_word  dx,dy;
}  SMILEY_AID;

static
SMILEY_AID  smiley[MAX_PLAYERS];

static
tm_void init_smileys(tm_void)
{
  tm_uword     i;
  tm_word      x0,y0;
  SMILEY_AID   *sm;
  PLAYER_DATA  *pl;

  x0=player[my_id].x;
  y0=player[my_id].y;
  for (pl=player, sm=smiley, i=0; i<players; i++, sm++, pl++)
  {
    sm->fx1=(pl->x-SMILEY_RAD+1)>>FIELD_SH;
    sm->fx2=(pl->x+SMILEY_RAD-1)>>FIELD_SH;
#if 0
    if (sm->fx1!=sm->fx2)
      if (akt_maze->data + sm->y1*akt_maze->x_size + sm->fx2)&OB_VISIBLE)
        sm->fx2=-1;
#endif
    sm->fy1=(pl->y-SMILEY_RAD+1)>>FIELD_SH;
    sm->fy2=(pl->y+SMILEY_RAD-1)>>FIELD_SH;
    sm->dx=pl->x-x0;
    sm->dy=pl->y-y0;
  }
  smiley[my_id].fx1=-1;
  smiley[my_id].fx2=-1;
}

#define VISIB_OFFS  ((141*SMILEY_RAD+50)/100)

static
tm_void check_smiley(tm_word x,tm_word y)
{
  tm_uword    i,j,k;
  SMILEY_AID  *sm;
  tm_uword    xf,found[MAX_SMILEY_PER_FIELD];
  tm_word     px,ph,fph[MAX_SMILEY_PER_FIELD],korr;

  j=0;
  for (sm=smiley, i=0; i<players; i++, sm++)
    if ((sm->fx1==x || sm->fx2==x) && (sm->fy1==y || sm->fy2==y))
    {
      found[j]=i;
      fph[j]=(tm_word)((sm->dx*sin + sm->dy*cos)>>15);
      j++;
    }

  for (i=1; i<j; i++)
    for (k=i; k>0; k--)
      if (fph[k]<fph[k-1])
      {
        ph=fph[k];   fph[k]=fph[k-1];     fph[k-1]=ph;
        xf=found[k]; found[k]=found[k-1]; found[k-1]=xf;
      }

  for (i=0; i<j; i++)
  {
    k=found[i];
    sm=smiley+k;
    sm->fx1=-1;
    sm->fx2=-1;
    px=(tm_word)((sm->dx*cos - sm->dy*sin)>>15);
    ph=fph[i];

    if (ph+VISIB_OFFS<px || ph+VISIB_OFFS<-px)
      continue;

    korr = 128+64*px/ph;
    px=(tm_word)((tm_long)c_fact_x*px/ph)+SCR_XOFFS;
    ph=(tm_word)((tm_long)c_fact_y*c_height/ph);
    if (px+ph/2>=c_findfirst(px-ph/2)) {
      if(korr<0)   korr=0;
      if(korr>255) korr=255;
      korr = atantab[korr];
      stack_smiley(px,ph,player[my_id].rot_z+korr-player[k].rot_z,player[k].col);
    }
  }
}


/* --- waende und smileys in der richtigen reihenfolge malen --- */

#define m_show_wall_xlr(d,x,y) {                                    \
  if ((d) & X_VISIBLE)                                              \
    {                                                               \
    draw_wall((xx[(x)  ]+yx[(y)  ])>>15,(xz[(x)  ]+yz[(y)  ])>>15,  \
              (xx[(x)+1]+yx[(y)  ])>>15,(xz[(x)+1]+yz[(y)  ])>>15,  \
              ((d)>>OB_X_SHIFT)&OB_MASK);                           \
    if (c_finished)                                                 \
      return;                                                       \
    }                                                               }

#define m_show_wall_xrl(d,x,y) {                                    \
  if ((d) & X_VISIBLE)                                              \
    {                                                               \
    draw_wall((xx[(x)+1]+yx[(y)  ])>>15,(xz[(x)+1]+yz[(y)  ])>>15,  \
              (xx[(x)  ]+yx[(y)  ])>>15,(xz[(x)  ]+yz[(y)  ])>>15,  \
              ((d)>>OB_X_SHIFT)&OB_MASK);                           \
    if (c_finished)                                                 \
      return;                                                       \
    }                                                               }

#define m_show_wall_ylr(d,x,y) {                                    \
  if ((d) & Y_VISIBLE)                                              \
    {                                                               \
    draw_wall((xx[(x)  ]+yx[(y)  ])>>15,(xz[(x)  ]+yz[(y)  ])>>15,  \
              (xx[(x)  ]+yx[(y)+1])>>15,(xz[(x)  ]+yz[(y)+1])>>15,  \
              ((d)>>OB_Y_SHIFT)&OB_MASK);                           \
    if (c_finished)                                                 \
      return;                                                       \
    }                                                               }

#define m_show_wall_yrl(d,x,y) {                                    \
  if ((d) & Y_VISIBLE)                                              \
    {                                                               \
    draw_wall((xx[(x)  ]+yx[(y)+1])>>15,(xz[(x)  ]+yz[(y)+1])>>15,  \
              (xx[(x)  ]+yx[(y)  ])>>15,(xz[(x)  ]+yz[(y)  ])>>15,  \
              ((d)>>OB_Y_SHIFT)&OB_MASK);                           \
    if (c_finished)                                                 \
      return;                                                       \
    }                                                               }

/* for debug */
#define M_SHOW_MASK  (0xFF)  /* Bit 7  6  5  4  3  2  1  0  */
                             /*     tl lt lb bl br rb rt tr */

static
tm_void m_show_tr(tm_void)  /* top right = 45..90 deg */
{
#if M_SHOW_MASK&0x01
  register tm_word  x,y;
  tm_word  ex,ey;
  tm_word  xl,xu,yu,m;
  register tm_uchar *p;
  tm_uchar *px,*py;

  ex=player[my_id].x>>FIELD_SH;
  ey=player[my_id].y>>FIELD_SH;
  px=py=akt_maze->data + ey*akt_maze->x_size + ex;

  check_smiley(ex,ey);
  m_show_wall_xrl(*py,ex,ey);
  m_show_wall_ylr(*px,ex,ey);
  xl=ex-1;
  xu=ex+1;
  yu=ey+1;
  px+=1;
  py+=akt_maze->x_size;
  m_show_wall_xlr(*py,ex,yu);
  m_show_wall_yrl(*px,xu,ey);

  for (;;)  /* immer groesserer abstand */
  {
    if (xl<0 && xu>=akt_maze->x_size && yu>=akt_maze->y_size)
      return;  /* nur wenn maze offen, sonst vorher c_finished */

    if (xu<akt_maze->x_size-1)  /* felder rechts */
    {
      check_smiley(xu,ey);
      m_show_wall_xrl(*px,xu,ey);
      if ((m=yu)>=akt_maze->y_size)
        m=akt_maze->y_size-1;
      for (p=px+akt_maze->x_size, y=ey+1; y<m; y++, p+=akt_maze->x_size)
      {
        m_show_wall_xlr(*p,xu,y);
        check_smiley(xu,y);
      }
    }

    if (yu<akt_maze->y_size-1)  /* felder oben */
    {
      if ((m=xl)<-1)  /* links */
        m=-1;
      for (p=py, x=ex; x>m; x--, p-=1)
      {
        check_smiley(x,yu);
        m_show_wall_ylr(*p,x,yu);
      }

      if (xl>=0)  /* spitze */
      {
        check_smiley(x,yu-1);
        m_show_wall_xlr(*p,x,yu);
        check_smiley(x,yu);
        m_show_wall_ylr(*p,x,yu);
      }

      if ((m=xu)>=akt_maze->x_size)  /* rechts */
        m=akt_maze->x_size-1;
      for (p=py+1, x=ex+1; x<m; x++, p+=1)
      {
        m_show_wall_yrl(*p,x,yu);
        check_smiley(x,yu);
      }

      if (xu<akt_maze->x_size-1)  /* ecke */
      {
        m_show_wall_xlr(*p,x,yu);
        m_show_wall_yrl(*p,x,yu);
        check_smiley(x,yu);
        x++;
        p+=1;
      }
      m_show_wall_yrl(*p,x,yu);
    }

    else
    {
      if (yu<akt_maze->y_size && xl>=0)  /* feld spitze */
      {
        p=py+xl-ex;
        check_smiley(xl,yu-1);
        m_show_wall_xlr(*p,xl,yu);
      }
      if (xu<akt_maze->x_size-1)  /* wand ecke */
      {
        p=px+(akt_maze->y_size-1-ey)*(tm_word)(akt_maze->x_size);
        m_show_wall_xlr(*p,xu,akt_maze->y_size-1);
      }
    }

    xl--;
    xu++;
    px+=1;

    if (xu<akt_maze->x_size)  /* rand rechts */
    {
      if ((m=yu)>=akt_maze->y_size)
        m=akt_maze->y_size-1;
      for (p=px, y=ey; y<m; y++, p+=akt_maze->x_size)
      {
        m_show_wall_yrl(*p,xu,y);
      }
    }

    yu++;
    py+=akt_maze->x_size;

    if (yu<akt_maze->y_size)  /* rand oben */
    {
      if ((m=xl)<-1)
        m=-1;
      for (p=py, x=ex; x>m; x--, p-=1)
      {
        m_show_wall_xlr(*p,x,yu);
      }

      if ((m=xu)>=akt_maze->x_size)
        m=akt_maze->x_size-1;
      for (p=py+1, x=ex+1; x<m; x++, p+=1)
      {
        m_show_wall_xlr(*p,x,yu);
      }
    }

  }  /* for (d) */
#endif
}

static
tm_void m_show_rt(tm_void)  /* right top = 0..45 deg */
{
#if M_SHOW_MASK&0x02
  register tm_word  x,y;
  tm_word  ex,ey;
  tm_word  xu,yl,yu,m;
  register tm_uchar *p;
  tm_uchar *px,*py;

  ex=player[my_id].x>>FIELD_SH;
  ey=player[my_id].y>>FIELD_SH;
  px=py=akt_maze->data + ey*akt_maze->x_size + ex;

  check_smiley(ex,ey);
  m_show_wall_xrl(*py,ex,ey);
  m_show_wall_ylr(*px,ex,ey);
  xu=ex+1;
  yl=ey-1;
  yu=ey+1;
  px+=1;
  py+=akt_maze->x_size;
  m_show_wall_xlr(*py,ex,yu);
  m_show_wall_yrl(*px,xu,ey);

  for (;;)  /* immer groesserer abstand */
  {
    if (xu>=akt_maze->x_size && yl<0 && yu>=akt_maze->y_size)
      return;  /* nur wenn maze offen, sonst vorher c_finished */

    if (yu<akt_maze->y_size-1)  /* felder oben */
    {
      check_smiley(ex,yu);
      m_show_wall_ylr(*py,ex,yu);
      if ((m=xu)>=akt_maze->x_size)
        m=akt_maze->x_size-1;
      for (p=py+1, x=ex+1; x<m; x++, p+=1)
      {
        m_show_wall_yrl(*p,x,yu);
        check_smiley(x,yu);
      }
    }

    if (xu<akt_maze->x_size-1)  /* felder rechts */
    {
      if ((m=yl)<-1)  /* unten */
        m=-1;
      for (p=px, y=ey; y>m; y--, p-=akt_maze->x_size)
      {
        check_smiley(xu,y);
        m_show_wall_xrl(*p,xu,y);
      }

      if (yl>=0)  /* spitze */
      {
        check_smiley(xu-1,y);
        m_show_wall_yrl(*p,xu,y);
        check_smiley(xu,y);
        m_show_wall_xrl(*p,xu,y);
      }

      if ((m=yu)>=akt_maze->y_size)  /* oben */
        m=akt_maze->y_size-1;
      for (p=px+akt_maze->x_size, y=ey+1; y<m; y++, p+=akt_maze->x_size)
      {
        m_show_wall_xlr(*p,xu,y);
        check_smiley(xu,y);
      }

      if (yu<akt_maze->y_size-1)  /* ecke */
      {
        m_show_wall_yrl(*p,xu,y);
        m_show_wall_xlr(*p,xu,y);
        check_smiley(xu,y);
        y++;
        p+=akt_maze->x_size;
      }
      m_show_wall_xlr(*p,xu,y);
    }

    else
    {
      if (xu<akt_maze->x_size && yl>=0)  /* feld spitze */
      {
        p=px+(yl-ey)*(tm_word)(akt_maze->x_size);
        check_smiley(xu-1,yl);
        m_show_wall_yrl(*p,xu,yl);
      }
      if (yu<akt_maze->y_size-1)  /* wand ecke */
      {
        p=py+akt_maze->x_size-1-ex;
        m_show_wall_yrl(*p,akt_maze->x_size-1,yu);
      }
    }

    yl--;
    yu++;
    py+=akt_maze->x_size;

    if (yu<akt_maze->y_size)  /* rand oben */
    {
      if ((m=xu)>=akt_maze->x_size)
        m=akt_maze->x_size-1;
      for (p=py, x=ex; x<m; x++, p+=1)
      {
        m_show_wall_xlr(*p,x,yu);
      }
    }

    xu++;
    px+=1;

    if (xu<akt_maze->x_size)  /* rand rechts */
    {
      if ((m=yl)<-1)
        m=-1;
      for (p=px, y=ey; y>m; y--, p-=akt_maze->x_size)
      {
        m_show_wall_yrl(*p,xu,y);
      }

      if ((m=yu)>=akt_maze->y_size)
        m=akt_maze->y_size-1;
      for (p=px+akt_maze->x_size, y=ey+1; y<m; y++, p+=akt_maze->x_size)
      {
        m_show_wall_yrl(*p,xu,y);
      }
    }

  }  /* for (d) */
#endif
}

static
tm_void m_show_rb(tm_void)  /* right bottom = 315..360 deg */
{
#if M_SHOW_MASK&0x04
  register tm_word  x,y;
  tm_word  ex,ey;
  tm_word  xu,yl,yu,m;
  register tm_uchar *p;
  tm_uchar *px,*py;

  ex=player[my_id].x>>FIELD_SH;
  ey=player[my_id].y>>FIELD_SH;
  px=py=akt_maze->data + ey*akt_maze->x_size + ex;

  check_smiley(ex,ey);
  m_show_wall_xrl(*py,ex,ey);
  m_show_wall_ylr(*px,ex,ey);
  xu=ex+1;
  yl=ey-1;
  yu=ey+1;
  m_show_wall_xlr(*(py+akt_maze->x_size),ex,yu);
  px+=1;
  py-=akt_maze->x_size;
  m_show_wall_yrl(*px,xu,ey);

  for (;;)  /* immer groesserer abstand */
  {
    if (xu>=akt_maze->x_size && yl<0 && yu>=akt_maze->y_size)
      return;  /* nur wenn maze offen, sonst vorher c_finished */

    if (yl>=0)  /* felder unten */
    {
      check_smiley(ex,yl);
      m_show_wall_ylr(*py,ex,yl);
      if ((m=xu)>=akt_maze->x_size)
        m=akt_maze->x_size-1;
      for (p=py+1, x=ex+1; x<m; x++, p+=1)
      {
        m_show_wall_yrl(*p,x,yl);
        check_smiley(x,yl);
      }
    }

    if (xu<akt_maze->x_size-1)  /* felder rechts */
    {
      if ((m=yl)<-1)  /* unten */
        m=-1;
      for (p=px, y=ey; y>m; y--, p-=akt_maze->x_size)
      {
        check_smiley(xu,y);
        m_show_wall_xrl(*p,xu,y);
      }

      if (yl>=0)  /* ecke */
      {
        m_show_wall_yrl(*p,xu,y);
        check_smiley(xu,y);
        m_show_wall_xrl(*p,xu,y);
      }

      if ((m=yu)>=akt_maze->y_size)  /* oben */
        m=akt_maze->y_size-1;
      for (p=px+akt_maze->x_size, y=ey+1; y<m; y++, p+=akt_maze->x_size)
      {
        m_show_wall_xlr(*p,xu,y);
        check_smiley(xu,y);
      }

      if (yu<akt_maze->y_size-1)  /* spitze */
      {
        check_smiley(xu-1,y);
        m_show_wall_yrl(*p,xu,y);
        m_show_wall_xlr(*p,xu,y);
        check_smiley(xu,y);
        y++;
        p+=akt_maze->x_size;
      }
      m_show_wall_xlr(*p,xu,y);
    }

    else
    {
      if (xu<akt_maze->x_size && yu<akt_maze->y_size-1)  /* feld spitze */
      {
        p=px+(yu-ey)*(tm_word)(akt_maze->x_size);
        check_smiley(xu-1,yu);
        m_show_wall_yrl(*p,xu,yu);
      }
      if (yl>=0)  /* wand ecke */
      {
        p=py+akt_maze->x_size-1-ex;
        m_show_wall_yrl(*p,akt_maze->x_size-1,yl);
      }
    }

    if (yl>=0)  /* rand unten */
    {
      if ((m=xu)>=akt_maze->x_size)
        m=akt_maze->x_size-1;
      for (p=py, x=ex; x<m; x++, p+=1)
      {
        m_show_wall_xrl(*p,x,yl);
      }
    }

    yl--;
    yu++;
    py-=akt_maze->x_size;

    xu++;
    px+=1;

    if (xu<akt_maze->x_size)  /* rand rechts */
    {
      if ((m=yl)<-1)
        m=-1;
      for (p=px, y=ey; y>m; y--, p-=akt_maze->x_size)
      {
        m_show_wall_yrl(*p,xu,y);
      }

      if ((m=yu)>=akt_maze->y_size)
        m=akt_maze->y_size-1;
      for (p=px+akt_maze->x_size, y=ey+1; y<m; y++, p+=akt_maze->x_size)
      {
        m_show_wall_yrl(*p,xu,y);
      }
    }

  }  /* for (d) */
#endif
}

static
tm_void m_show_br(tm_void)  /* bottom right = 270..315 deg */
{
#if M_SHOW_MASK&0x08
  register tm_word  x,y;
  tm_word  ex,ey;
  tm_word  xl,xu,yl,m;
  register tm_uchar *p;
  tm_uchar *px,*py;

  ex=player[my_id].x>>FIELD_SH;
  ey=player[my_id].y>>FIELD_SH;
  px=py=akt_maze->data + ey*akt_maze->x_size + ex;

  check_smiley(ex,ey);
  m_show_wall_xrl(*py,ex,ey);
  m_show_wall_ylr(*px,ex,ey);
  xl=ex-1;
  xu=ex+1;
  yl=ey-1;
  m_show_wall_xlr(*(py+akt_maze->x_size),ex,ey+1);
  px+=1;
  py-=akt_maze->x_size;
  m_show_wall_yrl(*px,xu,ey);

  for (;;)  /* immer groesserer abstand */
  {
    if (xl<0 && xu>=akt_maze->x_size && yl<0)
      return;  /* nur wenn maze offen, sonst vorher c_finished */

    if (yl>=0)  /* felder unten */
    {
      if ((m=xl)<-1)  /* links */
        m=-1;
      for (p=py, x=ex; x>m; x--, p-=1)
      {
        check_smiley(x,yl);
        m_show_wall_ylr(*p,x,yl);
      }

      if (xl>=0)  /* spitze */
      {
        check_smiley(x,yl+1);
        m_show_wall_xrl(*(p+akt_maze->x_size),x,yl+1);
        check_smiley(x,yl);
        m_show_wall_ylr(*p,x,yl);
      }

      if ((m=xu)>=akt_maze->x_size)  /* rechts */
        m=akt_maze->x_size-1;
      for (p=py+1, x=ex+1; x<m; x++, p+=1)
      {
        m_show_wall_yrl(*p,x,yl);
        check_smiley(x,yl);
      }
    }

    else
    {
      if (yl>-2 && xl>=0)  /* feld spitze */
      {
        p=py+xl-ex+akt_maze->x_size;
        y=yl+1;
        check_smiley(xl,y);
        m_show_wall_xrl(*p,xl,y);
      }
    }

    if (xu<akt_maze->x_size-1)  /* felder rechts */
    {
      check_smiley(xu,ey);
      m_show_wall_xlr(*(px+akt_maze->x_size),xu,ey+1);
      m_show_wall_xrl(*px,xu,ey);
      if ((m=yl)<-1)
        m=-1;
      for (p=px-akt_maze->x_size, y=ey-1; y>m; y--, p-=akt_maze->x_size)
      {
        check_smiley(xu,y);
        m_show_wall_xrl(*p,xu,y);
      }

      if (yl>=0)  /* ecke */
      {
        m_show_wall_yrl(*p,xu,y);
        check_smiley(xu,y);
        p+=1;
        m_show_wall_yrl(*p,xu+1,y);
      }
    }

    else
    {
      if (yl>=0)  /* wand ecke */
      {
        p=py+akt_maze->x_size-1-ex;
        m_show_wall_yrl(*p,akt_maze->x_size-1,yl);
      }
    }

    xl--;
    xu++;
    px+=1;

    if (xu<akt_maze->x_size)  /* rand rechts */
    {
      if ((m=yl)<-1)
        m=-1;
      for (p=px, y=ey; y>m; y--, p-=akt_maze->x_size)
      {
        m_show_wall_yrl(*p,xu,y);
      }
    }

    if (yl>=0)  /* rand unten */
    {
      if ((m=xl)<-1)
        m=-1;
      for (p=py, x=ex; x>m; x--, p-=1)
      {
        m_show_wall_xrl(*p,x,yl);
      }

      if ((m=xu)>=akt_maze->x_size)
        m=akt_maze->x_size-1;
      for (p=py+1, x=ex+1; x<m; x++, p+=1)
      {
        m_show_wall_xrl(*p,x,yl);
      }
    }

    yl--;
    py-=akt_maze->x_size;

  }  /* for (d) */
#endif
}

static
tm_void m_show_bl(tm_void)  /* bottom left = 225..270 deg */
{
#if M_SHOW_MASK&0x10
  register tm_word  x,y;
  tm_word  ex,ey;
  tm_word  xl,xu,yl,m;
  register tm_uchar *p;
  tm_uchar *px,*py;

  ex=player[my_id].x>>FIELD_SH;
  ey=player[my_id].y>>FIELD_SH;
  px=py=akt_maze->data + ey*akt_maze->x_size + ex;

  check_smiley(ex,ey);
  m_show_wall_xrl(*py,ex,ey);
  m_show_wall_ylr(*px,ex,ey);
  xl=ex-1;
  xu=ex+1;
  yl=ey-1;
  m_show_wall_xlr(*(py+akt_maze->x_size),ex,ey+1);
  m_show_wall_yrl(*(px+1),xu,ey);
  px-=1;
  py-=akt_maze->x_size;

  for (;;)  /* immer groesserer abstand */
  {
    if (xl<0 && xu>=akt_maze->x_size && yl<0)
      return;  /* nur wenn maze offen, sonst vorher c_finished */

    if (yl>=0)  /* felder unten */
    {
      if ((m=xl+1)<0)  /* links */
        m=0;
      for (p=py, x=ex; x>m; x--, p-=1)
      {
        check_smiley(x,yl);
        m_show_wall_ylr(*p,x,yl);
      }
      check_smiley(x,yl);

      if ((m=xu)>=akt_maze->x_size)  /* rechts */
        m=akt_maze->x_size-1;
      for (p=py+1, x=ex+1; x<m; x++, p+=1)
      {
        m_show_wall_yrl(*p,x,yl);
        check_smiley(x,yl);
      }

      if (xu<akt_maze->x_size-1)  /* spitze */
      {
        check_smiley(x,yl+1);
        m_show_wall_xrl(*(p+akt_maze->x_size),x,yl+1);
        m_show_wall_yrl(*p,x,yl);
        check_smiley(x,yl);
        x++;
        p+=1;
      }
      m_show_wall_yrl(*p,x,yl);
    }

    else
    {
      if (yl>-2 && xu<akt_maze->x_size-1)  /* feld spitze */
      {
        p=py+xu-ex+akt_maze->x_size;
        y=yl+1;
        check_smiley(xu,y);
        m_show_wall_xrl(*p,xu,y);
      }
    }

    if (xl>=0)  /* felder rechts */
    {
      check_smiley(xl,ey);
      m_show_wall_xlr(*(px+akt_maze->x_size),xl,ey+1);
      m_show_wall_xrl(*px,xl,ey);
      if ((m=yl)<-1)
        m=-1;
      for (p=px-akt_maze->x_size, y=ey-1; y>m; y--, p-=akt_maze->x_size)
      {
        check_smiley(xl,y);
        m_show_wall_xrl(*p,xl,y);
      }

      if (yl>=0)  /* ecke */
      {
        m_show_wall_ylr(*(p+1),xl+1,y);
        check_smiley(xl,y);
        m_show_wall_ylr(*p,xl,y);
      }
    }

    else
    {
      if (yl>=0)  /* wand ecke */
      {
        m_show_wall_ylr(*(py-ex),0,yl);
      }
    }

    if (xl>=0)  /* rand links */
    {
      if ((m=yl)<-1)
        m=-1;
      for (p=px, y=ey; y>m; y--, p-=akt_maze->x_size)
      {
        m_show_wall_ylr(*p,xl,y);
      }
    }

    xl--;
    xu++;
    px-=1;

    if (yl>=0)  /* rand unten */
    {
      if ((m=xl)<-1)
        m=-1;
      for (p=py, x=ex; x>m; x--, p-=1)
      {
        m_show_wall_xrl(*p,x,yl);
      }

      if ((m=xu)>=akt_maze->x_size)
        m=akt_maze->x_size-1;
      for (p=py+1, x=ex+1; x<m; x++, p+=1)
      {
        m_show_wall_xrl(*p,x,yl);
      }
    }

    yl--;
    py-=akt_maze->x_size;

  }  /* for (d) */
#endif
}

static
tm_void m_show_lb(tm_void)  /* left bottom = 180..225 deg */
{
#if M_SHOW_MASK&0x20
  register tm_word  x,y;
  tm_word  ex,ey;
  tm_word  xl,yl,yu,m;
  register tm_uchar *p;
  tm_uchar *px,*py;

  ex=player[my_id].x>>FIELD_SH;
  ey=player[my_id].y>>FIELD_SH;
  px=py=akt_maze->data + ey*akt_maze->x_size + ex;

  check_smiley(ex,ey);
  m_show_wall_xrl(*py,ex,ey);
  m_show_wall_ylr(*px,ex,ey);
  xl=ex-1;
  yl=ey-1;
  yu=ey+1;
  m_show_wall_xlr(*(py+akt_maze->x_size),ex,yu);
  m_show_wall_yrl(*(px+1),ex+1,ey);
  px-=1;
  py-=akt_maze->x_size;

  for (;;)  /* immer groesserer abstand */
  {
    if (xl<0 && yl<0 && yu>=akt_maze->y_size)
      return;  /* nur wenn maze offen, sonst vorher c_finished */

    if (xl>=0)  /* felder links */
    {
      if ((m=yl+1)<0)  /* unten */
        m=0;
      for (p=px, y=ey; y>m; y--, p-=akt_maze->x_size)
      {
        check_smiley(xl,y);
        m_show_wall_xrl(*p,xl,y);
      }
      check_smiley(xl,y);

      if ((m=yu)>=akt_maze->y_size)  /* oben */
        m=akt_maze->y_size-1;
      for (p=px+akt_maze->x_size, y=ey+1; y<m; y++, p+=akt_maze->x_size)
      {
        m_show_wall_xlr(*p,xl,y);
        check_smiley(xl,y);
      }

      if (yu<akt_maze->y_size-1)  /* spitze */
      {
        check_smiley(xl+1,y);
        m_show_wall_ylr(*(p+1),xl+1,y);
        m_show_wall_xlr(*p,xl,y);
        check_smiley(xl,y);
        y++;
        p+=akt_maze->x_size;
      }
      m_show_wall_xlr(*p,xl,y);
    }

    else
    {
      if (xl>-2 && yu<akt_maze->y_size-1)  /* feld spitze */
      {
        p=px+(yu-ey)*(tm_word)(akt_maze->x_size)+1;
        x=xl+1;
        check_smiley(x,yu);
        m_show_wall_ylr(*p,x,yu);
      }
    }

    if (yl>=0)  /* felder unten */
    {
      check_smiley(ex,yl);
      m_show_wall_yrl(*(py+1),ex+1,yl);
      m_show_wall_ylr(*py,ex,yl);
      if ((m=xl)<-1)
        m=-1;
      for (p=py-1, x=ex-1; x>m; x--, p-=1)
      {
        check_smiley(x,yl);
        m_show_wall_ylr(*p,x,yl);
      }

      if (xl>=0)  /* ecke */
      {
        m_show_wall_xrl(*(p+akt_maze->x_size),x,yl+1);
        check_smiley(x,yl);
        m_show_wall_xrl(*p,x,yl);
      }
    }

    else
    {
      if (xl>=0)  /* wand ecke */
      {
        p=px-ey*(tm_word)(akt_maze->x_size);
        m_show_wall_xrl(*p,xl,0);
      }
    }

    if (yl>=0)  /* rand unten */
    {
      if ((m=xl)<-1)
        m=-1;
      for (p=py, x=ex; x>m; x--, p-=1)
      {
        m_show_wall_xrl(*p,x,yl);
      }
    }

    yl--;
    yu++;
    py-=akt_maze->x_size;

    if (xl>=0)  /* rand links */
    {
      if ((m=yl)<-1)
        m=-1;
      for (p=px, y=ey; y>m; y--, p-=akt_maze->x_size)
      {
        m_show_wall_ylr(*p,xl,y);
      }

      if ((m=yu)>=akt_maze->y_size)
        m=akt_maze->y_size-1;
      for (p=px+akt_maze->x_size, y=ey+1; y<m; y++, p+=akt_maze->x_size)
      {
        m_show_wall_ylr(*p,xl,y);
      }
    }

    xl--;
    px-=1;

  }  /* for (d) */
#endif
}

static
tm_void m_show_lt(tm_void)  /* left top = 135..180 deg */
{
#if M_SHOW_MASK&0x40
  register tm_word  x,y;
  tm_word  ex,ey;
  tm_word  xl,yl,yu,m;
  register tm_uchar *p;
  tm_uchar *px,*py;

  ex=player[my_id].x>>FIELD_SH;
  ey=player[my_id].y>>FIELD_SH;
  px=py=akt_maze->data + ey*akt_maze->x_size + ex;

  check_smiley(ex,ey);
  m_show_wall_xrl(*py,ex,ey);
  m_show_wall_ylr(*px,ex,ey);
  xl=ex-1;
  yl=ey-1;
  yu=ey+1;
  m_show_wall_yrl(*(px+1),ex+1,ey);
  px-=1;
  py+=akt_maze->x_size;
  m_show_wall_xlr(*py,ex,yu);

  for (;;)  /* immer groesserer abstand */
  {
    if (xl<0 && yl<0 && yu>=akt_maze->y_size)
      return;  /* nur wenn maze offen, sonst vorher c_finished */

    if (xl>=0)  /* felder links */
    {
      if ((m=yl)<-1)  /* unten */
        m=-1;
      for (p=px, y=ey; y>m; y--, p-=akt_maze->x_size)
      {
        check_smiley(xl,y);
        m_show_wall_xrl(*p,xl,y);
      }

      if (yl>=0)  /* spitze */
      {
        check_smiley(xl+1,y);
        m_show_wall_ylr(*(p+1),xl+1,y);
        check_smiley(xl,y);
        m_show_wall_xrl(*p,xl,y);
      }

      if ((m=yu)>=akt_maze->y_size)  /* oben */
        m=akt_maze->y_size-1;
      for (p=px+akt_maze->x_size, y=ey+1; y<m; y++, p+=akt_maze->x_size)
      {
        m_show_wall_xlr(*p,xl,y);
        check_smiley(xl,y);
      }
    }

    else
    {
      if (xl>-2 && yl>=0)  /* feld spitze */
      {
        p=px+(yl-ey)*(tm_word)(akt_maze->x_size)+1;
        x=xl+1;
        check_smiley(x,yl);
        m_show_wall_ylr(*p,x,yl);
      }
    }

    if (yu<akt_maze->y_size-1)  /* felder oben */
    {
      check_smiley(ex,yu);
      m_show_wall_yrl(*(py+1),ex+1,yu);
      m_show_wall_ylr(*py,ex,yu);
      if ((m=xl)<-1)
        m=-1;
      for (p=py-1, x=ex-1; x>m; x--, p-=1)
      {
        check_smiley(x,yu);
        m_show_wall_ylr(*p,x,yu);
      }

      if (xl>=0)  /* ecke */
      {
        m_show_wall_xlr(*p,x,yu);
        check_smiley(x,yu);
        p+=akt_maze->x_size;
        m_show_wall_xlr(*p,x,yu+1);
      }
    }

    else
    {
      if (xl>=0)  /* wand ecke */
      {
        p=px+(akt_maze->y_size-1-ey)*(tm_word)(akt_maze->x_size);
        m_show_wall_xlr(*p,xl,akt_maze->y_size-1);
      }
    }

    yl--;
    yu++;
    py+=akt_maze->x_size;

    if (yu<akt_maze->y_size)  /* rand oben */
    {
      if ((m=xl)<-1)
        m=-1;
      for (p=py, x=ex; x>m; x--, p-=1)
      {
        m_show_wall_xlr(*p,x,yu);
      }
    }

    if (xl>=0)  /* rand links */
    {
      if ((m=yl)<-1)
        m=-1;
      for (p=px, y=ey; y>m; y--, p-=akt_maze->x_size)
      {
        m_show_wall_ylr(*p,xl,y);
      }

      if ((m=yu)>=akt_maze->y_size)
        m=akt_maze->y_size-1;
      for (p=px+akt_maze->x_size, y=ey+1; y<m; y++, p+=akt_maze->x_size)
      {
        m_show_wall_ylr(*p,xl,y);
      }
    }

    xl--;
    px-=1;

  }  /* for (d) */
#endif
}

static
tm_void m_show_tl(tm_void)  /* top left = 90..135 deg */
{
#if M_SHOW_MASK&0x80
  register tm_word  x,y;
  tm_word  ex,ey;
  tm_word  xl,xu,yu,m;
  register tm_uchar *p;
  tm_uchar *px,*py;

  ex=player[my_id].x>>FIELD_SH;
  ey=player[my_id].y>>FIELD_SH;
  px=py=akt_maze->data + ey*akt_maze->x_size + ex;

  check_smiley(ex,ey);
  m_show_wall_xrl(*py,ex,ey);
  m_show_wall_ylr(*px,ex,ey);
  xl=ex-1;
  xu=ex+1;
  yu=ey+1;
  m_show_wall_yrl(*(px+1),xu,ey);
  px-=1;
  py+=akt_maze->x_size;
  m_show_wall_xlr(*py,ex,yu);

  for (;;)  /* immer groesserer abstand */
  {
    if (xl<0 && xu>=akt_maze->x_size && yu>=akt_maze->y_size)
      return;  /* nur wenn maze offen, sonst vorher c_finished */

    if (xl>=0)  /* felder links */
    {
      check_smiley(xl,ey);
      m_show_wall_xrl(*px,xl,ey);
      if ((m=yu)>=akt_maze->y_size)
        m=akt_maze->y_size-1;
      for (p=px+akt_maze->x_size, y=ey+1; y<m; y++, p+=akt_maze->x_size)
      {
        m_show_wall_xlr(*p,xl,y);
        check_smiley(xl,y);
      }
    }

    if (yu<akt_maze->y_size-1)  /* felder oben */
    {
      if ((m=xl)<-1)  /* links */
        m=-1;
      for (p=py, x=ex; x>m; x--, p-=1)
      {
        check_smiley(x,yu);
        m_show_wall_ylr(*p,x,yu);
      }

      if (xl>=0)  /* ecke */
      {
        m_show_wall_xlr(*p,x,yu);
        check_smiley(x,yu);
        m_show_wall_ylr(*p,x,yu);
      }

      if ((m=xu)>=akt_maze->x_size)  /* rechts */
        m=akt_maze->x_size-1;
      for (p=py+1, x=ex+1; x<m; x++, p+=1)
      {
        m_show_wall_yrl(*p,x,yu);
        check_smiley(x,yu);
      }

      if (xu<akt_maze->x_size-1)  /* spitze */
      {
        check_smiley(x,yu-1);
        m_show_wall_xlr(*p,x,yu);
        m_show_wall_yrl(*p,x,yu);
        check_smiley(x,yu);
        x++;
        p+=1;
      }
      m_show_wall_yrl(*p,x,yu);
    }

    else
    {
      if (yu<akt_maze->y_size && xu<akt_maze->x_size-1)  /* feld spitze */
      {
        p=py+xu-ex;
        check_smiley(xu,yu-1);
        m_show_wall_xlr(*p,xu,yu);
      }
      if (xl>=0)  /* wand ecke */
      {
        p=px+(akt_maze->y_size-1-ey)*(tm_word)(akt_maze->x_size);
        m_show_wall_xlr(*p,xl,akt_maze->y_size-1);
      }
    }

    if (xl>=0)  /* rand links */
    {
      if ((m=yu)>=akt_maze->y_size)
        m=akt_maze->y_size-1;
      for (p=px, y=ey; y<m; y++, p+=akt_maze->x_size)
      {
        m_show_wall_ylr(*p,xl,y);
      }
    }

    xl--;
    xu++;
    px-=1;

    yu++;
    py+=akt_maze->x_size;

    if (yu<akt_maze->y_size)  /* rand oben */
    {
      if ((m=xl)<-1)
        m=-1;
      for (p=py, x=ex; x>m; x--, p-=1)
      {
        m_show_wall_xlr(*p,x,yu);
      }

      if ((m=xu)>=akt_maze->x_size)
        m=akt_maze->x_size-1;
      for (p=py+1, x=ex+1; x<m; x++, p+=1)
      {
        m_show_wall_xlr(*p,x,yu);
      }
    }

  }  /* for (d) */
#endif
}

/* --- die Hauptfunktion... */

tm_uword cross_col1,cross_col2;

tm_void draw_maze(tm_void)
{
  tm_word i;

  trans_maze();

  init_smileys();

  init_stack();
  c_init();

  if (cos>=0) {
    if (sin>=0) {
      if (sin>=cos)
        m_show_rt();
      else
        m_show_tr();
    } else {
      if (-sin>=cos)
        m_show_lt();
      else
        m_show_tl();
    }
  } else {
    if (sin>=0) {
      if (sin>=-cos)
        m_show_rb();
      else
        m_show_br();
    } else {
      if (-sin>=-cos)
        m_show_lb();
      else
        m_show_bl();
    }
  }

/* hier wird der hintergrund gemalt...
   spaeter vielleicht ein fertiges hintergrundbild kopieren,
   das z. b. schon spielstand u. ae. enthaelt...
   (auch die fertige kreisscheibe fuer den kompass!)
   */
#ifdef WALLS_LINES
  draw_rect(0,0,width,pf_height,BW_WHITE);
#endif
#ifdef WALLS_FILLED
  draw_rect(0,0,width,pf_height/2,heaven_color);
  draw_rect(0,pf_height/2,width,pf_height/2,floor_color);
#endif
#ifdef WALLS_SHADED
  draw_rect(0,0,width,pf_height/2,heaven_color);
  if(colors>=256)
    for(i=0;i<(pf_height/2);i++)
      draw_hline(0,(pf_height/2)+i,width,floor_color+(i*64/(pf_height/2)));
  else
    draw_rect(0,pf_height/2,width,pf_height/2,floor_color);
#endif

  /* Waende und Figuren zeichnen */
  draw_stack();

  /* Fadenkreuz zeichnen */
  draw_hline(SCR_XOFFS-CROSS,SCR_YOFFS-1,2*CROSS,cross_col1);
  draw_vline(SCR_XOFFS-1,SCR_YOFFS-CROSS,2*CROSS,cross_col1);
  draw_hline(SCR_XOFFS-CROSS,SCR_YOFFS,2*CROSS,cross_col2);
  draw_vline(SCR_XOFFS,SCR_YOFFS-CROSS,2*CROSS,cross_col2);

  draw_compass(player[my_id].rot_z);

  flip_screen();
}



