/*
   makefont.c

   Dieses Programm erzeugt aus einer via Metafont erstellten Pixel-
   Font-Datei eine Font-Bitmap als C-Source.

   Bisher werden dazu PackedFonts (*.PK aus PK.C) benutzt, eine 
   Umstellung auf *.GF (mit GF.C) sollte sehr einfach sein, da
   die dazu noetigen Funktionen einfach dem xdvi entnommen werden
   koennen.

   ???       Till  Erste Version
   21.11.96  Chip  Erweitert fuer TOS
   23.11.96  Chip  Nur noch eine Ebene
   26.11.96  Chip  Bitmaps auf 16 Bit aufgerundet
                   JAJA, wir wissen, dass man PK.C auf 16 Bit mit
                   MSB links umstellen kann! Wir hatten einfach
                   nicht mehr Lust auf die Arbeit...
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"

#ifdef __X11__
#include "x11.h"
#define DISP_ENABLE
#endif

#ifdef __TOS__
#include "gem.h"
#define DISP_ENABLE
#endif

#ifndef SEEK_SET
#define SEEK_SET 0
#endif

#define FALSE 0
#define TRUE  (!FALSE)

#define STEP 5       /* space between chars in portions of ' ' */
#define MAX_CHAR 128

int debug=0x0;

int main(short argc, char **argv)
{
  char  *inname,*outname;
#ifdef __TOS__
  char  argv1[40],argv2[40];
#endif
  char xtra[40];
  struct font	fontp;
  int ch,n,max_x=0,max_y=0,base_offset_x=0, base_offset_y=0;
  int unterkante=10000;
  FILE *outfile;
  int disp,step;

  inname=argv[1];
  outname=argv[2];
  strcpy(xtra,"");
#ifdef __TOS__
  init_gem();
  atexit(exit_gem);
  argv[0]="MAKEFONT";
  if (argc!=2 && argc!=3)
  {
    argc=get_parameter(argv1,argv2);
    inname=argv1;
    outname=argv2;
  }
#endif

  disp=FALSE;
#ifdef DISP_ENABLE
  if (argc!=3 && argc!=2 && argc!=4)
  {
    printf("usage: %s input.pk [output.c name_extension]\n",argv[0]);
    exit(1);
  }
  else if (argc==2)
    disp=TRUE;
#else
  if (argc>4)
  {
    printf("usage: %s input.pk output.c name_extension\n",argv[0]);
    exit(1);
  }
#endif

  /* extra name */
  if(argc == 4)
    strcpy(xtra, argv[3]);
  
  if(!disp) {
    outfile = fopen(outname,"w");
    if (outfile==NULL)
    {
      perror("outfile");
      exit(2);
    }
    fprintf(outfile,"/*\n"
                    "\t%s\n"
                    "\n"
                    "\tcreated by %s\n"
                    "*/\n"
                    "#ifdef __X11__\n"
                    "#include \"../ctypes.h\"\n"
                    "#include \"../grafik.h\"\n"
                    "#else\n"
                    "#include \"ctypes.h\"\n"
                    "#include \"grafik.h\"\n"
                    "#endif\n"
            ,outname,argv[0]);
  }

  /* los gehts */
  for(n=0;n<sizeof(struct font);n++) *(((char*)&fontp)+n)=0;

  fontp.filename = inname;
  fontp.file     = fopen(inname,"rb");
  if (fontp.file==NULL)
  {
    perror("infile");
    exit(3);
  }

  while( ((unsigned char)fgetc(fontp.file)) != 0x59);   /* PK-ID */

  /* Index lesen */
  read_PK_index(&fontp, 0);

  for(ch=0;ch<MAX_CHAR;ch++) {
    if(fontp.glyph[ch].addr != 0) {
      /* Buchstabenadresse ansteuern */
      fseek(fontp.file, fontp.glyph[ch].addr, SEEK_SET);
      fontp.read_char(&fontp, ch);

      /* maximale Zeichengroesse */
      if( fontp.glyph[ch].bitmap.w > max_x )
	max_x = fontp.glyph[ch].bitmap.w;
      if( fontp.glyph[ch].bitmap.h > max_y )
	max_y = fontp.glyph[ch].bitmap.h;

      /* maximale Offsets */
      if( fontp.glyph[ch].x > base_offset_x )
	base_offset_x = fontp.glyph[ch].x;
      if( fontp.glyph[ch].y > base_offset_y )
	base_offset_y = fontp.glyph[ch].y;

      /* minimale Unterkante */
      if( fontp.glyph[ch].y - fontp.glyph[ch].bitmap.h < unterkante )
	unterkante = fontp.glyph[ch].y - fontp.glyph[ch].bitmap.h;
    }
  }

  if(!disp) {
    fprintf(outfile,"CONST tm_uword font_height%s=%d;\n",xtra,base_offset_y-unterkante);
    step = fontp.glyph[' '].bitmap.w/STEP + 1;
  }

  /* SPACE (0x20) contains pixel data -> clear it */
  for(ch=0;ch<fontp.glyph[' '].bitmap.bytes_wide*
      fontp.glyph[' '].bitmap.h;ch++)
    ((char*)fontp.glyph[' '].bitmap.bits)[ch]=0;

#ifdef DISP_ENABLE
  if(disp)
    open_window(argc,argv,16*max_x,8*max_y+2,argv[0]);
#endif

  for(ch=0;ch<MAX_CHAR;ch++) {
    if(fontp.glyph[ch].addr != 0) {
      int x,y,pos,cnt;
      BMUNIT *data,mask;
      unsigned int out;

      if(!disp)
        fprintf(outfile,"CONST tm_ushort glyph%s_%d[]={\n  ",xtra,ch);

      data=(BMUNIT *)fontp.glyph[ch].bitmap.bits;
      mask=1; cnt=-1;
      for(y=0; y<fontp.glyph[ch].bitmap.h; y++) {
        pos = 0; out=0;
        for(x=0; x<fontp.glyph[ch].bitmap.w; x++) {
          if(mask == 0) { mask = 1; data++; }

#ifdef DISP_ENABLE
          if(disp)
            put_pixel((ch&0x0f)*max_x+x+max_x/2-fontp.glyph[ch].bitmap.w/2,
                      (ch>>4)*max_y+y+base_offset_y-fontp.glyph[ch].y,
                      (*data&mask)?1:0);
#endif

          if(*data&mask) out |= (1<<(15-pos));

          if((!disp)&&(++pos==16)) {
            if (++cnt==0)
              fprintf(outfile,"0x%x",out);
            else if (cnt==10)
            {
              fprintf(outfile,",\n  0x%x",out);
              cnt=0;
            }
            else
              fprintf(outfile,",0x%x",out);
            out=0; pos=0;
          }
          mask<<=1;
        }

        if((!disp)&&(pos!=0)) {
          if (++cnt==0)
            fprintf(outfile,"0x%x",out);
          else if (cnt==10)
          {
            fprintf(outfile,",\n  0x%x",out);
            cnt=0;
          }
          else
            fprintf(outfile,",0x%x",out);
        }
        mask = 0;
      }

      if(!disp)
        fprintf(outfile,"};\n",ch);
#ifdef DISP_ENABLE
      else
        draw_bitmap();
#endif
    }
  }

  if(!disp) {
    fprintf(outfile,"CONST CHARACTER charTable%s[]={\n",xtra);

    for(ch=0;ch<MAX_CHAR;ch++) {
      if(fontp.glyph[ch].addr != 0)
        fprintf(outfile,"  {%d,%d,%d,%d,glyph%s_%d}%s\n",
                fontp.glyph[ch].bitmap.w,
                fontp.glyph[ch].bitmap.h,
                base_offset_y-fontp.glyph[ch].y,
                fontp.glyph[ch].bitmap.w+step,
		xtra,
                ch,(ch==MAX_CHAR-1)?"":",");
      else
        fprintf(outfile,"  {0,0,0,0,(uchar*)NULL_PTR}%s\n",
                (ch==MAX_CHAR-1)?"":",");
    }

    fprintf(outfile,"};\n");
  }

#ifdef DISP_ENABLE
  if(disp) {
    while(get_key()==0);
    close_window();
  }
#endif

  return 0;
}
