/*

*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/keysym.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* ein paar Sachen fuers X11 */
Display  *display;
int      screen;
GC 	 gc;
Window   win;
XImage   *xi;
Colormap cmap;
Visual   *visual;
char     *screen_base;
Colormap map;

char icon_name[]="MakeFont";
#include "makefont.icn"

void draw_bitmap(void) 
{
  XPutImage(display, win, gc, xi, 0, 0, 0, 0, xi->width, xi->height);
}

void put_pixel(int x, int y, unsigned char color)
{
  XPutPixel(xi,x,y,color?BlackPixel(display,screen):WhitePixel(display,screen));
}

int open_window( int argc, char **argv, 
		int width, int height, 
		char *titel)
{
  /* und hier muss erstmal ein Fenster aufgemacht werden */
  Pixmap icon_pixmap;
  XSizeHints size_hints;
  char *display_name=NULL;
  int window_size=0,i;
  unsigned long valuemask=0;
  XGCValues values;
  XEvent report;
  XStandardColormap *std_map;
  XKeyboardControl key_control;
  XColor col;

  if((display=XOpenDisplay(display_name))==NULL) {
    fprintf(stderr,"%s: cannot connect to X server %s\n",
                         argv[0],XDisplayName(display_name));
    exit(-1);
  }

  screen = DefaultScreen(display);
  visual = DefaultVisual(display, screen);

  /* Window kreieren */
  win=XCreateSimpleWindow(display,RootWindow(display,screen),
			  25,15,width,height,
			  4,BlackPixel(display,screen),
			  WhitePixel(display,screen));

  /* Icon anmelden */
  icon_pixmap=XCreateBitmapFromData(display,win,icon_bitmap_bits,
				    icon_bitmap_width,icon_bitmap_height);

  size_hints.flags=PPosition|PSize|PMinSize;
  size_hints.x=25;
  size_hints.y=15;
  size_hints.min_width  = size_hints.width  = width;
  size_hints.min_height = size_hints.height = height;

  XSetStandardProperties(display,win,titel,icon_name,icon_pixmap,argv,
                         argc,&size_hints);

  XSelectInput(display,win,ExposureMask|KeyPressMask|KeyReleaseMask|
	       ButtonPressMask|ButtonReleaseMask|StructureNotifyMask);

  gc=XCreateGC(display,win,valuemask,&values);

  XSetForeground(display,gc,BlackPixel(display,screen));
  XSetBackground(display,gc,WhitePixel(display,screen));

  /* Window anzeigen */
  XMapWindow(display,win);
  
  /* warten, bis das Window da ist */
  do {
    XNextEvent(display,&report);
    if(report.type == Expose)
      while(XCheckTypedEvent(display,Expose,&report));
  }
  while(report.type != Expose);

  if((screen_base = malloc(width*height))==NULL) {
    puts("unable to alloc screen buffer");
  }

  /* Bitmap-Daten lesen und auf den Puffer schalten */
  xi = XCreateImage(display, visual, 1, XYBitmap, 0, 
		    screen_base, width, height, 8, 0);
  memset(screen_base, WhitePixel(display,screen), width*height/8);
  draw_bitmap();

  return(1);
}

void close_window(void)
{
  XDestroyImage(xi);
  /*    free(screen_base);  */
  XFreeGC(display,gc);
  XCloseDisplay(display);
}

int get_key(void)
{
  XEvent report;
  char buffer[20];
  int  bufsize=20;
  KeySym key=0;
  XComposeStatus compose;

  if(XPending(display))
  {
    XNextEvent(display,&report);

    switch(report.type)
    {
    case Expose:
      while(XCheckTypedEvent(display,Expose,&report));
      draw_bitmap();
      break;

    case ButtonPress:
      key=-1;
      break;

    case KeyPress:
      XLookupString((XKeyEvent*)&report, buffer, bufsize, &key, &compose);
      break;

    case MappingNotify:
      XRefreshKeyboardMapping((XMappingEvent*)&report);
      break;
    }
  }
  return(key);
}






