/*
  GIF decoder an viewer for TheMaze

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dialog.h"
#include "machine.h"
#include "ctypes.h"
#include "gr_prim.h"
#include "gif.h"
#include "machine.h"
#include "console.h"
#include "grafik.h"
#include "misc.h"
#include "maze.h"

#define noCHECKLEN
#define MAXX            (2048)

#define LOCAL_MAP       (0x80)
#define INTERLACE       (0x40)

char gif_prefix[]="TheMaze(GIF.C): ";

tm_uchar gif_grey[256];

#ifdef PICTURES_IN_EXEC
tm_char *file;
tm_long file_len;
#else
FILE    *file;
tm_uword pal[256][3];
#endif

tm_word bad_code_count;             /* global error count for decoder */

/*
	get_byte() -- return next byte of input file, used by decoder
*/

#ifdef PICTURES_IN_EXEC
#define get_byte() (--file_len>0)?((int)*file++)&0xff:-1
#else
#define get_byte() fgetc(file)
#endif

tm_char dithermatrix[][8]=
 {{  0, 64, 16, 80,  4, 68, 20, 84},
  { 96, 32,112, 48,100, 36,116, 52},
  { 24, 88,  8, 72, 28, 92, 12, 76},
  {120, 56,104, 40,124, 60,108, 44},
  {  6, 70, 22, 86,  2, 66, 18, 82},
  {102, 38,118, 54, 98, 34,116, 50},
  { 30, 94, 14, 78, 26, 90, 10, 74},
  {126, 62,110, 46,122, 58,106, 42},
  {  1, 65, 17, 81,  5, 69, 21, 85},
  { 97, 33,113, 49,101, 37,117, 53},
  { 25, 89,  9, 73, 29, 93, 13, 77},
  {121, 57,105, 41,125, 61,109, 45},
  {  7, 71, 23, 87,  3, 67, 19, 83},
  {103, 39,119, 55, 99, 35,117, 51},
  { 31, 95, 15, 79, 27, 91, 11, 75},
  {127, 63,111, 47,123, 59,107, 43}};

/*
	out_line() -- output next scan line, called by decoder
*/

tm_uchar *vmem;
#ifdef CHECKLEN
tm_long errcnt;
#endif

int     out_line(unsigned char pix[], int linelen)
{
  tm_word i;
  static tm_word z=0;
#ifdef CHECKLEN
  static tm_word cnt=0;
#endif

  if(colors>=256) {
    for(i=0;i<linelen;i++)
#ifdef CHECKLEN
      if(cnt++ < ((tm_long)width*(tm_long)height))
#endif
	*vmem++=*pix++;
#ifdef CHECKLEN
      else
	errcnt++;
#endif
  } else {
    z = (z+1)&7;
    for(i=0;i<linelen/8;i++) {
#ifdef CHECKLEN
      if(cnt++ < (width*height/8))
#endif
      *vmem++=
	((gif_grey[*pix++] > dithermatrix[z][0])?0x80:0) |
	((gif_grey[*pix++] > dithermatrix[z][1])?0x40:0) |
	((gif_grey[*pix++] > dithermatrix[z][2])?0x20:0) |
	((gif_grey[*pix++] > dithermatrix[z][3])?0x10:0) |
	((gif_grey[*pix++] > dithermatrix[z][4])?0x08:0) |
	((gif_grey[*pix++] > dithermatrix[z][5])?0x04:0) |
	((gif_grey[*pix++] > dithermatrix[z][6])?0x02:0) |
	((gif_grey[*pix++] > dithermatrix[z][7])?0x01:0);
    }
  }

#if 0
  /* check for odd line size */
  if(linelen != width) {
    fprintf(stderr, "Warning.  GIF decoder returned line of length %d.\n", linelen);
  }
#endif
  return 0;
}

/* DECODE.C - An LZW decoder for GIF
 * Copyright (C) 1987, by Steven A. Bennett
 *
 * Permission is given by the author to freely redistribute and include
 * this code in any program as long as this credit is given where due.
 *
 * In accordance with the above, I want to credit Steve Wilhite who wrote
 * the code which this is heavily inspired by...
 *
 * GIF and 'Graphics Interchange Format' are trademarks (tm) of
 * Compuserve, Incorporated, an H&R Block Company.
 *
 * Release Notes: This file contains a decoder routine for GIF images
 * which is similar, structurally, to the original routine by Steve Wilhite.
 * It is, however, somewhat noticably faster in most cases.
 *
 */

/* STD.H - My own standard header file...
 */

#define LOCAL static
#define IMPORT extern

#define FAST register

typedef short WORD;
typedef unsigned short UWORD;
typedef char TEXT;
typedef unsigned char UTINY;
typedef long LONG;
typedef unsigned long ULONG;
typedef int INT;

/* Various error codes used by decoder
 * and my own routines...   It's okay
 * for you to define whatever you want,
 * as long as it's negative...  It will be
 * returned intact up the various subroutine
 * levels...
 */
#define OUT_OF_MEMORY -10
#define BAD_CODE_SIZE -20
#define READ_ERROR -1
#define WRITE_ERROR -2
#define OPEN_ERROR -3
#define CREATE_ERROR -4


/* IMPORT INT get_byte()
 *
 *   - This external (machine specific) function is expected to return
 * either the next byte from the GIF file, or a negative number, as
 * defined in ERRS.H.
 */

/* IMPORT INT out_line(pixels, linelen)
 *     UBYTE pixels[];
 *     INT linelen;
 *
 *   - This function takes a full line of pixels (one byte per pixel) and
 * displays them (or does whatever your program wants with them...).  It
 * should return zero, or negative if an error or some other event occurs
 * which would require aborting the decode process...  Note that the length
 * passed will almost always be equal to the line length passed to the
 * decoder function, with the sole exception occurring when an ending code
 * occurs in an odd place in the GIF file...  In any case, linelen will be
 * equal to the number of pixels passed...
 */
IMPORT INT out_line();

/* IMPORT INT bad_code_count;
 *
 * This value is the only other global required by the using program, and
 * is incremented each time an out of range code is read by the decoder.
 * When this value is non-zero after a decode, your GIF file is probably
 * corrupt in some way...
 */

#define MAX_CODES   4095

/* Static variables */
LOCAL WORD curr_size;                     /* The current code size */
LOCAL WORD clear;                         /* Value for a clear code */
LOCAL WORD ending;                        /* Value for a ending code */
LOCAL WORD newcodes;                      /* First available code */
LOCAL WORD top_slot;                      /* Highest code for current size */
LOCAL WORD slot;                          /* Last read code */

/* The following static variables are used
 * for seperating out codes
 */
LOCAL WORD navail_bytes = 0;              /* # bytes left in block */
LOCAL WORD nbits_left = 0;                /* # bits left in current byte */
LOCAL UTINY b1;                           /* Current byte */
LOCAL UTINY byte_buff[257];               /* Current block */
LOCAL UTINY *pbytes;                      /* Pointer to next byte in block */

LOCAL LONG code_mask[13] = {
     0,
     0x0001, 0x0003,
     0x0007, 0x000F,
     0x001F, 0x003F,
     0x007F, 0x00FF,
     0x01FF, 0x03FF,
     0x07FF, 0x0FFF
     };


/* This function initializes the decoder for reading a new image.
 */
LOCAL WORD init_exp(WORD size)
   {
   curr_size = size + 1;
   top_slot = 1 << curr_size;
   clear = 1 << size;
   ending = clear + 1;
   slot = newcodes = ending + 1;
   navail_bytes = nbits_left = 0;
   return(0);
   }

/* get_next_code()
 * - gets the next code from the GIF file.  Returns the code, or else
 * a negative number in case of file errors...
 */
LOCAL WORD get_next_code(void)
   {
   WORD i, x;
   ULONG ret;

   if (nbits_left == 0)
      {
      if (navail_bytes <= 0)
	 {

	 /* Out of bytes in current block, so read next block
	  */
	 pbytes = byte_buff;
	 if ((navail_bytes = get_byte()) < 0)
	    return(navail_bytes);
	 else if (navail_bytes)
	    {
	    for (i = 0; i < navail_bytes; ++i)
	       {
	       if ((x = get_byte()) < 0)
		  return(x);
	       byte_buff[i] = x;
	       }
	    }
	 }
      b1 = *pbytes++;
      nbits_left = 8;
      --navail_bytes;
      }

   ret = b1 >> (8 - nbits_left);
   while (curr_size > nbits_left)
      {
      if (navail_bytes <= 0)
	 {

	 /* Out of bytes in current block, so read next block
	  */
	 pbytes = byte_buff;
	 if ((navail_bytes = get_byte()) < 0)
	    return(navail_bytes);
	 else if (navail_bytes)
	    {
	    for (i = 0; i < navail_bytes; ++i)
	       {
	       if ((x = get_byte()) < 0)
		  return(x);
               byte_buff[i] = x;
	       }
            }
	 }
      b1 = *pbytes++;
      ret |= b1 << nbits_left;
      nbits_left += 8;
      --navail_bytes;
      }
   nbits_left -= curr_size;
   ret &= code_mask[curr_size];
   return((WORD)(ret));
   }


/* The reason we have these seperated like this instead of using
 * a structure like the original Wilhite code did, is because this
 * stuff generally produces significantly faster code when compiled...
 * This code is full of similar speedups...  (For a good book on writing
 * C for speed or for space optomisation, see Efficient C by Tom Plum,
 * published by Plum-Hall Associates...)
 */
LOCAL UTINY stack[MAX_CODES + 1];            /* Stack for storing pixels */
LOCAL UTINY suffix[MAX_CODES + 1];           /* Suffix table */
LOCAL UWORD prefix[MAX_CODES + 1];           /* Prefix linked list */

/* WORD decoder(linewidth)
 *    WORD linewidth;               * Pixels per line of image *
 *
 * - This function decodes an LZW image, according to the method used
 * in the GIF spec.  Every *linewidth* "characters" (ie. pixels) decoded
 * will generate a call to out_line(), which is a user specific function
 * to display a line of pixels.  The function gets it's codes from
 * get_next_code() which is responsible for reading blocks of data and
 * seperating them into the proper size codes.  Finally, get_byte() is
 * the global routine to read the next byte from the GIF file.
 *
 * It is generally a good idea to have linewidth correspond to the actual
 * width of a line (as specified in the Image header) to make your own
 * code a bit simpler, but it isn't absolutely necessary.
 *
 * Returns: 0 if successful, else negative.  (See ERRS.H)
 *
 */

WORD decoder(WORD linewidth)
   {
   FAST UTINY *sp, *bufptr;
   UTINY *buf;
   FAST WORD code, fc, oc, bufcnt;
   WORD c, size, ret;

   /* Initialize for decoding a new image...
    */
   if ((size = get_byte()) < 0)
      return(size);
   if (size < 2 || 9 < size)
      return(BAD_CODE_SIZE);
   init_exp(size);

   /* Initialize in case they forgot to put in a clear code.
    * (This shouldn't happen, but we'll try and decode it anyway...)
    */
   oc = fc = 0;

   /* Allocate space for the decode buffer
    */
   if ((buf = (UTINY *)malloc(linewidth + 1)) == NULL)
      return(OUT_OF_MEMORY);

   /* Set up the stack pointer and decode buffer pointer
    */
   sp = stack;
   bufptr = buf;
   bufcnt = linewidth;

   /* This is the main loop.  For each code we get we pass through the
    * linked list of prefix codes, pushing the corresponding "character" for
    * each code onto the stack.  When the list reaches a single "character"
    * we push that on the stack too, and then start unstacking each
    * character for output in the correct order.  Special handling is
    * included for the clear code, and the whole thing ends when we get
    * an ending code.
    */
   while ((c = get_next_code()) != ending)
      {

      /* If we had a file error, return without completing the decode
       */
      if (c < 0)
	 {
         free(buf);
	 return(0);
	 }

      /* If the code is a clear code, reinitialize all necessary items.
       */
      if (c == clear)
         {
	 curr_size = size + 1;
	 slot = newcodes;
         top_slot = 1 << curr_size;

         /* Continue reading codes until we get a non-clear code
          * (Another unlikely, but possible case...)
	  */
         while ((c = get_next_code()) == clear)
            ;

         /* If we get an ending code immediately after a clear code
          * (Yet another unlikely case), then break out of the loop.
          */
         if (c == ending)
            break;

         /* Finally, if the code is beyond the range of already set codes,
          * (This one had better NOT happen...  I have no idea what will
          * result from this, but I doubt it will look good...) then set it
          * to color zero.
	  */
         if (c >= slot)
	    c = 0;

	 oc = fc = c;

	 /* And let us not forget to put the char into the buffer... And
	  * if, on the off chance, we were exactly one pixel from the end
	  * of the line, we have to send the buffer to the out_line()
	  * routine...
	  */
	 *bufptr++ = c;
	 if (--bufcnt == 0)
	    {
	    if ((ret = out_line(buf, linewidth)) < 0)
	       {
	       free(buf);
	       return(ret);
	       }
	    bufptr = buf;
	    bufcnt = linewidth;
	    }
	 }
      else
	 {

	 /* In this case, it's not a clear code or an ending code, so
	  * it must be a code code...  So we can now decode the code into
	  * a stack of character codes. (Clear as mud, right?)
	  */
	 code = c;

	 /* Here we go again with one of those off chances...  If, on the
	  * off chance, the code we got is beyond the range of those already
	  * set up (Another thing which had better NOT happen...) we trick
	  * the decoder into thinking it actually got the last code read.
	  * (Hmmn... I'm not sure why this works...  But it does...)
	  */
	 if (code >= slot)
	    {
	    if (code > slot)
	       ++bad_code_count;
	    code = oc;
	    *sp++ = fc;
	    }

	 /* Here we scan back along the linked list of prefixes, pushing
	  * helpless characters (ie. suffixes) onto the stack as we do so.
	  */
	 while (code >= newcodes)
	    {
	    *sp++ = suffix[code];
	    code = prefix[code];
	    }

	 /* Push the last character on the stack, and set up the new
	  * prefix and suffix, and if the required slot number is greater
	  * than that allowed by the current bit size, increase the bit
	  * size.  (NOTE - If we are all full, we *don't* save the new
	  * suffix and prefix...  I'm not certain if this is correct...
	  * it might be more proper to overwrite the last code...
	  */
	 *sp++ = code;
	 if (slot < top_slot)
	    {
	    suffix[slot] = fc = code;
	    prefix[slot++] = oc;
	    oc = c;
	    }
	 if (slot >= top_slot)
	    if (curr_size < 12)
	       {
	       top_slot <<= 1;
	       ++curr_size;
	       }

	 /* Now that we've pushed the decoded string (in reverse order)
	  * onto the stack, lets pop it off and put it into our decode
	  * buffer...  And when the decode buffer is full, write another
	  * line...
	  */
	 while (sp > stack)
	    {
	    *bufptr++ = *(--sp);
	    if (--bufcnt == 0)
	       {
	       if ((ret = out_line(buf, linewidth)) < 0)
		  {
		  free(buf);
		  return(ret);
		  }
	       bufptr = buf;
	       bufcnt = linewidth;
	       }
	    }
	 }
      }
   ret = 0;
   if (bufcnt != linewidth)
      ret = out_line(buf, (linewidth - bufcnt));
   free(buf);
   return(ret);
   }

#ifndef PICTURES_IN_EXEC
/* pictures are extern */
tm_bool show_gif(char *pname)
{
  tm_word  c, cols,width,height;
  tm_word  i,col[3];
  char     sig[6];         /* GIF signature */
  char     *s,*d,filename[128];

  for(s=PICTURE_PREFIX,d=filename;*s!=0;*d++=*s++);
  for(*d++=PATH_SEP,s=pname;*s!=0;*d++=*s++);
#ifdef TREE_D
  if(use_3d) { *d++='3'; *d++='d'; }
#endif
  *d++=(MAX_REZ_WIDTH/100)%10+'0';  *d++=(MAX_REZ_WIDTH/10)%10+'0';
  *d++=(MAX_REZ_HEIGHT/100)%10+'0'; *d++=(MAX_REZ_HEIGHT/10)%10+'0';
  for(s=PICTURE_SUFFIX;*s!=0;*d++=*s++); *d=0;

  if((file=fopen(filename,"rb"))==NULL) {
    console_printf("%sError opening image file \'%s\'.\n",gif_prefix,filename);
  } else {
    /* check gif signature */
    c = fread(sig, 1, 6, file);
    if(c!=6 || strncmp(sig, "GIF87a", 6)) {
      console_printf("%sError in GIF signature.\n",gif_prefix);
      fclose(file); return(ERROR);
    }

    /* get screen descriptor info */
    width = fgetc(file);  width += 256*fgetc(file);
    height = fgetc(file); height += 256*fgetc(file);

    /* check color map info */
    c = fgetc(file);
    if((c&LOCAL_MAP) == 0) {
      console_printf("%sError, image has local color map.\n",gif_prefix);
      fclose(file); return(ERROR);
    }

    cols = 1<<((c&0x07)+1); fgetc(file); fgetc(file);

    /* test width & height */
    if((width!=MAX_REZ_WIDTH)||(height!=MAX_REZ_HEIGHT)) {
      console_printf("%sIllegal picture size.\n",gif_prefix);
      fclose(file); return(ERROR);
    }

    /* all colors black for fadein */
    for(i=0;i<colors;i++) {
      col[0]=col[1]=col[2]=0;
      set_color(i,col);
    }

    /* get global color map */
    for(i=0; i<cols; i++) {
      pal[i][0] = (((tm_uword)fgetc(file))&0xff)*253/64;
      pal[i][1] = (((tm_uword)fgetc(file))&0xff)*253/64;
      pal[i][2] = (((tm_uword)fgetc(file))&0xff)*253/64;
    }

    /* get image separator */
    c = fgetc(file);
    if(c != ',') {
      console_printf("%sError, image separator character expected.\n",gif_prefix);
      fclose(file); return(ERROR);
    }

    /* waste left top width and height */
    for(i=0; i<8; i++)
      fgetc(file);

    /* check for interlace or local color map */
    c = fgetc(file);
    if(c & LOCAL_MAP) {
      console_printf("%sError, can't handle local color maps.\n",gif_prefix);
      fclose(file); return(ERROR);
    }
    if(c & INTERLACE) {
      console_printf("%sError, can't handle interlace.\n",gif_prefix);
      fclose(file); return(ERROR);
    }

    /* set color map */
    /* correct color depth? */
    if(colors==256) {
      for(c=0; c<colors; c++) {
	col[0]=(((tm_uword)pal[c][0])&0xff)*253/64;
	col[1]=(((tm_uword)pal[c][1])&0xff)*253/64;
	col[2]=(((tm_uword)pal[c][2])&0xff)*253/64;
	set_color(c,(tm_word*)col);
      }
    }
    for(c=0; c<colors; c++) {
      gif_grey[c]=pal[c][1]>>1;
    }

    /* whoa, hey, we're ready to decode this hummer */
    bad_code_count = 0;    vmem=mem;
    c = decoder(width);
    fclose(file);
    flip_screen();

    if(c<0)             console_printf("%sDecoder returned %d.\n",gif_prefix, c);
    if(bad_code_count)  console_printf("%sBad code count, possible corruption.\n",gif_prefix);
  }
  return OK;
}
#else
/* pictures are intern */

extern tm_picture pic_tm;
#ifdef THREE_D
extern tm_picture pic_tm3d;
#endif

char       *pic_names[]={"tm",""};
tm_picture *pic_data[]={&pic_tm};
#ifdef THREE_D
tm_picture *pic_3d_data[]={&pic_tm3d};
#endif

tm_word bad_code_count;             /* global error count for decoder */

tm_bool show_gif(char *pname, char *mem)
{
  tm_char *p;
  tm_word  c, cols;
  tm_uword i,col[3];
  tm_picture *pic;

  /* future release will have to search for name */
#ifdef THREE_D
  if(use_3d)
    pic=pic_3d_data[0];
  else
#endif
    pic=pic_data[0];

  for(c=0; c<pic->colors; c++)
    gif_grey[c]=
      (( 77*(tm_uword)(pic->color[c].r>>1))+
       (150*(tm_uword)(pic->color[c].g>>1))+
       ( 29*(tm_uword)(pic->color[c].b>>1)))/256;

  /* whoa, hey, we're ready to decode this hummer */
  file_len=pic->datalen; file=pic->data;
  bad_code_count = 0;    vmem=mem;
#ifdef CHECKLEN
  errcnt=0;
#endif
  c = decoder(pic->width);

#ifdef CHECKLEN
  if(errcnt!=0)
    printf("GIF length exceeded by %ld bytes\n",errcnt);
#endif

  if(c<0)             console_printf("%sDecoder returned %d.\n",gif_prefix, c);
  if(bad_code_count)  console_printf("%sBad code count, possible corruption.\n",gif_prefix);

  return OK;
}
#endif

