/*
  GR_PRIM.C

  die maschinenabhaengigen Grafik-Primitiva

  08.07.95 Till
  10.09.97 Till  auf Bodos Waende erweitert
*/

#include "ctypes.h"
#include "gr_prim.h"
#include "error.h"

#ifdef RAW2
#include "raw2/raw2.c"
#include "raw2/raw2wall.c"
#endif

#ifdef RAW256
#include "raw256/r256_pix.c"
#include "raw256/r256_lin.c"
#include "raw256/r256_cir.c"
#include "raw256/r256_blt.c"
#include "raw256/r256wall.c"
#endif

#ifdef RAW16M
#include "raw16m/r16m_pix.c"
#include "raw16m/r16m_lin.c"
#include "raw16m/r16m_cir.c"
#include "raw16m/r16m_blt.c"
#endif

/* this machine supports different color depths */
#ifdef MULTIMODE
tm_void  (*set_pixel)(tm_uword x, tm_uword y, tm_uword col);
tm_uword (*get_pixel)(tm_uword x, tm_uword y);
tm_void  (*put_bitmap)(tm_uword x, tm_uword y, tm_uword width, tm_uword height,
		      CONST tm_ushort *data, tm_uword col);
tm_void  (*save_screen)(void);
tm_void  (*restore_screen)(void);
tm_void  (*draw_line)(tm_uword x1, tm_uword y1, tm_uword x2, tm_uword y2, tm_uword col);
tm_void  (*draw_hline)(tm_uword x, tm_uword y, tm_uword len, tm_uword col);
tm_void  (*draw_vline)(tm_uword x, tm_uword y, tm_uword len, tm_uword col);
tm_void  (*draw_quadvline)(tm_uword x, tm_uword y, tm_uword len, tm_ulong cols);
tm_void  (*draw_vtexline)(tm_uword, tm_uword, tm_uword, tm_uword, tm_uchar *, tm_uword, tm_uchar *);
tm_void  (*draw_vdline)(tm_uword x, tm_uword y, tm_uword len, tm_uword e, tm_uword col1, tm_uword col2);
tm_void  (*draw_disc)(tm_uword x, tm_uword y, tm_uword rad, tm_uword int_col, tm_uword bd_col);
tm_void  (*draw_rect)(tm_uword x, tm_uword y, tm_uword w, tm_uword h, tm_uword col);
#ifdef USE_MALLOC
tm_void* (*save_rect)(tm_uword x, tm_uword y, tm_uword w, tm_uword h);
tm_void  (*restore_rect)(tm_void *);
tm_void  (*free_rect)(tm_void *);
#endif
tm_void  (*wall_2)(tm_uword x,tm_uword h, tm_uword wall_col,tm_uword border_col);
tm_void  (*wall_4)(tm_uword x1,tm_uword h1,
                          tm_uword x2,tm_uword h2,
                          tm_uword wall_col,tm_uword border_col);
tm_void  (*wall_6)(tm_uword x1,tm_uword h1,
                          tm_uword x2,tm_uword h2,
                          tm_uword x3,tm_uword h3,
                          tm_uword wall_col,tm_uword border_col);
#ifdef TEXTURE
tm_void  (*wall_4tex)(tm_word x1,tm_word h1,
                          tm_word x2,tm_word h2,
                          tm_uword wall,
		          tm_word xa, tm_word za, tm_word xb, tm_word zb);
#endif

void set_multimode(tm_uword mode)
{
  switch(mode) {
#ifdef RAW2
  case MULTI_RAW2:
    set_pixel         = r2_set_pixel;
    get_pixel         = r2_get_pixel;
    put_bitmap        = r2_put_bitmap;
    save_screen       = r2_save_screen;
    restore_screen    = r2_restore_screen;
    draw_line         = r2_draw_line;
    draw_hline        = r2_draw_hline;
    draw_vline        = r2_draw_vline;
    draw_quadvline    = r2_draw_quadvline;
    draw_vtexline     = r2_draw_vtexline;
    draw_vdline       = r2_draw_vdline;
    draw_disc         = r2_draw_disc;
    draw_rect         = r2_draw_rect;
    wall_2            = r2_wall_2;
    wall_4            = r2_wall_4;
    wall_6            = r2_wall_6;
    wall_4tex         = r2_wall_4tex;
#ifdef USE_MALLOC
    save_rect         = r2_save_rect;
    restore_rect      = r2_restore_rect;
    free_rect         = r2_free_rect;
#endif
#ifdef TREE_D
    draw_vline_3d     = r2_draw_vline_3d;
    draw_quadvline_3d = r2_draw_quadvline_3d;
#endif
    break;
#endif
#ifdef RAW256
  case MULTI_RAW256:
    set_pixel         = r256_set_pixel;
    get_pixel         = r256_get_pixel;
    put_bitmap        = r256_put_bitmap;
    save_screen       = r256_save_screen;
    restore_screen    = r256_restore_screen;
    draw_line         = r256_draw_line;
    draw_hline        = r256_draw_hline;
    draw_vline        = r256_draw_vline;
    draw_quadvline    = r256_draw_quadvline;
    draw_vtexline     = r256_draw_vtexline;
    draw_vdline       = r256_draw_vdline;
    draw_disc         = r256_draw_disc;
    draw_rect         = r256_draw_rect;
    wall_2            = r256_wall_2;
    wall_4            = r256_wall_4;
    wall_6            = r256_wall_6;
    wall_4tex         = r256_wall_4tex;
#ifdef USE_MALLOC
    save_rect         = r256_save_rect;
    restore_rect      = r256_restore_rect;
    free_rect         = r256_free_rect;
#endif
#ifdef TREE_D
    draw_vline_3d     = r256_draw_vline_3d;
    draw_quadvline_3d = r256_draw_quadvline_3d;
#endif
    break;
#endif
#ifdef RAW16M
  case MULTI_RAW16M:
    set_pixel         = r16m_set_pixel;
    get_pixel         = r16m_get_pixel;
    put_bitmap        = r16m_put_bitmap;
    save_screen       = r16m_save_screen;
    restore_screen    = r16m_restore_screen;
    draw_line         = r16m_draw_line;
    draw_hline        = r16m_draw_hline;
    draw_vline        = r16m_draw_vline;
    draw_quadvline    = r16m_draw_quadvline;
    draw_vtexline     = r16m_draw_vtexline;
    draw_vdline       = r16m_draw_vdline;
    draw_disc         = r16m_draw_disc;
    draw_rect         = r16m_draw_rect;
    wall_2            = r16m_wall_2;
    wall_4            = r16m_wall_4;
    wall_6            = r16m_wall_6;
    wall_4tex         = r16m_wall_4tex;
#ifdef USE_MALLOC
    save_rect         = r16m_save_rect;
    restore_rect      = r16m_restore_rect;
    free_rect         = r16m_free_rect;
#endif
#ifdef TREE_D
    draw_vline_3d     = r16m_draw_vline_3d;
    draw_quadvline_3d = r16m_draw_quadvline_3d;
#endif
    break;
#endif
  }
}
#endif /* MULTIMODE */

