/*
  GR_PRIM.H

  die maschinenabhaengigen Grafik-Primitiva

  8.7.95    Till
  13.11.96  Chip  display of title-pic changed
  14.11.96  Chip  put_bitmap nur durchsichtig
  26.11.96  Chip  put_bitmap mit 16 bit
  27.11.96  Chip  Typen auf tm_* umgestellt
  19.01.97  Chip  Copyrighttext als "richtige" Textausgabe
  07.04.97  Till  am 3D gebastelt
  10.09.97  Till  Wandmalroutinen eingebunden
  24.01.98  Till  Config-File
  15.02.98  Till  double-buffer
*/

#ifndef __GR_PRIM__
#define __GR_PRIM__

#include "ctypes.h"
#include "machine.h"

#ifdef THREE_D
#define LEFT_EYE   0
#define RIGHT_EYE  1
typedef enum {DEV_NONE=FALSE,DEV_3DMAX,DEV_CBMAX} tm_3d_dev;
extern tm_3d_dev use_3d;
#endif

#ifdef HAS_CONFIG_FILE
extern tm_void parse_config_file(tm_void);
#endif

extern tm_char video_driver_name[];

extern tm_bool open_screen(tm_uword argc, tm_char **argv,
                           tm_uword *width, tm_uword *height, tm_uword *colors);
extern tm_void close_screen(tm_void);
extern tm_void fade_title_pic(tm_char *copyright);
extern tm_void flip_screen(tm_void);
extern tm_void save_colors(tm_void);
extern tm_void restore_colors(tm_void);
extern tm_void set_color(tm_uword col, tm_uword prim[3]);
extern tm_void init_video_options(tm_void);
extern tm_void show_video_options(tm_void);
extern tm_bool set_video_option(tm_char *cmd);
extern tm_void update_double_buffer(tm_void);

/* in MULTIMODE we use function pointers   */
/* function pointers need other prototypes */
#ifdef MULTIMODE

#define MULTI_RAW2   0
#define MULTI_RAW256 1
#define MULTI_RAW16M 2
extern tm_void  set_multimode(tm_uword mode); 
extern tm_void  (*set_pixel)(tm_uword x, tm_uword y, tm_uword col);
extern tm_uword (*get_pixel)(tm_uword x, tm_uword y);
extern tm_void  (*put_bitmap)(tm_uword x, tm_uword y, tm_uword width, tm_uword height,
                              CONST tm_ushort *data, tm_uword col);
extern tm_void  (*fill_screen)(tm_uword col);
extern tm_void  (*save_screen)(tm_void);
extern tm_void  (*restore_screen)(tm_void);
extern tm_void  (*draw_line)(tm_uword x1, tm_uword y1, tm_uword x2, tm_uword y2, tm_uword col);
extern tm_void  (*draw_hline)(tm_uword x, tm_uword y, tm_uword len, tm_uword col);
extern tm_void  (*draw_vline)(tm_uword x, tm_uword y, tm_uword len, tm_uword col);
extern tm_void  (*draw_quadvline)(tm_uword x, tm_uword y, tm_uword len, tm_ulong cols);
extern tm_void  (*draw_vtexline)(tm_uword, tm_uword, tm_uword, tm_uword, tm_uchar *, tm_uword, tm_uchar *);
extern tm_void  (*draw_vdline)(tm_uword x, tm_uword y, tm_uword len, tm_uword e, tm_uword col1, tm_uword col2);
extern tm_void  (*draw_disc)(tm_uword x, tm_uword y, tm_uword rad, tm_uword int_col, tm_uword bd_col);
extern tm_void  (*draw_rect)(tm_uword x, tm_uword y, tm_uword w, tm_uword h, tm_uword col);
#ifdef USE_MALLOC
extern tm_void* (*save_rect)(tm_uword x, tm_uword y, tm_uword w, tm_uword h);
extern tm_void  (*restore_rect)(tm_void *);
extern tm_void  (*free_rect)(tm_void *);
#endif
extern tm_void  (*wall_2)(tm_uword x,tm_uword h, tm_uword wall_col,tm_uword border_col);
extern tm_void  (*wall_4)(tm_uword x1,tm_uword h1,
                          tm_uword x2,tm_uword h2,
                          tm_uword wall_col,tm_uword border_col);
extern tm_void  (*wall_6)(tm_uword x1,tm_uword h1,
                          tm_uword x2,tm_uword h2,
                          tm_uword x3,tm_uword h3,
                          tm_uword wall_col,tm_uword border_col);
#ifdef TEXTURE
tm_void  (*wall_4tex)(tm_word x1,tm_word h1,
                          tm_word x2,tm_word h2,
                          tm_uword wall,
		          tm_word xa, tm_word za, tm_word xb, tm_word zb);
#endif

#ifdef THREE_D
extern tm_void  (*draw_vline_3d)(tm_uword x, tm_uword y, tm_uword len, tm_uword col, tm_uword eye);
extern tm_void  (*draw_quadvline_3d)(tm_uword x, tm_uword y, tm_uword len, tm_ulong cols, tm_uword eye);
#endif

#else  /* MULTIMODE */

extern tm_void  set_pixel(tm_uword x, tm_uword y, tm_uword col);
extern tm_uword get_pixel(tm_uword x, tm_uword y);
extern tm_void  put_bitmap(tm_uword x, tm_uword y, tm_uword width, tm_uword height,
                           CONST tm_ushort *data, tm_uword col);
extern tm_void  fill_screen(tm_uword col);
extern tm_void  save_screen(tm_void);
extern tm_void  restore_screen(tm_void);
extern tm_void  draw_line(tm_uword x1, tm_uword y1, tm_uword x2, tm_uword y2, tm_uword col);
extern tm_void  draw_hline(tm_uword x, tm_uword y, tm_uword len, tm_uword col);
extern tm_void  draw_vline(tm_uword x, tm_uword y, tm_uword len, tm_uword col);
extern tm_void  draw_quadvline(tm_uword x, tm_uword y, tm_uword len, tm_ulong cols);
extern tm_void  draw_vtexline(tm_uword, tm_uword, tm_uword, tm_uword, tm_uchar *, tm_uword, tm_uchar *);
extern tm_void  draw_vdline(tm_uword x, tm_uword y, tm_uword len, tm_uword e, tm_uword col1, tm_uword col2);
extern tm_void  draw_disc(tm_uword x, tm_uword y, tm_uword rad, tm_uword int_col, tm_uword bd_col);
extern tm_void  draw_rect(tm_uword x, tm_uword y, tm_uword w, tm_uword h, tm_uword col);
#ifdef USE_MALLOC
extern tm_void* save_rect(tm_uword x, tm_uword y, tm_uword w, tm_uword h);
extern tm_void  restore_rect(tm_void *);
extern tm_void  free_rect(tm_void *);
#endif
extern tm_void  wall_2(tm_uword x,tm_uword h,
                       tm_uword wall_col,tm_uword border_col);
extern tm_void  wall_4(tm_uword x1,tm_uword h1,
                       tm_uword x2,tm_uword h2,
                       tm_uword wall_col,tm_uword border_col);
extern tm_void  wall_6(tm_uword x1,tm_uword h1,
                       tm_uword x2,tm_uword h2,
                       tm_uword x3,tm_uword h3,
                       tm_uword wall_col,tm_uword border_col);

#ifdef THREE_D
extern tm_void  draw_vline_3d(tm_uword x, tm_uword y, tm_uword len, tm_uword col, tm_uword eye);
extern tm_void  draw_quadvline_3d(tm_uword x, tm_uword y, tm_uword len, tm_ulong cols, tm_uword eye);
#endif

#endif /* else MULTIMODE */

#endif /* __GR_PRIM__ */
