/*

    Hauptprogramm fuer Maze

  history:
  ???       Till  creation
  13.11.96  Chip  display of title-pic changed
  26.11.96  Chip  do_error tested
  28.11.96  Chip  Typen auf tm_* umgestellt
  01.12.96  Chip  init_dialog verschoben (wegen Farben)
  19.01.97  Chip  Copyrighttext als "richtige" Textausgabe
  27.01.97  Till  diverser Kram wegen Netz / Optionen ohne '-'
  19.03.97  Chip  console jetzt systemabhaengig
  22.06.97  Chip  Fehlercodes in console.h
  24.01.98  Till  Config-File
  25.01.98  Till  Copyright auf '98 erweitert
  26.01.98  Till  Fehler in der Optionen-Abfrage
*/

#include <stdio.h>
#include "ctypes.h"
#include "grafik.h"
#include "gr_prim.h"
#include "joystick.h"
#include "dialog.h"
#include "text.h"
#include "timer.h"
#include "misc.h"
#include "maze.h"
#include "drawmaze.h"
#include "drawmap.h"
#include "move.h"
#include "audio.h"
#include "ninja.h"
#include "smiley.h"
#include "net.h"
#include "console.h"
#include "texture.h"

#define TIME_LIST  (5)    /* ueber soviele Frames wird die Zeit gemittelt */
#define Y_SPEED    (200*TIME_LIST)
#define X_SPEED    (10*TIME_LIST)

static tm_char  *copyright=
"Copyright 1995-98 by Bodo & Till";

int main(int argc, char **argv)
{
  tm_uword joy=0,i,quit;
  tm_word  arg_err,arg_help;
  tm_word  x,y,time_akt;
  tm_long  time_sum,time_ms[TIME_LIST];

  /* Sprache festlegen */
  language=LANGUAGE;

#ifdef MACHINE_HAS_CONSOLE
  init_net_options();       /* this has to be the first function call!!! */
  init_audio_options();
  init_video_options();
  init_joystick_options();

#ifdef HAS_CONFIG_FILE
  parse_config_file();
#endif

  /* parse command line options */
  for(arg_err=arg_help=FALSE,i=1;i<argc;i++) {
    if((argv[i][0]!='v')&&
       (argv[i][0]!='n')&&
       (argv[i][0]!='a')&&
       (argv[i][0]!='i')) {
      if(argv[i][0]=='?') {
	arg_help=TRUE;
      } else {
	console_printf("unknown option: '%s'\n",argv[i]);
	arg_err=TRUE;
      }
    } else {
      switch(argv[i][0]) {
      case 'a':
	if(!set_audio_option(&argv[i][1])) {
	  console_printf("audio option error: '%s'\n",argv[i]);
	  arg_err=TRUE;
	}
	break;
      case 'n':
	if(!set_net_option(&argv[i][1])) {
	  console_printf("net option error: '%s'\n",argv[i]);
	  arg_err=TRUE;
	}
	break;
      case 'v':
	if(!set_video_option(&argv[i][1])) {
	  console_printf("video option error: '%s'\n",argv[i]);
	  arg_err=TRUE;
	}
	break;
      case 'i':
	if(!set_joystick_option(&argv[i][1])) {
	  console_printf("input option error: '%s'\n",argv[i]);
	  arg_err=TRUE;
	}
	break;
      }
    }
  }

  if((arg_help)||(arg_err)) {
#ifdef MACHINE_ATARI_ST
    console_puts("usage: the_maze {options}");
#else
    console_printf("usage: %s {options}\n",argv[0]);
#endif
    console_puts("valid options:");
    console_puts("?\tgeneral help");
    console_puts("v...\tvideo options:");
    show_video_options();
    console_puts("n...\tnet options");
    show_net_options();
    console_puts("a...\taudio options");
    show_audio_options();
    console_puts("i...\tinput options");
    show_joystick_options();
#ifdef MACHINE_X11
    { extern tm_void kill_io_server(tm_void);
      kill_io_server(); }   /* nicht schoen, dass das hier steht ... */
#endif
    return(RET_HELP);
  }
#endif

  if(!init_screen((tm_uword)argc,(tm_char**)argv))
    return(RET_INIT_ERROR);
  if(!init_timer()) {
    exit_screen();
    return(RET_INIT_ERROR);
  }
  if(!init_joystick()) {
    exit_timer();
    exit_screen();
    return(RET_INIT_ERROR);
  }

  init_dialog();
  init_audio();
  init_net();

  /* hardware setup is done, let's view the title screen */
  fade_title_pic(copyright);

  /* play music and wait for button (and talk to io server) */
  do_music();

  init_maze_colors();

#ifdef TEXTURE
  init_texture();
#endif

  /* now wait for game to start */
  wait_net_start();

  /* intro msg */
  print_message(MSG_TOP,"Hallo Welt!",COL_RED);
  print_message(MSG_BOT,"Hallo Computer!",COL_YELLOW);

  /* and now play */
  init_smiley();
  init_maze((MAZE*)&maze_data[2]);
  init_player(&player[my_id]);
  init_ninjas();
/*
  draw_maze();         ueberfluessig ?
  beim st erscheint ein bild ohne waende...
  aber auch ohne diesen aufruf!
  hm. vielleicht auch nicht...
*/

  /* Zeitverwalter initialisieren */
  time_akt=0;
  for(i=0;i<TIME_LIST;time_ms[i++]=0);
  quit=FALSE;  

  do {
    net_update_game();

    time_ms[time_akt]=get_ticks();
    if(joy&JOY_FIRE2)
      draw_map();
    else
      draw_maze();

    joy=get_joystick_analog(&x,&y);
    time_ms[time_akt]=ticks2msec(get_ticks()-time_ms[time_akt]);
    if(++time_akt==TIME_LIST) time_akt=0;
    for(i=0,time_sum=0l;i<TIME_LIST;time_sum+=time_ms[i++]);
    move_player(&player[my_id],-y*time_sum/Y_SPEED,x*time_sum/X_SPEED);
    move_ninjas(time_sum);
    
    if(joy&JOY_FIRE1)      /* Menuetaste */
      quit = main_menu();

    if(joy&JOY_FIRE0)
      do_audio(SOUND_SHOT);
  }
  while(!quit);

  exit_ninjas();
  exit_net();
  exit_audio();
  exit_joystick();
  exit_timer();
  exit_screen();
  return(RET_OK);
}
