/*
   maze.c

   22. 7.95  Till   erste Version
    8.10.95  Till   sieht langsam gut aus
   10.10.95  Till   Labyrinth-Projektion funktioniert
   21.10.95  Till   Wandverwaltung optimiert
   28.10.95  Till   erweitert auf Rot-Gruen-3D
   23.08.96  Till   3D an Shutter-Brille angepasst
   10.11.96  Chip   unbenutzte Variablen geloescht
   13.11.96  Chip   Rechenfehler beseitigt
   14.11.96  Chip   farbvariablen, korrektes Schwarz und Weiss
   26.11.96  Chip   keine Multicolor-Fonts mehr
   27.11.96  Chip   Typen auf tm_* umgestellt
   01.12.96  Chip   display_maze_16()
   16.03.97  Chip   drawmaze.c und drawmap.c extrahiert
   10.09.97  Till   Shading
*/

#include "ctypes.h"
#include "mazes.h"
#include "maze.h"
#include "drawmaze.h"
#include "drawmap.h"
#include "misc.h"
#include "gr_prim.h"
#include "grafik.h"
#include "move.h"
#include "smiley.h"
#include "ninja.h"
#include "texture.h"
#include "machine.h"
#include "net.h"

#ifdef THREE_D
#include <mem.h>
#endif

tm_uword col_tab[16][3]={
  {0,0,0},           /* BLACK    */
  {1000,1000,1000},  /* WHITE    */
  {1000,0,0},        /* RED      */
  {0,1000,0},        /* GREEN    */
  {0,0,1000},        /* BLUE     */
  {1000,1000,0},     /* YELLOW   */
  {1000,0,1000},     /* MAGENTA  */
  {0,1000,1000},     /* CYAN     */
  {400,400,400},     /* DGREY    */
  {600,600,600},     /* LGREY    */
  {500,0,0},         /* DRED     */
  {0,500,0},         /* DGREEN   */
  {0,0,500},         /* DBLUE    */
  {500,500,0},       /* BROWN    */
  {500,0,500},       /* DMAGENTA */
  {0,500,500} };     /* DCYAN    */

MAZE      *akt_maze;

#ifdef WALLS_SHADED
tm_char shade_offset[MAX_REZ_WIDTH/2];
tm_char shade_dither[MAX_REZ_WIDTH/2];
#endif

tm_void
init_maze_colors(tm_void)
{
  tm_uword i,col[3];

  draw_rect(0,0,width,height,COL_BLACK);
  flip_screen();

  if(colors<16)
  {
#ifdef WALLS_LINES
    wall_color[OB_DOOR]=6;
    wall_color[OB_WALL]=BW_BLACK;
#else
    wall_color[OB_DOOR]=BW_DGREY;
    wall_color[OB_WALL]=BW_MGREY;
#endif
    wall_border=BW_BLACK;
    heaven_color=5;
    floor_color=6;

    cross_col1=BW_WHITE;
    cross_col2=BW_BLACK;
    border_col=BW_WHITE;
    inner_col=BW_BLACK;
    needle_col=BW_WHITE;

    map_color[OB_NOTHING]=BW_WHITE; /* fuer OB_FILLED */
    map_color[OB_DOOR]=BW_MGREY;
    map_color[OB_GLAS]=5;
    map_color[OB_WALL]=BW_WHITE;
  }
  else     /* colors < 16 */
  {
#ifdef WALLS_SHADED
    wall_color[OB_DOOR]=COL_VGREY;    /* 32 grau */
    wall_color[OB_WALL]=COL_VYELLOW;  /* 32 gelb */
    wall_border=COL_VRED;             /* 32 rot  */
    floor_color=COL_VCYAN;            /* 64 cyan */
    for(i=0;i<(width/2);i++) {
      if(i<((SHADE_LOWER_BOUND*width)/200)) {
        shade_offset[i]=0;
        shade_dither[i]=0;
      } else {
        if(i>((SHADE_UPPER_BOUND*width)/200)) {
          shade_offset[i]=31;
	  shade_dither[i]=0;
        } else {
          shade_offset[i]= 31*(i-((SHADE_LOWER_BOUND*width)/200))/
                  (((SHADE_UPPER_BOUND-SHADE_LOWER_BOUND)*width)/200);
	  /* die unteren Bits nochmal genauer aufgeloest */
	  shade_dither[i]= ((4*31*(i-((SHADE_LOWER_BOUND*width)/200))/
			    (((SHADE_UPPER_BOUND-SHADE_LOWER_BOUND)*width)/200))-
			    (4*(int)shade_offset[i]));
/*	  printf("%d(%d) ", shade_offset[i],shade_dither[i]); */
        }
      }
    }
#else
    wall_color[OB_DOOR]=COL_DGREY;
    wall_color[OB_WALL]=COL_YELLOW;
    wall_border=COL_RED;
    floor_color=COL_BROWN;
#endif
    heaven_color=COL_BLACK;

    cross_col1=COL_GREEN;
    cross_col2=COL_DGREEN;
    border_col=COL_MAGENTA;
    inner_col=COL_DMAGENTA;
    needle_col=COL_WHITE;

    map_color[OB_NOTHING]=COL_RED; /* fuer OB_FILLED */
    map_color[OB_DOOR]=COL_DGREY;
    map_color[OB_GLAS]=COL_BLUE;
    map_color[OB_WALL]=COL_YELLOW;

    for(i=0;i<16;i++)
      set_color(i,col_tab[i]);
  }

  /* bei 256 Farben soll es bunt werden */
  if(colors>=256) {
#ifdef ENABLE_RED_GREEN
    /* ab 256 Farben geht Brillen-3D */
    if(red_green_3d)  {
      /* bei 8 Bit pro Pixel haben wir 4 Bit ROT und 4 Bit GRUEN */
      for(i=0;i<16;i++) {
	for(j=0;j<16;j++) {
	  col[0] = i*200/3; /* 0 .. 15 -> 0 .. 1000 */
	  col[1] = j*200/3; /*         -"-          */
	  col[2] = 0;
	  set_color(i*16+j,col);
	}
      }
    } else
#endif
    {
      /* 64 Bodenfarben   (cyan) */
      for(i=0;i<64;i++) {
        col[0]=0; col[1]=col[2]=i*15; set_color(i+COL_VCYAN,col);
      }
      /* 32 Wandfarben    (gelb) */
      /* 32 Tuerfarben    (grau) */
      /* 32 Rahmenfarben  (rot)  */
      for(i=0;i<32;i++) {
        col[0]=col[1]=i*31; col[2]=0; set_color(i+COL_VYELLOW,col);
        col[0]=col[1]=col[2]=i*31;    set_color(i+COL_VGREY,col);
        col[0]=i*31; col[1]=col[2]=0; set_color(i+COL_VRED,col);
      }
    }
  }
}

tm_void init_maze(MAZE *this_maze)
{
  akt_maze = this_maze;
}
