/*
  misc.c

  diverse Routinen, die im Wesentlichen direkt auf
  hardwareabhaengige Teile aufsetzen, selbst aber
  nicht hardwareabhaengig sind

  18.7.95   Till  erste Version
  27.11.96  Chip  Typen auf tm_* umgestellt
*/

#include "ctypes.h"
#include "misc.h"
#include "timer.h"
#include "joystick.h"
#include "net.h"

/* diese Routine baut auf timer.c auf */
tm_void wait_msec(tm_ulong msec)
{
  tm_ulong waitfor=set_wait(msec);
  while(get_ticks()<waitfor);
}

/* diese Routinen bauen auf joystick.c auf */
tm_uword last_joy;
tm_bool  joy_delay;
tm_ulong del_time;

tm_void init_joystick_delay(tm_void)
{
  last_joy=0;
  joy_delay=NO;
  get_joystick_delay();   /* falls schon eine Taste gedrueckt ist */
}

tm_uword get_joystick_delay(tm_void)
{
  tm_uword joy,joy_x;

  /* ggf. ein wenig entprellen */
  if(kind_of_joystick()&JITTER_DEVICE) {
    do {
      joy_x=get_joystick();
      wait_msec(20);
    } while(joy_x!=get_joystick());
  } else
      joy_x=get_joystick();

  if(joy_x!=last_joy) {        /* Joystickzustand hat sich geaendert */
     joy_delay=NO;
     joy=joy_x;
  } else {
    if(!joy_delay) {
      joy_delay=YES;             /* bisher kein delay */
      del_time=set_wait(500);    /* eine halbe Sek verzoegern */
      joy=0;                     /* nix tun */
    } else {
      if(get_ticks()>del_time) {
        joy=joy_x;
        del_time=set_wait(100);  /* 1/10 Sek verzoegern */
      } else
        joy=0;
    }
  }
  last_joy=joy_x;
  return(joy);
}

/* der Zufallsgenerator von Gunter */
/* ein linear feedback shift register */
tm_ulong rand_seed;
tm_void 
srnd(tm_ulong initval) {
  rand_seed=initval;
}

tm_ulong 
rnd(tm_void) {
  rand_seed=(rand_seed<<16)|((((rand_seed<<1)^(rand_seed<<4))>>16)&0xffff);
  return rand_seed;
}

/* auf Joystick/Maus/Tastatur-Taste warten */
tm_void 
wait_4_button(tm_void) {
  tm_uword joy;

  do {
    joy=get_joystick();
    net_handle_messages();
  } while(joy&JOY_FIRE);
  wait_msec(10);
  do {
    joy=get_joystick();
    net_handle_messages();
  } while(!(joy&JOY_FIRE));
}

/* set_wait liefert den Wert den get_ticks in */
/* msec Millisekunden liefern wird            */
tm_ulong set_wait(tm_ulong msec)
{
  return(get_ticks()+msec2ticks(msec)+1);
}
