/*
    MOVE.C

    Spielerbewegung steuern

    ???       Till  erste Versionen
    27.11.96  Chip  Typen auf tm_* umgestellt
    01.03.97  Chip  Fluten wird beim Design gemacht
    02.03.97  Chip  Kollision erstmal 'raus, nur auf min/max begrenzen
    11.05.97  Chip  Kollisionen mit Waenden
    27.12.97  Chip  Kollisionen mit anderen (Waende neu)
*/

#include "ctypes.h"
#include "mazes.h"
#include "maze.h"
#include "misc.h"
#include "audio.h"
#include "net.h"
#include "grafik.h"
#include "gr_prim.h"
#include "move.h"
#include "ninja.h"

#include "console.h"

tm_void init_player(PLAYER_DATA *pl)
{
  tm_uword  x,y;

  do {
    x=rnd() % (akt_maze->x_size-1);
    y=rnd() % (akt_maze->y_size-1);
  } while (akt_maze->data[y*akt_maze->x_size+x] & (OB_FILLED | OB_FLAG));

  pl->x=x*FIELD+FIELD/2;
  pl->y=y*FIELD+FIELD/2;
  pl->rot_z=rnd()&0xffff;

/* nur fuer test... */
  switch ((int)rnd() & 3) {
  case 0:    pl->x += FIELD/4;    break;
  case 1:    pl->x -= FIELD/4;    break;
  default:   break;  }
  switch ((int)rnd() & 3) {
  case 0:    pl->y += FIELD/4;    break;
  case 1:    pl->y -= FIELD/4;    break;
  default:   break;  }
/* ...nur fuer test */

  akt_maze->data[y*akt_maze->x_size+x]|=OB_FLAG;
}

/* ----------------------------------------------------------------- */
/* testen, ob es in der entsprechenden Richtung eine Wand gibt       */

#define wall_x(f_offs)          \
        ((akt_maze->data[f_offs]&(OB_SOLID<<OB_X_SHIFT))!=0)
#define wall_y(f_offs)          \
        ((akt_maze->data[f_offs]&(OB_SOLID<<OB_Y_SHIFT))!=0)

/* ---------------------------------------------------------------- */
/* Abstand zu einer anderen Figur pruefen und begrenzen             */

static
tm_void check_player(tm_word  ox,      tm_word  oy,
                     tm_uword dx,      tm_uword dy,
                     tm_bool  *block_x,tm_bool  *block_y,
                     tm_uword *ddx,    tm_uword *ddy)
{
  if (ox==SMILEY_RAD-2)
  {
    if (oy==SMILEY_RAD-2)
    {
      if (dx>dy)
      {
        *block_y=TRUE;
        if (*ddx>4*SMILEY_RAD-4)  *ddx=4*SMILEY_RAD-4;
      }
      else
      {
        *block_x=TRUE;
        if (*ddy>4*SMILEY_RAD-4)  *ddy=4*SMILEY_RAD-4;
      }
    }
    else if (oy<(tm_word)(SMILEY_RAD-2))
    {
      *block_x=TRUE;
      if (*ddy>oy+3*SMILEY_RAD-2)  *ddy=oy+3*SMILEY_RAD-2;
    }
    else
      *ddy=oy-SMILEY_RAD+2;
  }
  else if (ox<(tm_word)(SMILEY_RAD-2))
  {
    if (oy==SMILEY_RAD-2)
    {
      *block_y=TRUE;
      if (*ddx>ox+3*SMILEY_RAD-2)  *ddx=ox+3*SMILEY_RAD-2;
    }
    else if (oy>(tm_word)(SMILEY_RAD-2))
      *ddy=oy-SMILEY_RAD+2;
#if 0
    else
/* hierher nur, wenn die Spieler ueberlappen! */
      console_puts("check_player(): Ueberlappung gefunden!");
#endif
  }
  else
  {
    *ddx=ox-SMILEY_RAD+2;
    if (oy>(tm_word)(SMILEY_RAD-2))
      *ddy=oy-SMILEY_RAD+2;
  }
}

/* ---------------------------------------------------------------- */
/* naechsten Schritt bei Bedarf kuerzen                             */

static
tm_void limit_move(tm_uword dx,     tm_uword dy,
                   tm_uword *ddx,   tm_uword *ddy)
{
  tm_uword sx,sy;

  sx=*ddx;
  sy=*ddy;

  if (sx>dx)
  {
    if (sy>dy)
    {
      sy=dy;
      sx=dx;
    }
    else
    /* Feldgrenze in y erreicht */
    {
      sx=(tm_uword)(((tm_ulong)sy*dx)/dy);
      if (sx>dx) sx=dx;
    }
  }

  else if (sy>dy)
  /* Feldgrenze in x erreicht */
  {
    sy=(tm_uword)(((tm_ulong)sx*dy)/dx);
    if (sy>dy) sy=dy;
  }

  else if ( ((tm_ulong)sx*dy) < ((tm_ulong)sy*dx) )
  /* Feldgrenze in x eher als in y erreicht */
  {
    sy=(tm_uword)(((tm_ulong)sx*dy)/dx);
  }

  else
  /* Feldgrenze in y eher als in x oder gleich mit x erreicht */
  {
    sx=(tm_uword)(((tm_ulong)sy*dx)/dy);
  }

  *ddx=sx;
  *ddy=sy;
}

/* ------------------------------------------------------ */
/* einen Schritt in den entsprechenden Quandranten machen */

static
tm_void move_pxpy(PLAYER_DATA *pl,tm_uword dx,tm_uword dy)
{
  tm_uword x,y,fx,fy,ddx,ddy;
  tm_word  ox,oy;
  tm_uword f_offs;
  tm_bool  block_x,block_y;
  tm_uword i;

  x=pl->x + SMILEY_RAD;
  y=pl->y + SMILEY_RAD;

  while (dx!=0 || dy!=0)
  {
    ox=x&(FIELD-1);  oy=y&(FIELD-1);   /* aktueller Feldoffset */
    fx=x/FIELD;      fy=y/FIELD;       /* aktueller Feldindex */
    f_offs=fy*akt_maze->x_size+fx;

    ddx=FIELD-ox;
    ddy=FIELD-oy;                      /* maximale Schrittkoord. */

    block_x=(dx==0); block_y=(dy==0);

    if (ox==0)
    {
      if (oy==0)
      {
        if (wall_y(f_offs-akt_maze->x_size)) block_x=TRUE;
        if (wall_x(f_offs-1))                block_y=TRUE;
        if (dx>=dy)
        {
          if (!block_x &&
              (wall_x(f_offs) ||
               wall_y(f_offs)))  block_y=TRUE;
        }
        else
        {
          if (!block_y &&
              (wall_x(f_offs) ||
               wall_y(f_offs)))  block_x=TRUE;
        }
        if (block_x)   ddy=2*SMILEY_RAD;
        if (block_y)   ddx=2*SMILEY_RAD;
      }
      else /* oy!=0 */
      {
        if (wall_y(f_offs))   block_x=TRUE;
        if (oy<2*SMILEY_RAD)
        {
          if (wall_y(f_offs-akt_maze->x_size) ||
              wall_x(f_offs))   block_x=TRUE;
          if (block_x)   ddy=2*SMILEY_RAD-oy;
        }
      }
    }
    else if (oy==0) /* ox!=0 */
    {
      if (wall_x(f_offs))   block_y=TRUE;
      if (ox<2*SMILEY_RAD)
      {
        if (wall_x(f_offs-1) ||
            wall_y(f_offs))   block_y=TRUE;
        if (block_y)   ddx=2*SMILEY_RAD-ox;
      }
    }

    if (block_x && block_y)  break;

    for (i=0; i<players; i++)
    {
      if (pl==&(player[i]))  continue;
      ox=player[i].x-x;
      oy=player[i].y-y;
      if (ox<=-(tm_word)(3*SMILEY_RAD-2)  ||
          oy<=-(tm_word)(3*SMILEY_RAD-2)  ||
          ox>=(tm_word)(ddx+SMILEY_RAD-2) ||
          oy>=(tm_word)(ddy+SMILEY_RAD-2)) continue;
      check_player(ox,oy,dx,dy,&block_x,&block_y,&ddx,&ddy);
      if (block_x && block_y)  break;
    }

    if (block_x && block_y)  break;

    limit_move(dx,dy,&ddx,&ddy);

    if (!block_x) x+=ddx;
    if (!block_y) y+=ddy;
    dx-=ddx;
    dy-=ddy;
  }

  pl->x=x - SMILEY_RAD;
  pl->y=y - SMILEY_RAD;
}

static
tm_void move_pxmy(PLAYER_DATA *pl,tm_uword dx,tm_uword dy)
{
  tm_uword x,y,fx,fy,ddx,ddy;
  tm_word  ox,oy;
  tm_uword f_offs;
  tm_bool  block_x,block_y;
  tm_uword i;

  x=pl->x + SMILEY_RAD;
  y=pl->y - SMILEY_RAD;

  while (dx!=0 || dy!=0)
  {
    ox=x&(FIELD-1);  oy=y&(FIELD-1);   /* aktueller Feldoffset */
    fx=x/FIELD;      fy=y/FIELD;       /* aktueller Feldindex */
    f_offs=fy*akt_maze->x_size+fx;

    ddx=FIELD-ox;
    ddy=(oy==0)?FIELD:oy;              /* maximale Schrittkoord. */

    block_x=(dx==0); block_y=FALSE;

    if (ox==0)
    {
      if (oy==0)
      {
        if (wall_y(f_offs))    block_x=TRUE;
        if (wall_x(f_offs-1))  block_y=TRUE;
        if (dx>=dy)
        {
          if (!block_x &&
              (wall_x(f_offs-akt_maze->x_size) ||
               wall_y(f_offs-akt_maze->x_size)))  block_y=TRUE;
        }
        else
        {
          if (!block_y &&
              (wall_x(f_offs-akt_maze->x_size) ||
               wall_y(f_offs-akt_maze->x_size)))  block_x=TRUE;
        }
        if (block_x)   ddy=2*SMILEY_RAD;
        if (block_y)   ddx=2*SMILEY_RAD;
      }
      else /* oy!=0 */
      {
        if (wall_y(f_offs))   block_x=TRUE;
        if (oy>=FIELD-2*SMILEY_RAD)
        {
          if (wall_y(f_offs+akt_maze->x_size) ||
              wall_x(f_offs+akt_maze->x_size))   block_x=TRUE;
          if (block_x)   ddy=oy+2*SMILEY_RAD-FIELD+1;
        }
      }
    }
    else if (oy==0) /* ox!=0 */
    {
      if (wall_x(f_offs))   block_y=TRUE;
      if (ox<2*SMILEY_RAD)
      {
        if (wall_x(f_offs-1) ||
            wall_y(f_offs-akt_maze->x_size))   block_y=TRUE;
        if (block_y)   ddx=2*SMILEY_RAD-ox;
      }
    }

    if (block_x && block_y)  break;

    for (i=0; i<players; i++)
    {
      if (pl==&(player[i]))  continue;
      ox=player[i].x-x;
      oy=y-player[i].y;
      if (ox<=-(tm_word)(3*SMILEY_RAD-2)  ||
          oy<=-(tm_word)(3*SMILEY_RAD-2)  ||
          ox>=(tm_word)(ddx+SMILEY_RAD-2) ||
          oy>=(tm_word)(ddy+SMILEY_RAD-2)) continue;
      check_player(ox,oy,dx,dy,&block_x,&block_y,&ddx,&ddy);
      if (block_x && block_y)  break;
    }

    if (block_x && block_y)  break;

    limit_move(dx,dy,&ddx,&ddy);

    if (!block_x) x+=ddx;
    if (!block_y) y-=ddy;
    dx-=ddx;
    dy-=ddy;
  }

  pl->x=x - SMILEY_RAD;
  pl->y=y + SMILEY_RAD;
}

static
tm_void move_mxpy(PLAYER_DATA *pl,tm_uword dx,tm_uword dy)
{
  tm_uword x,y,fx,fy,ddx,ddy;
  tm_word  ox,oy;
  tm_uword f_offs;
  tm_bool  block_x,block_y;
  tm_uword i;

  x=pl->x - SMILEY_RAD;
  y=pl->y + SMILEY_RAD;

  while (dx!=0 || dy!=0)
  {
    ox=x&(FIELD-1);  oy=y&(FIELD-1);   /* aktueller Feldoffset */
    fx=x/FIELD;      fy=y/FIELD;       /* aktueller Feldindex */
    f_offs=fy*akt_maze->x_size+fx;

    ddx=(ox==0)?FIELD:ox;
    ddy=FIELD-oy;                      /* maximale Schrittkoord. */

    block_x=FALSE; block_y=(dy==0);

    if (ox==0)
    {
      if (oy==0)
      {
        if (wall_y(f_offs-akt_maze->x_size)) block_x=TRUE;
        if (wall_x(f_offs))                  block_y=TRUE;
        if (dx>=dy)
        {
          if (!block_x &&
              (wall_x(f_offs-1) ||
               wall_y(f_offs)))  block_y=TRUE;
        }
        else
        {
          if (!block_y &&
              (wall_x(f_offs-1) ||
               wall_y(f_offs)))  block_x=TRUE;
        }
        if (block_x)   ddy=2*SMILEY_RAD;
        if (block_y)   ddx=2*SMILEY_RAD;
      }
      else /* oy!=0 */
      {
        if (wall_y(f_offs))   block_x=TRUE;
        if (oy<2*SMILEY_RAD)
        {
          if (wall_y(f_offs-akt_maze->x_size) ||
              wall_x(f_offs-1))   block_x=TRUE;
          if (block_x)   ddy=2*SMILEY_RAD-oy;
        }
      }
    }
    else if (oy==0) /* ox!=0 */
    {
      if (wall_x(f_offs))   block_y=TRUE;
      if (ox>=FIELD-2*SMILEY_RAD)
      {
        if (wall_x(f_offs+1) ||
            wall_y(f_offs+1))   block_y=TRUE;
        if (block_y)   ddx=ox+2*SMILEY_RAD-FIELD+1;
      }
    }

    if (block_x && block_y)  break;

    for (i=0; i<players; i++)
    {
      if (pl==&(player[i]))  continue;
      ox=x-player[i].x;
      oy=player[i].y-y;
      if (ox<=-(tm_word)(3*SMILEY_RAD-2)  ||
          oy<=-(tm_word)(3*SMILEY_RAD-2)  ||
          ox>=(tm_word)(ddx+SMILEY_RAD-2) ||
          oy>=(tm_word)(ddy+SMILEY_RAD-2)) continue;
      check_player(ox,oy,dx,dy,&block_x,&block_y,&ddx,&ddy);
      if (block_x && block_y)  break;
    }

    if (block_x && block_y)  break;

    limit_move(dx,dy,&ddx,&ddy);

    if (!block_x) x-=ddx;
    if (!block_y) y+=ddy;
    dx-=ddx;
    dy-=ddy;
  }

  pl->x=x + SMILEY_RAD;
  pl->y=y - SMILEY_RAD;
}

static
tm_void move_mxmy(PLAYER_DATA *pl,tm_uword dx,tm_uword dy)
{
  tm_uword x,y,fx,fy,ddx,ddy;
  tm_word  ox,oy;
  tm_uword f_offs;
  tm_bool  block_x,block_y;
  tm_uword i;

  x=pl->x - SMILEY_RAD;
  y=pl->y - SMILEY_RAD;

  while (dx!=0 || dy!=0)
  {
    ox=x&(FIELD-1);  oy=y&(FIELD-1);   /* aktueller Feldoffset */
    fx=x/FIELD;      fy=y/FIELD;       /* aktueller Feldindex */
    f_offs=fy*akt_maze->x_size+fx;

    ddx=(ox==0)?FIELD:ox;
    ddy=(oy==0)?FIELD:oy;              /* maximale Schrittkoord. */

    block_x=FALSE; block_y=FALSE;

    if (ox==0)
    {
      if (oy==0)
      {
        if (wall_y(f_offs))  block_x=TRUE;
        if (wall_x(f_offs))  block_y=TRUE;
        if (dx>=dy)
        {
          if (!block_x &&
              (wall_x(f_offs-1) ||
               wall_y(f_offs-akt_maze->x_size)))  block_y=TRUE;
        }
        else
        {
          if (!block_y &&
              (wall_x(f_offs-1) ||
               wall_y(f_offs-akt_maze->x_size)))  block_x=TRUE;
        }
        if (block_x)   ddy=2*SMILEY_RAD;
        if (block_y)   ddx=2*SMILEY_RAD;
      }
      else /* oy!=0 */
      {
        if (wall_y(f_offs))   block_x=TRUE;
        if (oy>=FIELD-2*SMILEY_RAD)
        {
          if (wall_y(f_offs+akt_maze->x_size) ||
              wall_x(f_offs+akt_maze->x_size-1))   block_x=TRUE;
          if (block_x)   ddy=oy+2*SMILEY_RAD-FIELD+1;
        }
      }
    }
    else if (oy==0) /* ox!=0 */
    {
      if (wall_x(f_offs))   block_y=TRUE;
      if (ox>=FIELD-2*SMILEY_RAD)
      {
        if (wall_x(f_offs+1) ||
            wall_y(f_offs-akt_maze->x_size+1))   block_y=TRUE;
        if (block_y)   ddx=ox+2*SMILEY_RAD-FIELD+1;
      }
    }

    if (block_x && block_y)  break;

    for (i=0; i<players; i++)
    {
      if (pl==&(player[i]))  continue;
      ox=x-player[i].x;
      oy=y-player[i].y;
      if (ox<=-(tm_word)(3*SMILEY_RAD-2)  ||
          oy<=-(tm_word)(3*SMILEY_RAD-2)  ||
          ox>=(tm_word)(ddx+SMILEY_RAD-2) ||
          oy>=(tm_word)(ddy+SMILEY_RAD-2)) continue;
      check_player(ox,oy,dx,dy,&block_x,&block_y,&ddx,&ddy);
      if (block_x && block_y)  break;
    }

    if (block_x && block_y)  break;

    limit_move(dx,dy,&ddx,&ddy);

    if (!block_x) x-=ddx;
    if (!block_y) y-=ddy;
    dx-=ddx;
    dy-=ddy;
  }

  pl->x=x + SMILEY_RAD;
  pl->y=y + SMILEY_RAD;
}

/* -------------------------------------------------------------- */
/* Spieler bewegen, falls er das kann ;-)                         */

tm_void 
move_player(PLAYER_DATA *pl, tm_word forward, tm_word rot)
{
  tm_word dx,dy;

  pl->rot_z+=rot;  /* Spieler drehen */

  dx=(tm_word)(((tm_long)forward*sintab[pl->rot_z>>4])>>15);
  dy=(tm_word)(((tm_long)forward*costab[pl->rot_z>>4])>>15);

  if (dx >= 0)
    if (dy >= 0)
      move_pxpy(pl,+dx,+dy);
    else
      move_pxmy(pl,+dx,-dy);
  else
    if (dy > 0)
      move_mxpy(pl,-dx,+dy);
    else
      move_mxmy(pl,-dx,-dy);
}
