/*
  fft.c - part of TheMaze music converter

  (c) 1996 Till Harbaum
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "fft.h"

#define	TWO_PI	((double)2.0 * M_PI)
#define noDEBUG

int 
permute(int index, int power) {
  int ib=0,n1=power;
  while(n1--) { ib=(ib<<1)+(index&1); index>>=1; }
  return(ib);
}

double 
magnitude(int n, double *real, double *imag, int power) {
  n = permute(n, power);
  return (sqrt(real[n] * real[n] + imag[n] * imag[n]));
}

double
max_amp(int samples, int power, double *real, double *imag)
{
  double max;
  int	loop;

  max = (double)0.0;
  for (loop = 0; loop < samples; loop++)  {
    if ( magnitude(loop, real, imag, power) > max)
      max  = magnitude(loop,real,imag, power);
  }
  return(max);
}

void 
fft(int samples, int power, double *real, double *imag)
{
  unsigned i1, i2, i3, i4, y;
  int	 loop, loop1, loop2;
  double	 a1, a2, b1, b2, z1, z2, v;
  
  /* Scale the data */
  
  for (loop = 0; loop < samples; loop++)  {
    real[loop] /= (double)samples;
    imag[loop] /= (double)samples;
  }
  
  i1 = samples >> 1;
  i2 = 1;
  v = TWO_PI * ((double)1.0 / (double)samples);
  
  for (loop = 0; loop < power; loop++)  {
    i3 = 0;
    i4 = i1;
    
    for (loop1 = 0; loop1 < i2; loop1++)  {
      y = permute(i3 / i1, power);
      z1 =  cos(v * y);
      z2 = -sin(v * y);
      
      for (loop2 = i3; loop2 < i4; loop2++)  {
	a1 = real[loop2];
	a2 = imag[loop2];
	
	b1 = z1*real[loop2+i1] - z2*imag[loop2+i1];
	b2 = z2*real[loop2+i1] + z1*imag[loop2+i1];
	
	real[loop2]      = a1 + b1;
	imag[loop2]      = a2 + b2;
	
	real[loop2 + i1] = a1 - b1;
	imag[loop2 + i1] = a2 - b2;
      }
      
      i3 += (i1 << 1);
      i4 += (i1 << 1);
    }
    
    i1 >>= 1;
    i2 <<= 1;
  }
}

/* naechstgroessere Zweierpotenz der Samplezahl bilden */
int 
next_pow(int m, int *p) {
  int i=0;
  *p=0;
  do {
    if(m&1) i++;
    m>>=1;
    (*p)++;

  } while(m!=0);

  if(i<2) (*p)--;
  return(1<<*p);
}

double
base_frequency(char *filename, double freq)
{
  FILE   *file;
  long   len;
  short  *inb,*k;
  double step,t;
  int    i,j;
  int    samples, power;
  double *real, *imag, max, max_f;

  /* .sw-Datei laden und volle Zweierpotenz umbasteln */
  if((file = fopen(filename,"rb"))==NULL) {
    fprintf(stderr,"Error opening input file '%s'\n", filename);
    return(-1);
  }

  /* Laenge der Datei ermitteln (stat() geht nicht auf Atari...) */
  fseek(file,0,2);    /* ans Ende der Datei */
  len=ftell(file);
  fseek(file,0,0);    /* und wieder an den Anfang der Datei */
  
#ifdef DEBUG
  printf("converting '%s' with %ld samples (%f sec)\n",filename,len/2,(len/2.0)/freq);
#endif
  samples=next_pow(len/2,&power);

#ifdef DEBUG
  printf("expanding to 2^%d=%d samples\n",power,samples);
#endif

  /* Datei laden */
  k=inb=malloc(len);

#if 0
  fread(inb,2,len/2,file);        /* this is endian dependent */
#else
  while((i=fgetc(file))!=EOF) {   /* this not ... */
    *k++=fgetc(file)+256*i;
  }
#endif

  fclose(file);

  /* in zweipotenzigen Double-Puffer wandeln */
  step=(double)(len/2)/(double)samples;
  if(((real = (double *)malloc(samples*sizeof(double)))==NULL)||
     ((imag = (double *)malloc(samples*sizeof(double)))==NULL)) {
    fprintf(stderr,"Out of memory\n");
    return(-1);
  }

  /* samples skalieren */
  for(i=0;i<samples;i++) {
    real[i]=((double)inb[(int)(i*step)])/32768.0;
    imag[i]=0.0;
  }

#ifdef DEBUG
  printf("Maxamp=%f\n",max_amp(samples, power, real, imag));
#endif

  /* Forwaerts-FFT */
  fft(samples, power, real, imag);

  /* Frequenz mit maximalem Anteil suchen */
  max = 0.0;
  for (i = 0; i < samples/2; i++)  {
    if ( magnitude(i, real, imag, power) > max) { 
      max  = magnitude(i,real,imag,power); max_f=((double)i/ ((double)samples/freq) /step); }
  }

#ifdef DEBUG
  printf("base frequency: %f Hz\n",max_f);
#endif

  /* den brauchen wir nicht mehr */
  free(inb);
  free(imag);
  free(real);

  return(max_f);
}


/* EOF */
