/*
  mus2c

  (c) 1996 Till Harbaum

 */


#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "fft.h"
#include "machine.h"

#ifdef __MSDOS__
#define PREFIX "music\\"
#else
#define PREFIX
#endif

#define COMMENT ';'
#define MAX_INSTR 8    /* maximum 8 instruments */

#define noDEBUG

#define is_id(a)    (isalpha(a)||isdigit(a)||(a=='_')||(a=='.')||(a=='-')||(a=='#'))
#define is_white(a) ((a=='\t')||(a==' '))

const  char  notes[25][2]={"c","c#","d","d#","e","f","f#","g","g#","a","a#","h",
                           ".","db",".","eb",".",".","gb",".","ab",".","b",".","p"}; 
const  double trans[12]={65.4,69.3,73.44,77.76,82.4,87.3,92.5,98,102.84,110,116.5,123.44};
double freq[MAX_INSTR];  /* per fft ermittelte Hauptfrequenz der Instrumente */
int    counter[MAX_INSTR],sample_len[MAX_INSTR],sample_index[MAX_INSTR];
int    last_cnt[MAX_INSTR];
long   tone_factor[MAX_INSTR][10][12];  /* instr/octave/tone */

int
main(int argc, char *argv[])
{
  FILE *ifile,*ofile,*tfile,*hfile;
  char filename[128],line[128],instr[32],s[8];
  int  pos,instruments=0,i,column,j,k,h,len,vol,wert,wi,lin,c;
  int  beatlen,max_sample_len,sfreq,rc;
  short m;

  if(argc!=3) {
    fprintf(stderr,"usage: %s infile output_rate\n",argv[0]);
    exit(1);
  }

  /* get play back frequency */
  sfreq=atoi(argv[2]);

  if((ifile=fopen(argv[1],"r"))==NULL) {
    fprintf(stderr,"error opening input file '%s'\n",argv[1]);
    exit(1);
  }

  if((ofile=fopen("the_song.c","w"))==NULL) {
    fprintf(stderr,"error opening output file 'the_song.c'\n");
    fclose(ifile);
    exit(1);
  }

  if((hfile=fopen("the_song.h","w"))==NULL) {
    fprintf(stderr,"error opening output file 'the_song.h'\n");
    fclose(ofile);
    fclose(ifile);
    exit(1);
  }

  fprintf(ofile,"/* the_song.c - automatically generated from %s by %s */\n\n",argv[1],argv[0]);
  fprintf(hfile,"/* the_song.h - automatically generated from %s by %s */\n\n",argv[1],argv[0]);
    
  /* no sound support */
  if(sfreq==0) {
    fprintf(ofile,"/* this file is intentionally left blank */\n"
	          "/* (due to lack of sound support for this system) */\n");
    fprintf(hfile,"/* this file is intentionally left blank */\n"
                  "/* (due to lack of sound support for this system) */\n");

    fclose(hfile);
    fclose(ofile);
    fclose(ifile);
    exit(0);
  }

  fprintf(ofile,"#include \"ctypes.h\"\n#include \"music.h\"\n");

  /* clear tone array */
  for(i=0;i<MAX_INSTR;i++)
    for(k=0;k<10;k++)
      for(h=0;h<12;h++)
	tone_factor[i][k][h]=0;

  beatlen=32; lin=0; max_sample_len=0; rc=0; c=32;
  while(fgets(line,sizeof(line),ifile)!=NULL) {
    for(c=0;c<MAX_INSTR;c++) last_cnt[c]=counter[c];
    pos=0; column=0; c=32; wi=0; lin++;
    while((line[pos]!=0)&&(line[pos]!=COMMENT)&&(line[pos]!='\n')) {

      /* erste Textzeile enthaelt Instrumentenliste */
      if(!instruments) {
        while(is_id(line[pos])) {
	  i=0; while(is_id(line[pos])) instr[i++]=line[pos++];
	  instr[i]='\0';
	  
	  if(isdigit(instr[0])) {
#ifdef DEBUG
	    printf("beat length = %d/4\n",instr[0]-'0');
#endif
	    beatlen=(instr[0]-'0')*8;    /* in n/32 */
	  } else {
	    /* reuse previous instrument */
	    if(instr[0]=='#') {
	      sample_index[rc++]=instruments-1;
	    } else {
	      /* frequenz bestimmen */
	      sprintf(filename,PREFIX "instrmnt%c%s.sw",PATH_SEP,instr);
	      freq[instruments++]=base_frequency(filename, sfreq);
	      sample_index[rc++]=instruments-1;
	      
#ifdef DEBUG
	      printf("instrument%d is '%s' at %f Hz base frequency\n",
		     instruments-1,instr,freq[instruments-1]);
#endif
	      /* Instrument in C-Source uebernehmen */
	      fprintf(ofile,"CONST tm_short mus_sam_%d[]={\n",instruments-1);
	      tfile=fopen(filename,"rb");
	      i=1; while((m=fgetc(tfile))!=EOF) {
		fprintf(ofile,"%s0x%04x%s",(i==1)?"":",",fgetc(tfile)+256*m,(!(i&7))?"\n":"");
		i++;
	      }
	      fclose(tfile);
	      fprintf(ofile,"};\n");
	      sample_len[instruments-1]=i-1;
	    }
	  }
	  /* whitespaces ueberlesen */
	  while(is_white(line[pos])) pos++;
	}
	/* Adressen der Samples */
	fprintf(ofile,"\nCONST tm_short *mus_sample[]={");
	for(i=0;i<rc;i++) {
	  fprintf(ofile,"%smus_sam_%d",(i!=0)?",":"",sample_index[i]);
	}
	/* Laenge der Samples */
	fprintf(ofile,"};\nCONST tm_ushort mus_sample_len[]={");
	for(i=0;i<rc;i++) {
	  fprintf(ofile,"%s%u",(i!=0)?",":"",sample_len[sample_index[i]]);
	  counter[i]=0;
	}
	fprintf(ofile,"};\n\nCONST tm_uchar mus_song[]={%d,\n",beatlen);
      } else {
	/* whitespaces ueberlesen */
	while(is_white(line[pos])) { 
	  column++;
	  if(line[pos++]=='\t')
	    column=((column-1)&~7)+8;
	}

	/* folgt ein sinnvolles Zeichen ? */
	if(is_id(line[pos])) {

	  wi=1; /* diese Zeile enthaelt Musikdaten */
	  j=column;
	  /* Zeichenkette lesen */
	  i=0; while(is_id(line[pos])) { 
	    instr[i++]=tolower(line[pos++]); 
	    column++;
	  };
	  instr[i]='\0';
	  
	  /* Wort muss auf 10ter Spalte beginnen */
	  if((j%10)==0) {
	    j/=10;
	    for(k=-1,i=0;i<25;i++) {
	      if((isdigit(instr[1]))||(instr[1]=='-')) {
		if((instr[0]==notes[i][0])&&
		   (notes[i][1]=='\0'))k=i;
	      } else {
		if((instr[0]==notes[i][0])&&
		   (instr[1]==notes[i][1])) k=i;
	      }
#if 0		
	      if(strncmp(notes[i],instr,strlen(notes[i]))==0)
		k=i;
#endif
	    }

	    if((k==-1)&&(instr[0]=='e')&&(instr[1]=='#'))   k=5;
	    if((k==-1)&&(instr[0]=='h')&&(instr[1]=='#')) { k=0; instr[2]+=1; }

	    if(k==-1) {
	      fprintf(stderr,"line %d: unknown note '%s' for instrument %d\n",lin,instr,j);
	      exit(1);
	    }
	    
	    if(k==24) {
	      if(counter[j]!=0) {
		fprintf(stderr,"line %d: counter overrun for instrument %d in pause\n",lin,j);
		exit(1);
	      }
	      counter[j]=len=atoi(&instr[2]);
	      last_cnt[j]=10000;
	      if(len<c) c=len;
#ifdef DEBUG
	      printf("Pause %d fuer Inst %d\n",len,j);
#endif
	    } else {
	      /* fuehrende Buchstaben ueberlesen */
	      i=0; while(!isdigit(instr[i])) i++;
	      /* Notenwert extrahieren */
	      h=0; while(isdigit(instr[i])) s[h++]=instr[i++]; s[h]='\0';
	      wert=atoi(s);
	      /* Laenge extrahieren */
	      i++; h=0; while(isdigit(instr[i])) s[h++]=instr[i++]; s[h]='\0';
	      if(counter[j]!=0) {
		fprintf(stderr,"line %d: counter overrun for instrument %d\n",lin,j);
		exit(1);
	      }
	      counter[j]=len=atoi(s);
	      last_cnt[j]=10000;
	      if(len<c) c=len;
	      /* ggf. Lautstaerke extrahieren */
	      if(instr[i]=='-') {
		i++; h=0; while(isdigit(instr[i])) s[h++]=instr[i++]; s[h]='\0';
		vol=atoi(s)*255/100;
	      } else
		vol=255;

	      i=65536*((trans[k%12]*pow(2,wert))/freq[sample_index[j]]);
#ifdef DEBUG
	      printf("Note %s%d/L:%.2f/V:%3d%%(%3.2f Hz) fuer Inst %d ->Faktor=%5d -> len=%d\n",
		     notes[k],wert,len/32.0,vol*100/255,
		     (trans[k%12]*pow(2,wert)),j,i,
		     (int)(sample_len[sample_index[j]]/((trans[k%12]*pow(2,wert))/freq[sample_index[j]])));
#endif
	      /* ergebende Samplelaenge berechnen */
	      if((sample_len[sample_index[j]]*65536/i)>max_sample_len)
		max_sample_len=sample_len[sample_index[j]]*65536/i;

	      /* Tondaten ausgeben */
	      fprintf(ofile,"%d,%3d,%3d,%3d,%3d,0x%02x,0x%02x, ",
		      j,(i>>24)&0xff,(i>>16)&0xff,(i>>8)&0xff,i&0xff,vol,len);
	    }
	  } else {
	    fprintf(stderr,"line %d: '%s' not at tab position (%d)\n",lin,instr,j);
	    exit(1);
	  }
	}
      }
    }

    if(wi) {
      for(i=0;i<rc;i++) {
	if(last_cnt[i]<c)
	  c=last_cnt[i];
      }

      /* neue Zaehlerstaende berechnen */
      for(i=0;i<rc;i++) {
	counter[i]-=c;
#ifdef DEBUG
	printf("cnt[%d] = %d\n",i,counter[i]);
#endif
	if(counter[i]<0) {
	  fprintf(stderr,"line %d: counter underrun for instrument %d\n",lin,i);

	  for(i=0;i<rc;i++) {
	    fprintf(stderr,"line %d: counter%d = %d\n",lin,i,counter[i]);
	  }

	  exit(1);
	}	       
      }
      /* Pause ausgeben */
      fprintf(ofile,"MUS_WAIT,%2d,\n",c);
    }
  }

  /* Zaehlerstaende checken */
  for(i=0;i<rc;i++) {
    if(counter[i]!=0) {
      fprintf(stderr,"line %d: asynchronous ending at instrument %d\n",lin,i);
      exit(1);
    }	       
  }

  fprintf(ofile,"MUS_END};\n");
  fprintf(hfile,"#define MUS_INSTRUMENTS %d\n#define MUS_MAX_PRECALC %d\n",
	  rc,max_sample_len);

  /* alle Dateien schliessen */
  fclose(hfile);
  fclose(ifile);
  fclose(ofile);

  return(0);
}


/* EOF */
