/*
  net.c - high level net routines


 */

#include "ctypes.h"
#include "net.h"
#include "grafik.h"
#include "dialog.h"
#include "gr_prim.h"
#include "timer.h"
#include "misc.h"
#include "machine.h"
#include "ninja.h"

#include <stdio.h>
#include <string.h>

#define noNET2_DEBUG

/* host <-> net type conversion */
#ifdef LITTLE_ENDIAN
#define SWAP(a) (((a&0x00ff)<<8)|((a&0xff00)>>8))
#else
#define SWAP(a) (a)
#endif

/* positions of all players */
PLAYER_DATA player[MAX_PLAYERS];
tm_bool game_running=FALSE;
tm_word my_id=-1,players=0;
tm_word net_ok=OK;

tm_short
net_ping(tm_void) {
  tm_char msg[10]={NET_SYNC,0};
  tm_ulong set,diff;

  diff=get_ticks();
  if((net_ok=net_send(msg,2))!=0) {
    set=set_wait(1000);                       /* 1 sec timeout */
    while((!net_stat())&&(get_ticks()<set));  /* auf Antwort warten */
    if(net_stat()) net_recv(msg);
    diff=get_ticks()-diff;
#ifdef NET2_DEBUG
    printf("Delay: %d msec\n",(int)ticks2msec(diff));
#endif
    return(msg[2]+1);
  }
  return(0);
}

tm_void
init_gameplay(tm_void) {
  tm_word i;

  /* first do all other players at default pos */
  for(i=0;i<players;i++)
    if(i!=my_id)
      player[i].x=player[i].y=player[i].rot_z=0;
}

tm_void
exit_gameplay(tm_void) {

}

tm_void
net_handle_messages(tm_void) {
  tm_char   msg[MAX_PKT_LEN];
  tm_ushort splay[3];

  while(net_stat()) {
    net_recv(msg);

    switch(msg[0]) {
    case NET_SYNC:   /* only the master sends syncs */
#ifdef NET2_DEBUG
      printf("NET_SYNC %d\n",msg[2]);
#endif
      if(!i_am_master) my_id=msg[2];
      break;

    case NET_START:
#ifdef NET2_DEBUG
      printf("NET_START\n");
#endif
      if(!game_running) {
	init_gameplay();
	game_running=TRUE;
      }
      break;

    case NET_STOP:
#ifdef NET2_DEBUG
      printf("NET_STOP\n");
#endif
      if(game_running) {
	exit_gameplay();
	game_running=FALSE;
      }
      break;

    case NET_SIZE:
#ifdef NET2_DEBUG
      printf("NET_SIZE %d\n",msg[2]);
#endif
      players=msg[2];
      break;

    case NET_PLR_POS:
#ifdef NET2_DEBUG
      printf("NET_PLR_POS %d\n",msg[1]);
#endif
      if(msg[1]!=my_id) {
	memcpy(splay,&msg[2],sizeof(PLAYER_DATA));
	player[msg[1]].x=SWAP(splay[0]);
	player[msg[1]].y=SWAP(splay[1]);
	player[msg[1]].rot_z=SWAP(splay[2]);
      }
      break;
    }
  }
}

tm_void
net_update_game(tm_void) {
  tm_char msg[40];
  tm_ushort splay[3];

  /* send current game status */
  /* send player position */
  splay[0]=SWAP(player[my_id].x);
  splay[1]=SWAP(player[my_id].y);
  splay[2]=SWAP(player[my_id].rot_z);
  msg[0]=NET_PLR_POS;
  memcpy(&msg[1],splay,sizeof(PLAYER_DATA));
  net_send(msg,sizeof(PLAYER_DATA)+1);

  /* receive current game status */
  net_handle_messages();
}

tm_void
wait_net_start(tm_void) {
  tm_uword joy,mch,del;
  tm_char msg[100];

  /* wait for key release */
  do
    joy=get_joystick();
  while(joy&JOY_FIRE);

  if(i_am_master) {
    my_id=0;

    do {
      mch=net_ping();
      sprintf(msg,"Anzahl Rechner im Ring: %d|Spiel starten mit Feuertaste",mch);
      draw_rect(0,0,width,height,COL_BLACK);
      text_box("Master",msg);
      del=set_wait(100);
      do
	joy=get_joystick();
      while((!(joy&JOY_FIRE))&&(get_ticks()<del));
    } while((!(joy&JOY_FIRE))||((mch==0))&&(net_ok));   /* wait until ring closed */

    /* wait for joystick glitches */
    wait_msec(100);

    /* wait for release */
    do
      joy=get_joystick();
    while(joy&JOY_FIRE);

    if(net_ok) {
      /* send number of players (net size) */
      msg[0]=NET_SIZE;
      msg[1]=mch;
      do {
	net_send(msg,2);    /* send size message */
	wait_msec(100);
	net_handle_messages();
      } while(players==0);

      /* start communication */
      msg[0]=NET_START;
      do {
	net_send(msg,1);    /* send start message */
	wait_msec(100);
	net_handle_messages();
      } while(!game_running);
    } else {
      players=1; /* net is not up, play single player game */
    }
  } else {
    text_box("Slave","warte auf Spielstart ...");
    /* the master does some pings and then a 'start game' */
    while(!game_running) net_handle_messages();
  }

  /* are we alone? lets get some ninjas */
  /* may be we'll let run some ninjas even in the other case */
  if(players==1)
    ninja_count=MAX_PLAYERS-1;
  else
    ninja_count=0;
}
