/*
   net.h

   ???       Till  erste Version
   27.11.96  Chip  Typen auf tm_* umgestellt
*/

#ifndef __NET__
#define __NET__
#include "ctypes.h"
#include "maze.h"
#include "machine.h"

#define NET_IN  0
#define NET_OUT 1

#define MAX_PLAYERS 32   /* 16 players and 16 shots */

/* type of net */
#define NET_NONE     0
#define NET_IP_UDP   1   /* Internet Datagram Connection */
#define NET_IP_TCP   2   /* Internet Stream Connection   */
#define NET_RS232    3   /* 19200 bit/s serial RS232 */
#define NET_CENTR    4   /* parallel printer port */
#define NET_MIDI     5   /* 31250 bit/s serial MIDI */
#define NET_ALINK    6   /* 20 Mbit/s serial transputer link */
#define NET_IPX      7   /* Novell netware ipx driver */
#define NET_NETBIOS  8   /* Microsoft netbios driver */

/* type of packet */
#define NET_LINK     125
#define NET_CONNECT  126 /* connect (only for udp/tcp/ipx/netbios) */
#define NET_DISC     127 /* dosconnect (only -""- ) */
#define NET_ACK      (MAX_PLAYERS+0)
#define NET_SYNC     (MAX_PLAYERS+1) /* synchronisation frame */
#define NET_PLR_POS  (MAX_PLAYERS+2) /* player position */
#define NET_SHT_POS  (MAX_PLAYERS+3) /* shot position */
#define NET_HIT      (MAX_PLAYERS+4) /* player was hit */
#define NET_KILL     (MAX_PLAYERS+5) /* player is dead */
#define NET_NEW      (MAX_PLAYERS+6) /* player is alive (again) */
#define NET_START    (MAX_PLAYERS+7) /* start a session */
#define NET_STOP     (MAX_PLAYERS+8) /* end a session */
#define NET_SIZE     (MAX_PLAYERS+9) /* number of players in current game */

#define ENV_ENTRY_NAME "TMUSERNAME"  /* name for environment */

/* the following is at least for udp/ip */
#define MAX_MESSAGE  1000  /* buffer up to 1000 messages */
#define MAX_PKT_LEN    48  /* max. packet size */
typedef struct {
  tm_uchar type;
  tm_uchar player;
  tm_uword x,y;    /* player/shot position */
} UDP_POS;

/* from net.c */
extern tm_bool init_net(tm_void);
extern tm_void exit_net(tm_void);
extern tm_void net_update_game(tm_void);
extern tm_void net_handle_messages(tm_void);
extern tm_void wait_net_start(tm_void);
#if 0
extern tm_short net_ping(tm_void);
extern tm_void init_gameplay(tm_void);
extern tm_void exit_gameplay(tm_void);
#endif

/* from net driver */
#ifdef I_KNOW_YOUR_NAME
extern tm_char *get_player_name(tm_char *name, tm_word maxlen);
#endif

extern tm_char net_driver_name[];
extern tm_bool i_am_master;
extern tm_word my_id,players;
extern PLAYER_DATA player[MAX_PLAYERS];
extern tm_bool game_running;

extern tm_bool  net_send(tm_char *data, tm_short len);
extern tm_bool  net_stat(tm_void);
extern tm_short net_recv(tm_char *data);

extern tm_void init_net_options(tm_void);
extern tm_void show_net_options(tm_void);
extern tm_bool set_net_option(tm_char *cmd);
#endif /* __NET__ */
