/*
  net_hl.c - high level net routines

  13.09.97  Till  lt. net.txt angefangen
  17.09.97  Till  weitergemacht

*/

#include "ctypes.h"
#include "net.h"
#include "grafik.h"
#include "dialog.h"
#include "gr_prim.h"
#include "timer.h"
#include "misc.h"
#include "machine.h"
#include "ninja.h"

#include <stdio.h>
#include <string.h>

#define  NETHL_DEBUG    /* enable debug options */

#ifdef NETHL_DEBUG
#define DEBUG(a)  console_puts(a)
#else
#define DEBUG(a)
#endif

/* host <-> net type conversion */
#ifdef LITTLE_ENDIAN
#define SWAP(a) (((a&0x00ff)<<8)|((a&0xff00)>>8))
#else
#define SWAP(a) (a)
#endif

/* net state */
tm_word  net_state=NET_SETUP;
tm_ulong net_rnd;
tm_bool  net_exceed=FALSE;

/* player */
PLAYER_DATA player[MAX_PLAYERS];
tm_word my_id=-1,players=0;

#define MY_ID  (my_id<<4)

tm_uchar*
net_check_inport(tm_void) {
  static tm_uchar input_queue[4],check;
  static tm_word  cnt=0,i;
  
  /* data received */
  if(receive_data(&input_queue[cnt])) {
    /* one paket (hopefully) complete? */
    /* paket driver always receives complete paket! */
    if((cnt++==4)||(packet_receiver)) {
      /* check check */
      check = input_queue[0]^input_queue[1]^
              input_queue[2]^input_queue[3];
      check = (check&0x0f)^((check&0xf0)>>4);

      /* check valid? */
      if(check==NET_CHECK) {
        /* it is */
        cnt=0;
        /* success!! */
        return(&input_queue[0]);
      } else { 
        /* nope */
        if(!packet_receiver) { 
          /* try to resync stream driver */
          for(i=0;i<3;i++)
            input_queue[i]=input_queue[i+1];
          cnt--;
        } else {
          /* you just can't resync a packet driver */
          cnt=0;
        } 
#ifdef NETHL_DEBUG
        console_puts("net_hl.c: invalid packet check");
#endif
      }
    }
  }
  return((tm_uchar*)0);  /* nothing received */   
}

/* handle incoming paket, may be called from receiver interrupt */
tm_void 
net_handle_incoming_packet(tm_uchar *byte) {
  tm_ulong i;

  switch(net_state) {

  /* no connection up to now */
  case NET_SETUP:
    /* a setup id packet!! */
    if((byte[0]==SETUP)&&(byte[1]==ID)) {
      if(byte[2]!=255) {
        /* ring not closed -> forward/increase packet */
        send_packet_dir(SETUP,ID,byte[2]+1,0);
      } else {
        /* ring closed -> send random sync and become temporary master */
        net_rnd=rnd();
        send_packet_dir(SETUP,SYNC|(net_rnd&0x7f000)>>12,
                       (net_rnd&0xff0)>>4,(net_rnd&0xf)<<4);
        net_state=NET_TMASTER;
      }
    }
    /* a sync packet */
    if((byte[0]==SETUP)&&(byte[1]&SYNC)) {
      /* ring closed by other machine -> become temporary slave */
      net_state=NET_TSLAVE; 
    }
    
    /* no in this state known packet -> perhaps linked in */
    /* running game?? just repeat!! */
    if((byte[0]!=SETUP)||((byte[1]!=ID)&&(!byte[1]&SYNC))) {
      net_state=NET_PASSIVE;
      send_packet(byte);
    }
    break;

  case NET_PASSIVE:
    /* ring setup?? */
    if((byte[0]==SETUP&&(byte[1]==ID)) {
      if(byte[2]!=255) {
        /* ring not closed -> forward/increase packet */
        send_packet_dir(SETUP,ID,byte[2]+1,0);
        net_state=NET_SETUP;
      } else {
        /* ring closed -> send random sync and become temporary master */
        net_rnd=rnd();
        send_packet_dir(SETUP,SYNC|(net_rnd&0x7f000)>>12,
                       (net_rnd&0xff0)>>4,(net_rnd&0xf)<<4);
        net_state=NET_TMASTER;
      }
    } else {
       /* no ring setup, just repeat */
       send_packet(byte);
    }
    break;

  case NET_TMASTER:
    /* waiting for own sync to return */
    if((byte[0]==SETUP)&&(byte[1]&SYNC)) {
      /* extract random number */
      i=((((tm_ulong)byte[1])&0x7f)<<12)|
        ( ((tm_ulong)byte[2])      << 4)|
        ( ((tm_ulong)byte[3])&0xf0)>> 4);
      
      /* own number?? */
      if(i==net_rnd) {
        /* ok, start enumeration */
        send_packet_dir(SETUP,ENUM,0,0);
        net_state=NET_TMESENT;
      } else {
        /* no, check priority */
        if(i>net_rnd) {
          /* just repeat and become slave */
          send_packet(byte);
          net_state=NET_TSLAVE;
        }
      }
    }
    break;

    /* temporary master enumeration sent */
  case NET_TMESENT:
    /* wait for enum packet */
    if((byte[0]==SETUP)&&(byte[1]==ENUM)) {
      /* max 16 player */
      if(byte[2]<16) {
        my_id=0;  /* temporary master */ 
        net_state=NET_MSELECT;
        net_exceed=FALSE;
      } else {
        /* ring size exceeded */
        send_packet_dir(SETUP,EXCEED,0,0);
        net_state=NET_EXCEED;
        net_exceed=TRUE;
      }
    }
    break;

    /* wait for exceed to return */
  case NET_EXCEED:
    if((byte[0]==SETUP)&&(byte[1]==EXCEED))
      net_state=NET_SETUP;
    break;

  case NET_TSLAVE:
    break;

  default:
#ifdef NETHL_DEBUG
    console_puts("net_hl.c: illegal net state");
#endif
    break;    
  }
}

/* detect if timeout occured */
tm_bool
net_timeout(tm_ulong tout) {
  static tm_ulong timer=0;
  
  if((ticks2msec(get_ticks())-timer)>tout) {
    timer=ticks2msec(get_ticks());
    return(TRUE);
  }
  return(FALSE);
}

tm_word
net_handle_messages(tm_void) {
  tm_uchar *byte;

  /* have to send something?? */
  switch(net_state) {

  /* in disc state ping periodically */
  case NET_SETUP:
    /* ping every 100msec */
    if(net_timeout(100)) {
      send_packet_dir(SETUP,ID,0,0);
      DEBUG("<PING>");
    } break;

  /* temporary master sync every sec until answer */
  case NET_TMASTER:
    /* resend SYNC every sec */
    if(net_timeout(1000)) {
      send_packet_dir(SETUP,SYNC|(net_rnd&0x7f000)>>12,(net_rnd&0xff0)>>4,(net_rnd&0xf)<<4);
      DEBUG("resending sync");
    } break;

  /* temporary master enumeration sent */
  case NET_TMESENT:
    /* resend ENUM every sec */
    if(net_timeout(1000)) {
      send_packet_dir(SETUP,ENUM,0,0);
      DEBUG("resending enumeration request");
    } break;

  /* ring size exeeded */
  case NET_EXCEED:
    /* resend EXCEED every sec */
    if(net_timeout(1000)) {
      send_packet_dir(SETUP,EXCEED,0,0);
      DEBUG("resending exceed message");
    } break;
  
  /* waiting for master request */
  case NET_MSELECT:
    DEBUG("waiting for master request");
    break;

  /* i want to be master */
  case NET_MREQUEST:
    if(net_timeout(1000)) {
      send_packet_dir(SETUP,MREQUEST,0,MY_ID);
      DEBUG("resending master request");
    } break;
  }

  /* are there bytes available?? */
  byte=check_inport();

  /* valid packet received */
  while(byte!=(tm_uchar*)0)
  {
    net_handle_incoming_packet(byte);
    /* are there still bytes available?? */
    byte=check_inport();
  }
}

tm_void
init_gameplay(tm_void) {
  tm_word i;

  /* first do all other players at default pos */
  for(i=0;i<players;i++)
    if(i!=my_id)
      player[i].x=player[i].y=player[i].rot_z=0;
}

tm_void
exit_gameplay(tm_void) {
}

tm_void
net_update_game(tm_void) {
  tm_char msg[40];
  tm_ushort splay[3];

  /* send current game status */
  /* send player position */
  splay[0]=SWAP(player[my_id].x);
  splay[1]=SWAP(player[my_id].y);
  splay[2]=SWAP(player[my_id].rot_z);
  msg[0]=NET_PLR_POS;
  memcpy(&msg[1],splay,sizeof(PLAYER_DATA));
  net_send(msg,sizeof(PLAYER_DATA)+1);

  /* receive current game status */
  net_handle_messages();
}

tm_void
wait_net_start(tm_void) {
  sprintf(msg,"Anzahl Rechner im Ring: %d|Spiel starten mit Feuertaste",mch);
  text_box("Master",msg);

  /* are we alone? lets get some ninjas */
  /* may be we'll let run some ninjas even in the other case */
  if(players==1)
    ninja_count=MAX_PLAYERS-1;
  else
    ninja_count=0;
}
