/*
  ninja.c

  ???       Till  erste Version
  13.11.96  Chip  WIDTH -> width
  27.11.96  Chip  Typen auf tm_* umgestellt
  13.07.97  Chip  nur noch eine Spielerliste
  15.07.97  Chip  Smileys erstmal primitiv sichtbar
*/

#include <stdlib.h>
#include <string.h>

#include "ctypes.h"
#include "grafik.h"
#include "mazes.h"
#include "ninja.h"
#include "move.h"
#include "misc.h"
#include "smiley.h"
#include "drawmaze.h"
#include "machine.h"
#include "gr_prim.h"
#include "net.h"

#define noSTATIC_NINJAS   /* Ninjas don't move */

extern tm_short atantab[];

tm_uword    ninja_count;

static tm_bool  ninja_rot[MAX_PLAYERS];
static tm_uword start_ninja_id; /* first index of ninjas */

tm_bool init_ninjas(tm_void)
{
  tm_word i;
  start_ninja_id = players;

  for(i=0;i<ninja_count;i++) {
    players ++;

    init_player(&player[start_ninja_id+i]);

    /* Ninjas bewegen sich nur im 90 Grad-Raster */
    /* Richtungen: 0x0000 0x4000 0x8000 0xc000   */
    player[i].rot_z = player[i].rot_z & 0xc000;
    ninja_rot[i]=rnd()&1;
  }

  return(OK);
}

tm_void exit_ninjas(tm_void)
{
}

#define TIME_LIST  (5)    /* aus main.c */
#define NINJA_SPEED    (200*TIME_LIST)

tm_void move_ninjas(tm_long time_sum)
{
#ifdef STATIC_NINJAS
  (tm_void) time_sum;
#else
  tm_uword i,ox,oy;

  for(i=start_ninja_id;i<players;i++) {
    if((rnd()&0xffff)>0x1000)
      ninja_rot[i] = rnd()&1;

    ox=player[i].x; oy=player[i].y;

    if((rnd()&0xffff)>0x100) {  /* 97.5 % ein Stueck laufen */
      move_player(&player[i],100*time_sum/NINJA_SPEED,0);
    }

    if((ox==player[i].x)||(oy==player[i].y))
      if(ninja_rot[i])  player[i].rot_z += rnd()&0x2fff;
      else              player[i].rot_z -= rnd()&0x2fff;
  }
#endif
}

#if 0 /* erstmal weg wegen neuer wand-darstellung */

/* achtung, bei monochrom-schirmen noch anzupassen... */
/* die farben koennen sehr gut in init_ninjas() eingestellt werden... */
#if 0
static
tm_uword ninja_col[NINJAS]={COL_YELLOW,COL_DGREEN,COL_DMAGENTA,COL_DRED,
			    COL_BLUE,COL_DCYAN,COL_GREEN,COL_MAGENTA};
#endif

/* achtung, nur noch eine spielerliste! */

tm_void draw_ninjas_players(tm_word x, tm_word y, tm_uword rot)
{
  tm_word i,j,k,cos,sin,rad,korr;
  tm_long xx,yy,offset;
  tm_long dist;
  tm_word dx,dy;
  tm_bool smiley_drawn[NINJAS+MAX_PLAYERS-1];
  PLAYER_DATA smiley_data[NINJAS+MAX_PLAYERS-1];
  tm_uword scol[NINJAS+MAX_PLAYERS-1];
#ifdef THREE_D
  tm_word cos2,sin2;
#endif

  j=0;

  /* add players to list */
  for(i=0;i<players;i++) {
    if(i!=my_id) {
      memcpy(&smiley_data[j],&player[i],sizeof(PLAYER_DATA));
      scol[j++]=ninja_col[i];
    }
  }

  /* add ninjas to list */
  for(i=0;i<ninja_count;i++) {
    memcpy(&smiley_data[j],&ninja[i],sizeof(PLAYER_DATA));
    scol[j++]=ninja_col[i];
  }

#ifdef THREE_D
  if(use_3d) {
    cos  = costab[(rot+ROTATE_3D)>>4];
    sin  = sintab[(rot+ROTATE_3D)>>4];
    cos2 = costab[(rot-ROTATE_3D)>>4];
    sin2 = sintab[(rot-ROTATE_3D)>>4];
  } else {
#endif
    cos = costab[rot>>4];
    sin = sintab[rot>>4];
#ifdef THREE_D
  }
#endif

  /* bisher ist kein Smiley gemalt worden */
  for(j=0;j<(ninja_count+players-1);j++) {
    smiley_drawn[j]=FALSE;
  }

  for(j=0;j<(ninja_count+players-1);j++) {
    /* Smiley mit groesstem Abstand suchen */
    for(dist=0,k=0;k<(ninja_count+players-1);k++) {
      if((!smiley_drawn[k])&&
	 ((((tm_long)(smiley_data[k].x-player[my_id].x)*
	    (tm_long)(smiley_data[k].x-player[my_id].x)+
	    (tm_long)(smiley_data[k].y-player[my_id].y)*
	    (tm_long)(smiley_data[k].y-player[my_id].y)))>dist)) {
	dist=((tm_long)(smiley_data[k].x-player[my_id].x)*
	      (tm_long)(smiley_data[k].x-player[my_id].x)+
	      (tm_long)(smiley_data[k].y-player[my_id].y)*
	      (tm_long)(smiley_data[k].y-player[my_id].y));
	i=k;
      }
    }

    smiley_drawn[i]=TRUE;

#ifdef THREE_D
    if(!use_3d) {
#endif
      /* rotate and translate ninja */
      xx=(tm_long)smiley_data[i].x-x;  
      yy=(tm_long)smiley_data[i].y-y;
      dx=(xx*cos-yy*sin)>>15; 
      dy=(xx*sin+yy*cos)>>15;

      if(dy>0) {
	/* project ninja */
	offset=(SCR_DIST*(tm_long)dx/dy);
	rad=(((tm_long)(SCR_DIST*2)<<6)/dy);

	/* draw ninja */
#ifdef __TOS__
	if(labs(offset)<width) {    
#else
	if(abs(offset)<width) {    
#endif
	  korr=128+64*dx/dy;
	  if(korr<0)   korr=0;
	  if(korr>255) korr=255;
	  korr=atantab[korr];
	  draw_smiley(width/2+offset,dy,rad,(player[my_id].rot_z+korr-smiley_data[i].rot_z),
		      (colors>=16)?scol[i]:BW_DGREY);
	  /* printf("%d %f\n",korr, 32768.0/M_PI*atan((float)dx/(float)dy)); */
	}
      }
#ifdef THREE_D
    } else {
      /* display left eyes smiley */
      /* rotate and translate ninja */
      xx=(tm_long)smiley_data[i].x-x;              
      yy=(tm_long)smiley_data[i].y-y;
      dx=((xx*cos-yy*sin)>>15)+OFFSET_3D; 
      dy=(xx*sin+yy*cos)>>15;

      if(dy>0) {
	/* project ninja */
	offset=(SCR_DIST*(tm_long)dx/dy);
	rad=(((tm_long)(SCR_DIST*3)<<6)/dy);

	/* draw ninja */
	if(abs(offset)<width)
	  korr=128+64*dx/dy;
	  if(korr<0)   korr=0;
	  if(korr>255) korr=255;
	  korr=atantab[korr];
	  draw_smiley_3d(width/2+offset,dy,rad,player[my_id].rot_z+korr-smiley_data[i].rot_z+ROTATE_3D,
			 (colors>=16)?scol[i]:BW_DGREY,LEFT_EYE);
      }

      /* right eye */
      xx=(tm_long)smiley_data[i].x-x;                
      yy=(tm_long)smiley_data[i].y-y;
      dx=((xx*cos2-yy*sin2)>>15)-OFFSET_3D; 
      dy=(xx*sin2+yy*cos2)>>15;

      if(dy>0) {
	/* project ninja */
	offset=(SCR_DIST*(tm_long)dx/dy);
	rad=(((tm_long)(SCR_DIST*3)<<6)/dy);

	/* draw ninja */
	if(abs(offset)<width)
	  korr=128+64*dx/dy;
	  if(korr<0)   korr=0;
	  if(korr>255) korr=255;
	  korr=atantab[korr];
	  draw_smiley_3d(width/2+offset,dy,rad,player[my_id].rot_z+korr-smiley_data[i].rot_z-ROTATE_3D,
			 (colors>=16)?scol[i]:BW_DGREY,RIGHT_EYE);
      }
    }
#endif
  }
}

#endif
