/*
  gif2c.c  
  gif picture to c source converter for TheMaze

  (c) 1996 Till Harbaum
*/

#include <stdio.h>
#include <string.h>

#define LOCAL_MAP       (0x80)
#define INTERLACE       (0x40)

int
main(int argc, char *argv[])
{
  int     i, c, colors, width, height;
  char    sig[6];         /* GIF signature */
  FILE *infp,*outfp;
  char infile[128],outfile[128],buf[16],*name;
  int     pal[256][3];
  long    len=0;

  name=argv[1];
  if(strrchr(argv[1],'/')!=NULL)  name=strrchr(argv[1],'/')+1;
  if(strrchr(argv[1],'\\')!=NULL) name=strrchr(argv[1],'\\')+1;

  if(argc != 2) {
    fprintf(stderr, "Usage:  %s file\n", argv[0]);
    exit(1);
  }

  strcpy(infile, argv[1]);
  strcat(infile, ".gif");
  infp = fopen(infile, "rb");
  if(!infp) {
    fprintf(stderr, "Error opening file %s for input.\n", infile);
    exit(1);
  }
  strcpy(outfile, argv[1]);
  strcat(outfile, ".c");
  outfp = fopen(outfile, "wb");
  if(!outfp) {
    fprintf(stderr, "Error opening file %s for output.\n", outfile);
    exit(1);
  }

  /* check gif signature */
  c = fread(sig, 1, 6, infp);
  if(c!=6 || strncmp(sig, "GIF87a", 6)) {
    fprintf(stderr, "Error in GIF signature.\n");
    exit(1);
  }
  /* get screen descriptor info */

  width  = fgetc(infp); width  += 256*fgetc(infp);
  height = fgetc(infp); height += 256*fgetc(infp);
  
  /* output width, height, start, stop, and depth */  
  printf("Image size = %dx%d\n", width, height);

  /* check color map info */  
  c = fgetc(infp);
  if((c&LOCAL_MAP) == 0) {
    fprintf(stderr, "Oops, I can only handle images with global color maps.\n");
    exit(1);
  }
  colors = 1<<((c&0x07)+1);
  fgetc(infp);
  fgetc(infp);
  
  /* get global color map */
  
  for(i=0; i<colors; i++) {
    pal[i][0] = fgetc(infp);
    pal[i][1] = fgetc(infp);
    pal[i][2] = fgetc(infp);
  }

  /* get image separator */  
  c = fgetc(infp);
  if(c != ',') {
    fprintf(stderr, "Oops, image separator character expected.\n");
    exit(1);
  }

  /* waste left top width and height */  
  for(i=0; i<8; i++)
    fgetc(infp);
  
  /* check for interlace or local color map */  
  c = fgetc(infp);
  if(c & LOCAL_MAP) {
    fprintf(stderr, "Oops, can't handle local color maps.\n");
    exit(1);
  }
  if(c & INTERLACE) {
    fprintf(stderr, "Oops, can't handle interlace. Try converting with Vpic.\n");
    exit(1);
  }

  fprintf(outfp,"/*\n\t%s\n\tautomatically generated from %s by %s\n*/\n\n",outfile,infile,argv[0]);
  fprintf(outfp,"\n#include \"ctypes.h\"\n#include \"gif.h\"\n");

  fprintf(outfp,"tm_color col_%s[]={\n\t",name);
  for(i=0;i<colors;i++)
    fprintf(outfp,"{%3d,%3d,%3d}%s%s",pal[i][0],pal[i][1],pal[i][2],(i==colors-1)?"":", ",((i&3)==3)?"\n\t":"");
  fprintf(outfp,"};\n\n");

  fprintf(outfp,"tm_char data_%s[]={\n\t",name);
  for(i=0;(c=fgetc(infp))!=EOF;) {
    fprintf(outfp,"0x%02x,%s",c&0xff,((i++&7)==7)?"\n\t":"");
    len++;
  }
  fprintf(outfp,"0};\n\n");

  strcpy(buf,name);
  buf[strlen(buf)-4]=0;
  fprintf(outfp,"tm_picture pic_%s={\n",buf);
  fprintf(outfp,"\t%d,%d,\n",width,height);
  fprintf(outfp,"\t%d,col_%s,\n",colors,name);
  fprintf(outfp,"\tdata_%s,%ld};\n/* -- EOF -- */",name,len);

  fclose(infp);
  fclose(outfp);
  return 0;
}
