/*
   img2c.c

   konvertiert eine GEM-IMG-datei
   in c-quelltexte

   ACHTUNG, die daten muessen noch von benutzenden
   programm auf das notwendige format umgesetzt werden
   (tip: bei farbbildern liegen die daten so vor:
    zeile 1 plane 1, zeile 1 plane 2, ...
    zeile 2 plane 1, zeile 2 plane 2, ... usw.)

   13.11.96   chip    quick and dirty als pic2c
   30.11.96   chip    umgesetzt als img2c
*/

#include <stdio.h>
#include <stdlib.h>

typedef unsigned char  uchar;
typedef unsigned short ushort;

struct {
  ushort  version;
  ushort  length;
  ushort  nplanes;		/* Anzahl Farbebenen */
  ushort  patlen;		/* Musterlaenge in Bytes */
  ushort  pixwidth;
  ushort  pixheight;
  ushort  scanwidth;		/* Bildbreite */
  ushort  nlines;		/* Bildhoehe */
  char    x_id[4];
  ushort  col_model;
} img_header;

void read_error(void)
{
  fprintf(stderr,"Lesefehler!\n");
  exit(2);
}

void line_error(void)
{
  fprintf(stderr,"Zeile zu lang!\n");
  exit(2);
}

int main(int argc,char *argv[])
{
  FILE    *s,*dc,*dh;
  ushort  i,j,ncols,bpl,bnr,lnr,nl;
  ushort  (*col_tab)[3];
  uchar   *line,*pat;
  int     b;
  long    end;

  if (argc!=4)
  {
    fputs("Aufruf: IMG2C <img-quelle> <c-ziel> <h-ziel>\n",stderr);
    return(1);
  }

  if ((s=fopen(argv[1],"rb"))==NULL)
  {
    fprintf(stderr,"Kann \"%s\" nicht als Quelle oeffnen!\n",argv[1]);
    return(1);
  }
  if ((dc=fopen(argv[2],"w"))==NULL)
  {
    fprintf(stderr,"Kann \"%s\" nicht als Ziel oeffnen!\n",argv[2]);
    return(1);
  }
  if ((dh=fopen(argv[3],"w"))==NULL)
  {
    fprintf(stderr,"Kann \"%s\" nicht als Ziel oeffnen!\n",argv[3]);
    return(1);
  }

  fprintf(dc,
    "/*\n"
    "   picture %s\n\n"
    "   automatically generated from %s\n"
    "*/\n"
    "#ifdef __X11__\n"
    "#include \"..\\ctypes.h\"\n"
    "#else\n"
    "#include \"ctypes.h\"\n"
    "#endif\n"
    ,argv[2],argv[1]);
  fprintf(dh,
    "/*\n"
    "   picture-header %s\n\n"
    "   automatically generated from %s\n"
    "*/\n"
    ,argv[3],argv[1]);

  /* So, der Header kann gelesen und geprueft werden */
  if (fread(&img_header,sizeof(img_header),(size_t)1,s)!=(size_t)1)
    read_error();

  ncols=1<<img_header.nplanes;
  bpl=(img_header.scanwidth+7)>>3;
  bpl*=img_header.nplanes;
  if (ncols<2 || bpl==0)
  {
    fprintf(stderr,"Kein korrekter Header in \"%s\"!\n",argv[1]);
    return(3);
  }
  col_tab=(ushort (*)[])malloc(ncols*3*sizeof(ushort));
  line=(uchar *)malloc(bpl);
  pat=(uchar *)malloc(img_header.patlen);
  if (col_tab==NULL || line==NULL || pat==NULL)
  {
    perror("Speicher fuer Farb-, Zeilen- oder Musterpuffer");
    return(10);
  }

  /* alter IMG-Header, sehr beliebt bei monochromen Bildern */
  if (img_header.length==8)
  {
    static  img_col[16][3]={
      {1000,1000,1000}, {   0,   0,   0}, {1000,   0,   0}, {   0, 857,   0},
      {   0,   0,1000}, {   0,   0, 714}, { 714, 286,   0}, {   0, 714,   0},
      { 714, 714, 714}, { 286, 286, 286}, {   0,1000,1000}, {   0, 714, 714},
      {1000,   0,1000}, { 714,   0, 714}, { 714, 714,   0}, {1000,1000,   0} };
    col_tab[0][0]=img_col[0][0];
    col_tab[0][1]=img_col[0][1];
    col_tab[0][2]=img_col[0][2];
    col_tab[1][0]=img_col[1][0];
    col_tab[1][1]=img_col[1][1];
    col_tab[1][2]=img_col[1][2];
    for (i=2; i<ncols; i++)
    {
      col_tab[i][0]=img_col[i][0];
      col_tab[i][1]=img_col[i][1];
      col_tab[i][2]=img_col[i][2];
    }
  }
  /* neuer XIMG-Header, enthaelt die Farbtabelle */
  else if (img_header.length==8+3+ncols*3)
  {
    if (img_header.x_id[0]!='X' || img_header.x_id[1]!='I' ||
        img_header.x_id[2]!='M' || img_header.x_id[3]!='G')
    {
      fprintf(stderr,"Kein korrekter Header in \"%s\"!\n",argv[1]);
      return(3);
    }
    if (img_header.col_model!=0)
    {
      fprintf(stderr,"Unterstuetzung nur von RGB-Tabelle...\n");
      return(4);
    }
    if (fread(col_tab,sizeof(ushort)*3,(size_t)ncols,s)!=(size_t)ncols)
      read_error();
  }
  else
  {
    fprintf(stderr,"Kein korrekter Header in \"%s\"!\n",argv[1]);
    return(3);
  }

  /* Farben ablegen */
  fprintf(dc,"tm_uword title_pic_col[][3]={\n");
  for (i=0; i<ncols-1; i++)
    fprintf(dc,"  {%u,%u,%u},\n",col_tab[i][0],col_tab[i][1],col_tab[i][2]);
  fprintf(dc,"  {%u,%u,%u}\n};\n",col_tab[i][0],col_tab[i][1],col_tab[i][2]);
  if (ferror(dc))
  {
    fprintf(stderr,"Schreibfehler!\n");
    return(2);
  }

  /* Binding speichern */
  fprintf(dh,"#define TITLE_PIC_WIDTH %u\n"
             "#define TITLE_PIC_HEIGHT %u\n"
             "#define TITLE_PIC_COLORS %u\n"
             "extern tm_uword title_pic_col[][3];\n"
             "extern tm_ubyte title_pic_data[];\n",
             img_header.scanwidth,img_header.nlines,ncols);
  if (ferror(dh))
  {
    fprintf(stderr,"Schreibfehler!\n");
    return(2);
  }
  fclose(dh);

  /* jetzt die Daten lesen */
  fseek(s,(long)img_header.length*sizeof(ushort),SEEK_SET);
  if (ftell(s)!=(long)img_header.length*sizeof(ushort))
  {
    perror(argv[1]);
    return(2);
  }

  fprintf(dc,"tm_ubyte title_pic_data[]={\n");
  bnr=0;
  nl=1;
  for (lnr=0; lnr<img_header.nlines; )
  {
    switch (b=fgetc(s))
    {
    case EOF:
      read_error();
    case 0x00:
      if ((b=fgetc(s))==EOF)
        read_error();
      if (b==0)
      {
        if ((b=fgetc(s))==EOF)
          read_error();
        if (b!=0xFF)
        {
          fprintf(stderr,"Ungueltiger Code in \"%s\"!\n",argv[1]);
          return(2);
        }
        /* vertical replication */
        if ((b=fgetc(s))==EOF)
          read_error();
        nl=(ushort)b;
      }
      else
      {
        /* pattern run */
        i=(ushort)b;
        for (j=0; j<img_header.patlen; j++)
        {
          if ((b=fgetc(s))==EOF)
            read_error();
          pat[j]=(uchar)b;
        }
        while (i--)
          for (j=0; j<img_header.patlen; j++)
          {
            if (bnr==bpl)
              line_error();
            line[bnr++]=pat[j];
          }
      }
      break;
    case 0x80:
      /* bit string */
      if ((b=fgetc(s))==EOF)
        read_error();
      for (i=(ushort)b; i--; )
      {
        if ((b=fgetc(s))==EOF)
          read_error();
        if (bnr==bpl)
          line_error();
        line[bnr++]=(uchar)b;
      }
      break;
    default:
      /* solid run */
      j = (b&0x80) ? 0xFF : 0x00;
      for (i=b&0x7F; i--; )
      {
        if (bnr==bpl)
          line_error();
        line[bnr++]=(uchar)j;
      }
      break;
    }
    if (bnr==bpl)
    {
      while (nl--)
      {
        if (++lnr>img_header.nlines)
        {
          fprintf(stderr,"Zuviele Zeilen in \"%s\"!\n",argv[1]);
          return(2);
        }
        for (i=0; i<bpl-1; i++)
          if ((i%16)==15)
            fprintf(dc,"0x%X,\n",line[i]);
          else
            fprintf(dc,"0x%X,",line[i]);
        if (lnr==img_header.nlines)
          fprintf(dc,"0x%X };\n",line[i]);
        else
          fprintf(dc,"0x%X,\n",line[i]);
      }
      if (ferror(dc))
      {
        fprintf(stderr,"Schreibfehler!\n");
        return(2);
      }
      bnr=0;
      nl=1;
    }
  }

  end=ftell(s);
  fseek(s,0L,SEEK_END);
  if (ftell(s)!=end)
    fprintf(stderr,"Quelldatei zu lang!\n");

  free(col_tab);
  free(line);
  fclose(s);
  fclose(dc);
  return(0);
}
