/*
  raw2\raw2.c

ACHTUNG, Teile fuer REV_BIT_ORDER vorbereitet, aber nicht fertig...

  Treiber fuer monochrome Displays (Bit 7 ist MSB)

  Farbdefinitionen stehen in "grafik.h"!

  Muss per #include kompiliert werden, dabei muessen
  im einfuegenden Modul vorkommen:

  static tm_ulong screen_base,screen_show;
  tm_ubyte offscreen_buffer[WIDTH*HEIGHT/8];
  #define WIDTH  (...L)
  #define HEIGHT (...L)
  #define USE_CLIB             ...wenn memset() etc. gut sind

  Optional sind:
  #define BIG_ENDIAN

  03.03.96  Chip  erste Version
  05.03.96  Till  erweitert auf ein paar Graustufen
  08.11.96  Chip  weiter bearbeitet
  13.11.96  Chip  off-screen-buffer ausserhalb (wegen titelbild)
  14.11.96  Chip  put_bitmap nur durchsichtig
  17.11.96  Chip  korrektes Schwarz und Weiss
  26.11.96  Chip  put_bitmap mit 16 bit
  27.11.96  Chip  Typen auf tm_* umgestellt
  19.01.97  Chip  verbesserte Funktionen eingebunden, Muster
  26.01.97  Chip  save_screen() korrigiert
  07.03.97  Chip  etwas fehlertoleranter
  16.03.97  Chip  draw_line() und andere koennen jetzt auch muster
  18.09.97  Till  fill_screen rausgeworfen 
  23.01.98  Till  height/width in draw_bitmap ersetzt
*/

/* --- general & miscellaneous --- */

#define  BWIDTH  (WIDTH/8)
#define  SWIDTH  (WIDTH/16)
#define  LWIDTH  (WIDTH/32)

#ifdef BIG_ENDIAN

static CONST tm_ulong r2_l_tab[]={
	0xFFFFFFFFL,0x7FFFFFFFL,0x3FFFFFFFL,0x1FFFFFFFL,
	0x0FFFFFFFL,0x07FFFFFFL,0x03FFFFFFL,0x01FFFFFFL,
	0x00FFFFFFL,0x007FFFFFL,0x003FFFFFL,0x001FFFFFL,
	0x000FFFFFL,0x0007FFFFL,0x0003FFFFL,0x0001FFFFL,
	0x0000FFFFL,0x00007FFFL,0x00003FFFL,0x00001FFFL,
	0x00000FFFL,0x000007FFL,0x000003FFL,0x000001FFL,
	0x000000FFL,0x0000007FL,0x0000003FL,0x0000001FL,
	0x0000000FL,0x00000007L,0x00000003L,0x00000001L
	};
static CONST tm_ulong r2_r_tab[]={
	0x80000000L,0xC0000000L,0xE0000000L,0xF0000000L,
	0xF8000000L,0xFC000000L,0xFE000000L,0xFF000000L,
	0xFF800000L,0xFFC00000L,0xFFE00000L,0xFFF00000L,
	0xFFF80000L,0xFFFC0000L,0xFFFE0000L,0xFFFF0000L,
	0xFFFF8000L,0xFFFFC000L,0xFFFFE000L,0xFFFFF000L,
	0xFFFFF800L,0xFFFFFC00L,0xFFFFFE00L,0xFFFFFF00L,
	0xFFFFFF80L,0xFFFFFFC0L,0xFFFFFFE0L,0xFFFFFFF0L,
	0xFFFFFFF8L,0xFFFFFFFCL,0xFFFFFFFEL,0xFFFFFFFFL
	};

#else	/* BIG_ENDIAN */

static CONST tm_ulong r2_l_tab[]={
	0xFFFFFFFFL,0xFFFFFF7FL,0xFFFFFF3FL,0xFFFFFF1FL,
	0xFFFFFF0FL,0xFFFFFF07L,0xFFFFFF03L,0xFFFFFF01L,
	0xFFFFFF00L,0xFFFF7F00L,0xFFFF3F00L,0xFFFF1F00L,
	0xFFFF0F00L,0xFFFF0700L,0xFFFF0300L,0xFFFF0100L,
	0xFFFF0000L,0xFF7F0000L,0xFF3F0000L,0xFF1F0000L,
	0xFF0F0000L,0xFF070000L,0xFF030000L,0xFF010000L,
	0xFF000000L,0x7F000000L,0x3F000000L,0x1F000000L,
	0x0F000000L,0x07000000L,0x03000000L,0x01000000L
	};
static CONST tm_ulong r2_r_tab[]={
	0x00000080L,0x000000C0L,0x000000E0L,0x000000F0L,
	0x000000F8L,0x000000FCL,0x000000FEL,0x000000FFL,
	0x000080FFL,0x0000C0FFL,0x0000E0FFL,0x0000F0FFL,
	0x0000F8FFL,0x0000FCFFL,0x0000FEFFL,0x0000FFFFL,
	0x0080FFFFL,0x00C0FFFFL,0x00E0FFFFL,0x00F0FFFFL,
	0x00F8FFFFL,0x00FCFFFFL,0x00FEFFFFL,0x00FFFFFFL,
	0x80FFFFFFL,0xC0FFFFFFL,0xE0FFFFFFL,0xF0FFFFFFL,
	0xF8FFFFFFL,0xFCFFFFFFL,0xFEFFFFFFL,0xFFFFFFFFL
	};

#endif /* #ifdef BIG_ENDIAN */

/*
   MUSTER:
   jedes muster besteht aus einem 4-mal wiederholten
   8 x 8 pixelfeld, jeweils fuer h-lines und v-lines.
   die indizes und damit die zu erzielenden effekte sind
   in "grafik.h" definiert
*/

static CONST tm_ulong r2_h_pattern[16][8]={
  {0x00000000L,0x00000000L,0x00000000L,0x00000000L,0x00000000L,0x00000000L,0x00000000L,0x00000000L},
  {0xFFFFFFFFL,0xFFFFFFFFL,0xFFFFFFFFL,0xFFFFFFFFL,0xFFFFFFFFL,0xFFFFFFFFL,0xFFFFFFFFL,0xFFFFFFFFL},
  {0x00000000L,0xAAAAAAAAL,0x00000000L,0xAAAAAAAAL,0x00000000L,0xAAAAAAAAL,0x00000000L,0xAAAAAAAAL},
  {0x55555555L,0xAAAAAAAAL,0x55555555L,0xAAAAAAAAL,0x55555555L,0xAAAAAAAAL,0x55555555L,0xAAAAAAAAL},
  {0xFFFFFFFFL,0xAAAAAAAAL,0xFFFFFFFFL,0xAAAAAAAAL,0xFFFFFFFFL,0xAAAAAAAAL,0xFFFFFFFFL,0xAAAAAAAAL},
  {0x88888888L,0x00000000L,0x22222222L,0x00000000L,0x88888888L,0x00000000L,0x22222222L,0x00000000L},
  {0x77777777L,0xFFFFFFFFL,0xDDDDDDDDL,0xFFFFFFFFL,0x77777777L,0xFFFFFFFFL,0xDDDDDDDDL,0xFFFFFFFFL},
  {0x00000000L,0x88888888L,0x00000000L,0xAAAAAAAAL,0x00000000L,0x88888888L,0x00000000L,0xAAAAAAAAL},
  {0xFFFFFFFFL,0x77777777L,0xFFFFFFFFL,0x55555555L,0xFFFFFFFFL,0x77777777L,0xFFFFFFFFL,0x55555555L},
  {0xFFFFFFFFL,0x88888888L,0x88888888L,0x88888888L,0xFFFFFFFFL,0x88888888L,0x88888888L,0x88888888L},
  {0x88888888L,0x50505050L,0x20202020L,0x50505050L,0x88888888L,0x05050505L,0x02020202L,0x05050505L},
  {0x00000000L,0x77777777L,0x77777777L,0x77777777L,0x00000000L,0x77777777L,0x77777777L,0x77777777L},
  {0x77777777L,0xAFAFAFAFL,0xDFDFDFDFL,0xAFAFAFAFL,0x77777777L,0xFAFAFAFAL,0xFDFDFDFDL,0xFAFAFAFAL},
  {0x0F0F0F0FL,0x0F0F0F0FL,0x0F0F0F0FL,0x0F0F0F0FL,0xF0F0F0F0L,0xF0F0F0F0L,0xF0F0F0F0L,0xF0F0F0F0L},
  {0xFEFEFEFEL,0x7C7C7C7CL,0x38383838L,0x10101010L,0x38383838L,0x7C7C7C7CL,0xFEFEFEFEL,0xFFFFFFFFL},
  {0x38383838L,0x44444444L,0xAAAAAAAAL,0x82828282L,0xBABABABAL,0x44444444L,0x38383838L,0x00000000L} };

static CONST tm_ulong r2_v_pattern[16][8]={
  {0x00000000L,0x00000000L,0x00000000L,0x00000000L,0x00000000L,0x00000000L,0x00000000L,0x00000000L},
  {0xFFFFFFFFL,0xFFFFFFFFL,0xFFFFFFFFL,0xFFFFFFFFL,0xFFFFFFFFL,0xFFFFFFFFL,0xFFFFFFFFL,0xFFFFFFFFL},
  {0xAAAAAAAAL,0x00000000L,0xAAAAAAAAL,0x00000000L,0xAAAAAAAAL,0x00000000L,0xAAAAAAAAL,0x00000000L},
  {0xAAAAAAAAL,0x55555555L,0xAAAAAAAAL,0x55555555L,0xAAAAAAAAL,0x55555555L,0xAAAAAAAAL,0x55555555L},
  {0xFFFFFFFFL,0x55555555L,0xFFFFFFFFL,0x55555555L,0xFFFFFFFFL,0x55555555L,0xFFFFFFFFL,0x55555555L},
  {0x11111111L,0x00000000L,0x44444444L,0x00000000L,0x11111111L,0x00000000L,0x44444444L,0x00000000L},
  {0xEEEEEEEEL,0xFFFFFFFFL,0xBBBBBBBBL,0xFFFFFFFFL,0xEEEEEEEEL,0xFFFFFFFFL,0xBBBBBBBBL,0xFFFFFFFFL},
  {0xAAAAAAAAL,0x00000000L,0x88888888L,0x00000000L,0xAAAAAAAAL,0x00000000L,0x88888888L,0x00000000L},
  {0x55555555L,0xFFFFFFFFL,0x77777777L,0xFFFFFFFFL,0x55555555L,0xFFFFFFFFL,0x77777777L,0xFFFFFFFFL},
  {0xFFFFFFFFL,0x11111111L,0x11111111L,0x11111111L,0xFFFFFFFFL,0x11111111L,0x11111111L,0x11111111L},
  {0x11111111L,0x0A0A0A0AL,0x04040404L,0x0A0A0A0AL,0x11111111L,0xA0A0A0A0L,0x40404040L,0xA0A0A0A0L},
  {0x00000000L,0xEEEEEEEEL,0xEEEEEEEEL,0xEEEEEEEEL,0x00000000L,0xEEEEEEEEL,0xEEEEEEEEL,0xEEEEEEEEL},
  {0xEEEEEEEEL,0xF5F5F5F5L,0xFBFBFBFBL,0xF5F5F5F5L,0xEEEEEEEEL,0x5F5F5F5FL,0xBFBFBFBFL,0x5F5F5F5FL},
  {0xF0F0F0F0L,0xF0F0F0F0L,0xF0F0F0F0L,0xF0F0F0F0L,0x0F0F0F0FL,0x0F0F0F0FL,0x0F0F0F0FL,0x0F0F0F0FL},
  {0xC1C1C1C1L,0xE3E3E3E3L,0xF7F7F7F7L,0xFFFFFFFFL,0xF7F7F7F7L,0xE3E3E3E3L,0xC1C1C1C1L,0x80808080L},
  {0x1C1C1C1CL,0x22222222L,0x55555555L,0x51515151L,0x55555555L,0x22222222L,0x1C1C1C1CL,0x00000000L} };


/* --- pixel (*_PIX) --- */

#ifndef X_R2_SET_PIXEL
#ifdef MULTIMODE
tm_void r2_set_pixel(tm_uword x, tm_uword y, tm_uword col)
#else
tm_void set_pixel(tm_uword x, tm_uword y, tm_uword col)
#endif
{
  tm_ubyte  m;
#ifndef REV_BIT_ORDER
  m=0x80>>(x&7);
#else
  m=0x01<<(x&7);
#endif
  if (r2_h_pattern[col][y&7] & m)
    *(tm_ubyte*)(screen_base+y*BWIDTH+(x>>3))|=m;
  else
    *(tm_ubyte*)(screen_base+y*BWIDTH+(x>>3))&=~m;
}
#endif

#ifndef X_R2_GET_PIXEL
#ifdef MULTIMODE
tm_uword r2_get_pixel(tm_uword x, tm_uword y)
#else
tm_uword get_pixel(tm_uword x, tm_uword y)
#endif
{
  tm_ubyte  col;
  col=*(tm_ubyte*)(screen_base+y*BWIDTH+(x>>3));
  return ( col&(0x01<<(x&7)) ? 1 : 0);
}
#endif

/* --- lines (*_LIN) --- */

#ifndef X_R2_DRAW_LINE
#ifdef MULTIMODE
tm_void r2_draw_line(tm_uword x1, tm_uword y1, tm_uword x2, tm_uword y2, tm_uword col)
#else
tm_void draw_line(tm_uword x1, tm_uword y1, tm_uword x2, tm_uword y2, tm_uword col)
#endif
{
  tm_word   dp,e;
  tm_ubyte  *p,m;
  tm_ubyte  c,cm;

  p=(tm_ubyte *)screen_base;
  c=(tm_ubyte)(r2_h_pattern[col][0]); /* hier muster... */
  cm=0x80;

  if (x1<x2)
  {
    m=0x80>>(x1&7);
    p+=x1>>3;
    x2=x2-x1;
    p+=y1*BWIDTH;
    if (y1<y2)
    {
      y2=y2-y1;
      dp=+BWIDTH;
    }
    else
    {
      y2=y1-y2;
      dp=-BWIDTH;
    }
  }
  else  {
    m=0x80>>(x2&7);
    p+=x2>>3;
    x2=x1-x2;
    p+=y2*BWIDTH;
    if (y1<y2)
    {
      y2=y2-y1;
      dp=-BWIDTH;
    }
    else
    {
      y2=y1-y2;
      dp=+BWIDTH;
    }
  }

  if (x2>=y2)
  {
    x1=x2+1;
    y2+=y2;
    e=y2-x2;
    x2+=x2;
    if (col==BW_WHITE)
      do
      {
        *p|=m;
        if (e>=0) { p+=dp; e-=x2; }
        if ((m>>=1)==0) {  m=0x80; p++; }
        e+=y2;
      }
      while (--x1);
    else if (col==BW_BLACK)
      do
      {
        *p&=~m;
        if (e>=0) { p+=dp; e-=x2; }
        if ((m>>=1)==0) {  m=0x80; p++; }
        e+=y2;
      }
      while (--x1);
    else
      do
      {
        if (c&cm)
          *p|=m;
        else
          *p&=~m;
        if ((cm>>=1)==0) cm=0x80;
        if (e>=0) { p+=dp; e-=x2; }
        if ((m>>=1)==0) {  m=0x80; p++; }
        e+=y2;
      }
      while (--x1);
  }
  else
  {
    x1=y2+1;
    x2+=x2;
    e=x2-y2;
    y2+=y2;
    if (col==BW_WHITE)
      do
      {
        *p|=m;
        if (e>=0)
        {
          if ((m>>=1)==0) { m=0x80; p++; }
          e-=y2;
        }
        p+=dp;
        e+=x2;
      }
      while (--x1);
    else if (col==BW_BLACK)
      do
      {
        *p&=~m;
        if (e>=0)
        {
          if ((m>>=1)==0) { m=0x80; p++; }
          e-=y2;
        }
        p+=dp;
        e+=x2;
      }
      while (--x1);
    else
      do
      {
        if (c&cm)
          *p|=m;
        else
          *p&=~m;
        if ((cm>>=1)==0) cm=0x80;
        if (e>=0)
        {
          if ((m>>=1)==0) { m=0x80; p++; }
          e-=y2;
        }
        p+=dp;
        e+=x2;
      }
      while (--x1);
  }
}
#endif

#ifndef X_R2_DRAW_HLINE
#ifdef MULTIMODE
tm_void r2_draw_hline(tm_uword x, tm_uword y, tm_uword len, tm_uword col)
#else
tm_void draw_hline(tm_uword x, tm_uword y, tm_uword len, tm_uword col)
#endif
{
  tm_ulong  mm;
  tm_ulong  ml,mr;
  tm_ulong  *p;

  if (len==0)
    return;

  mm=r2_h_pattern[col][y&7];
  p=(tm_ulong *)screen_base+y*LWIDTH;
  len+=x-1;
  ml=r2_l_tab[x&0x1F];
  mr=r2_r_tab[len&0x1F];
  x>>=5;
  p+=x;
  len>>=5;
  len-=x;
  if (len==0)
  {
    ml&=mr;
    *p&=~ml;
    *p|=ml&mm;
  }
  else
  {
    *p++=(*p&~ml)|(ml&mm);
    while (--len)
      *p++=mm;
    *p=(*p&~mr)|(mr&mm);
  }
}
#endif

#ifndef X_R2_DRAW_VLINE
#ifdef MULTIMODE
tm_void r2_draw_vline(tm_uword x, tm_uword y, tm_uword len, tm_uword col)
#else
tm_void draw_vline(tm_uword x, tm_uword y, tm_uword len, tm_uword col)
#endif
{
  tm_ubyte  pat,s,r;
  tm_ubyte  *p;

  if (len==0)
    return;

  pat=(tm_uchar)(r2_v_pattern[col][x&7] >> (y&7));
  s=0x80>>(x&7);
  r=~s;
  x>>=3;
  p=(tm_ubyte *)screen_base+x+y*BWIDTH;

  for (;;) {
    if (pat&0x01) *p|=s; else *p&=r;  p+=BWIDTH;  if (--len==0) break;
    if (pat&0x02) *p|=s; else *p&=r;  p+=BWIDTH;  if (--len==0) break;
    if (pat&0x04) *p|=s; else *p&=r;  p+=BWIDTH;  if (--len==0) break;
    if (pat&0x08) *p|=s; else *p&=r;  p+=BWIDTH;  if (--len==0) break;
    if (pat&0x10) *p|=s; else *p&=r;  p+=BWIDTH;  if (--len==0) break;
    if (pat&0x20) *p|=s; else *p&=r;  p+=BWIDTH;  if (--len==0) break;
    if (pat&0x40) *p|=s; else *p&=r;  p+=BWIDTH;  if (--len==0) break;
    if (pat&0x80) *p|=s; else *p&=r;  p+=BWIDTH;  if (--len==0) break;
  }
}
#endif

#ifndef X_R2_DRAW_VTEXLINE
/* draw texturized vertical line */
#ifdef MULTIMODE
tm_void r2_draw_vtexline(tm_uword x, tm_uword y, tm_uword len, tm_uword real_len, 
		   tm_uchar *texture, tm_uword tex_height, tm_uchar *colortab)
#else
tm_void draw_vtexline(tm_uword x, tm_uword y, tm_uword len, tm_uword real_len, 
		   tm_uchar *texture, tm_uword tex_height, tm_uchar *colortab)
#endif
{
  (tm_void)x;
  (tm_void)y;
  (tm_void)len;
  (tm_void)real_len;
  (tm_void)texture;
  (tm_void)tex_height;
  (tm_void)colortab;
  do_error(1,"r2_draw_vtexline does not exist");
}
#endif

#ifndef X_R2_DRAW_QUADVLINE
/* draw four byte width vertical line */ 
#ifdef MULTIMODE
void r2_draw_quadvline(tm_uword x, tm_uword y, tm_uword len, tm_ulong cols)
#else
void draw_quadvline(tm_uword x, tm_uword y, tm_uword len, tm_ulong cols)
#endif
{
  (tm_void)x;
  (tm_void)y;
  (tm_void)len;
  (tm_void)cols;
  do_error(1,"r2_draw_quadvline does not exist");
}
#endif

#ifndef X_R2_DRAW_VDLINE
#ifdef MULTIMODE
void r2_draw_vdline(tm_uword x, tm_uword y, tm_uword len, tm_uword e, tm_uword col1, tm_uword col2)
#else
void draw_vdline(tm_uword x, tm_uword y, tm_uword len, tm_uword e, tm_uword col1, tm_uword col2)
#endif
{
  (tm_void)x;
  (tm_void)y;
  (tm_void)len;
  (tm_void)e;
  (tm_void)col1;
  (tm_void)col2;
  do_error(1,"r2_draw_vdline does not exist");
}
#endif

/* --- arcs (*_CIR) --- */

#define init_bit(x)  (0x80>>((x)&7))
#define shl_bit(x,p) if ((x<<=1)==0) { x=0x01; p--; }
#define shr_bit(x,p) if ((x>>=1)==0) { x=0x80; p++; }

#ifndef X_R2_DRAW_DISC
#ifdef MULTIMODE
tm_void r2_draw_disc(tm_uword x, tm_uword y, tm_uword rad, tm_uword int_col, tm_uword bd_col)
#else
tm_void draw_disc(tm_uword x, tm_uword y, tm_uword rad, tm_uword int_col, tm_uword bd_col)
#endif
{
/* funktioniert erstmal nur voll-schwarz oder voll-weiss */

  tm_ulong upper;
  tm_ubyte *u_run_l,*u_run_r;
  tm_ulong lower;
  tm_ubyte *l_run_l,*l_run_r;
  tm_word  counter=rad-1,line,last=0;
  tm_uword xval=0,yval=rad;
  tm_ubyte u_bit_l,u_bit_r,l_bit_l,l_bit_r;

  upper=(y-rad)*WIDTH+x;
  lower=(y+rad)*WIDTH+x;
  do {
    if(counter<0) {
      counter+=2*--yval;

      upper+=WIDTH;
      lower-=WIDTH;

      u_run_l=(tm_ubyte *)(screen_base+(upper-xval)/8);
      u_run_r=(tm_ubyte *)(screen_base+(upper+xval-1)/8);
      l_run_l=(tm_ubyte *)(screen_base+(lower-xval)/8);
      l_run_r=(tm_ubyte *)(screen_base+(lower+xval-1)/8);

      u_bit_l=init_bit(upper-xval);
      u_bit_r=init_bit(upper+xval-1);
      l_bit_l=init_bit(lower-xval);
      l_bit_r=init_bit(lower+xval-1);

      for(line=0;line<1+xval-last;line++) {
        if (bd_col) {
          *u_run_l |= u_bit_l;
          *l_run_l |= l_bit_l;
          *u_run_r |= u_bit_r;
          *l_run_r |= l_bit_r;
        } else {
          *u_run_l &= ~u_bit_l;
          *l_run_l &= ~l_bit_l;
          *u_run_r &= ~u_bit_r;
          *l_run_r &= ~l_bit_r;
          }
        shr_bit(u_bit_l,u_run_l);
        shr_bit(l_bit_l,l_run_l);
        shl_bit(u_bit_r,u_run_r);
        shl_bit(l_bit_r,l_run_r);
        }
      for(line=0;line<last-1;line++) {
        if (int_col) {
          *u_run_l |= u_bit_l;
          *l_run_l |= l_bit_l;
          *u_run_r |= u_bit_r;
          *l_run_r |= l_bit_r;
        } else {
          *u_run_l &= ~u_bit_l;
          *l_run_l &= ~l_bit_l;
          *u_run_r &= ~u_bit_r;
          *l_run_r &= ~l_bit_r;
          }
        shr_bit(u_bit_l,u_run_l);
        shr_bit(l_bit_l,l_run_l);
        shl_bit(u_bit_r,u_run_r);
        shl_bit(l_bit_r,l_run_r);
        }

      last=xval;
    }
    else
      counter-=2*xval++;
  }
  while(yval>1);

  upper+=WIDTH;
  u_run_l=(tm_ubyte *)(screen_base+(upper-xval)/8);
  u_run_r=(tm_ubyte *)(screen_base+(upper+xval-1)/8);
  u_bit_l=init_bit(upper-xval);
  u_bit_r=init_bit(upper+xval-1);
  for(line=0;line<1+xval-last;line++) {
    if (bd_col) {
      *u_run_l |= u_bit_l;
      *u_run_r |= u_bit_r;
    } else {
      *u_run_l &= ~u_bit_l;
      *u_run_r &= ~u_bit_r;
      }
    shr_bit(u_bit_l,u_run_l);
    shl_bit(u_bit_r,u_run_r);
    }
  for(line=0;line<last-1;line++) {
    if (int_col) {
      *u_run_l |= u_bit_l;
      *u_run_r |= u_bit_r;
    } else {
      *u_run_l &= ~u_bit_l;
      *u_run_r &= ~u_bit_r;
      }
    shr_bit(u_bit_l,u_run_l);
    shl_bit(u_bit_r,u_run_r);
    }
#endif
}

/* --- bit blocks (*_BLT) --- */

#ifndef X_R2_SAVE_SCREEN
/* angezeigtes Bild sichern */
#ifdef MULTIMODE
tm_void r2_save_screen(tm_void)
#else
tm_void save_screen(tm_void)
#endif
{
#ifndef COPY_TO_SCREEN
  tm_ulong *dest=(tm_ulong*)offscreen_buffer,*src=(tm_ulong*)screen_show;
#else
  tm_ulong *dest=(tm_ulong*)offscreen_buffer,*src=(tm_ulong*)screen_base;
#endif

#ifdef USE_CLIB
  memcpy(dest,src,(size_t)(BWIDTH*HEIGHT));
#else
  tm_ulong cnt;
  for(cnt=0;cnt<(LWIDTH*HEIGHT);cnt++)
    *dest++=*src++;
#endif
}
#endif

#ifndef X_R2_RESTORE_SCREEN
/* Wiederherstellen des Bildinhaltes */
#ifdef MULTIMODE
tm_void r2_restore_screen(tm_void)
#else
tm_void restore_screen(tm_void)
#endif
{
  tm_ulong *src=(tm_ulong*)offscreen_buffer,*dest=(tm_ulong*)screen_base;

#ifdef USE_CLIB
  memcpy(dest,src,(size_t)(BWIDTH*HEIGHT));
#else
  tm_ulong cnt;
  for(cnt=0;cnt<(LWIDTH*HEIGHT);cnt++)
    *dest++=*src++;
#endif
}
#endif


#ifndef X_R2_PUT_BITMAP

#ifdef BIG_ENDIAN

#ifdef MULTIMODE
tm_void r2_put_bitmap(tm_uword x, tm_uword y, tm_uword w, tm_uword h,
                      CONST tm_ushort *data, tm_uword col)
#else
tm_void put_bitmap(tm_uword x, tm_uword y, tm_uword w, tm_uword h,
                   CONST tm_ushort *data, tm_uword col)
#endif
{
  tm_ushort  *p1,*p2;
  tm_uword   i,n;
  tm_uword   d,sr,sl,me;

  if (w==0 || h==0)
    return;

  p1=(tm_ushort *)screen_base;
  p1+=y*SWIDTH+(x>>4);
  h++;
  n=(w+15)>>4;
  sr=x&0xF;
  sl=16-sr;
  me=0xFFFFU;
  if (w&0xF)
    me=~(me>>(w&0xF));

  if (col)
  {
    while (--h)
    {
      p2=p1;
      d=0;
      for (i=n; --i; )
      {
        *p2++|=(d<<sl)|(*data>>sr);
        d=*data++;
      }
      *p2|=(d<<sl)|((*data&me)>>sr);
      if ((d=(*data++&me)<<sl)!=0)
        *++p2|=d;
      p1+=SWIDTH;
    }
  }
  else
  {
    while (--h)
    {
      p2=p1;
      d=0;
      for (i=n; --i; )
      {
        *p2++&=~((d<<sl)|(*data>>sr));
        d=*data++;
      }
      *p2&=~((d<<sl)|((*data&me)>>sr));
      if ((d=(*data++&me)<<sl)!=0)
        *++p2&=~d;
      p1+=SWIDTH;
    }
  }
}

#else /* BIG_ENDIAN */

tm_ushort wswap(tm_ushort a) { return( (a<<8)|(a>>8)); }

#ifdef MULTIMODE
tm_void r2_put_bitmap(tm_uword x, tm_uword y, tm_uword w, tm_uword h,
                      CONST tm_ushort *data, tm_uword col)
#else
tm_void put_bitmap(tm_uword x, tm_uword y, tm_uword w, tm_uword h,
                   CONST tm_ushort *data, tm_uword col)
#endif
{
  tm_ushort  *p1,*p2;
  tm_uword   i,n;
  tm_uword   d,sr,sl,me;

  if (w==0 || h==0)
    return;

  p1=(tm_ushort *)screen_base;
  p1+=y*SWIDTH+(x>>4);
  h++;
  n=(w+15)>>4;
  sr=x&0xF;
  sl=16-sr;
  me=0xFFFFU;
  if (w&0xF)
    me=~(me>>(w&0xF));

  if (col)
  {
    while (--h)
    {
      p2=p1;
      d=0;
      for (i=n; --i; )
      {
        *p2++|=wswap((d<<sl)|(*data>>sr));
        d=*data++;
      }
      *p2|=wswap((d<<sl)|((*data&me)>>sr));
      if ((d=(*data++&me)<<sl)!=0)
        *++p2|=wswap(d);
      p1+=SWIDTH;
    }
  }
  else
  {
    while (--h)
    {
      p2=p1;
      d=0;
      for (i=n; --i; )
      {
        *p2++&=~wswap(((d<<sl)|(*data>>sr)));
        d=*data++;
      }
      *p2&=~wswap(((d<<sl)|((*data&me)>>sr)));
      if ((d=(*data++&me)<<sl)!=0)
        *++p2&=~wswap(d);
      p1+=SWIDTH;
    }
  }
}
#endif /* #ifdef BIG_ENDIAN */

#endif /* #ifndef X_R2_PUT_BITMAP */

static
tm_void hline_pattern(tm_uword x, tm_uword y, tm_uword len, tm_ulong pat)
{
  tm_ulong  ml,mr;
  tm_ulong  *p;

  p=(tm_ulong *)screen_base+y*LWIDTH;
  len+=x-1;
  ml=r2_l_tab[x&0x1F];
  mr=r2_r_tab[len&0x1F];
  x>>=5;
  p+=x;
  len>>=5;
  len-=x;
  if (len==0)
  {
    ml&=mr;
    *p&=~ml;
    *p|=ml&pat;
  }
  else
  {
    *p++=(*p&~ml)|(ml&pat);
    while (--len)
      *p++=pat;
    *p=(*p&~mr)|(mr&pat);
  }
}

#ifndef X_R2_DRAW_RECT
#ifdef MULTIMODE
tm_void r2_draw_rect(tm_uword x, tm_uword y, tm_uword w, tm_uword h, tm_uword col)
#else
tm_void draw_rect(tm_uword x, tm_uword y, tm_uword w, tm_uword h, tm_uword col)
#endif
{
  tm_uword  yy=y;
  CONST tm_ulong  *pat;

  pat=r2_h_pattern[col];
  while (h--)
  {
    yy&=7;
    hline_pattern(x,y++,w,pat[yy++]);
  }
}
#endif

#ifdef USE_MALLOC
#ifndef USE_CLIB
#error Missing code for use without USE_CLIB!
#endif

#ifndef X_R2_SAVE_RECT
#ifdef MULTIMODE
tm_void* r2_save_rect(tm_uword x, tm_uword y, tm_uword w, tm_uword h)
#else
tm_void* save_rect(tm_uword x, tm_uword y, tm_uword w, tm_uword h)
#endif
{
  tm_uword i;
  tm_uchar *p,*d,*data;

  /* horizontally round to next byte boundary */
  w+=x&3; x=x/8; if((w&~3)!=w) w=((w&~3)/8)+1;
  
  data = (tm_char*)malloc( w*h + 4 * sizeof(tm_uword));

  if(data==NULL) {
    console_printf("out of memory in r2_save_rect\n");
  } else {
    *((tm_uword*)(data+0*sizeof(tm_uword)))=x;
    *((tm_uword*)(data+1*sizeof(tm_uword)))=y;
    *((tm_uword*)(data+2*sizeof(tm_uword)))=w;
    *((tm_uword*)(data+3*sizeof(tm_uword)))=h;

    p=(tm_uchar*)(screen_base+y*(WIDTH/8)+x);
    d=data+4*sizeof(tm_uword);
    for(i=y;i<(y+h);i++) {
      memcpy(d, p, w);
      d+=w;
      p+=(WIDTH/8);
    }
  }
  return ((void*)data);
}
#endif

#ifndef X_R2_RESTORE_RECT
#ifdef MULTIMODE
tm_void r2_restore_rect(tm_void *data)
#else
tm_void restore_rect(tm_void *data)
#endif
{
  tm_uword i,x,y,w,h;
  tm_uchar *p,*d;

  if(data!=NULL) {
    x=*((tm_uword*)(data+0*sizeof(tm_uword)));
    y=*((tm_uword*)(data+1*sizeof(tm_uword)));
    w=*((tm_uword*)(data+2*sizeof(tm_uword)));
    h=*((tm_uword*)(data+3*sizeof(tm_uword)));

    p=(tm_uchar*)(screen_base+y*(WIDTH/8)+x);
    d=data+4*sizeof(tm_uword);
    for(i=y;i<(y+h);i++) {
      memcpy(p, d, w);
      d+=w;
      p+=(WIDTH/8);
    }
  }
}
#endif

#ifndef X_R2_FREE_RECT
#ifdef MULTIMODE
tm_void r2_free_rect(tm_void *data)
#else
tm_void free_rect(tm_void *data)
#endif
{
  if(data!=NULL)
    free(data);
}
#endif

#endif

/* --- end --- */
