/*
  raw256\r256_blt.c

  Bit-Block-Treiber fuer einfache 256-Farben-Displays!

  10.7.95  Till
  17.9.97  Till fill_screen rausgeworfen
  21.1.98  Till frei skalierbar
*/

#ifndef X_R256_SAVE_SCREEN
/* Bildspeicher sichern */
#ifdef MULTIMODE
void r256_save_screen(void)
#else
void save_screen(void)
#endif
{
  tm_ulong *dest=(tm_ulong*)offscreen_buffer,*src=(tm_ulong*)screen_base;

#ifdef USE_CLIB
  memcpy(dest,src,(size_t)(WIDTH*HEIGHT));
#else
  tm_ulong cnt;
  for(cnt=0;cnt<(WIDTH*HEIGHT/sizeof(tm_ulong));cnt++)
    *dest++=*src++;
#endif
}
#endif

#ifndef X_R256_RESTORE_SCREEN
/* Wiederherstellen des Bildinhaltes */
#ifdef MULTIMODE
void r256_restore_screen(void)
#else
void restore_screen(void)
#endif
{
  tm_ulong *src=(tm_ulong*)offscreen_buffer,*dest=(tm_ulong*)screen_base;

#ifdef USE_CLIB
  memcpy(dest,src,(size_t)(WIDTH*HEIGHT));
#else
  tm_ulong cnt;
  for(cnt=0;cnt<(WIDTH*HEIGHT/sizeof(tm_ulong));cnt++)
    *dest++=*src++;
#endif

}
#endif

#ifndef X_R256_PUT_BITMAP
tm_ushort r256_bit_tab[]={0x8000,0x4000,0x2000,0x1000,0x0800,0x0400,0x0200,0x0100,
			  0x0080,0x0040,0x0020,0x0010,0x0008,0x0004,0x0002,0x0001};

/* negative Farben sind durchsichtig! */
#ifdef MULTIMODE
void r256_put_bitmap(tm_uword x, tm_uword y, tm_uword w, tm_uword h, CONST tm_ushort *data,
		     tm_uword col)
#else
void put_bitmap(tm_uword x, tm_uword y, tm_uword w, tm_uword h, CONST tm_ushort *data,
		     tm_uword col)
#endif
{
  tm_word   wc,hc,sw,byte=0;
  tm_uchar *adr=(tm_uchar*)(screen_base+y*WIDTH+x);
  sw=((w-1)/16)+1;        /* Breite in Words  */
  
  for(hc=0;hc<h;hc++) {     /* Zeilen durchgehen */
    byte=hc*sw;
    for(wc=0;wc<w;wc++) {       /* alle Pixel durchgehen */
      if(data[(wc>>4)+byte] & r256_bit_tab[wc&0x0f])
	  *adr=col;
      adr++;
    }
    adr+=((tm_uword)WIDTH-w);
  }
}
#endif

#ifndef X_R256_DRAW_RECT
#ifdef MULTIMODE
void r256_draw_rect(tm_uword x, tm_uword y, tm_uword w, tm_uword h, tm_uword col)
#else
void draw_rect(tm_uword x, tm_uword y, tm_uword w, tm_uword h, tm_uword col)
#endif
{
  tm_uword yc;
  tm_uchar *adr=(tm_uchar*)(screen_base+y*WIDTH+x);
#ifndef USE_CLIB
  tm_uword  xc;
#endif 

  for(yc=0;yc<h;yc++) {
#ifdef USE_CLIB
    memset(adr,col,(size_t)w);
    adr+=(tm_uword)WIDTH;
#else
    for(xc=0;xc<w;xc++)
      *adr++=col;
    
    adr+=(WIDTH-w);
#endif
  }
}
#endif

#ifdef USE_MALLOC
#ifndef USE_CLIB
#error Missing code for use without USE_CLIB!
#endif

#ifndef X_R256_SAVE_RECT
#ifdef MULTIMODE
tm_void* r256_save_rect(tm_uword x, tm_uword y, tm_uword w, tm_uword h)
#else
tm_void* save_rect(tm_uword x, tm_uword y, tm_uword w, tm_uword h)
#endif
{
  tm_uword i;
  tm_uchar *p,*d,*data=(tm_char*)malloc( w*h + 4 * sizeof(tm_uword));

  if(data==NULL) {
    console_printf("out of memory in r256_save_rect\n");
  } else {
    *((tm_uword*)(data+0*sizeof(tm_uword)))=x;
    *((tm_uword*)(data+1*sizeof(tm_uword)))=y;
    *((tm_uword*)(data+2*sizeof(tm_uword)))=w;
    *((tm_uword*)(data+3*sizeof(tm_uword)))=h;

    p=(tm_uchar*)(screen_base+y*WIDTH+x);
    d=data+4*sizeof(tm_uword);
    for(i=y;i<(y+h);i++) {
      memcpy(d, p, w);
      d+=w;
      p+=WIDTH;
    }
  }
  return ((void*)data);
}
#endif

#ifndef X_R256_RESTORE_RECT
#ifdef MULTIMODE
tm_void r256_restore_rect(tm_void *data)
#else
tm_void restore_rect(tm_void *data)
#endif
{
  tm_uword i,x,y,w,h;
  tm_uchar *p,*d;

  if(data!=NULL) {
    x=*((tm_uword*)(data+0*sizeof(tm_uword)));
    y=*((tm_uword*)(data+1*sizeof(tm_uword)));
    w=*((tm_uword*)(data+2*sizeof(tm_uword)));
    h=*((tm_uword*)(data+3*sizeof(tm_uword)));

    p=(tm_uchar*)(screen_base+y*WIDTH+x);
    d=data+4*sizeof(tm_uword);
    for(i=y;i<(y+h);i++) {
      memcpy(p, d, w);
      d+=w;
      p+=WIDTH;
    }
  }
}
#endif

#ifndef X_R256_FREE_RECT
#ifdef MULTIMODE
tm_void r256_free_rect(tm_void *data)
#else
tm_void free_rect(tm_void *data)
#endif
{
  if(data!=NULL)
    free(data);
}
#endif

#endif
