/*
  raw256\r256_lin.c

  Linientreiber fuer einfache 256-Farben-Displays!

  12.7.95  Till
*/

#ifndef X_R256_DRAW_LINE
#ifdef MULTIMODE
void r256_draw_line(tm_uword x1, tm_uword y1, tm_uword x2, tm_uword y2, tm_uword col)
#else
void draw_line(tm_uword x1, tm_uword y1, tm_uword x2, tm_uword y2, tm_uword col)
#endif
{
  register tm_word  LC,M,dx,dy;
  tm_word  t;
  register tm_uchar *start;

  /* ggf. Koordinaten tauschen */
  /* (immer von links nach rechts zeichnen) */
  if(x2<x1) 
    { t=x2; x2=x1; x1=t; t=y2; y2=y1; y1=t; }

  /* Differenzen errechnen */
  dx=x2-x1;
  dy=y2-y1; if(dy<0) dy=-dy;
  
  start=(tm_uchar*)(screen_base+y1*WIDTH+x1);
  
  if(dx>dy) {  /* Horizontal zeichnen */
    M=dx/2;
    if(y1<=y2) { /* Linie senkt sich */
      for(LC=0;LC<=dx;LC++) {
        if(M>dx) { M-=dx; start+=(tm_uword)WIDTH; }
	M+=dy;
	*start++=col;
      }
    } else { /* Linie steigt */
      for(LC=0;LC<=dx;LC++) {
        if(M>dx) { M-=dx; start-=(tm_uword)WIDTH; }
	M+=dy;
	*start++=col;
      }
    }
  } else {     /* Vertikal zeichnen */
    M=dy/2;

    if(y1>=y2) { /* Linie steigt */
      for(LC=0;LC<=dy;LC++) {
        if(M>dy) { M-=dy; start++; }
	M+=dx;
	*start=col;
	start-=(tm_uword)WIDTH;
      }
    } else { /* Linie senkt sich */
      for(LC=0;LC<=dy;LC++) {
        if(M>dy) { M-=dy; start++; }
        M+=dx;
        *start=col;
        start+=(tm_uword)WIDTH;
      }
    }
  }
}
#endif

#ifndef X_R256_DRAW_LINE
#ifdef MULTIMODE
void r256_draw_hline(tm_uword x, tm_uword y, tm_uword len, tm_uword col)
#else
void draw_hline(tm_uword x, tm_uword y, tm_uword len, tm_uword col)
#endif
{
  tm_uchar *start=(tm_uchar*)(screen_base+y*WIDTH+x);

#ifdef USE_CLIB
  memset(start,col,(size_t)len);
#else
  tm_uword cnt;

  for(cnt=0;cnt<len;cnt++)
    *start++=col;
#endif
}
#endif

#ifndef X_R256_DRAW_VLINE
#ifdef MULTIMODE
void r256_draw_vline(tm_uword x, tm_uword y, tm_uword len, tm_uword col)
#else
void draw_vline(tm_uword x, tm_uword y, tm_uword len, tm_uword col)
#endif
{
  tm_uchar *start=(tm_uchar*)(screen_base+y*WIDTH+x);
  tm_uword cnt;

  for(cnt=0;cnt<len;cnt++) {
    *start=col;
    start+=(tm_uword)WIDTH;
  }
}
#endif

#ifndef X_R256_DRAW_QUADVLINE
/* draw four byte width vertical line */ 
#ifdef MULTIMODE
void r256_draw_quadvline(tm_uword x, tm_uword y, tm_uword len, tm_ulong cols)
#else
void draw_quadvline(tm_uword x, tm_uword y, tm_uword len, tm_ulong cols)
#endif
{
  tm_ulong *start=(tm_ulong*)(screen_base+y*WIDTH+x);
  tm_uword cnt;

  for(cnt=0;cnt<len;cnt++) {
    *start=cols;
    start+=(tm_uword)(WIDTH/4);
  }
}
#endif

#ifndef X_R256_DRAW_VLINE_3D
#ifdef MULTIMODE
void r256_draw_vline_3d(tm_uword x, tm_uword y, tm_uword len, tm_uword col, tm_uword eye)
#else
void draw_vline_3d(tm_uword x, tm_uword y, tm_uword len, tm_uword col, tm_uword eye)
#endif
{
  tm_uchar *start=(tm_uchar*)(screen_base+(y&~1)*WIDTH+x+WIDTH*eye);
  tm_uword cnt;

  for(cnt=0;cnt<(len/2);cnt++) {
    *start=col;
    start+=(tm_uword)(2*WIDTH);
  }
}
#endif

#ifndef X_R256_DRAW_QUADVLINE_3D
/* draw four byte width vertical line */
#ifdef MULTIMODE
void r256_draw_quadvline_3d(tm_uword x, tm_uword y, tm_uword len, tm_ulong cols, tm_uword eye)
#else
void draw_quadvline_3d(tm_uword x, tm_uword y, tm_uword len, tm_ulong cols, tm_uword eye)
#endif
{
  tm_ulong *start=(tm_ulong*)(screen_base+(y&~1)*WIDTH+x+WIDTH*eye);
  tm_uword cnt;

  for(cnt=0;cnt<(len/2);cnt++) {
    *start=cols;
    start+=(tm_uword)(WIDTH/2);
  }
}
#endif

#ifndef X_R256_DRAW_VTEXLINE
/* draw texturized vertical line */
/* Linie wird von der Mitte aus gezeichnet */
/* Musterhoehe muss gerade sein!! */
#ifdef MULTIMODE
void r256_draw_vtexline(tm_uword x, tm_uword y, tm_uword len, tm_uword real_len, 
		   tm_uchar *texture, tm_uword tex_height, tm_uchar *coltab)
#else
void draw_vtexline(tm_uword x, tm_uword y, tm_uword len, tm_uword real_len, 
		   tm_uchar *texture, tm_uword tex_height, tm_uchar *coltab)
#endif
{
  tm_uchar *start1,*start2;
  tm_uword cnt,n1,n2,M;

  start2=WIDTH+(start1=(tm_uchar*)(screen_base+(y+(len>>1))*WIDTH+x));
	
  /* schon wieder ein wenig Bresenham */
  n1=(tex_height>>1)-1; n2=tex_height>>1;
  for(M=0,cnt=0;cnt<(len>>1);cnt++) {
    while(M>real_len) { M-=real_len; n1--; n2++; }  /* This is only needed if the texture is greater than the wall */
    M+=tex_height;
    *start1=coltab[texture[n1]];  *start2=coltab[texture[n2]];
    start1-=(tm_uword)WIDTH;      start2+=(tm_uword)WIDTH;
  }
}
#endif

#ifndef X_R256_DRAW_VDLINE
/* some kind of dithered vertical line (unused) */
#ifdef MULTIMODE
void r256_draw_vdline(tm_uword x, tm_uword y, tm_uword len, tm_uword e, tm_uword col1, tm_uword col2)
#else
void draw_vdline(tm_uword x, tm_uword y, tm_uword len, tm_uword e, tm_uword col1, tm_uword col2)
#endif
{
  tm_uchar *start=(tm_uchar*)(screen_base+y*WIDTH+x);
  tm_uword cnt;

  for(cnt=0;cnt<len;cnt++) {
    if((cnt&3)<=e)
      *start=col1;
    else
      *start=col2;
    start+=(tm_uword)WIDTH;
  }
}
#endif


