/*
   r256wall.c

   Waende malen in allen Varianten

   08.03.97   chip   first version
   10.09.97   Till   umgemodelt auf 256 Farben und shading
   28.10.97   Till   shading mit dithern

  ACHTUNG! Diese Quelle muss hinter r256_xxx.c "included" werden...
  Weitere Bemerkungen siehe dort!
*/

#include "machine.h"
#include "drawmaze.h"
#include "mazes.h"
#include "texture.h"

#ifndef SCR_XOFFS
#define SCR_XOFFS  (width/2)
#define SCR_YOFFS  (pf_height/2)
#endif

#ifdef  WALLS_SHADED
#define SHADE_OFFSET(a) (shade_offset[a])
#else
#define SHADE_OFFSET(a) (0)
#endif

#ifdef WALLS_LINES

#ifndef X_R256_WALL_2
#ifdef MULTIMODE
tm_void r256_wall_2(tm_uword x,tm_uword h,
               tm_uword wall_col,tm_uword border_col)
#else
tm_void wall_2(tm_uword x,tm_uword h,
               tm_uword wall_col,tm_uword border_col)
#endif
{
  draw_vline(x,SCR_YOFFS-h,2*h+1,border_col);
}
#endif

#ifndef X_R256_WALL_4
#ifdef MULTIMODE
tm_void r256_wall_4(tm_uword x1,tm_uword h1,
               tm_uword x2,tm_uword h2,
               tm_uword wall_col,tm_uword border_col)
#else
tm_void wall_4(tm_uword x1,tm_uword h1,
               tm_uword x2,tm_uword h2,
               tm_uword wall_col,tm_uword border_col)
#endif
{
  draw_line(x1,SCR_YOFFS-h1,x2,SCR_YOFFS-h2,wall_col);
  draw_line(x1,SCR_YOFFS+h1,x2,SCR_YOFFS+h2,wall_col);
  draw_vline(x1,SCR_YOFFS-h1,2*h1+1,border_col);
  draw_vline(x2,SCR_YOFFS-h2,2*h2+1,border_col);
}
#endif

#ifndef X_R256_WALL_6
#ifdef MULTIMODE
tm_void r256_wall_6(tm_uword x1,tm_uword h1,
               tm_uword x2,tm_uword h2,
               tm_uword x3,tm_uword h3,
               tm_uword wall_col,tm_uword border_col)
#else
tm_void wall_6(tm_uword x1,tm_uword h1,
               tm_uword x2,tm_uword h2,
               tm_uword x3,tm_uword h3,
               tm_uword wall_col,tm_uword border_col)
#endif
{
  if (h1>h3)
  {
    draw_hline(x1,SCR_YOFFS-h1,x2-x1+1,wall_col);
    draw_hline(x1,SCR_YOFFS+h1,x2-x1+1,wall_col);
    draw_line(x2,SCR_YOFFS-h2,x3,SCR_YOFFS-h3,wall_col);
    draw_line(x2,SCR_YOFFS+h2,x3,SCR_YOFFS+h3,wall_col);
  }
  else /* h1>h3 */
  {
    draw_line(x1,SCR_YOFFS-h1,x2,SCR_YOFFS-h2,wall_col);
    draw_line(x1,SCR_YOFFS+h1,x2,SCR_YOFFS+h2,wall_col);
    draw_hline(x2,SCR_YOFFS-h3,x3-x2+1,wall_col);
    draw_hline(x2,SCR_YOFFS+h3,x3-x2+1,wall_col);
  }
  draw_vline(x1,SCR_YOFFS-h1,2*h1+1,border_col);
  draw_vline(x3,SCR_YOFFS-h3,2*h3+1,border_col);
}
#endif

#endif /* WALLS_LINES */


#if defined(WALLS_FILLED) || defined(WALLS_SHADED)

#define V_BORDER  /* Waende mit senkrechten Raendern */
#define noH_BORDER  /* Waende mit waagerechten Raendern */

#ifdef WALLS_FILLED
tm_uword global_wall_col;
#endif

#ifdef WALLS_SHADED
const tm_uword dither[4][2][4]={ 
 { { 0,0,0,0 }, { 0,0,0,0 } },
 { { 1,0,1,0 }, { 0,0,0,0 } },
 { { 1,0,1,0 }, { 0,1,0,1 } },
 { { 1,0,1,0 }, { 1,1,1,1 } } };

tm_char shade_buffer[4][MAX_REZ_WIDTH];
static tm_void init_wall_hline256(tm_uword base, tm_bool triple,
                               tm_uword x1, tm_uword x2, tm_uword x3,
                               tm_uword h1, tm_uword h2, tm_uword h3)
{
  tm_uword n,col,cnt,col_dir;
  tm_uword ramp_start,ramp_end,ramp_width;
  tm_uword col_start,col_end,col_width;

  /* nur zwei Punkte gegeben (wall4) */
  if(!triple)
  {
    ramp_start = x1;
    ramp_end   = x2;
    col_start  = h1;
    col_end    = h2;
  }
  else /* drei Punkte gegeben (wall6) */
  {
    if(h1==h2)
    {
      for(n=x1;n<=x2;n++) 
	shade_buffer[0][n]=shade_buffer[1][n]=base+shade_offset[h1];
      ramp_start = x2;
      ramp_end   = x3;
      col_start  = h2;
      col_end    = h3;
    } 
    else /* h2==h3 */
    {
      ramp_start = x1;
      ramp_end   = x2;
      col_start  = h1;
      col_end    = h2;
      for(n=x2;n<=x3;n++) 
	shade_buffer[0][n]=shade_buffer[1][n]=base+shade_offset[h2];
    }
  }

  ramp_width = ramp_end-ramp_start+1;
  col_width  = (col_start<col_end)?(col_end-col_start+1):(col_start-col_end+1);
  col_dir    = (col_start<col_end)?1:-1;

  /* Verlauf erzeugen */
  for(cnt=0,col=col_start,n=ramp_start;n<=ramp_end;n++) {
    while(cnt>ramp_width) { 
      cnt-=ramp_width; 
      col+=col_dir; 
    }
    cnt+=col_width;

    /* die Zeilen der Dithermatrix fuellen */
    shade_buffer[0][n]=base+shade_offset[col]+dither[shade_dither[col]][0][n&3] ;
    shade_buffer[1][n]=base+shade_offset[col]+dither[shade_dither[col]][1][n&3] ;
  }
}
#endif

static tm_void wall_hline256(tm_uword x1,tm_uword x2,tm_uword y)
{
#ifdef WALLS_FILLED
  draw_hline(x1,y,x2-x1+1,global_wall_col);
#else
  memcpy((char*)screen_base+y*width+x1,&shade_buffer[y&1][x1],x2-x1+1);
#endif
}

#ifdef MULTIMODE
tm_void r256_wall_2(tm_uword x,tm_uword h,
               tm_uword wall_col,tm_uword border_col)
#else
tm_void wall_2(tm_uword x,tm_uword h,
               tm_uword wall_col,tm_uword border_col)
#endif
{
  draw_vline(x,SCR_YOFFS-h,2*h+1,border_col+SHADE_OFFSET(h));
}

#ifdef MULTIMODE
tm_void r256_wall_4(tm_uword x1,tm_uword h1,
               tm_uword x2,tm_uword h2,
               tm_uword wall_col,tm_uword border_col)
#else
tm_void wall_4(tm_uword x1,tm_uword h1,
               tm_uword x2,tm_uword h2,
               tm_uword wall_col,tm_uword border_col)
#endif
{
/* x1<x2, h1!=0 || h2!=0 */

  tm_uword  i,xx,hh,co1,co2;
  tm_word   e;

  if (x1>=x2 || (h1==0 && h2==0))
    return;

#ifdef WALLS_FILLED
  global_wall_col = wall_col;
#else
  init_wall_hline256(wall_col,FALSE,x1,x2,0,h1,h2,0);
#endif

#ifdef V_BORDER
  draw_vline(x1,SCR_YOFFS-h1,2*h1+1,border_col+SHADE_OFFSET(h1));
  draw_vline(x2,SCR_YOFFS-h2,2*h2+1,border_col+SHADE_OFFSET(h2));

  if (x1==x2-1)
    return;
#endif

  if (h1<h2)
  {
    hh=h2-h1;
    xx=x2-x1;
#ifdef V_BORDER
    x2--;
#endif
    i=h1+h1;
    h1+=SCR_YOFFS;
    h2=h1;

    if (i!=0)
    {
#ifdef V_BORDER
      x1++;
#endif
      while (--i)
        wall_hline256(x1,x2,--h2);
#ifdef V_BORDER
      x1--;
#endif
      h2--;
    }

    if (xx>=hh)
    {
      i=xx+1;
      hh+=hh;
      e=(tm_word)hh-(tm_word)xx;
      xx+=xx;
      co1=x1;
      if (hh==0)
        x1=x2;
      do
      {
        if (x1++==x2)
          break;
        if (e>=0)
        {
#ifdef H_BORDER
          draw_hline(co1,h1,x1-co1,border_col);
          draw_hline(co1,h2,x1-co1,border_col);
          wall_hline256(x1,x2,h1++);
          wall_hline256(x1,x2,h2--);
#else
          wall_hline256(co1,x2,h1++);
          wall_hline256(co1,x2,h2--);
#endif
          co1=x1;
          e-=xx;
        }
        e+=hh;
      }
      while (--i);
      if (co1!=x1)
      {
#ifdef H_BORDER
        draw_hline(co1,h1,x2-co1+1,border_col);
        draw_hline(co1,h2,x2-co1+1,border_col);
#else
        wall_hline256(co1,x2,h1);
        wall_hline256(co1,x2,h2);
#endif
      }
    }
    else /* xx>=hh */
    {
      i=hh+1;
      xx+=xx;
      e=(tm_word)xx-(tm_word)hh;
      hh+=hh;
#ifdef H_BORDER
      co1=h1;
      co2=h2;
#endif
      do
      {
#ifdef H_BORDER
        if (x1<x2)
        {
          wall_hline256(x1+1,x2,h1);
          wall_hline256(x1+1,x2,h2);
        }
        h1++;
        h2--;
#else
        wall_hline256(x1,x2,h1++);
        wall_hline256(x1,x2,h2--);
#endif
        if (e>=0)
        {
#ifdef H_BORDER
          draw_vline(x1,co1,h1-co1,border_col+SHADE_OFFSET(h1));
          draw_vline(x1,h2+1,co2-h2,border_col+SHADE_OFFSET(h2));
          co1=h1;
          co2=h2;
#endif
          if (x1++==x2)
            break;
          e-=hh;
        }
        e+=xx;
      }
      while (--i);
#ifdef H_BORDER
      if (co1!=h1)
      {
        draw_vline(x1,co1,h1-co1,border_col+SHADE_OFFSET(h1));
        draw_vline(x1,h2+1,co2-h2,border_col+SHADE_OFFSET(h2));
      }
#endif
    }
  }
  else /* h1<h2 */
  {
    hh=h1-h2;
    xx=x2-x1;
#ifdef V_BORDER
    x1++;
#endif
    i=h2+h2;
    h2+=SCR_YOFFS;
    h1=h2;

    if (i!=0)
    {
#ifdef V_BORDER
      x2--;
#endif
      while (--i)
        wall_hline256(x1,x2,--h2);
#ifdef V_BORDER
      x2++;
#endif
      h2--;
    }

    if (xx>=hh)
    {
      i=xx+1;
      hh+=hh;
      e=(tm_word)hh-(tm_word)xx;
      xx+=xx;
      co2=x2;
      if (hh==0)
        x2=x1;
      do
      {
        if (x2--==x1)
          break;
        if (e>=0)
        {
#ifdef H_BORDER
          draw_hline(x2,h1,co2-x2+1,border_col);
          draw_hline(x2,h2,co2-x2+1,border_col);
          wall_hline256(x1,x2,h1++);
          wall_hline256(x1,x2,h2--);
#else
          wall_hline256(x1,co2,h1++);
          wall_hline256(x1,co2,h2--);
#endif
          co2=x2;
          e-=xx;
        }
        e+=hh;
      }
      while (--i);
      if (co2!=x2)
      {
#ifdef H_BORDER
        draw_hline(x1,h1,co2-x1+1,border_col);
        draw_hline(x1,h2,co2-x1+1,border_col);
#else
        wall_hline256(x1,co2,h1);
        wall_hline256(x1,co2,h2);
#endif
      }
    }
    else /* xx>=hh */
    {
      i=hh+1;
      xx+=xx;
      e=(tm_word)xx-(tm_word)hh;
      hh+=hh;
#ifdef H_BORDER
      co1=h1;
      co2=h2;
#endif
      do
      {
#ifdef H_BORDER
        if (x1<x2)
        {
          wall_hline256(x1,x2-1,h1);
          wall_hline256(x1,x2-1,h2);
        }
        h1++;
        h2--;
#else
        wall_hline256(x1,x2,h1++);
        wall_hline256(x1,x2,h2--);
#endif
        if (e>=0)
        {
#ifdef H_BORDER
          draw_vline(x2,co1,h1-co1,border_col+SHADE_OFFSET(h1));
          draw_vline(x2,h2+1,co2-h2,border_col+SHADE_OFFSET(h2));
          co1=h1;
          co2=h2;
#endif
          if (x2--==x1)
            break;
          e-=hh;
        }
        e+=xx;
      }
      while (--i);
#ifdef H_BORDER
      if (co1!=h1)
      {
        draw_vline(x2,co1,h1-co1,border_col+SHADE_OFFSET(h1));
        draw_vline(x2,h2+1,co2-h2,border_col+SHADE_OFFSET(h2));
      }
#endif
    }
  }
}

#ifdef MULTIMODE
tm_void r256_wall_6(tm_uword x1,tm_uword h1,
               tm_uword x2,tm_uword h2,
               tm_uword x3,tm_uword h3,
               tm_uword wall_col,tm_uword border_col)
#else
tm_void wall_6(tm_uword x1,tm_uword h1,
               tm_uword x2,tm_uword h2,
               tm_uword x3,tm_uword h3,
               tm_uword wall_col,tm_uword border_col)
#endif
{
/* x1<x2<x3, h2!=0, h1==h2 || h2==h3 */

  tm_uword  i,xx,hh,co1,co2;
  tm_word   e;

  if (x1>=x2 || x2>=x3 || h2==0)
    return;

#ifdef WALLS_FILLED
  global_wall_col = wall_col;
#else
  init_wall_hline256(wall_col,TRUE,x1,x2,x3,h1,h2,h3);
#endif

#ifdef V_BORDER
  draw_vline(x1,SCR_YOFFS-h1,2*h1+1,border_col+SHADE_OFFSET(h1));
  draw_vline(x3,SCR_YOFFS-h3,2*h3+1,border_col+SHADE_OFFSET(h2));
#endif

  if (h1>h3)
  {
    hh=h2-h3;
    xx=x3-x2;
#ifdef V_BORDER
    x1++;
#endif
    i=h3+h3;
    h3+=SCR_YOFFS;
    h2=h3;

    if (i!=0)
    {
#ifdef V_BORDER
      x3--;
#endif
      while (--i)
        wall_hline256(x1,x3,--h2);
#ifdef V_BORDER
      x3++;
#endif
      h2--;
    }

    if (xx>=hh)
    {
      i=xx+1;
      hh+=hh;
      e=(tm_word)hh-(tm_word)xx;
      xx+=xx;
      co2=x3;
      if (hh==0)
        x3=x2;
      while (--i)
      {
        if (x3--==x2)
          break;
        if (e>=0)
        {
#ifdef H_BORDER
          draw_hline(x3+1,h2,co2-x3,border_col);
          draw_hline(x3+1,h3,co2-x3,border_col);
          wall_hline256(x1,x3,h2--);
          wall_hline256(x1,x3,h3++);
#else
          wall_hline256(x1,co2,h2--);
          wall_hline256(x1,co2,h3++);
#endif
          co2=x3;
          e-=xx;
        }
        e+=hh;
      }
#ifdef H_BORDER
      draw_hline(x1,h2,co2-x1+1,border_col);
      draw_hline(x1,h3,co2-x1+1,border_col);
#else
      wall_hline256(x1,co2,h2);
      wall_hline256(x1,co2,h3);
#endif
    }
    else /* xx>=hh */
    {
      i=hh+1;
      xx+=xx;
      e=(tm_word)xx-(tm_word)hh;
      hh+=hh;
#ifdef H_BORDER
      co1=h2;
      co2=h3;
#endif
      while (--i)
      {
#ifdef H_BORDER
        wall_hline256(x1,x3-1,h2--);
        wall_hline256(x1,x3-1,h3++);
#else
        wall_hline256(x1,x3,h2--);
        wall_hline256(x1,x3,h3++);
#endif
        if (e>=0)
        {
#ifdef H_BORDER
          draw_vline(x3,h2+1,co1-h2,border_col+SHADE_OFFSET(h2));
          draw_vline(x3,co2,h3-co2,border_col+SHADE_OFFSET(h3));
          co1=h2;
          co2=h3;
#endif
          if (x3--==x2)
            break;
          e-=hh;
        }
        e+=xx;
      }
#ifdef H_BORDER
      if (co1!=h2)
      {
        draw_vline(x3,h2+1,co1-h2,border_col+SHADE_OFFSET(h2));
        draw_vline(x3,co2,h3-co2,border_col+SHADE_OFFSET(h3));
      }
      draw_hline(x1,h2,x3-x1+1,border_col);
      draw_hline(x1,h3,x3-x1+1,border_col);
#else
      wall_hline256(x1,x3,h2);
      wall_hline256(x1,x3,h3);
#endif
    }
  }
  else /* h1>h3 */
  {
    hh=h2-h1;
    xx=x2-x1;
#ifdef V_BORDER
    x3--;
#endif
    i=h1+h1;
    h1+=SCR_YOFFS;
    h2=h1;

    if (i!=0)
    {
#ifdef V_BORDER
      x1++;
#endif
      while (--i)
        wall_hline256(x1,x3,--h2);
#ifdef V_BORDER
      x1--;
#endif
      h2--;
    }

    if (xx>=hh)
    {
      i=xx+1;
      hh+=hh;
      e=(tm_word)hh-(tm_word)xx;
      xx+=xx;
      co1=x1;
      if (hh==0)
        x1=x2;
      while (--i)
      {
        if (x1++==x2)
          break;
        if (e>=0)
        {
#ifdef H_BORDER
          draw_hline(co1,h1,x1-co1,border_col);
          draw_hline(co1,h2,x1-co1,border_col);
          wall_hline256(x1,x3,h1++);
          wall_hline256(x1,x3,h2--);
#else
          wall_hline256(co1,x3,h1++);
          wall_hline256(co1,x3,h2--);
#endif
          co1=x1;
          e-=xx;
        }
        e+=hh;
      }
#ifdef H_BORDER
      draw_hline(co1,h1,x3-co1+1,border_col);
      draw_hline(co1,h2,x3-co1+1,border_col);
#else
      wall_hline256(co1,x3,h1);
      wall_hline256(co1,x3,h2);
#endif
    }
    else /* xx>=hh */
    {
      i=hh+1;
      xx+=xx;
      e=(tm_word)xx-(tm_word)hh;
      hh+=hh;
#ifdef H_BORDER
      co1=h1;
      co2=h2;
#endif
      while (--i)
      {
#ifdef H_BORDER
        wall_hline256(x1+1,x3,h1++);
        wall_hline256(x1+1,x3,h2--);
#else
        wall_hline256(x1,x3,h1++);
        wall_hline256(x1,x3,h2--);
#endif
        if (e>=0)
        {
#ifdef H_BORDER
          draw_vline(x1,co1,h1-co1,border_col+SHADE_OFFSET(h1));
          draw_vline(x1,h2+1,co2-h2,border_col+SHADE_OFFSET(h2));
          co1=h1;
          co2=h2;
#endif
          if (x1++==x2)
            break;
          e-=hh;
        }
        e+=xx;
      }
#ifdef H_BORDER
      if (co1!=h1)
      {
        draw_vline(x1,co1,h1-co1,border_col+SHADE_OFFSET(h1));
        draw_vline(x1,h2+1,co2-h2,border_col+SHADE_OFFSET(h2));
      }
      draw_hline(x1,h1,x3-x1+1,border_col);
      draw_hline(x1,h2,x3-x1+1,border_col);
#else
      wall_hline256(x1,x3,h1);
      wall_hline256(x1,x3,h2);
#endif
    }
  }
}

#endif /* WALLS_FILLED */

#ifdef TEXTURE
#include "texture.h"

static tm_word wall_intersect(tm_word x, tm_word h, tm_word wall, 
			      tm_word x1, tm_word z1, tm_word x2, tm_word z2) 
{
  tm_word n, nenner, zaehler, m, t, rh;

  nenner  = ((x-SCR_XOFFS) * (z2-z1) - SCR_XOFFS * (x2-x1));

  if(nenner != 0) {
    zaehler = ((x-SCR_XOFFS) * z1 - SCR_XOFFS * x1);

    /* vertical clip */
    rh = (h<SCR_YOFFS)?h:SCR_YOFFS;

#ifdef TEXTURE_MIR
    draw_vtexline(x, SCR_YOFFS-rh-1, 2*rh, 2*h, 
		       textures[wall>>1][((-2*TEXTURE_WIDTH * zaehler)/nenner)&(2*TEXTURE_WIDTH-1)], 
		       TEXTURE_HEIGHT, texture_colmap[wall>>1][shade_offset[rh]]);
#else
    draw_vtexline(x, SCR_YOFFS-rh-1, 2*rh, 2*h, 
		       textures[wall>>1][((-TEXTURE_WIDTH * zaehler)/nenner)&(TEXTURE_WIDTH-1)], 
		       TEXTURE_HEIGHT, texture_colmap[wall>>1][shade_offset[rh]]);
#endif
  }
}

#ifndef X_R256_WALL_4TEX
#ifdef MULTIMODE
tm_void r256_wall_4tex(tm_word xl, tm_word hl,
		       tm_word xr, tm_word hr,
		       tm_uword wall,
		       tm_word xa, tm_word za, tm_word xb, tm_word zb)
#else
tm_void wall_4tex(tm_word xl, tm_word hl,
		  tm_word xr, tm_word hr,
		  tm_uword wall,
		  tm_word xa, tm_word za, tm_word xb, tm_word zb)
#endif
{
  tm_word N,n,hwid,vwid,i,dir;

  /* draw all columns */
  hwid = xr-xl;
  vwid = hr-hl;

  /* incr/decr wall */
  if(vwid<0) { vwid = -vwid; dir = -1;} else { dir =1; }
  
  for(i=xl,n=hl,N=0;i<=xr;i++) {
    while(N>hwid) { N-=hwid; n+=dir; } N+=vwid;
    wall_intersect(i, n, wall, xa,za,xb,zb);
  }
}
#endif  /* X_R256_WALL_4TEX */
#endif  /* TEXTURE */
