/*
    smiley.c

    ????      Till  erste Version
    13.11.96  Chip  MAX_SIZE, HEIGHT -> height, WIDTH -> width
    28.11.96  Chip  Typen auf tm_* umgestellt
    09.01.98  Till  nochmal von vorne begonnen
    15.01.98  Till  Darstellung unabhaengig von SMILESTP und SMILERAD immer gleich
    18.01.98  Till  Smileys frei skalierbar mit Clipping
    18.03.98  Till  beliebig geclippte Smileys fuer Ueberschneidungen mit Tueren
*/

#include "ctypes.h"
#include "smiley.h"
#include "grafik.h"
#include "gr_prim.h"
#include "drawmaze.h"
#include "machine.h"
#include "maze.h"
#include "net.h"
#include "misc.h"
#include "move.h"

#define SHADOW_VRAD    6          /* shadows vertical radius is rad/6 */
#define SMILEY_SCALE_FACTOR  8    /* reduce size of smiley by 8/10 */ 

extern tm_ushort smiley_lines[];  /* lines array with rle encoded smiley lines */
extern tm_ushort smiley[];        /* disc array with offsets into lines array */
extern tm_ushort smiley_table[];  /* offset into disc array */

#define RAD_SCALE(n)   ((SMILEY_SCALE_FACTOR*n)/10)
#define SMILESTEPS     (1<<(SMILESTP))

tm_void 
init_smiley(tm_void)
{
  tm_word i;

  /* smiley colors */
  if(colors>=16) {
    for(i=0;i<MAX_PLAYERS;i++)
      player[i].col=i%16;
  } else {
    for(i=0;i<MAX_PLAYERS;i++)
      player[i].col=BW_LGREY;
  }

  /* init players */
  for(i=0;i<MAX_PLAYERS/2;i++) {
    player[i].is_shot = FALSE;
    player[i].radius  = SMILEY_RAD;
  }

  /* init shots */
  for(i=MAX_PLAYERS/2;i<MAX_PLAYERS;i++) {
    player[i].is_shot = TRUE;
    player[i].radius  = SHOT_RAD;
  }
}

tm_void 
draw_hline_hclip(tm_word x, tm_word y, tm_word len, tm_word color) 
{
  /* crossing left border? */
  if(x<0) {
    len+=x;
    if(len<=0) return;
    x=0;
  }

  /* crossing right border? */
  if(x+len>width) {
    len = width-x;
    if(len<=0) return;
  }

  /* ok draw clipped line */
  draw_hline(x,y,len,color);
}

tm_void 
draw_smiley(tm_word x, tm_uword rad, tm_ushort rot, tm_uword color)
{
  tm_word  i,j,len,m,M,svrad,k;
  tm_long  sx,n;
  tm_ushort *sm,base;

  /* first - the shadow (shape of smiley with rot == 0)*/
  base = smiley_table[0];
  svrad = RAD_SCALE((rad/SHADOW_VRAD)<<1);
  j = (pf_height>>1)+rad-(svrad>>1);

  rad = RAD_SCALE(rad);

  /* clip shadow vertical? */
  if(j+svrad < pf_height) {

    /* no do not clip vertically, clip horizontally?  */
    if((x-(tm_word)rad<0)||(x+(tm_word)rad>width)) {
      /* do the bresenham */
      for(m=0,M=0,i=0;i<svrad;i++) {
	while(M>svrad) { M-=svrad; m++; } M += 2*SMILERAD;
	sm = &smiley_lines[smiley[base+m]];
	if((*sm>>8)!=0) {
	  sx = (tm_long)SMILERAD * (tm_long)x - (tm_long)SMILERAD * (tm_long)rad + 
	    (tm_long)(*sm++&0xff) * (tm_long)rad ;  /* left offset */
	  draw_hline_hclip(sx/SMILERAD,j+i,
			   (((sx%SMILERAD)+((tm_long)(*sm++&0xff) * (tm_long)rad))/SMILERAD), COL_BLACK );
	}
      }
    } else {
      /* no, do not clip horizontally */
      for(m=0,M=0,i=0;i<svrad;i++) {
	while(M>svrad) { M-=svrad; m++; } M += 2*SMILERAD;
	sm = &smiley_lines[smiley[base+m]];
	if((*sm>>8)!=0) {
	  sx = (tm_long)SMILERAD * (tm_long)x - (tm_long)SMILERAD * (tm_long)rad + 
	    (tm_long)(*sm++&0xff) * (tm_long)rad ;  /* left offset */
	  draw_hline(sx/SMILERAD,j+i,
		     (((sx%SMILERAD)+((tm_long)(*sm++&0xff) * (tm_long)rad))/SMILERAD), COL_BLACK );
	}
      }
    }
  } else {
    /* do vertical clipping */

    for(m=0,M=0,i=0;i<svrad;i++) {
      while(M>svrad) { M-=svrad; m++; } M += 2*SMILERAD;
      sm = &smiley_lines[smiley[base+m]];
      if((*sm>>8)!=0) {
	sx = (tm_long)SMILERAD * (tm_long)x - (tm_long)SMILERAD * (tm_long)rad + 
	  (tm_long)(*sm++&0xff) * (tm_long)rad ;  /* left offset */
	if(j+i<pf_height)
	  draw_hline_hclip(sx/SMILERAD,j+i,
		     (((sx%SMILERAD)+((tm_long)(*sm++&0xff) * (tm_long)rad))/SMILERAD), COL_BLACK );
	else
	  sm++;
      }
    }
  }

  /* get base in disc array for this particular angle */
  base = smiley_table[rot>>(16-SMILESTP)];

  /* clip vertically? */
  if(rad<(pf_height/2)) {
    
    /* no do not clip vertically, clip horizontally?  */
    if((x-(tm_word)rad<0)||(x+(tm_word)rad>width)) {

      /* do the bresenham again */
      for(m=0,M=0,i=0;i<(rad<<1);i++) {
	while(M>(rad<<1)) { M-=(rad<<1); m++; } M += 2*SMILERAD;
	sm = &smiley_lines[smiley[base+m]];
	sx = (tm_long)SMILERAD * (tm_long)x - (tm_long)SMILERAD * (tm_long)rad + 
	  (tm_long)(*sm&0xff) * (tm_long)rad ;  /* left offset */
	len=*sm++;
	for(j=0;j<len>>8;j++) {
	  n = (tm_long)(*sm&0xff) * (tm_long)rad;
	  svrad = *sm++>>8;
	  draw_hline_hclip(sx/SMILERAD,pf_height/2-rad+i, (((sx%SMILERAD)+n)/SMILERAD), 
		   (svrad==COL_YELLOW)?color:svrad );
	  sx+=n;
	}
      }

    } else {

      /* do the bresenham again */
      for(m=0,M=0,i=0;i<(rad<<1);i++) {
	while(M>(rad<<1)) { M-=(rad<<1); m++; } M += 2*SMILERAD;
	sm = &smiley_lines[smiley[base+m]];
	sx = (tm_long)SMILERAD * (tm_long)x - (tm_long)SMILERAD * (tm_long)rad + 
	  (tm_long)(*sm&0xff) * (tm_long)rad ;  /* left offset */
	len=*sm++;
	for(j=0;j<len>>8;j++) {
	  n = (tm_long)(*sm&0xff) * (tm_long)rad;
	  svrad = *sm++>>8;
	  draw_hline(sx/SMILERAD,pf_height/2-rad+i, (((sx%SMILERAD)+n)/SMILERAD), 
		   (svrad==COL_YELLOW)?color:svrad );
	  sx+=n;
	}
      }
    }
  } else {
    /* clip vertically */
    /* do the bresenham again */
    for(m=0,M=0,i=0;i<(rad<<1);i++) {
      while(M>(rad<<1)) { M-=(rad<<1); m++; } M += 2*SMILERAD;
      sm = &smiley_lines[smiley[base+m]];
      sx = (tm_long)SMILERAD * (tm_long)x - (tm_long)SMILERAD * (tm_long)rad + 
	(tm_long)(*sm&0xff) * (tm_long)rad ;  /* left offset */
      len=*sm++;
      for(j=0;j<len>>8;j++) {
	k=pf_height/2-rad+i;
	if((k>=0)&&(k<pf_height)) {
	  n = (tm_long)(*sm&0xff) * (tm_long)rad;
	  svrad = *sm++>>8;
	  draw_hline_hclip(sx/SMILERAD,pf_height/2-rad+i, (((sx%SMILERAD)+n)/SMILERAD), 
			   (svrad==COL_YELLOW)?color:svrad );
	  sx+=n;
	} else
	  sm++;
      }
    }
  }
}

tm_void 
draw_hline_clip(tm_word x, tm_word y, tm_word len, tm_word color, tm_word left_clip, tm_word right_clip) 
{
  /* crossing left border? */
  if(x<left_clip) {
    len+=(x-left_clip);
    if(len<=0) return;
    x=left_clip;
  }

  /* crossing right border? */
  if(x+len>right_clip) {
    len = right_clip-x;
    if(len<=0) return;
  }

  /* ok draw clipped line */
  draw_hline(x,y,len,color);
}

/* draw a clipped smiley */
tm_void 
draw_smiley_clip(tm_word x, tm_uword rad, tm_ushort rot, 
		 tm_uword color, tm_word left_clip, tm_word right_clip)
{
  tm_word  i,j,len,m,M,svrad,k;
  tm_long  sx,n;
  tm_ushort *sm,base;

  /* correct clipping */
  if(left_clip  < 0)          left_clip  = 0;
  if(right_clip > width)      right_clip = width;
  if(left_clip >= right_clip) return;

  /* first - the shadow (shape of smiley with rot == 0)*/
  base = smiley_table[0];
  svrad = RAD_SCALE((rad/SHADOW_VRAD)<<1);
  j = (pf_height>>1)+rad-(svrad>>1);

  rad = RAD_SCALE(rad);

  /* clip shadow vertical? */
  if(j+svrad < pf_height) {

    /* no do not clip vertically, clip horizontally?  */
    if((x-(tm_word)rad<left_clip)||(x+(tm_word)rad>right_clip)) {
      /* do the bresenham */
      for(m=0,M=0,i=0;i<svrad;i++) {
	while(M>svrad) { M-=svrad; m++; } M += 2*SMILERAD;
	sm = &smiley_lines[smiley[base+m]];
	if((*sm>>8)!=0) {
	  sx = (tm_long)SMILERAD * (tm_long)x - (tm_long)SMILERAD * (tm_long)rad + 
	    (tm_long)(*sm++&0xff) * (tm_long)rad ;  /* left offset */
	  draw_hline_clip(sx/SMILERAD,j+i,
			  (((sx%SMILERAD)+((tm_long)(*sm++&0xff) * (tm_long)rad))/SMILERAD), 
			  COL_BLACK, left_clip, right_clip );
	}
      }
    } else {
      /* no, do not clip horizontally */
      for(m=0,M=0,i=0;i<svrad;i++) {
	while(M>svrad) { M-=svrad; m++; } M += 2*SMILERAD;
	sm = &smiley_lines[smiley[base+m]];
	if((*sm>>8)!=0) {
	  sx = (tm_long)SMILERAD * (tm_long)x - (tm_long)SMILERAD * (tm_long)rad + 
	    (tm_long)(*sm++&0xff) * (tm_long)rad ;  /* left offset */
	  draw_hline(sx/SMILERAD,j+i,
		     (((sx%SMILERAD)+((tm_long)(*sm++&0xff) * (tm_long)rad))/SMILERAD), COL_BLACK );
	}
      }
    }
  } else {
    /* do vertical clipping */

    for(m=0,M=0,i=0;i<svrad;i++) {
      while(M>svrad) { M-=svrad; m++; } M += 2*SMILERAD;
      sm = &smiley_lines[smiley[base+m]];
      if((*sm>>8)!=0) {
	sx = (tm_long)SMILERAD * (tm_long)x - (tm_long)SMILERAD * (tm_long)rad + 
	  (tm_long)(*sm++&0xff) * (tm_long)rad ;  /* left offset */
	if(j+i<pf_height)
	  draw_hline_clip(sx/SMILERAD,j+i,
		     (((sx%SMILERAD)+((tm_long)(*sm++&0xff) * (tm_long)rad))/SMILERAD), 
			   COL_BLACK, left_clip, right_clip );
	else
	  sm++;
      }
    }
  }

  /* get base in disc array for this particular angle */
  base = smiley_table[rot>>(16-SMILESTP)];

  /* clip vertically? */
  if(rad<(pf_height/2)) {
    
    /* no do not clip vertically, clip horizontally?  */
    if((x-(tm_word)rad<left_clip)||(x+(tm_word)rad>right_clip)) {

      /* do the bresenham again */
      for(m=0,M=0,i=0;i<(rad<<1);i++) {
	while(M>(rad<<1)) { M-=(rad<<1); m++; } M += 2*SMILERAD;
	sm = &smiley_lines[smiley[base+m]];
	sx = (tm_long)SMILERAD * (tm_long)x - (tm_long)SMILERAD * (tm_long)rad + 
	  (tm_long)(*sm&0xff) * (tm_long)rad ;  /* left offset */
	len=*sm++;
	for(j=0;j<len>>8;j++) {
	  n = (tm_long)(*sm&0xff) * (tm_long)rad;
	  svrad = *sm++>>8;
	  draw_hline_clip(sx/SMILERAD,pf_height/2-rad+i, (((sx%SMILERAD)+n)/SMILERAD), 
			  (svrad==COL_YELLOW)?color:svrad,
			  left_clip, right_clip);
	  sx+=n;
	}
      }

    } else {

      /* do the bresenham again */
      for(m=0,M=0,i=0;i<(rad<<1);i++) {
	while(M>(rad<<1)) { M-=(rad<<1); m++; } M += 2*SMILERAD;
	sm = &smiley_lines[smiley[base+m]];
	sx = (tm_long)SMILERAD * (tm_long)x - (tm_long)SMILERAD * (tm_long)rad + 
	  (tm_long)(*sm&0xff) * (tm_long)rad ;  /* left offset */
	len=*sm++;
	for(j=0;j<len>>8;j++) {
	  n = (tm_long)(*sm&0xff) * (tm_long)rad;
	  svrad = *sm++>>8;
	  draw_hline(sx/SMILERAD,pf_height/2-rad+i, (((sx%SMILERAD)+n)/SMILERAD), 
		   (svrad==COL_YELLOW)?color:svrad );
	  sx+=n;
	}
      }
    }
  } else {
    /* clip vertically */
    /* do the bresenham again */
    for(m=0,M=0,i=0;i<(rad<<1);i++) {
      while(M>(rad<<1)) { M-=(rad<<1); m++; } M += 2*SMILERAD;
      sm = &smiley_lines[smiley[base+m]];
      sx = (tm_long)SMILERAD * (tm_long)x - (tm_long)SMILERAD * (tm_long)rad + 
	(tm_long)(*sm&0xff) * (tm_long)rad ;  /* left offset */
      len=*sm++;
      for(j=0;j<len>>8;j++) {
	k=pf_height/2-rad+i;
	if((k>=0)&&(k<pf_height)) {
	  n = (tm_long)(*sm&0xff) * (tm_long)rad;
	  svrad = *sm++>>8;
	  draw_hline_clip(sx/SMILERAD,pf_height/2-rad+i, (((sx%SMILERAD)+n)/SMILERAD), 
			  (svrad==COL_YELLOW)?color:svrad,
			  left_clip, right_clip);
	  sx+=n;
	} else
	  sm++;
      }
    }
  }
}






