/*
   MK_SMILE.C

   creates tabulars for fast smiley generation

   TODO:
   + identische Zeilen suchen und nur einmal speichern
   + identische Smileys suchen und nur einmal speichern
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define BLACK   0
#define WHITE   1
#define RED     2
#define GREEN   3
#define BLUE    4
#define YELLOW  5
#define MAGENTA 6
#define CYAN    7

#ifdef __X11__
#include "x11.h"
#define DISP_ENABLE
#endif

#ifdef __TOS__
#include "gem.h"
#define DISP_ENABLE
#endif

#ifndef M_PI
#define M_PI (3.14159254f)
#endif

#ifndef M_PI_2
#define M_PI_2 (M_PI/2)
#endif

#define FALSE   (0)
#define TRUE    (!FALSE)

#define RAD   100
#define DELTA 4096

/* 1.3 = fieser Grinser   1.5 / 1.75 / 2.0  */
#define  F_MOR   (1.5f)
#define  F_MUR   (F_MOR-0.6f)
#define  F_MOR_2 (F_MOR*F_MOR)
#define  F_MUR_2 (F_MUR*F_MUR)
#define  F_MAX   (0.6f)
#define  F_MAX_2 (F_MAX*F_MAX)
#define  F_MY    (0.6f)
#define  F_DIST  ((F_MAX_2/F_MUR_2-F_MAX_2/F_MOR_2)/2.0f)

float pos[12];
int col[12];

int get_line(float y, float rot) {
  int n,i;
  float cy,cu,co;

  /* values for outer square */
  pos[0]=-3.0; col[0]=BLACK;   /* transparent left offset */
  pos[1]= 3.0; col[1]=YELLOW;  /* disc */
  n=2;

  /* now add the eyes (two circles) */
  if((y<0.7)&&(y>-0.1)) {
    /* first the white part */
    pos[5]=pos[1]; col[5]=col[3]=col[1]; n=6;
    cy = (y-0.3)*(1/0.4);
    cy = sqrt(1-cy*cy)*0.4/(M_PI_2);
    pos[1] = -.5/M_PI_2 - cy;
    pos[2] = -.5/M_PI_2 + cy; col[2]=WHITE;
    pos[3] =  .5/M_PI_2 - cy;
    pos[4] =  .5/M_PI_2 + cy; col[4]=WHITE;

    /* and now the pupil (two circles) */
    if((y<0.4)&&(y>0.0)) {
      pos[9]=pos[5]; col[9]=col[5];
      pos[5]=pos[3]; col[5]=col[3];
      pos[8]=pos[4]; col[8]=col[4];
      pos[4]=pos[2]; col[4]=col[2]; n=10;
      cy = (y-0.2)*(1/0.2);
      cy = sqrt(1-cy*cy)*0.2/(M_PI_2);
      pos[2] = -.4/M_PI_2 - cy; col[2]=WHITE;
      pos[3] = -.4/M_PI_2 + cy; col[3]=BLACK;
      pos[6] =  .4/M_PI_2 - cy; col[6]=WHITE;
      pos[7] =  .4/M_PI_2 + cy; col[7]=BLACK;
    }
  }

  /* the mouth (area between two parabels) */
  if(y> -F_MY-F_DIST) {
    cu = sqrt( (y + F_MY + F_DIST) * F_MUR_2 );
    if( y> -F_MY+F_DIST) {
      co = sqrt( (y + F_MY - F_DIST) * F_MOR_2 );
      if(co<cu) {
	/* both parabels visible */
	pos[5] = pos[1]; col[5] = col[3] = col[1]; n=6; 
	pos[1] = -cu/M_PI_2;
	pos[2] = -co/M_PI_2; col[2]=RED;
	pos[3] =  co/M_PI_2;
	pos[4] =  cu/M_PI_2; col[4]=RED;
      }
    } else {
      /* only one parabel visible */
      pos[3] = pos[1]; col[3] = col[1]; n=4;
      pos[1] = -cu/M_PI_2;
      pos[2] =  cu/M_PI_2; col[2]=RED;
    }
  }

  /* shift (rotation after projection) */
  for(i=0;i<n;i++) {
    pos[i] += rot / 90.0;
    if(pos[i]<-1.0) pos[i]=-1.0;
    if(pos[i]> 1.0) pos[i]= 1.0;
  }

  /* project square to circle */
  for(i=0;i<n;i++)
    pos[i] = sin(M_PI_2 * pos[i]) * sqrt(1.0 - y*y);

  return n;
}

int main(int argc, char **argv)
{
  int y,i,j,r,h,n,s,t,hit;
  int rad, steps, intpos[12];
  FILE *file;

  /* Sachen zum Suchen identischer Zeilen */
  int *colmem, *posmem;
  int *stepmem, *rowmem;
  int *lenmem, mem, *rowlst, *steplst;

  /* Sachen zum Suchen identischer Smileys-Steps */
  int *sm_step, *sm_row, *sm_lst, sm, *sm_num;

  int *sm_data, *sm_line;

  if(argc!=3) {
#ifdef DISP_ENABLE
    open_window(0, argv, 2*RAD+10, 2*RAD+10, "MakeSmiley");
    
    do { 
      for(r=-180;r<180;r+=2) {
	for(y=-RAD;y<RAD;y++) {
	  n=get_line((float)y/RAD,r);
	  for(j=0,i=0;i<n;i++) {
	    for(;j<RAD+RAD*pos[i];j++) {
	      put_pixel(5+j,RAD-y+5, col[i]);
	    }
	  }
	}
	draw_bitmap();
      }
    } while(get_key()==0);
    
    close_window();
#else
    printf("usage: mk_smile radius steps\n");
#endif
  } else {

    if((file=fopen("smiletab.c","w"))==NULL) {
      printf("unable to open smiletab.c\n");
      exit(1);
    }

    rad   = atoi(argv[1]);
    steps = atoi(argv[2]);

    fprintf(file,"/* smiletab.c - auto generated smiley tables for TheMaze */\n");
    fprintf(file,"/* contains smiley data for %d smileys with radius %d */\n\n", steps, rad);

    fprintf(file,"#include \"../ctypes.h\"\n\n");

    /*  file=stdout; */

    /* Speicher fuer alle moeglichen Zeilen besorgen */
    posmem  = (int*)malloc(steps*2*rad*12*sizeof(int));  /* Platz fuer Offsets */
    colmem  = (int*)malloc(steps*2*rad*12*sizeof(int));  /* Platz fuer Farben  */
    lenmem  = (int*)malloc(steps*2*rad   *sizeof(int));  /* Platz fuer Anzahl  */
    stepmem = (int*)malloc(steps*2*rad   *sizeof(int));  /* Platz fuer Schritt */
    rowmem  = (int*)malloc(steps*2*rad   *sizeof(int));  /* Platz fuer Zeile   */
    steplst = (int*)malloc(      2*rad   *sizeof(int));
    rowlst  = (int*)malloc(      2*rad   *sizeof(int));
    mem=0;   /* keine Zeile im Speicher */

    /* Speicher fuer alle moeglichen Smileys besorgen */
    sm_step = (int*)malloc(steps*2*rad*sizeof(int));
    sm_row  = (int*)malloc(steps*2*rad*sizeof(int));
    sm_num  = (int*)malloc(steps      *sizeof(int));
    sm_lst  = (int*)malloc(steps      *sizeof(int));
    sm=0;    /* kein Smiley im Speicher */

    /* smiley data table */
    sm_line = (int*)malloc(steps*2*rad*sizeof(int));
    sm_data = (int*)malloc(steps*sizeof(int));

    fprintf(file,"tm_ushort *smiley_rows[]={\n");

    for(i=0;i<steps;i++) {
      for(y=0;y<2*rad;y++) {
	j=get_line(((float)y-(float)rad)/(float)-rad,(360.0*(float)i)/(float)steps -180.0);

	/* scale radius */
	for(r=0;r<j;r++)
	  intpos[r] = rad + pos[r] * rad;

	/* remove identical fields */
	for(r=1;r<j;r++) {
	  if(intpos[r-1] == intpos[r]) {
	    for(h=r;h<j;h++) {
	      intpos[h]=intpos[h+1];
	      col[h]=col[h+1];
	    }
	    j--; r--;
	  }
	}

	/* search in line buffer for identical line */ 
	for(hit=0,s=0;(s<mem)&&(!hit);s++) {
	  /* correct length?? */
	  if(lenmem[s]==j) {
	    /* posfield and colfield match?? */
	    if(( memcmp(&posmem[12*s],intpos,j*sizeof(int)) == 0 ) &&
	       ( memcmp(&colmem[12*s],   col,j*sizeof(int)) == 0 )) {
	      steplst[y] = stepmem[s]; 
	      rowlst[y]  = rowmem[s];
	      hit=1;
	    }
	  }
	}

	if(!hit) {
	  /* output line */
	  fprintf(file,"{0x%04x%s",((j-1)<<8)|intpos[0], (j==1)?"":"," );
	  for(r=1;r<j;r++)
	    fprintf(file,"0x%04x%s",(col[r]<<8)|(intpos[r]-intpos[r-1]), (r==j-1)?"":",");
	  fprintf(file,"},\n");

	  /* and save line to buffer */
	  memcpy(&posmem[12*mem], intpos, j*sizeof(int));
	  memcpy(&colmem[12*mem],    col, j*sizeof(int));
	  stepmem[mem] = i;
	  rowmem[mem] = y;
	  lenmem[mem] = j;
	  mem++;

	  /* output info */
	  steplst[y] = i;
	  rowlst[y] = y;

	  sm_line[i+2*rad*y] = mem-1;
	} else
	  sm_line[i+2*rad*y] = s;
      }

      /* search for smiley in buffer */
      for(hit=0,s=0;s<sm;s++) {
	if((memcmp(&sm_step[2*rad*s], steplst, 2*rad*sizeof(int)) == 0)&&
	   (memcmp( &sm_row[2*rad*s],  rowlst, 2*rad*sizeof(int)) == 0)) {
	  sm_lst[i] = sm_num[s];
	  hit=1;
	}
      }

      if(!hit) {
#if 0
	/* output line */
	fprintf(file,"\ntm_ushort *sm_%02d[]={",i);
	for(y=0;y<2*rad;y++)
	  fprintf(file,"sm_%02d_%02d%s",steplst[y],rowlst[y],(y==2*rad-1)?"":",");
	fprintf(file,"};\n\n");
#endif
	/* store smiley in buffer */
	memcpy(&sm_step[2*rad*sm], steplst, 2*rad*sizeof(int));
	memcpy( &sm_row[2*rad*sm],  rowlst, 2*rad*sizeof(int));
	sm_num[sm]=i;
	sm++;

	/* output smiley */
	sm_lst[i] = i;

	sm_data[i] = sm-1;
      } else
	sm_data[i] = s;
    }

    fseek(file, -(strlen(";\n")), 1);  /* remove ;\n */
    fprintf(file,"\n};\n\n");

    fprintf(file,"tm_ushort *smiley_lines[]={");


    fprintf(file,"tm_ushort **smiley_data[]={");
    
#if 0
    fprintf(file,"tm_ushort **smiley_data[]={");
    for(i=0;i<steps;i++)
      fprintf(file,"sm_%02d%s",sm_lst[i],(i==steps-1)?"":",");
    fprintf(file,"};\n");
#endif
    
    /* Pufferspeicher freigeben */
    /* Zeilen */
    free(posmem); free(colmem); free(stepmem); free(rowmem); free(lenmem); 
    free(steplst); free(rowlst);

    /* Smiley */
    free(sm_step); free(sm_row); free(sm_lst); free(sm_num);

    free(sm_data); free(sm_line);

    fclose(file);
  }

  return 0;
}
