/*
   MK_SMILE.C

   creates tabulars for fast smiley generation

   TODO:
   + identische Zeilen suchen und nur einmal speichern
   + identische Smileys suchen und nur einmal speichern
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define BLACK   0
#define WHITE   1
#define RED     2
#define GREEN   3
#define BLUE    4
#define YELLOW  5
#define MAGENTA 6
#define CYAN    7

#ifdef __X11__
#include "x11.h"
#define DISP_ENABLE
#endif

#ifdef __TOS__
#include "gem.h"
#define DISP_ENABLE
#endif

#ifndef M_PI
#define M_PI (3.14159254f)
#endif

#ifndef M_PI_2
#define M_PI_2 (M_PI/2)
#endif

#define FALSE   (0)
#define TRUE    (!FALSE)

#define RAD   100
#define DELTA 4096

/* 1.3 = fieser Grinser   1.5 / 1.75 / 2.0  */
#define  F_MOR   (1.5f)
#define  F_MUR   (F_MOR-0.6f)
#define  F_MOR_2 (F_MOR*F_MOR)
#define  F_MUR_2 (F_MUR*F_MUR)
#define  F_MAX   (0.6f)
#define  F_MAX_2 (F_MAX*F_MAX)
#define  F_MY    (0.6f)
#define  F_DIST  ((F_MAX_2/F_MUR_2-F_MAX_2/F_MOR_2)/2.0f)

float pos[12];
int col[12];

int get_line(float y, float rot) {
  int n,i;
  float cy,cu,co;

  /* values for outer square */
  pos[0]=-3.0; col[0]=BLACK;   /* transparent left offset */
  pos[1]= 3.0; col[1]=YELLOW;  /* disc */
  n=2;

  /* now add the eyes (two circles) */
  if((y<0.7)&&(y>-0.1)) {
    /* first the white part */
    pos[5]=pos[1]; col[5]=col[3]=col[1]; n=6;
    cy = (y-0.3)*(1/0.4);
    cy = sqrt(1-cy*cy)*0.4/(M_PI_2);
    pos[1] = -.5/M_PI_2 - cy;
    pos[2] = -.5/M_PI_2 + cy; col[2]=WHITE;
    pos[3] =  .5/M_PI_2 - cy;
    pos[4] =  .5/M_PI_2 + cy; col[4]=WHITE;

    /* and now the pupil (two circles) */
    if((y<0.4)&&(y>0.0)) {
      pos[9]=pos[5]; col[9]=col[5];
      pos[5]=pos[3]; col[5]=col[3];
      pos[8]=pos[4]; col[8]=col[4];
      pos[4]=pos[2]; col[4]=col[2]; n=10;
      cy = (y-0.2)*(1/0.2);
      cy = sqrt(1-cy*cy)*0.2/(M_PI_2);
      pos[2] = -.4/M_PI_2 - cy; col[2]=WHITE;
      pos[3] = -.4/M_PI_2 + cy; col[3]=BLACK;
      pos[6] =  .4/M_PI_2 - cy; col[6]=WHITE;
      pos[7] =  .4/M_PI_2 + cy; col[7]=BLACK;
    }
  }

  /* the mouth (area between two parabels) */
  if(y> -F_MY-F_DIST) {
    cu = sqrt( (y + F_MY + F_DIST) * F_MUR_2 );
    if( y> -F_MY+F_DIST) {
      co = sqrt( (y + F_MY - F_DIST) * F_MOR_2 );
      if(co<cu) {
	/* both parabels visible */
	pos[5] = pos[1]; col[5] = col[3] = col[1]; n=6; 
	pos[1] = -cu/M_PI_2;
	pos[2] = -co/M_PI_2; col[2]=RED;
	pos[3] =  co/M_PI_2;
	pos[4] =  cu/M_PI_2; col[4]=RED;
      }
    } else {
      /* only one parabel visible */
      pos[3] = pos[1]; col[3] = col[1]; n=4;
      pos[1] = -cu/M_PI_2;
      pos[2] =  cu/M_PI_2; col[2]=RED;
    }
  }

  /* shift (rotation after projection) */
  for(i=0;i<n;i++) {
    pos[i] += rot / 90.0;
    if(pos[i]<-1.0) pos[i]=-1.0;
    if(pos[i]> 1.0) pos[i]= 1.0;
  }

  /* project square to circle */
  for(i=0;i<n;i++)
    pos[i] = sin(M_PI_2 * pos[i]) * sqrt(1.0 - y*y);

  return n;
}

int main(int argc, char **argv)
{
  int y,i,j,r,h,n,s,t,hit;
  int rad, steps, intpos[12];
  FILE *file;

  /* Sachen zum Suchen identischer Zeilen */
  int *line_cnt, *line_col, *line_pos, lines, *smile_line;

  /* Sachen zum Suchen identischer Smileys-Steps */
  int *smiley, *smiley_table, smiles;

  if(argc!=3) {
#ifdef DISP_ENABLE
    open_window(0, argv, 2*RAD+10, 2*RAD+10, "MakeSmiley");
    
    do { 
      for(r=-180;r<180;r+=2) {
	for(y=-RAD;y<RAD;y++) {
	  n=get_line((float)y/RAD,r);
	  for(j=0,i=0;i<n;i++) {
	    for(;j<RAD+RAD*pos[i];j++) {
	      put_pixel(5+j,RAD-y+5, col[i]);
	    }
	  }
	}
	draw_bitmap();
      }
    } while(get_key()==0);
    
    close_window();
#else
    printf("usage: mk_smile radius steps\n");
#endif
  } else {

    rad   = atoi(argv[1]);
    steps = 1<<atoi(argv[2]);

    /* Speicher fuer alle moeglichen Zeilen besorgen */
    line_pos   = (int*)malloc(steps*2*rad*12*sizeof(int));  /* Platz fuer Offsets */
    line_col   = (int*)malloc(steps*2*rad*12*sizeof(int));  /* Platz fuer Farben  */
    line_cnt   = (int*)malloc(steps*2*rad   *sizeof(int));  /* Platz fuer Anzahl  */
    smile_line = (int*)malloc(      2*rad   *sizeof(int));
    lines=0;   /* keine Zeile im Speicher */

    /* Speicher fuer alle moeglichen Smileys besorgen */
    smiley = (int*)malloc(steps*2*rad*sizeof(int));
    smiles=0;    /* kein Smiley im Speicher */

    smiley_table  = (int*)malloc(steps      *sizeof(int));

    for(i=0;i<steps;i++) {
      for(y=0;y<2*rad;y++) {
	j=get_line(((float)y-(float)rad)/(float)-rad,(360.0*(float)i)/(float)steps -180.0);

	/* scale radius */
	for(r=0;r<j;r++)
	  intpos[r] = rad + pos[r] * rad;

	/* remove identical fields */
	for(r=1;r<j;r++) {
	  if(intpos[r-1] == intpos[r]) {
	    for(h=r;h<j;h++) {
	      intpos[h]=intpos[h+1];
	      col[h]=col[h+1];
	    }
	    j--; r--;
	  }
	}

	/* search in line buffer for identical line */ 
	for(hit=0,s=0;(s<lines)&&(!hit);s++) {
	  /* correct length?? */
	  if(line_cnt[s]==j) {
	    /* posfield and colfield match?? */
	    if(( memcmp(&line_pos[12*s],intpos,j*sizeof(int)) == 0 ) &&
	       ( memcmp(&line_col[12*s],   col,j*sizeof(int)) == 0 )) {
	      smile_line[y] = s;
	      hit=1;
	    }
	  }
	}

	if(!hit) {
	  /* and save line to buffer */
	  memcpy(&line_pos[12*lines], intpos, j*sizeof(int));
	  memcpy(&line_col[12*lines],    col, j*sizeof(int));
	  line_cnt[lines] = j;
	  smile_line[y] = lines++;
	}
      }

      /* search for smiley in buffer */
      for(hit=0,s=0;s<smiles;s++) {
	if(memcmp(&smiley[2*rad*s], smile_line, 2*rad*sizeof(int)) == 0) {
	  smiley_table[i] = s;
	  hit=1;
	}
      }

      if(!hit) {
	/* store smiley in buffer */
	memcpy(&smiley[2*rad*smiles], smile_line, 2*rad*sizeof(int));
	smiley_table[i]=smiles++;
      }

      printf(" %2d%%\r",(100*i)/steps);
      fflush(stdout);
    }

    printf("%d(%d) lines in %d(%d) smileys\n",lines,2*rad*steps,smiles,steps);

    if((file=fopen("smiletab.c","w"))==NULL) {
      printf("unable to open smiletab.c\n");
      exit(1);
    }

    fprintf(file,"/* smiletab.c - auto generated smiley tables for TheMaze */\n");
    fprintf(file,"/* contains smiley data for %d smileys with radius %d */\n\n", steps, rad);

    fprintf(file,"#ifdef __X11__\n#include \"../ctypes.h\"\n#else\n#include \"ctypes.h\"\n#endif\n\n");

    /*  file=stdout; */

    /* output lines */
    fprintf(file,"tm_ushort smiley_lines[]={\n"); 
    for(s=0,i=0;i<lines;i++) {
      j=line_cnt[i];
      fprintf(file,"0x%04x%s",
	      ((j-1)<<8)|line_pos[12*i],((j!=1)||(i!=lines-1))?",":"" );
      for(r=1;r<j;r++)
	fprintf(file,"0x%04x%s",
		(line_col[12*i+r]<<8)|(line_pos[12*i+r]-line_pos[12*i+r-1]),
		((r!=j-1)||(i!=lines-1))?",":"");
      line_cnt[i]=s;
      s+=j;
      fprintf(file,"\n");
    }
    fprintf(file,"};\n\ntm_ushort smiley[]={\n");
    for(i=0;i<smiles;i++) {
      for(y=0;y<2*rad;y++)
	fprintf(file,"0x%04x%s",line_cnt[smiley[i*2*rad+y]],((i!=smiles-1)||(y!=(2*rad)-1))?",":"");
      fprintf(file,"\n");
    }
    fprintf(file,"};\n\ntm_ushort smiley_table[]={");
    for(i=0;i<steps;i++)
      fprintf(file,"%d%s",2*rad*smiley_table[i],(i==steps-1)?"":",");
    fprintf(file,"};\n/* --- EOF --- */\n");

    /* Pufferspeicher freigeben */
    /* Zeilen */
    free(line_cnt); free(line_col); free(line_pos); free(smile_line);

    /* Smiley */
    free(smiley); free(smiley_table);

    fclose(file);
  }

  return 0;
}
