/*
    texture.c  -  texture mapping for TheMaze 

    initialisation and preparation of textures used by drawmaze.c and r256_lin.c

    TODO:
    + shading (black fog)
    - mipmapping
    - floor
    - sky
*/

/*
 * Portions Copyright (C) 1989, 1991 by Jef Poskanzer.  See copyright notice
 * below, at the beginning of the relevant code.
 */

#include "machine.h"
#ifdef TEXTURE
#include <stdlib.h>
#include "ctypes.h"
#include "texture.h"
#include "grafik.h"
#include "net.h"

/* include textures (xv is known to create xpm's that cannot be compiled */
/* with a c compiler, use xpaint instead to convert those xpm's) */
#include "textures/st_wall.xpm"
#include "textures/st_door.xpm"

#undef DEBUG
/* the following values need fine tuning */
#define TEX_BLACKLEVEL  (10)     /* minimum brightness after shading */
#define TEX_LOG                  /* use logarithmic color weighting */
#define TEX_DARK_WEIGHT (95)     /* weight reduces to this % every step into the depth */
          
typedef unsigned char pixval;

#define PPM_MAXMAXVAL 255
typedef struct { pixval r, g, b; } pixel;

static char **tex_xpm[TEXTURES]={st_door_xpm, st_wall_xpm};
#ifdef TEXTURE_MIR
tm_char textures[TEXTURES][2*TEXTURE_WIDTH][TEXTURE_HEIGHT];   /* at least texture for wall and door */
#else
tm_char textures[TEXTURES][TEXTURE_WIDTH][TEXTURE_HEIGHT];
#endif

/* one colormap for each texture */
pixel   texture_colors[TEXTURES][256];
tm_char texture_col_allocated[TEXTURES][256];

/* one colormap for arch texture/shade */
tm_uchar texture_colmap[TEXTURES][TEX_SHADES][256];

static tm_bool texture_quant(tm_void);

tm_word tex_cols[TEXTURES], tex_rows[TEXTURES], tex_colors[TEXTURES];

tm_void 
init_texture(tm_void) {
  tm_word  i,n, bytes, col_cnt, x, y, idx, m, j, k;
  tm_uword rgb[3], rawcols, gescols;
  tm_word mindist;
  tm_char chr;

  if(use_texture) {

    col_cnt = TEX_COL_START;

    for(i=0;i<TEXTURES;i++) {

      /* clear colortable flags */
      for(n=0; n<256; n++)
	texture_col_allocated[i][n]=FALSE;

      sscanf(tex_xpm[i][0], "%d %d %d %d", &tex_cols[i], &tex_rows[i], &tex_colors[i], &bytes);

      /* up to now all textures have the same size (TEXTURE_WIDTH * TEXTURE_HEIGHT) */
      /* but this may change sometime, so save the size ... */

      if((tex_rows[i]!=TEXTURE_HEIGHT)||(tex_cols[i]!=TEXTURE_WIDTH)||(bytes>1))
	printf("texture.c: illegal texture bitmap format\n");

      /* read colors from xpm */
      for(n=0;n<tex_colors[i];n++) {
	sscanf(tex_xpm[i][n+1], "%c c #%2x%*2x%2x%*2x%2x%*2x\n",&chr, &rgb[0], &rgb[1], &rgb[2] );

#ifdef DEBUG
	 printf("%c(%3d) red=0x%02x green=0x%02x blue=0x%02x\n",
		chr, (tm_word)chr, rgb[0],rgb[1],rgb[2]);
#endif

	 /* save colormap entry */
	 texture_colors[i][chr].r = rgb[0];
	 texture_colors[i][chr].g = rgb[1];
	 texture_colors[i][chr].b = rgb[2];
	 texture_col_allocated[i][chr] = TRUE;
      }
    }

    /* quantize textures */
    texture_quant();

    /* create texture maps */
    for(i=0;i<TEXTURES;i++) {
      /* ok, translate the bitmap */
      for(y=0;y<tex_rows[i];y++)
	for(x=0;x<tex_cols[i];x++)
#ifdef TEXTURE_MIR
	  textures[i][(2*TEXTURE_WIDTH)-1-x][y]=
#endif
	    textures[i][x][y]=tex_xpm[i][y+tex_colors[i]+1][x];
    }
  }
}


/***************************************************************/
/* The following code based on code from the 'pbmplus' package */
/* written by Jef Poskanzer                                    */
/***************************************************************/


/* ppmquant.c - quantize the colors in a pixmap down to a specified number
**
** Copyright (C) 1989, 1991 by Jef Poskanzer.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.
*/


#define PPM_GETR(p) ((p).r)
#define PPM_GETG(p) ((p).g)
#define PPM_GETB(p) ((p).b)

#define PPM_ASSIGN(p,red,grn,blu) \
  { (p).r = (red); (p).g = (grn); (p).b = (blu); }

#define PPM_EQUAL(p,q) ( (p).r == (q).r && (p).g == (q).g && (p).b == (q).b )


/* Color scaling macro -- to make writing ppmtowhatever easier. */

#define PPM_DEPTH(newp,p,oldmaxval,newmaxval) \
    PPM_ASSIGN( (newp), \
	        ((int) PPM_GETR(p)) * ((int)newmaxval) / ((int)oldmaxval), \
	        ((int) PPM_GETG(p)) * ((int)newmaxval) / ((int)oldmaxval), \
	        ((int) PPM_GETB(p)) * ((int)newmaxval) / ((int)oldmaxval) )


/* Luminance macro. */

/* 
 * #define PPM_LUMIN(p) \
 *   ( 0.299 * PPM_GETR(p) + 0.587 * PPM_GETG(p) + 0.114 * PPM_GETB(p) )
 */

/* Luminance macro, using only integer ops.  Returns an int (*256)  JHB */
#define PPM_LUMIN(p) \
  ( 77 * PPM_GETR(p) + 150 * PPM_GETG(p) + 29 * PPM_GETB(p) )

/* Color histogram stuff. */

typedef struct chist_item* chist_vec;
struct chist_item { pixel color;
			int value;
		      };

typedef struct chist_list_item* chist_list;
struct chist_list_item { struct chist_item ch;
			     chist_list next;
			   };

typedef chist_list* chash_table;

typedef struct box* box_vector;
struct box {
  int index;
  int colors;
  int sum;
};

int pixel_sum;

#define MAXCOLORS 32767
#define CLUSTER_MAXVAL 63

#define LARGE_LUM
#define REP_AVERAGE_PIXELS

#define FS_SCALE 1024

#define HASH_SIZE 6553

#define ppm_hashpixel(p) ((((int) PPM_GETR(p) * 33023 +    \
			    (int) PPM_GETG(p) * 30013 +    \
			    (int) PPM_GETB(p) * 27011) & 0x7fffffff)   \
			  % HASH_SIZE)

static chist_vec   mediancut(chist_vec, int, int, int);
static int         redcompare(const void *, const void *);
static int         greencompare(const void *, const void *);
static int         bluecompare(const void *, const void *);
static int         sumcompare(const void *, const void *);
static chist_vec   ppm_computechist(int,int *);
static chash_table ppm_computechash(int,int *);
static chist_vec   ppm_chashtochist(chash_table, int);
static chash_table ppm_allocchash(void);
static void        ppm_freechist(chist_vec);
static void        ppm_freechash(chash_table);


/****************************************************************************/

static tm_bool
texture_quant(tm_void)
{
  pixel**          pixels;
  register pixel*  pP;
  int              row;
  register int     col, limitcol;
  pixval           maxval, newmaxval;
  int              colors;
  register int     index;
  chist_vec chv, colormap;
  chash_table  cht;
  int              i;
  unsigned char    *picptr;
  static char      *fn = "ppmquant()";

  int tex, shade, color;
  tm_word rgb[3];

  index = 0;
  maxval = 255;

  /*
   *  attempt to make a histogram of the colors, unclustered.
   *  If at first we don't succeed, lower maxval to increase color
   *  coherence and try again.  This will eventually terminate, with
   *  maxval at worst 15, since 32^3 is approximately MAXCOLORS.
   */

  for ( ; ; ) {
#ifdef DEBUG
    printf("%s: making histogram\n", fn);
#endif

    chv = ppm_computechist(MAXCOLORS, &colors);
    if (chv != (chist_vec) 0) break;
    
    printf("%s: too many colors!\n", fn);

    newmaxval = maxval / 2;
#ifdef DEBUG
    printf("%s: rescaling colors (maxval=%d) %s\n",
	   fn, newmaxval, "to improve clustering");
#endif
    printf("texture.c: internal error xxx\n");
#if 0
    for (row=0; row<rows; ++row)
      for (col=0, pP=pixels[row]; col<cols; ++col, ++pP)
	PPM_DEPTH( *pP, *pP, maxval, newmaxval );
    maxval = newmaxval;
#endif
  }

#ifdef DEBUG
  printf("%s: %d colors found\n", fn, colors);
#endif


  /*
   * Step 3: apply median-cut to histogram, making the new colormap.
   */

#ifdef DEBUG
  printf("%s: choosing %d colors\n", fn, TEX_COL_NUM);
#endif
  colormap = mediancut(chv, colors, pixel_sum, maxval);
  ppm_freechist(chv);

  /*
   *  Step 4: map the colors in the image to their closest match in the
   *  new colormap, and write 'em out.
   */

#ifdef DEBUG
  printf("%s: mapping image to new colors\n", fn);
#endif
  cht = ppm_allocchash();

  /* find closest colors in colormap */

  for(tex=0; tex<TEXTURES; tex++) {
    for(color=0; color<256; color++) {
      if(texture_col_allocated[tex][color]) {

	for(shade=0;shade<TEX_SHADES;shade++) {
	  /* search colormap for closest match. */
	  register int i, r1, g1, b1, r2, g2, b2;
	  register long dist, newdist;

	  /* search for this shade of the color */
	  r1 = texture_colors[tex][color].r - (((((TEX_SHADES-1-shade) * texture_colors[tex][color].r) / TEX_SHADES)*(100-TEX_BLACKLEVEL))/100); 
	  g1 = texture_colors[tex][color].g - (((((TEX_SHADES-1-shade) * texture_colors[tex][color].g) / TEX_SHADES)*(100-TEX_BLACKLEVEL))/100); 
	  b1 = texture_colors[tex][color].b - (((((TEX_SHADES-1-shade) * texture_colors[tex][color].b) / TEX_SHADES)*(100-TEX_BLACKLEVEL))/100); 
#if 0
	  r1 = (texture_colors[tex][color].r * shade)/TEX_SHADES ;
	  g1 = (texture_colors[tex][color].g * shade)/TEX_SHADES ;
	  b1 = (texture_colors[tex][color].b * shade)/TEX_SHADES ;
#endif
	  dist = 2000000000;
	
	  for (i=0; i< TEX_COL_NUM; i++) {
	    r2 = PPM_GETR( colormap[i].color );
	    g2 = PPM_GETG( colormap[i].color );
	    b2 = PPM_GETB( colormap[i].color );

	    newdist = ( r1 - r2 ) * ( r1 - r2 ) +
	              ( g1 - g2 ) * ( g1 - g2 ) +
	              ( b1 - b2 ) * ( b1 - b2 );
      
	    if (newdist<dist) { index = i;  dist = newdist; }
	  }
	  /* enter index into color table */
	  texture_colmap[tex][shade][color] = index+TEX_COL_START;
	}
      }
    }
  }

  /* set colormap */
  for(color=0;color<TEX_COL_NUM;color++) {
    rgb[0] = (colormap[color].color.r *200)/51;
    rgb[1] = (colormap[color].color.g *200)/51;
    rgb[2] = (colormap[color].color.b *200)/51;
    set_color(color+TEX_COL_START, rgb);   /* create color entry */
  }

  /* free cht and colormap */
  ppm_freechist(colormap);
  ppm_freechash(cht);

  return 0;
}



/*
** Here is the fun part, the median-cut colormap generator.  This is based
** on Paul Heckbert's paper "Color Image Quantization for Frame Buffer
** Display", SIGGRAPH '82 Proceedings, page 297.
*/



/****************************************************************************/
static chist_vec mediancut( chv, colors, sum, maxval)
     chist_vec chv;
     int colors, sum;
     int maxval;
{
  chist_vec colormap;
  box_vector bv;
  register int bi, i;
  int boxes;

  bv = (box_vector) malloc(sizeof(struct box) * TEX_COL_NUM);
  colormap = (chist_vec) 
             malloc(sizeof(struct chist_item) * TEX_COL_NUM );

  if (!bv || !colormap) puts("unable to malloc in mediancut()");

  for (i=0; i<TEX_COL_NUM; i++)
    PPM_ASSIGN(colormap[i].color, 0, 0, 0);

  /*
   *  Set up the initial box.
   */
  bv[0].index = 0;
  bv[0].colors = colors;
  bv[0].sum = sum;
  boxes = 1;


  /*
   ** Main loop: split boxes until we have enough.
   */

  while ( boxes < TEX_COL_NUM ) {
    register int indx, clrs;
    int sm;
    register int minr, maxr, ming, maxg, minb, maxb, v;
    int halfsum, lowersum;

    /*
     ** Find the first splittable box.
     */
    for (bi=0; bv[bi].colors<2 && bi<boxes; bi++) ;
    if (bi == boxes) break;	/* ran out of colors! */

    indx = bv[bi].index;
    clrs = bv[bi].colors;
    sm = bv[bi].sum;

    /*
     ** Go through the box finding the minimum and maximum of each
     ** component - the boundaries of the box.
     */
    minr = maxr = PPM_GETR( chv[indx].color );
    ming = maxg = PPM_GETG( chv[indx].color );
    minb = maxb = PPM_GETB( chv[indx].color );

    for (i=1; i<clrs; i++) {
      v = PPM_GETR( chv[indx + i].color );
      if (v < minr) minr = v;
      if (v > maxr) maxr = v;

      v = PPM_GETG( chv[indx + i].color );
      if (v < ming) ming = v;
      if (v > maxg) maxg = v;

      v = PPM_GETB( chv[indx + i].color );
      if (v < minb) minb = v;
      if (v > maxb) maxb = v;
    }

    /*
     ** Find the largest dimension, and sort by that component.  I have
     ** included two methods for determining the "largest" dimension;
     ** first by simply comparing the range in RGB space, and second
     ** by transforming into luminosities before the comparison.  You
     ** can switch which method is used by switching the commenting on
     ** the LARGE_ defines at the beginning of this source file.
     */
    {
      /* LARGE_LUM version */

      pixel p;
      int rl, gl, bl;

      PPM_ASSIGN(p, maxr - minr, 0, 0);
      rl = PPM_LUMIN(p);

      PPM_ASSIGN(p, 0, maxg - ming, 0);
      gl = PPM_LUMIN(p);

      PPM_ASSIGN(p, 0, 0, maxb - minb);
      bl = PPM_LUMIN(p);

      if (rl >= gl && rl >= bl)
	qsort((char*) &(chv[indx]), (size_t) clrs, sizeof(struct chist_item),
	      redcompare );
      else if (gl >= bl)
	qsort((char*) &(chv[indx]), (size_t) clrs, sizeof(struct chist_item),
	      greencompare );
      else 
	qsort((char*) &(chv[indx]), (size_t) clrs, sizeof(struct chist_item),
	      bluecompare );
    }

    /*
     ** Now find the median based on the counts, so that about half the
     ** pixels (not colors, pixels) are in each subdivision.
     */
    lowersum = chv[indx].value;
    halfsum = sm / 2;
    for (i=1; i<clrs-1; i++) {
      if (lowersum >= halfsum) break;
      lowersum += chv[indx + i].value;
    }

    /*
     ** Split the box, and sort to bring the biggest boxes to the top.
     */
    bv[bi].colors = i;
    bv[bi].sum = lowersum;
    bv[boxes].index = indx + i;
    bv[boxes].colors = clrs - i;
    bv[boxes].sum = sm - lowersum;
    ++boxes;
    qsort((char*) bv, (size_t) boxes, sizeof(struct box), sumcompare);
  }  /* while (boxes ... */
  
  /*
   ** Ok, we've got enough boxes.  Now choose a representative color for
   ** each box.  There are a number of possible ways to make this choice.
   ** One would be to choose the center of the box; this ignores any structure
   ** within the boxes.  Another method would be to average all the colors in
   ** the box - this is the method specified in Heckbert's paper.  A third
   ** method is to average all the pixels in the box.  You can switch which
   ** method is used by switching the commenting on the REP_ defines at
   ** the beginning of this source file.
   */
  
  for (bi=0; bi<boxes; bi++) {
    /* REP_AVERAGE_PIXELS version */
    register int indx = bv[bi].index;
    register int clrs = bv[bi].colors;
    register long r = 0, g = 0, b = 0, sum = 0;

    for (i=0; i<clrs; i++) {
      r += PPM_GETR( chv[indx + i].color ) * chv[indx + i].value;
      g += PPM_GETG( chv[indx + i].color ) * chv[indx + i].value;
      b += PPM_GETB( chv[indx + i].color ) * chv[indx + i].value;
      sum += chv[indx + i].value;
    }

    r = r / sum;  if (r>maxval) r = maxval;	/* avoid math errors */
    g = g / sum;  if (g>maxval) g = maxval;
    b = b / sum;  if (b>maxval) b = maxval;

    PPM_ASSIGN( colormap[bi].color, r, g, b );
  }

  free(bv);
  return colormap;
}


/**********************************/
static int redcompare(p1, p2)
     const void *p1, *p2;
{
  return (int) PPM_GETR( ((chist_vec)p1)->color ) - 
         (int) PPM_GETR( ((chist_vec)p2)->color );
}

/**********************************/
static int greencompare(p1, p2)
     const void *p1, *p2;
{
  return (int) PPM_GETG( ((chist_vec)p1)->color ) - 
         (int) PPM_GETG( ((chist_vec)p2)->color );
}

/**********************************/
static int bluecompare(p1, p2)
     const void *p1, *p2;
{
  return (int) PPM_GETB( ((chist_vec)p1)->color ) - 
         (int) PPM_GETB( ((chist_vec)p2)->color );
}

/**********************************/
static int sumcompare(p1, p2)
     const void *p1, *p2;
{
  return ((box_vector) p2)->sum - ((box_vector) p1)->sum;
}



/****************************************************************************/
static chist_vec 
  ppm_computechist(maxcolors, colorsP)
     int maxcolors;
     int* colorsP;
{
  chash_table cht;
  chist_vec chv;

  cht = ppm_computechash(maxcolors, colorsP);
  if (!cht) return (chist_vec) 0;

  chv = ppm_chashtochist(cht, maxcolors);
  ppm_freechash(cht);
  return chv;
}

/****************************************************************************/
static chash_table ppm_computechash(maxcolors, colorsP )
     int maxcolors;
     int* colorsP;
{
  chash_table cht;
  register pixel pP, pOrg;
  chist_list chl;
  int col, row, hash, tex, shade;
#ifdef TEX_LOG
  int r;
#endif

  cht = ppm_allocchash( );
  *colorsP = 0;
  pixel_sum = 0;

  /* scan all textures */
  for(tex=0;tex<TEXTURES;tex++) {

    /* Go through the entire image, building a hash table of colors. */
    for (row=0; row<tex_rows[tex]; row++)
      for (col=0; col<tex_cols[tex]; col++) {

	/* get pixel value */
	pOrg = texture_colors[tex][tex_xpm[tex][row+tex_colors[tex]+1][col]];

#ifdef TEX_LOG
          /* with increasing depth lesser pixels are added */
          /* to the histogram. This gives (hopefully) better */
          /* color resolution for the foreground colors */
#define INCR (r)
	for(r=999,shade=0;shade<TEX_SHADES;shade++,r=MAX((TEX_DARK_WEIGHT*r)/100,1)) {
#else
#define INCR (1)
	for(shade=0;shade<TEX_SHADES;shade++) {
#endif

	  pP.r = pOrg.r - ((((shade * pOrg.r) / TEX_SHADES)*(100-TEX_BLACKLEVEL))/100); 
	  pP.g = pOrg.g - ((((shade * pOrg.g) / TEX_SHADES)*(100-TEX_BLACKLEVEL))/100); 
	  pP.b = pOrg.b - ((((shade * pOrg.b) / TEX_SHADES)*(100-TEX_BLACKLEVEL))/100); 

	  hash = ppm_hashpixel(pP);
          pixel_sum+=INCR;

          /* search color in hash table */
	  for (chl = cht[hash]; chl != (chist_list) 0; chl = chl->next)
	    if (PPM_EQUAL(chl->ch.color, pP)) break;
      
	  if (chl != (chist_list) 0) 
            (chl->ch.value)+=INCR;
	  else {
	    if ((*colorsP)++ > maxcolors) {
	      ppm_freechash(cht);
	      return (chash_table) 0;
	    }
	
	    chl = (chist_list) malloc(sizeof(struct chist_list_item));
	    if (!chl) puts("ran out of memory computing hash table");
	  
            /* create new entry */
	    chl->ch.color = pP;
	    chl->ch.value = INCR;
	    chl->next = cht[hash];
	    cht[hash] = chl;
	  }
	}
      }
    }
  
  return cht;
}


/****************************************************************************/
static chash_table ppm_allocchash()
{
  chash_table cht;
  int i;

  cht = (chash_table) malloc( HASH_SIZE * sizeof(chist_list) );
  if (!cht) puts("ran out of memory allocating hash table");

  for (i=0; i<HASH_SIZE; i++ )
    cht[i] = (chist_list) 0;

  return cht;
}


/****************************************************************************/
static chist_vec ppm_chashtochist( cht, maxcolors )
     chash_table cht;
     int maxcolors;
{
  chist_vec chv;
  chist_list chl;
  int i, j;

  /* Now collate the hash table into a simple chist array. */
  chv = (chist_vec) malloc( maxcolors * sizeof(struct chist_item) );

  /* (Leave room for expansion by caller.) */
  if (!chv) puts("ran out of memory generating histogram");

  /* Loop through the hash table. */
  j = 0;
  for (i=0; i<HASH_SIZE; i++)
    for (chl = cht[i];  chl != (chist_list) 0;  chl = chl->next) {
      /* Add the new entry. */
      chv[j] = chl->ch;
      ++j;
    }

  return chv;
}


/****************************************************************************/
static void ppm_freechist( chv )
     chist_vec chv;
{
  free( (char*) chv );
}


/****************************************************************************/
static void ppm_freechash( cht )
     chash_table cht;
{
  int i;
  chist_list chl, chlnext;

  for (i=0; i<HASH_SIZE; i++)
    for (chl = cht[i];  chl != (chist_list) 0; chl = chlnext) {
      chlnext = chl->next;
      free( (char*) chl );
    }

  free( (char*) cht );
}

#endif /* TEXTURE */

