/*
   console.c

   Ein-/Ausgabe auf der Konsole taetigen (so vorhanden)

   8.7.95    Till  erste Version
   27.11.96  Chip  Typen auf tm_* umgestellt
   11.05.97  Chip  TOS-Version vervollstaendigt
   12.07.97  Chip  Verhalten korrigiert
*/

#include "ctypes.h"
#include "machine.h"
#include "console.h"
#include "error.h"
#include "misc.h"
#include "tos\con.h"
#include "tos\screen.h"

#include <ext.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <tos.h>

/* this functon is for signal() */
static
void error_exit(int no)
{
  tm_char msg[30];

  sprintf(msg,"signal #%d caught!\n",no);
  do_error((tm_uword)no,msg);
  wait_msec(500);
  exit(no);
}

static tm_bool  console_used;

tm_void console_init(tm_void)
{
  console_used=FALSE;
  signal(SIGSEGV,error_exit);
  signal(SIGILL,error_exit);
  signal(SIGABRT,error_exit);
  signal(SIGADR,error_exit);
  signal(SIGDIV0,error_exit);
  signal(SIGPRIV,error_exit);
}

/* Achtung, wird bei normalem Ende mehrfach aufgerufen! */
tm_void console_exit(tm_void)
{
  if (console_used && isatty(fileno(stderr)))
  {
    fputs("<Enter>",stderr);
    getchar();
  }
  console_used=FALSE;
}

tm_void console_printf(tm_char *fmt, ...) 
{
  va_list args;
  void    *now;

  if (!console_used && isatty(fileno(stderr)))
    fputs("\033E\033v",stderr);

  console_used=TRUE;
  now=Logbase();
  if (logbase!=NULL)
    Setscreen(logbase,(void*)-1L,-1);

  va_start(args,fmt);
  vfprintf(stderr,fmt,args);
  va_end(args);

  Setscreen(now,(void*)-1L,-1);
}

tm_void console_puts(tm_char *string)
{
  void    *now;

  if (!console_used && isatty(fileno(stderr)))
    fputs("\033E",stderr);

  console_used=TRUE;
  now=Logbase();
  if (logbase!=NULL)
    Setscreen(logbase,(void*)-1L,-1);

  fputs(string,stderr);
  fputs("\n",stderr);

  Setscreen(now,(void*)-1L,-1);
}

tm_void console_waitkey(tm_void)
{
  void    *lnow,*pnow;

  lnow=Logbase();
  pnow=Physbase();
  if (logbase!=NULL && physbase!=NULL)
    Setscreen(logbase,physbase,-1);

  if (isatty(fileno(stderr)))
    wait_4_button();

  Setscreen(lnow,pnow,-1);
}
