/*
  joystick.c

  Joystick-Treiber fr ST

  03.03.96  Chip   Maustreiber
  10.11.96  Chip   Verbesserungen
  27.11.96  Chip   Typen auf tm_* umgestellt
  14.03.97  Chip   sauberes Ende bei Absturz versucht
  20.03.97  Chip   alle Treiber fertig
*/

/*
   ACHTUNG, muss wegen der Interrupt-Funktion
   ohne Stackueberpruefung compiliert werden!
*/

#include <stdlib.h>
#include <tos.h>

#include "ctypes.h"
#include "joystick.h"
#include "console.h"

tm_char joystick_driver_name[]="IKBD driver V1.1";

/*
   prinzipiell waeren alle Treiber parallel moeglich,
   das beisst sich aber mit der Angabe, ob das device
   relativ oder absolut ist...
*/

static tm_uword joy_inp;
enum { JOY_MOUSE=0, JOY_KBD, JOY_JOY };

tm_void show_joystick_options(tm_void) 
{
  console_printf("%s input driver options:\n",joystick_driver_name);
  console_puts("id=d\tset input device to mouse (d=m), keyboard (d=k) or\n"
               "\tjoystick 1 (d=j), default is mouse");
}

tm_void init_joystick_options(tm_void)
{
  joy_inp=JOY_MOUSE;
}

tm_bool set_joystick_option(tm_char *cmd)
{
  if (cmd[0]=='d' && cmd[1]=='=' && cmd[3]=='\0')
    switch (cmd[2])
    {
    case 'm': joy_inp=JOY_MOUSE;  return(OK);
    case 'k': joy_inp=JOY_KBD;    return(OK);
    case 'j': joy_inp=JOY_JOY;    return(OK);
    }
  return(FALSE);
}

static  tm_uword   fkeys;
static  tm_uword   m_joy,k_joy,j_joy;
static  tm_word    mx,my;

static	KBDVBASE   *vbase;
static	tm_void    (*old_ikbdsys)(tm_void);
static	tm_ubyte   state;
enum { FIRST=0, MX, MY, JS };	/* constants for state */

static
tm_void my_ikbdsys(tm_void)
{
  register tm_ubyte st;
  register tm_ubyte da;

  st=*((tm_ubyte *)0xFFFFFC00L);
  da=*((tm_ubyte *)0xFFFFFC02L);
  if (st&0x70)			/* nearly no error service... */
    return;

  if (state==MX) {			/* second mouse byte */
    mx+=(tm_byte)da;
    state=MY;
  } else if (state==MY) {		/* third mouse byte */
    my+=(tm_byte)da;
    state=FIRST;

  } else if (state==JS) {		/* second joystick byte */
    if (da&0x01)  j_joy|=JOY_UP;     else  j_joy&=~JOY_UP;
    if (da&0x02)  j_joy|=JOY_DOWN;   else  j_joy&=~JOY_DOWN;
    if (da&0x04)  j_joy|=JOY_LEFT;   else  j_joy&=~JOY_LEFT;
    if (da&0x08)  j_joy|=JOY_RIGHT;  else  j_joy&=~JOY_RIGHT;
    /* the fire key is reported by mouse packets */
    state=FIRST;

  } else {			/* first byte */
    state=FIRST;
    if (da&0x80) {		/* special / key released */
      switch (da) {
      case 0xFE:
      case 0xFF:
        state=JS;
        return;
      case 0xFB:
        m_joy|=JOY_FIRE0;
        m_joy|=JOY_FIRE1;
        state=MX;
        return;
      case 0xFA:
        m_joy|=JOY_FIRE0;
        m_joy&=~JOY_FIRE1;
        state=MX;
        return;
      case 0xF9:
        m_joy&=~JOY_FIRE0;
        m_joy|=JOY_FIRE1;
        j_joy|=JOY_FIRE0;
        state=MX;
        return;
      case 0xF8:
        m_joy&=~JOY_FIRE0;
        m_joy&=~JOY_FIRE1;
        j_joy&=~JOY_FIRE0;
        state=MX;
        return;
      case 0xB8: k_joy &= ~JOY_FIRE0; return;  /* ALT */
      case 0xAA: k_joy &= ~JOY_FIRE2; return;  /* SHIFT-L */
      case 0xB6: k_joy &= ~JOY_FIRE2; return;  /* SHIFT-R */
      case 0x9D: k_joy &= ~JOY_FIRE1; return;  /* CTRL */
      case 0xC8: k_joy &= ~JOY_UP;    return;
      case 0xD0: k_joy &= ~JOY_DOWN;  return;
      case 0xCB: k_joy &= ~JOY_LEFT;  return;
      case 0xCD: k_joy &= ~JOY_RIGHT; return;
      default:
        if (da>=0xBB && da<=0xC4)
          fkeys &= ~(1<<(da-0x3B));
        return;
        }

    } else {			/* key pressed */

      switch (da) {
      case 0x38: k_joy |= JOY_FIRE0; return;  /* ALT */
      case 0x2A: k_joy |= JOY_FIRE2; return;  /* SHIFT-L */
      case 0x36: k_joy |= JOY_FIRE2; return;  /* SHIFT-R */
      case 0x1D: k_joy |= JOY_FIRE1; return;  /* CTRL */
      case 0x48: k_joy |= JOY_UP;    return;
      case 0x50: k_joy |= JOY_DOWN;  return;
      case 0x4B: k_joy |= JOY_LEFT;  return;
      case 0x4D: k_joy |= JOY_RIGHT; return;
      default:
        if (da>=0x3B && da<=0x44)
          fkeys |= 1<<(da-0x3B);
        return;
        }
      }
    }
}

static
tm_void init_ikbdsys(tm_void)
{
  state=FIRST;

  vbase=Kbdvbase();
  old_ikbdsys=vbase->kb_kbdsys;
  vbase->kb_kbdsys=my_ikbdsys;
}

/* Achtung, wird bei normalem Ende zweimal ausgefhrt! */
static
tm_void exit_ikbdsys(tm_void)
{
  vbase->kb_kbdsys=old_ikbdsys;
}


/* Funktionstasten */

tm_bool init_fkeys(tm_void)
{
  fkeys = 0;
  return(FUNKEYS);
}

tm_void exit_fkeys(tm_void)
{
}

tm_uword get_fkeys(tm_void)
{
  return(fkeys);
}

/* der Haupttreiber (LOS, VORWAERTS!) */

#define X_MAX        (100)
#define Y_MAX        (100)
#define X_THRESHOLD  (3*X_MAX/4)
#define Y_THRESHOLD  (3*Y_MAX/4)

tm_bool init_joystick(tm_void)
{
  m_joy=0;
  k_joy=0;
  j_joy=0;
  mx=0;
  my=0;

  init_ikbdsys();
  atexit(exit_ikbdsys);
  if (joy_inp==JOY_MOUSE)
    return(ABSOLUTE_DEVICE|JITTER_DEVICE);
  else
    return(RELATIVE_DEVICE);
}

tm_void exit_joystick(tm_void)
{
  exit_ikbdsys();
}

tm_uword get_joystick(tm_void)
{
  tm_uword joy;

  switch (joy_inp)
  {
  case JOY_MOUSE:
    joy=(m_joy&JOY_FIRE) | (k_joy&JOY_FIRE);

    if (mx<-X_MAX)  mx=-X_MAX;  else if (mx>+X_MAX)  mx=+X_MAX;
    if (my<-Y_MAX)  my=-Y_MAX;  else if (my>+Y_MAX)  my=+Y_MAX;

    if (mx<-X_THRESHOLD)       joy|=JOY_LEFT;
    else if (mx>+X_THRESHOLD)  joy|=JOY_RIGHT;
    if (my<-Y_THRESHOLD)       joy|=JOY_UP;
    else if (my>+Y_THRESHOLD)  joy|=JOY_DOWN;
    return(joy);
  case JOY_KBD:
    return(k_joy);
  default:
    return(j_joy | (k_joy&JOY_FIRE));
  }
}

tm_uword get_joystick_analog(tm_word *x, tm_word *y)
{
  tm_uword joy=get_joystick();

  if (joy_inp==JOY_MOUSE)
  {
    *x=mx;
    *y=my;
  }
  else
  {
    *x=0;
    *y=0;
    if(joy&JOY_LEFT)       *x=-100;
    else if(joy&JOY_RIGHT) *x=+100;
    if(joy&JOY_UP)         *y=-100;
    else if(joy&JOY_DOWN)  *y=+100;
  }
  return(joy);
}

tm_uword kind_of_joystick(tm_void)
{
  if (joy_inp==JOY_MOUSE)
    return(ABSOLUTE_DEVICE|JITTER_DEVICE);
  else
    return(RELATIVE_DEVICE);
}
