/*
   scrraw2.c

   Bildschirm-Treiber fuer Atari ST monochrome 640 x 400

   03.03.96  Chip  erste Version
   12.11.96  Chip  weiter bearbeitet
   13.11.96  Chip  Titelbildbearbeitung geaendert
   14.11.96  Chip  korrektes Schwarz und Weiss
   26.11.96  Chip  put_bitmap mit 16 Bit
   28.11.96  Chip  Typen auf tm_* umgestellt
   01.12.96  Chip  BIG_ENDIAN und LITTLE_ENDIAN unterstuetzt
   26.01.97  Chip  Copyrighttext als "richtige" Textausgabe
   05.03.97  Chip  eigene Wandfunktionen
   22.06.97  Chip  console... und error... korrigiert
*/

#define noNO_FLIP

#include <aes.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <tos.h>
#include <vdi.h>

#include "ctypes.h"
#include "error.h"
#include "gr_prim.h"
#include "grafik.h"
#include "text.h"
#include "tos\picraw2.h"
#include "tos\screen.h"
#include "tos\con.h"

tm_char video_driver_name[]="Intern high V1.3";

#define USE_CLIB    /* Grafikprimitiva duerfen string.h benutzen */

#define WIDTH     (640L)
#define HEIGHT    (400L)
#define COLORS    (2)

#if TITLE_PIC_WIDTH!=WIDTH || TITLE_PIC_HEIGHT!=HEIGHT	\
    || TITLE_PIC_COLORS!=COLORS
#error "Falsches Titelbild!"
#endif

tm_void show_video_options(tm_void) 
{
}

tm_void init_video_options(tm_void)
{
  console_init();
  atexit(console_exit);
}

tm_bool set_video_option(tm_char *cmd)
{
  (tm_void)cmd;
  return(FALSE);
}

tm_uword open_rate;   /* wie weit bei open_screen() gekommen? */

int	vdi_handle;	/* fuer GEM, ganz primitiv */

static  tm_ulong  screen_base;	/* Bild zum Beschreiben */
static  tm_ulong  screen_show;	/* angezeigtes Bild */

static  tm_ubyte  screen_buffer[WIDTH*HEIGHT/8*2+0xFF];

static  tm_ubyte  *offscreen_buffer;

void    *physbase=NULL,*logbase=NULL;	/* gerettete Zeiger */

static  tm_uword  col0,col1;		/* gerettete Farben */

static tm_void alert(CONST tm_char *s)
{
  tm_char text[200];

  sprintf(text,"[3][%s|][OK]",s);
  graf_mouse(ARROW,NULL);
  form_alert(1,text);
}

tm_void flip_screen(tm_void)
{
  tm_long t;

  t=screen_base;
  screen_base=screen_show;
  screen_show=t;
  Setscreen((void*)screen_base,(void*)screen_show,-1);
  Vsync();
}

tm_bool open_screen(tm_uword argc, tm_char **argv,
                    tm_uword *width, tm_uword *height, tm_uword *colors )
{
  int      i,work_in[11],work_out[57];
  tm_ubyte *p;

  (tm_void)argc;
  (tm_void)argv;

  /* momentane Zeiger retten */
  physbase=Physbase();
  logbase =Logbase();

  open_rate=0;
  atexit(close_screen);

#ifdef NOFLIP
  (tm_void)screen_buffer;
  screen_base=screen_show=(tm_long)physbase;
#else
  screen_base=((tm_long)screen_buffer+0xFF)&(~0xFFL);
  screen_show=screen_base+WIDTH*HEIGHT/8;
#endif

  /* beim AES anmelden */
  if (appl_init()<0) {
    alert(no_appl_init[language]);
    return(FALSE);
    }
  open_rate=1;

  vdi_handle=graf_handle(&i,&i,&i,&i);
  for (i=0; i<10; work_in[i++]=1);
  work_in[10]=2;
  v_opnvwk(work_in,&vdi_handle,work_out);
  open_rate=2;
  if (work_out[0]+1!=WIDTH || work_out[1]+1!=HEIGHT) {
    alert(wrong_display[language]);
    return(FALSE);
    }
  wind_update(BEG_UPDATE);
  open_rate=3;
  graf_mouse(M_OFF,NULL);
  open_rate=4;

  Setscreen((void*)-1L,(void*)screen_show,-1);
  if ((tm_long)Physbase()!=screen_show) {
    Setscreen((void*)-1L,physbase,-1);
    graf_mouse(M_ON,NULL);
    wind_update(END_UPDATE);
    alert(wrong_display[language]);
    return(FALSE);
    }

  Setscreen((void*)screen_base,(void*)-1L,-1);
  if ((tm_long)Logbase()!=screen_base) {
    Setscreen(logbase,physbase,-1);
    graf_mouse(M_ON,NULL);
    wind_update(END_UPDATE);
    alert(wrong_display[language]);
    return(FALSE);
    }

  /* Titelbild vorbereiten */
  offscreen_buffer=title_pic_data;
  if (title_pic_col[0][0]!=0)
    for (p=title_pic_data, i=0; i<WIDTH*HEIGHT/8; p++, i++)
      *p=~*p;

  *width  = WIDTH;
  *height = HEIGHT;
  *colors = COLORS;

  return(OK);
}

/* Achtung, wird bei normalem Ende zweimal aufgerufen! */
tm_void close_screen(tm_void)
{
  Setscreen(logbase,physbase,-1);

  switch (open_rate)
  {
  case 4:
    graf_mouse(M_ON,NULL);
  case 3:
    wind_update(END_UPDATE);
  case 2:
    v_clsvwk(vdi_handle);
  case 1:
    appl_exit();
  default:
    open_rate=0;
  }
}

tm_void fade_title_pic(tm_char *copyright)
{
  Setcolor(0,0);
  restore_screen();
  draw_text((WIDTH-get_text_width(copyright))/2,HEIGHT/6,
            copyright,BW_WHITE);
  flip_screen();
}

/* gibt's nicht bei monochromen schirmen... */
tm_void set_color(tm_uword col, tm_uword prim[3])
{
  (tm_void)col;
  (tm_void)prim;
}

tm_void save_colors(tm_void)
{
  col0=Setcolor(0,-1);
  col1=Setcolor(1,-1);
  atexit(restore_colors);
}

/* Achtung, wird bei normalem Ende zweimal aufgerufen! */
tm_void restore_colors(tm_void)
{
  Setcolor(0,col0);
  Setcolor(1,col1);
}

#include "raw2\raw2.c"

#include "raw2\raw2wall.c"

/* --- end --- */
