/*
   scrvdi16.c

   Bildschirm-Treiber fuer Atari ST niedrige Aufloesung ueber VDI

   01.12.96  Chip  kopiert von SCRRAW2.C und RAW2.C
   26.01.97  Chip  Copyrighttext als "richtige" Textausgabe
   16.03.97  Chip  eigene Wandfunktionen
   22.06.97  Chip  console... und error... korrigiert
*/

#define noNO_FLIP

#include <aes.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <tos.h>
#include <vdi.h>

#include "ctypes.h"
#include "error.h"
#include "gr_prim.h"
#include "grafik.h"
#include "text.h"
#include "tos\picvdi16.h"
#include "tos\con.h"

tm_char video_driver_name[]="Intern low(VDI) V1.0";

#define USE_CLIB    /* Grafikprimitiva duerfen string.h benutzen */

#define WIDTH     (320L)
#define HEIGHT    (200L)
#define COLORS    (16)
#define PLANES    (4)

#if TITLE_PIC_WIDTH!=WIDTH || TITLE_PIC_HEIGHT!=HEIGHT    \
    || TITLE_PIC_COLORS!=COLORS
#error "Falsches Titelbild!"
#endif

tm_void show_video_options(tm_void) 
{
}

tm_void init_video_options(tm_void)
{
  console_init();
  atexit(console_exit);
}

tm_bool set_video_option(tm_char *cmd)
{
  (tm_void)cmd;
  return(FALSE);
}

tm_uword open_rate;   /* wie weit bei open_screen() gekommen? */

int	vdi_handle;	/* fuer GEM, ganz primitiv */

static tm_ulong screen_base;	/* Bild zum Beschreiben */
static tm_ulong screen_show;	/* angezeigtes Bild */

static tm_ubyte screen_buffer[PLANES*WIDTH*HEIGHT/8*2+0xFF];

static tm_ubyte *offscreen_buffer;

void *physbase=NULL,*logbase=NULL;	/* gerettete Zeiger */

static tm_ushort old_col[COLORS];	/* gerettete Farben */

static tm_void alert(CONST tm_char *s)
{
  tm_char text[200];

  sprintf(text,"[3][%s|][OK]",s);
  graf_mouse(ARROW,NULL);
  form_alert(1,text);
}

tm_void flip_screen(tm_void)
{
  tm_long t;

  t=screen_base;
  screen_base=screen_show;
  screen_show=t;
  Setscreen((void*)screen_base,(void*)screen_show,-1);
  Vsync();
}

tm_bool open_screen(tm_uword argc, tm_char **argv,
                    tm_uword *width, tm_uword *height, tm_uword *colors )
{
  int       i,j,k,work_in[11],work_out[57];
  tm_ushort *p,l[PLANES*WIDTH/16];

  (tm_void)argc;
  (tm_void)argv;

  /* momentane Zeiger retten */
  physbase=Physbase();
  logbase =Logbase();

#ifdef NOFLIP
  (tm_void)screen_buffer;
  screen_base=screen_show=(tm_long)physbase;
#else
  screen_base=((tm_long)screen_buffer+0xFF)&(~0xFFL);
  screen_show=screen_base+PLANES*WIDTH*HEIGHT/8;
#endif

  /* beim AES anmelden */
  if (appl_init()<0) {
    alert(no_appl_init[language]);
    return(FALSE);
    }
  open_rate=1;

  vdi_handle=graf_handle(&i,&i,&i,&i);
  for (i=0; i<10; work_in[i++]=1);
  work_in[10]=2;
  v_opnvwk(work_in,&vdi_handle,work_out);
  open_rate=2;
  if (work_out[0]+1!=WIDTH || work_out[1]+1!=HEIGHT ||
      work_out[13]!=COLORS) {
    alert(wrong_display[language]);
    return(FALSE);
    }
  wind_update(BEG_UPDATE);
  open_rate=3;
  graf_mouse(M_OFF,NULL);
  open_rate=4;

  Setscreen((void*)-1L,(void*)screen_show,-1);
  if ((tm_long)Physbase()!=screen_show) {
    Setscreen((void*)-1L,physbase,-1);
    graf_mouse(M_ON,NULL);
    wind_update(END_UPDATE);
    alert(wrong_display[language]);
    return(FALSE);
    }

  Setscreen((void*)screen_base,(void*)-1L,-1);
  if ((tm_long)Logbase()!=screen_base) {
    Setscreen(logbase,physbase,-1);
    graf_mouse(M_ON,NULL);
    wind_update(END_UPDATE);
    alert(wrong_display[language]);
    return(FALSE);
    }

  /* ein paar GEM-Einstellungen */
  vswr_mode(vdi_handle,MD_REPLACE);
  vsf_perimeter(vdi_handle,0);

  /* Titelbild vorbereiten */
  offscreen_buffer=title_pic_data;
  p=(tm_ushort *)title_pic_data;
  for (i=0; i<HEIGHT; i++) {
    for (j=0; j<PLANES*WIDTH/16; j++)
      l[j]=*p++;
    p-=PLANES*WIDTH/16;
    for (j=0; j<WIDTH/16; j++)
      for (k=0; k<PLANES; k++)
        *p++=l[j+k*WIDTH/16];
    }

  *width  = WIDTH;
  *height = HEIGHT;
  *colors = COLORS;

  return(OK);
}

/* Achtung, wird bei normalem Ende zweimal aufgerufen! */
tm_void close_screen(tm_void)
{
  Setscreen(logbase,physbase,-1);

  switch (open_rate)
  {
  case 4:
    graf_mouse(M_ON,NULL);
  case 3:
    wind_update(END_UPDATE);
  case 2:
    v_clsvwk(vdi_handle);
  case 1:
    appl_exit();
  default:
    open_rate=0;
  }
}

tm_void fade_title_pic(tm_char *copyright)
{
  int  i,j,c;

  for (i=0; i<COLORS; i++)
    Setcolor(i,0);
  restore_screen();
  draw_text((WIDTH-get_text_width(copyright))/2,HEIGHT/6,copyright,COL_RED);
  flip_screen();
  for (j=0; j<=100; j++) {
    for (i=0; i<COLORS; i++) {
      c =(int)((7L*title_pic_col[i][0]*j/100000L)<<8);
      c|=(int)((7L*title_pic_col[i][1]*j/100000L)<<4);
      c|=(int)( 7L*title_pic_col[i][2]*j/100000L);
      Setcolor(i,c);
      }
    Vsync();
    }
}

tm_void set_color(tm_uword col, tm_uword prim[3])
{
  /* besser ueber VDI, damit die Indizes stimmen... */
  vs_color(vdi_handle,col,(int *)prim);
}

tm_void save_colors(tm_void)
{
  int  i;

  for (i=0; i<COLORS; i++)
    old_col[i]=Setcolor(i,-1);
}

/* Achtung, wird bei normalem Ende zweimal aufgerufen! */
tm_void restore_colors(tm_void)
{
  int  i;

  for (i=0; i<COLORS; i++)
    Setcolor(i,old_col[i]);
}

/* ab hier jetzt der eigentliche Treiber... */

int line_color=-1,area_color=-1;

#if 0

/* --- pixel (*_PIX) --- */

tm_void set_pixel(tm_uword x, tm_uword y, tm_uword col)
{
  int  pxy[4];

  if (col!=line_color)
  {
    line_color=col;
    vsl_color(vdi_handle,col);
  }
  pxy[0]=x;
  pxy[1]=y;
  pxy[2]=x;
  pxy[3]=y;
  v_pline(vdi_handle,2,pxy);
}

tm_uword get_pixel(tm_uword x, tm_uword y)
{
  int dummy,col;

  v_get_pixel(vdi_handle,x,y,&dummy,&col);
  return(col);
}

#endif /* #if 0 */

/* --- lines (*_LIN) --- */

tm_void draw_line(tm_uword x1, tm_uword y1, tm_uword x2, tm_uword y2, tm_uword col)
{
  int  pxy[4];

  if (col!=line_color)
  {
    line_color=col;
    vsl_color(vdi_handle,col);
  }
  pxy[0]=x1;
  pxy[1]=y1;
  pxy[2]=x2;
  pxy[3]=y2;
  v_pline(vdi_handle,2,pxy);
}

tm_void draw_hline(tm_uword x, tm_uword y, tm_uword len, tm_uword col)
{
  int  pxy[4];

  if (col!=line_color)
  {
    line_color=col;
    vsl_color(vdi_handle,col);
  }
  pxy[0]=x;
  pxy[1]=y;
  pxy[2]=x+len-1;
  pxy[3]=y;
  v_pline(vdi_handle,2,pxy);
}

tm_void draw_vline(tm_uword x, tm_uword y, tm_uword len, tm_uword col)
{
  int  pxy[4];

  if (col!=line_color)
  {
    line_color=col;
    vsl_color(vdi_handle,col);
  }
  pxy[0]=x;
  pxy[1]=y;
  pxy[2]=x;
  pxy[3]=y+len-1;
  v_pline(vdi_handle,2,pxy);
}

tm_void draw_vtexline(tm_uword x, tm_uword y, tm_uword len, tm_uword real_len, 
                      tm_uchar *texture, tm_uword tex_height)
{
  (tm_void)x;
  (tm_void)y;
  (tm_void)len;
  (tm_void)real_len;
  (tm_void)texture;
  (tm_void)tex_height;
  do_error(1,"draw_vtexline does not exist");
}

void draw_quadvline(tm_uword x, tm_uword y, tm_uword len, tm_ulong cols)
{
  (tm_void)x;
  (tm_void)y;
  (tm_void)len;
  (tm_void)cols;
  do_error(2,"draw_quadvline does not exist");
}

#if 0

void draw_vdline(tm_uword x, tm_uword y, tm_uword len, tm_uword e, tm_uword col1, tm_uword col2)
{
  (tm_void)x;
  (tm_void)y;
  (tm_void)len;
  (tm_void)e;
  (tm_void)col1;
  (tm_void)col2;
  do_error(3,"draw_vdline does not exist");
}

#endif /* #if 0 */

/* --- arcs (*_CIR) --- */

tm_void draw_disc(tm_uword x, tm_uword y, tm_uword rad, tm_uword int_col, tm_uword bd_col)
{
  if (bd_col!=line_color)
  {
    line_color=bd_col;
    vsl_color(vdi_handle,bd_col);
  }
  if (int_col!=area_color)
  {
    area_color=int_col;
    vsf_color(vdi_handle,int_col);
  }
  v_circle(vdi_handle,x,y,rad);
  v_arc(vdi_handle,x,y,rad,0,3600);
}

/* --- bit blocks (*_BLT) --- */

tm_void save_screen(tm_void)
{
  tm_ulong *dest=(tm_ulong*)offscreen_buffer,*src=(tm_ulong*)screen_show;

#ifdef USE_CLIB
  memcpy(dest,src,(size_t)(PLANES*WIDTH*HEIGHT/8));
#else
  tm_ulong cnt;
  for(cnt=0;cnt<(PLANES*WIDTH*HEIGHT/32);cnt++)
    *dest++=*src++;
#endif
}

tm_void restore_screen(tm_void)
{
  tm_ulong *src=(tm_ulong*)offscreen_buffer,*dest=(tm_ulong*)screen_base;

#ifdef USE_CLIB
  memcpy(dest,src,(size_t)(PLANES*WIDTH*HEIGHT/8));
#else
  tm_ulong cnt;
  for(cnt=0;cnt<(PLANES*WIDTH*HEIGHT/32);cnt++)
    *dest++=*src++;
#endif
}

tm_void put_bitmap(tm_uword x, tm_uword y, tm_uword width, tm_uword height,
                   CONST tm_ushort *data, tm_uword col)
{
  MFDB  mfdb_s,mfdb_d;
  int   pxy[8],c[2];

  mfdb_s.fd_addr=data;
  mfdb_s.fd_w=width;
  mfdb_s.fd_h=height;
  mfdb_s.fd_wdwidth=(width+15)/16;
  mfdb_s.fd_stand=1;
  mfdb_s.fd_nplanes=1;

  mfdb_d.fd_addr=NULL;

  pxy[0]=0;
  pxy[1]=0;
  pxy[2]=width-1;
  pxy[3]=height-1;
  pxy[4]=x;
  pxy[5]=y;
  pxy[6]=x+width-1;
  pxy[7]=y+height-1;

  c[0]=col;
  c[1]=0;

  vrt_cpyfm(vdi_handle,MD_TRANS,pxy,&mfdb_s,&mfdb_d,c);
}

tm_void fill_screen(tm_uword col)
{
  int  pxy[4];

  if (col!=area_color)
  {
    area_color=col;
    vsf_color(vdi_handle,col);
  }
  pxy[0]=0;
  pxy[1]=0;
  pxy[2]=(int)(WIDTH-1);
  pxy[3]=(int)(HEIGHT-1);
  v_bar(vdi_handle,pxy);
}

tm_void draw_rect(tm_uword x, tm_uword y, tm_uword w, tm_uword h, tm_uword col)
{
  int  pxy[4];

  if (col!=area_color)
  {
    area_color=col;
    vsf_color(vdi_handle,col);
  }
  pxy[0]=x;
  pxy[1]=y;
  pxy[2]=x+w-1;
  pxy[3]=y+h-1;
  v_bar(vdi_handle,pxy);
}

/* --- die wand-funktionen --- */

#define SCR_YOFFS (HEIGHT/2)

#ifdef WALLS_LINES

tm_void wall_2(tm_uword x,tm_uword h,
               tm_uword wall_col,tm_uword border_col)
{
  int  pxy[4];

  (tm_void)wall_col;

  if (border_col!=line_color)
  {
    line_color=border_col;
    vsl_color(vdi_handle,border_col);
  }
  pxy[0]=x;  pxy[1]=(int)(SCR_YOFFS-h);
  pxy[2]=x;  pxy[3]=(int)(SCR_YOFFS+h);
  v_pline(vdi_handle,2,pxy);
}

tm_void wall_4(tm_uword x1,tm_uword h1,
               tm_uword x2,tm_uword h2,
               tm_uword wall_col,tm_uword border_col)
{
  int  pxy[10];

  pxy[0]=x1;     pxy[1]=(int)(SCR_YOFFS-h1);
  pxy[2]=x1;     pxy[3]=(int)(SCR_YOFFS+h1);
  pxy[4]=x2;     pxy[5]=(int)(SCR_YOFFS+h2);
  pxy[6]=x2;     pxy[7]=(int)(SCR_YOFFS-h2);
  pxy[8]=pxy[0]; pxy[9]=pxy[1];

  if (wall_col!=line_color)
  {
    line_color=wall_col;
    vsl_color(vdi_handle,wall_col);
  }
  v_pline(vdi_handle,2,pxy+2);
  v_pline(vdi_handle,2,pxy+6);

  if (border_col!=line_color)
  {
    line_color=border_col;
    vsl_color(vdi_handle,border_col);
  }
  v_pline(vdi_handle,2,pxy);
  v_pline(vdi_handle,2,pxy+4);
}

tm_void wall_6(tm_uword x1,tm_uword h1,
               tm_uword x2,tm_uword h2,
               tm_uword x3,tm_uword h3,
               tm_uword wall_col,tm_uword border_col)
{
  int  pxy[14];

  pxy[0] =x1;     pxy[1] =(int)(SCR_YOFFS-h1);
  pxy[2] =x1;     pxy[3] =(int)(SCR_YOFFS+h1);
  pxy[4] =x2;     pxy[5] =(int)(SCR_YOFFS+h2);
  pxy[6] =x3;     pxy[7] =(int)(SCR_YOFFS+h3);
  pxy[8] =x3;     pxy[9] =(int)(SCR_YOFFS-h3);
  pxy[10]=x2;     pxy[11]=(int)(SCR_YOFFS-h2);
  pxy[12]=pxy[0]; pxy[13]=pxy[1];

  if (wall_col!=line_color)
  {
    line_color=wall_col;
    vsl_color(vdi_handle,wall_col);
  }
  v_pline(vdi_handle,2,pxy+2);
  v_pline(vdi_handle,2,pxy+4);
  v_pline(vdi_handle,2,pxy+8);
  v_pline(vdi_handle,2,pxy+10);

  if (border_col!=line_color)
  {
    line_color=border_col;
    vsl_color(vdi_handle,border_col);
  }
  v_pline(vdi_handle,2,pxy);
  v_pline(vdi_handle,2,pxy+6);
}

#endif

#ifdef WALLS_FILLED

tm_void wall_2(tm_uword x,tm_uword h,
               tm_uword wall_col,tm_uword border_col)
{
  int  pxy[4];

  (tm_void)wall_col;

  if (border_col!=line_color)
  {
    line_color=border_col;
    vsl_color(vdi_handle,border_col);
  }
  pxy[0]=x;  pxy[1]=(int)(SCR_YOFFS-h);
  pxy[2]=x;  pxy[3]=(int)(SCR_YOFFS+h);
  v_pline(vdi_handle,2,pxy);
}

tm_void wall_4(tm_uword x1,tm_uword h1,
               tm_uword x2,tm_uword h2,
               tm_uword wall_col,tm_uword border_col)
{
  int  pxy[8];

  if (wall_col!=area_color)
  {
    area_color=wall_col;
    vsf_color(vdi_handle,wall_col);
  }
  if (border_col!=line_color)
  {
    line_color=border_col;
    vsl_color(vdi_handle,border_col);
  }
  pxy[0] =x1;  pxy[1] =(int)(SCR_YOFFS-h1);
  pxy[2] =x1;  pxy[3] =(int)(SCR_YOFFS+h1);
  pxy[4] =x2;  pxy[5] =(int)(SCR_YOFFS+h2);
  pxy[6] =x2;  pxy[7] =(int)(SCR_YOFFS-h2);
  v_fillarea(vdi_handle,4,pxy);
  v_pline(vdi_handle,2,pxy);
  v_pline(vdi_handle,2,pxy+4);
}

tm_void wall_6(tm_uword x1,tm_uword h1,
               tm_uword x2,tm_uword h2,
               tm_uword x3,tm_uword h3,
               tm_uword wall_col,tm_uword border_col)
{
  int  pxy[12];

  if (wall_col!=area_color)
  {
    area_color=wall_col;
    vsf_color(vdi_handle,wall_col);
  }
  if (border_col!=line_color)
  {
    line_color=border_col;
    vsl_color(vdi_handle,border_col);
  }
  pxy[0] =x1;  pxy[1] =(int)(SCR_YOFFS-h1);
  pxy[2] =x1;  pxy[3] =(int)(SCR_YOFFS+h1);
  pxy[4] =x2;  pxy[5] =(int)(SCR_YOFFS+h2);
  pxy[6] =x3;  pxy[7] =(int)(SCR_YOFFS+h3);
  pxy[8] =x3;  pxy[9] =(int)(SCR_YOFFS-h3);
  pxy[10]=x2;  pxy[11]=(int)(SCR_YOFFS-h2);
  v_fillarea(vdi_handle,6,pxy);
  v_pline(vdi_handle,2,pxy);
  v_pline(vdi_handle,2,pxy+6);
}

#endif

/* --- end --- */
