/*
   libaudio.c

   Unix audio driver via libaudio

*/

#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/signal.h>

#include <stropts.h>
#include <sys/ioctl.h>

#include <multimedia/libaudio.h>
#include <multimedia/audio_device.h>
#include <multimedia/audio_hdr.h>
#ifndef _MULTIMEDIA_AUDIO_ENCODE_H
#include <multimedia/ulaw2linear.h>
#endif

#include "../dialog.h"
#include "../ctypes.h"
#include "../audio.h"
#include "../console.h"
#include "../joystick.h"
#include "../music/music.h"
#include "../music/the_song.h"
#include "../net.h"

extern tm_char *tm_sounds[];
extern tm_word tm_sound_len[];

tm_char audio_driver_name[]="LibAudio V0.7";

#define AUDIO_DEV     "/dev/audio"

char *sound_files[]= { "boing","click","beam-the","bonus","shoot" };

int             Audio_fd = -1;          /* file descriptor for audio device */
Audio_hdr       Dev_hdr;                /* audio header for device */
tm_uword        org_volume;
tm_bool         aon = ON;
tm_uchar        mus_linear2ulaw[65536];

tm_void 
show_audio_options(tm_void)
{
  console_printf("%s audio driver options:\n",audio_driver_name);
  console_puts("as=n\tswitch audio output on (n=1) or off (n=0), default is on");
  console_puts("av=n\tset audio volume to n (0..100), default is system volume");
}

tm_bool  g_on_f,g_on,g_vol_f;
tm_uword g_vol;

tm_void init_audio_options(tm_void)
{
  g_on_f  = FALSE;
  g_vol_f = FALSE;
}

tm_bool 
set_audio_option(tm_char *cmd)
{
  tm_uword val=atoi(&cmd[2]);

  if(cmd[1]=='=') {
    if((cmd[0]=='s')&&((val==0)||(val==1))) {
      g_on_f=TRUE;
      g_on=(tm_bool)val;
      return(OK);
    }
    if((cmd[0]=='v')&&(val>=0)&&(val<=100)) {
      g_vol_f=TRUE;
      g_vol=val;
      return(OK);
    }
  }
  return(FALSE);
}

tm_bool 
init_audio(void)
{
  int err,i;
  struct stat     st;

  err = stat(AUDIO_DEV, &st);
  if (err < 0) {
    console_printf ("cannot stat "AUDIO_DEV"\n");
    return(ERROR);
  }
  if (!S_ISCHR(st.st_mode)) {
    console_printf (AUDIO_DEV " is not an audio device\n");
    return(ERROR);
  }
  /* Try it quickly, first */
  Audio_fd = open(AUDIO_DEV, O_WRONLY | O_NDELAY);
  if ((Audio_fd < 0) && (errno == EBUSY)) {
    console_printf (AUDIO_DEV " is busy\n");
    return(ERROR);
  }

  if (Audio_fd < 0) {
    console_printf ("error opening "AUDIO_DEV"\n");
    return(ERROR);
  }
  
  /* Get the device output encoding configuration */
  if (audio_get_play_config(Audio_fd, &Dev_hdr) != AUDIO_SUCCESS) {
    console_printf (AUDIO_DEV" is not an audio device\n");
    return(ERROR);
  }

  vol_audio(&org_volume, PARM_GET);

  /* command line options */
  if(g_on_f)  audio_on(&g_on,PARM_SET);
  if(g_vol_f) vol_audio(&g_vol,PARM_SET);

  /* create ulaw conversion table */
  for(i=0;i<65536;i++)
    mus_linear2ulaw[i]=audio_s2u((short)i);

  return(OK);
}

/* dumm-di-dumm trallala */
#define MUSIC_ENDLESS
#define MUSIC_PREFILL         (4)     /* have 3 beats in queue */
#define MUSIC_OUTPUT_RATE     (8000)  /* this is 8kHz u-law */
#define MUSIC_BEATS_PER_MIN   (80)    /* try techno polka! */
#define MUSIC_OUTPUT_SIZE     (60*MUSIC_OUTPUT_RATE/MUSIC_BEATS_PER_MIN)
#define MUSIC_BUFFER_SIZE     (MUS_MAX_PRECALC+MUSIC_OUTPUT_SIZE)
tm_long  music_buffer[MUSIC_BUFFER_SIZE];
tm_uchar music_output_buffer[MUSIC_OUTPUT_SIZE];

tm_void
do_music(tm_void) {
  tm_uchar *pos=(tm_uchar*)mus_song,sample;
  tm_word  i,time,vol,len,joy,beatlen;
  unsigned int eof_cnt,old_eof=-MUSIC_PREFILL;
  tm_ulong val,factor;
  tm_long  *dest;
  tm_short *src;
  tm_void flush_audio(tm_void);

  if((aon)&&(Audio_fd>=0)) {
    /* clear buffer */
    memset(music_buffer,0,sizeof(music_buffer));

    /* get length of beat in n/32 */
    beatlen=*pos++;

    do {
      time=0;  /* start a new beat */
      while((time<beatlen)&&(*pos!=MUS_END)) {   /* one beat completed ? */
	
	sample=*pos++;                      /* get sample no */

	if(sample==MUS_WAIT) {
	  time+=*pos++;     /* time in n/32 beats */
	} else {
	  /* get frequency factor  */
	  factor=*pos++*256*256*256; factor+=*pos++*256*256; factor+=*pos++*256; factor+=*pos++;
	  vol=*pos++; len=*pos++; /* get volume and length */
	
	  /* compose samples in buffer */
	  dest=&music_buffer[time*(MUSIC_OUTPUT_SIZE/beatlen)]; src=(tm_short*)mus_sample[sample];
	  for(i=0;i<((tm_word)mus_sample_len[sample]<<16)/factor;i++)
	    *dest+++=(vol*src[(i*factor)>>16])>>8;
	}
#ifdef MUSIC_ENDLESS
	if(*pos==MUS_END)
	  pos=(tm_uchar*)mus_song+1;
#endif
      }

      /* play this beat */
      /* output buffer (4*long -> signed short -> ulaw) */
      for(i=0;i<MUSIC_OUTPUT_SIZE;i++)
	music_output_buffer[i]=mus_linear2ulaw[(music_buffer[i]/MUS_INSTRUMENTS/8)&0xffff];
      /* shift music buffer down */
      /* bcopy(&music_buffer[MUSIC_OUTPUT_SIZE],music_buffer,(MUSIC_BUFFER_SIZE-MUSIC_OUTPUT_SIZE)*sizeof(tm_long)); */
      memcpy(music_buffer,&music_buffer[MUSIC_OUTPUT_SIZE],(MUSIC_BUFFER_SIZE-MUSIC_OUTPUT_SIZE)*sizeof(tm_long));
      /* clear buffer end */
      memset(&music_buffer[MUSIC_BUFFER_SIZE-MUSIC_OUTPUT_SIZE],0,MUSIC_OUTPUT_SIZE*sizeof(tm_long));
      
      /* and down to the device */
      write(Audio_fd,music_output_buffer,MUSIC_OUTPUT_SIZE);
      audio_play_eof(Audio_fd);
      
      /* wait for end of beat */
      do {
	audio_get_play_eof(Audio_fd,&eof_cnt);
	joy=get_joystick();
      } while((eof_cnt==old_eof)&&!(joy&JOY_FIRE));
      old_eof=eof_cnt;
      
      net_handle_messages();
      /* music running and nobody pressed fire and game not running? */
    } while((*pos!=MUS_END)&&!(joy&JOY_FIRE)&&!(game_running));
    
    /* wait for button release */
    if(joy&JOY_FIRE) {
      flush_audio();
      wait_msec(10);
      do
	joy=get_joystick();
      while(joy&JOY_FIRE);
    } else {
      if(!game_running)
	/* wait two beats for music to stop */
	wait_msec(2000*MUSIC_BEATS_PER_MIN/60);
      else
	flush_audio();
    }
  } else {
    do {
      joy=get_joystick();
      net_handle_messages();
    } while(!(joy&JOY_FIRE)&&(!game_running));
    while(joy&JOY_FIRE) joy=get_joystick();
  }
}

tm_bool 
audio_on(tm_bool *on, tm_bool get) {
  if(get) *on = aon;
  else    aon = *on;
  return(OK);
}

/* volume ranges from 0..100 */
tm_bool 
vol_audio(tm_uword *volume, tm_bool get) 
{
  double vol;
  int err;

  if(Audio_fd >= 0) {
    if(get) {
      err = audio_get_play_gain(Audio_fd, &vol);
      if (err != AUDIO_SUCCESS) {
	console_printf ("could not get output volume from "AUDIO_DEV"\n");
	return(ERROR);
      }
      *volume = vol * 100.0;
    } else {
      vol = *volume / 100.0;
      err = audio_set_play_gain(Audio_fd, &vol);
      if (err != AUDIO_SUCCESS) {
	console_printf ("could not set output volume for "AUDIO_DEV"\n");
	return(ERROR);
      }
    }
  }
  return(OK);
}

tm_bool 
exit_audio(tm_void)
{
  vol_audio(&org_volume, PARM_SET);

  if(Audio_fd >= 0)
    (void) close(Audio_fd);                 /* close output */

  return(OK);
}

tm_void 
flush_audio(tm_void)
{
  int err;
  
  err = audio_flush_play(Audio_fd);
  if (err != AUDIO_SUCCESS) {
    console_printf ("could not flush audio for "AUDIO_DEV"\n");
  }
}

#define AUDIO_BUFFER_SIZE 32768
char buffer[AUDIO_BUFFER_SIZE];

tm_void 
do_audio(tm_sound sound)
{
  if((Audio_fd >= 0)&&(aon)) {
    flush_audio();
    write(Audio_fd, tm_sounds[sound], tm_sound_len[sound]);
  }
}

tm_void 
test_audio(tm_void)
{
  do_audio(SOUND_TEST);
  audio_drain(Audio_fd,FALSE);
}
