/*
   net_udp.c 

   Netztreiber fuer TheMaze

   unterstuetzt Internet-UDP und Sparc-SIO

*/

#define noNET_DEBUG

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <pwd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "../ctypes.h"
#include "../net.h"

#define PORT      5555

#define SIO_DEV "/dev/ttya"

tm_char net_driver_name[]="IP-UDP V0.3";

#define MAX_XTIMEOUT (10)

#ifndef hpux
#define TM_FD_ZERO(s)    FD_ZERO(s)
#define TM_FD_SET(e,s)   FD_SET(e,s)
#define TM_FD_ISSET(e,s) FD_ISSET(e,s)
#define TM_LIMIT         (ulimit(4,0))
typedef fd_set           tm_fd_set;
#else
/* looks like hpux was made by microsoft ... */
#define TM_FD_ZERO(s)	 {*s =  0;}
#define TM_FD_SET(e,s)   {*s |= (1<<e);}
#define TM_FD_ISSET(e,s) (*s &  (1<<e))
#define TM_LIMIT         (31)
typedef int              tm_fd_set;
#endif

int    sock;
struct sockaddr_in addr;

/* Prototypen der lokalen Funktionen */
tm_char* ip2a(long ip);
tm_bool  open_udp(tm_void);
tm_bool  close_udp(tm_void);
tm_word  recv_udp(char *data, tm_word len);
tm_bool  send_udp(tm_char *data, tm_word len);
tm_bool  stat_udp(tm_word timeout);

/* Tabellen, falls wir Server sind */
tm_bool addr_assigned[MAX_PLAYERS];
int sockets[2],child;

tm_long next_ip;      /* ip of the next machine in the chain */
tm_long my_ip;        /* my ip address */
tm_bool i_am_master=TRUE;
tm_char my_no=0;      /* our number in the net, starting with master = 0 */

/* messages between TheMaze and IOserver */
#define IO_QUIT    (1)
#define IO_ACK     (2)
#define IO_OPEN    (3)
#define IO_CONNECT (4)
#define IO_SEND    (5)
#define IO_STAT    (6)
#define IO_RECV    (7)
#define IO_DISC    (8)
#define IO_MYIP    (9)

/* write data between TheMaze <-> IO server */ 
tm_bool
io_write_byte(tm_char data) {
  if(write(sockets[child?1:0], &data, 1)<0) {
    perror("io server: writing byte");
    return(ERROR);
  }
  return(OK);
}

/* write data between TheMaze <-> IO server */ 
tm_bool
io_write(tm_char *data, tm_word len) {
  if(write(sockets[child?1:0], data, len)<0) {
    perror("io server: writing data");
    return(ERROR);
  }
  return(OK);
}

tm_char  msg_buffer[MAX_MESSAGE][MAX_PKT_LEN];  /* 1000 * 48 */
tm_short msg_len[MAX_MESSAGE];
tm_word read_index, write_index;

/* startup IO server */
tm_bool
init_io_server(tm_void) {
  char buf[256];
  struct timeval timeout;
  tm_fd_set  fdset;
  int i,used;
  long ip;
  struct sockaddr_in inaddr;
  static xtimeout=0;

  if(socketpair(AF_UNIX,SOCK_DGRAM,0,sockets)<0) {
    perror("opening stream socket pair");
    return(ERROR);
  }

  /* fork io server */
  if((child=fork())==-1)
    perror("fork io server");
  else { 
    if(child) {
      /* This is the parent */
      close(sockets[0]);
    } else {
      /* This is the child */
      close(sockets[1]);
      read_index=write_index=0;  /* no message at now */
#ifdef NET_DEBUG
      printf("io server ready\n");
#endif
      while(!child) {
	timeout.tv_usec=0; timeout.tv_sec=1;

	TM_FD_ZERO(&fdset);
	TM_FD_SET(sockets[0],&fdset);
	TM_FD_SET(sock,&fdset);
	if(select(TM_LIMIT,&fdset,NULL,NULL,&timeout)>0) {
	  xtimeout=0;
	  if(TM_FD_ISSET(sockets[0],&fdset)) {
	    if((used=read(sockets[0],buf,sizeof(buf)))<0)
	      perror("io server: reading message");

	    /* */
	    switch(buf[0]) {
	    case IO_MYIP:
#ifdef NET_DEBUG
	      printf("io server: my_ip request\n");
#endif	      /* extract ip address */
	      memcpy(&next_ip,&buf[1],sizeof(next_ip));	      
	      addr.sin_addr.s_addr=next_ip;
	      break;

	      /* mother wants us to die */
	    case IO_QUIT: 
#ifdef NET_DEBUG
	      printf("io server: quit request\n");
#endif
	      io_write_byte(IO_QUIT);
	      child=1;
	      break;
	      
	    case IO_OPEN:
#ifdef NET_DEBUG
	      printf("io server: open request\n");
#endif
	      /* extract ip address */
	      memcpy(&next_ip,&buf[1],sizeof(next_ip));

	      /* open udp socket */
	      open_udp();
	      break;
	      
	      /* connect to named machine */
	    case IO_CONNECT:
	      /* the master never does a connect request */
	      i_am_master=FALSE;
	      my_no=-1;   /* don't assume we are master */

	      /* destination address */
	      memcpy(&addr.sin_addr.s_addr,&buf[1],sizeof(long));
	      /* my address */
	      memcpy(&ip,&buf[1+sizeof(long)],sizeof(long));

#ifdef NET_DEBUG
	      printf("io server: connect request (from=%s/",ip2a(ip));
#endif
	      memcpy(&ip,&buf[1],sizeof(long));
#ifdef NET_DEBUG
	      printf("to=%s)\n",ip2a(addr.sin_addr.s_addr));
#endif
	      buf[0]=NET_CONNECT;
	      memcpy(&ip,&buf[1+sizeof(long)],sizeof(long)); ip=htonl(ip);
	      memcpy(&buf[1],&ip,sizeof(long));
	      send_udp(buf,sizeof(long)+1);
	      break;
	      
	      /* stat */
	    case IO_STAT:
#ifdef XNET_DEBUG
	      printf("io server: status request\n");
#endif
	      if(write_index!=read_index) io_write_byte(TRUE);
	      else                        io_write_byte(FALSE);
	      break;

	    case IO_SEND:
#ifdef NET_DEBUG
	      printf("io server: send request (1+%d bytes)\n",used-1);
#endif
	      /* send (overwrite IO_SEND with source id) */
	      buf[0]=buf[1]; buf[1]=my_no;
	      send_udp(buf,used);
	      break;

	      /* receive */
	    case IO_RECV:
#ifdef NET_DEBUG
	      printf("io server: receive request\n");
#endif
	      if(write_index==read_index)
		fprintf(stderr,"receive: no data available\n");
	      else {
		/* write != read -> read points to data */
		memcpy(buf,msg_buffer[read_index],msg_len[read_index]);
		io_write(buf,msg_len[read_index]);
		read_index=(read_index+1)%MAX_MESSAGE;
	      }
	      break;	      
	      	    
	    default:
#ifdef NET_DEBUG
	      printf("io server: unknown request\n");
#endif
	      break;
	    }
	  }
	  if(TM_FD_ISSET(sock,&fdset)) {
#ifdef NET_DEBUG
	    printf("io server: message via udp\n");
#endif
	    used = sizeof(buf) - 1;
	    i = sizeof inaddr;
	    if ((used = recvfrom(sock, buf, used, 0,
				    (struct sockaddr *)&inaddr, &i)) <= 0) {
	      perror("recvfrom");
	    } else {
#ifdef NET_DEBUG
	      printf("io server: got %d bytes from %s\n",
		     used, ip2a(inaddr.sin_addr.s_addr));
#endif
	      
	      /*  {  int i; for(i=0;i<bufused;i++) printf("%02x,",udp_buf[i]); } */
	      
	      switch(buf[0]) {
	      case NET_CONNECT:
		memcpy(&ip,&buf[1],sizeof(ip));
		ip=ntohl(ip);
#ifdef NET_DEBUG
		printf("NET_CONNECT from %s\n",ip2a(ip));
		printf("next_ip changes from %s ",ip2a(next_ip));
		printf("to %s\n",ip2a(inaddr.sin_addr.s_addr));
#endif
		/* now send own ip to new next */
		ip=next_ip;
		memcpy(&addr,&inaddr,sizeof(inaddr));
		next_ip=inaddr.sin_addr.s_addr;
		
		/* reply with own next_ip */
		buf[0]=NET_LINK; ip=htonl(ip);
		memcpy(&buf[1],&ip,sizeof(ip));
		send_udp(buf,sizeof(ip)+1);
		break;
		
		/* receive new next_ip */
	      case NET_LINK:
#ifdef NET_DEBUG
		printf("setting next_ip from %s ",ip2a(next_ip));
#endif
		memcpy(&next_ip,&buf[1],sizeof(ip));
		next_ip=ntohl(next_ip);
#ifdef NET_DEBUG
		printf("to %s\n",ip2a(next_ip));
#endif
		addr.sin_addr.s_addr=next_ip;

		/* and say hello to mom */
		buf[0]=IO_ACK; 
		memcpy(&buf[1],&next_ip,sizeof(next_ip));
		io_write(buf,sizeof(long)+1);
		break;
		
		/* if we can't handle it ourself, give it to mom */
	      default:
		/* if it's not from us forward it to next machine */
		if(buf[1]!=my_no) {
		  /* perhaps it's the SYNC-COMMAND? */
		  /* -> update counter */
		  if(buf[0]==NET_SYNC) { 
		    my_no=++buf[2]; 
#ifdef NET_DEBUG
		    printf("got new id: %d\n",my_no);
#endif
		  }
		  send_udp(buf,used);
		}

		/* buffer overrun ? */
		if( ((write_index+1)%MAX_MESSAGE)==read_index)
		  fprintf(stderr,"buffer overrun\n");
		else {
		  memcpy(msg_buffer[write_index],buf,used);
		  msg_len[write_index]=used;
		  write_index=(write_index+1)%MAX_MESSAGE;
		}
		break;
	      }
	    }	 
	  }
	} else {
#ifdef NET_DEBUG
	  printf("io server: Timeout!\n");
#endif
	  if(xtimeout++==MAX_XTIMEOUT) {
	    fprintf(stderr,"io server: Panic, mom doesn't talk anymore, doing suicide!\n");
	    child=1;
	  }
	}
      }
      /* close the unix domain socket */
      close(sockets[0]);
      /* close the udp socket */
#ifdef NET_DEBUG
      printf("io server: closing udp socket\n");
#endif
      close_udp();
      exit(0);  /* io server end here */
    }
  }
  return(OK);
}

/* first Byte = 0..31=destination, otherwise broadcast */
tm_bool
net_send(tm_char *data, tm_short len) {
  tm_char sbuf[MAX_PKT_LEN+1];
  sbuf[0]=IO_SEND;
  memcpy(&sbuf[1],data,len);

  /* send data to io server */
  io_write(sbuf,len+1);

  return(OK);
}

/* is data availlable? */
tm_bool
net_stat(tm_void) {
  struct timeval timeout;
  tm_fd_set fdset;
  tm_char stat;

  /* request net status */
  io_write_byte(IO_STAT);

  /* wait for answer */
  timeout.tv_usec=0; timeout.tv_sec=1;
  TM_FD_ZERO(&fdset);
  TM_FD_SET(sockets[1],&fdset);
  if(select(TM_LIMIT,&fdset,NULL,NULL,&timeout)>0) {
    read(sockets[1],&stat,1);
  } else {
    fprintf(stderr,"net_stat: io server not responing\n");
    return(0);
  }
  return(stat);
}

/* read one packet */
tm_short
net_recv(tm_char *data) {
  struct timeval timeout;
  tm_fd_set fdset;
  tm_char stat;
  tm_short len;

  /* request packet */
  io_write_byte(IO_RECV);

  /* wait for answer */
  timeout.tv_usec=0; timeout.tv_sec=1;
  TM_FD_ZERO(&fdset);
  TM_FD_SET(sockets[1],&fdset);
  if(select(TM_LIMIT,&fdset,NULL,NULL,&timeout)>0) {
    len=read(sockets[1],data,MAX_PKT_LEN);
  } else {
    fprintf(stderr,"net_recv: io server not responing\n");
    return(0);
  }
  return(len);
}

tm_void
kill_io_server(tm_void) {
  struct timeval timeout;
  tm_fd_set fdset;
  char   answer=0;

  /* ask server to die */
  io_write_byte(IO_QUIT);

  /* wait for answer */
  timeout.tv_usec=0; timeout.tv_sec=1;
  TM_FD_ZERO(&fdset);
  TM_FD_SET(sockets[1],&fdset);
  if(select(TM_LIMIT,&fdset,NULL,NULL,&timeout)>0)
    read(sockets[1],&answer,1);

  if(answer!=IO_QUIT) {
    /* kill io-server */
    fprintf(stderr,"io server not responding, killing io server\n");
    kill(child,SIGKILL);
  } 
#ifdef NET_DEBUG
  else {
    printf("io server died peacefully\n");
  }
#endif

  /* close socket to child */
  close(sockets[1]);
}

tm_void 
show_net_options(tm_void) 
{
  tm_char buf[128];

  if (gethostname(buf, sizeof buf - 1)) strcpy(buf,"unknown");
  console_printf("%s net driver options:\n",net_driver_name);
  console_printf("np=n\tset net port no to n (1000 - 9999), default is %d\n",PORT);
  console_printf("nm=n\tset n as receiver machine name, default is local (%s)\n",buf);
}

tm_word g_port;
tm_char g_name[128];

tm_void 
init_net_options(tm_void)
{
  g_port=PORT;
  /* get my machines name */
  if (gethostname(g_name, sizeof g_name - 1)) strcpy(g_name,"unknown");

  init_io_server();
}

tm_bool 
set_net_option(tm_char *cmd)
{
  tm_word val=atoi(&cmd[2]);

  if(cmd[1]=='=') {
    if((cmd[0]=='p')&&(val>=1000)&&(val<=9999)) {
      g_port=val;
      return(OK);
    }
    if(cmd[0]=='m') {
      strcpy(g_name,&cmd[2]);
      return(OK);
    }
  }
  return(FALSE);
}

tm_char*
get_player_name(tm_char *buffer, tm_word len)
{
  struct passwd *pass;
  tm_char *comma, *cp1, *cp2;
  static tm_char fullname[80];

  /* get environment entry */
  if((cp1=getenv(ENV_ENTRY_NAME))!=NULL) {
    strncpy(buffer,cp1,len);
    return(buffer);
  }
  
  /* Get user information from password file */
  if (!(pass = getpwuid(getuid())))
    return(buffer);    /* unknown user */

  /* find a comma indicating further info after name */
  comma = strchr(pass->pw_gecos, ',');
  
  /* NULL out the comma */
  if (comma) *comma = '\0';
  
  /* Use the nickname if not null otherwise password file name */
  cp1 = pass->pw_gecos;
  cp2 = fullname;
  
  /* Search through the gecos field looking for an '&' which on very
   * old UNIX systems is supposed to be the users user name with the
   * first letter uppercased.
   */
  while(*cp1)
    {
      /* Look for the '&' symbol */
      if(*cp1 != '&')
	*cp2++ = *cp1++;
      else {
	/* A ha. Now copy the users name to be in place of '&' */
	strcpy(cp2, pass->pw_name);
	
	/* Convert the first letter to uppercase. */
	if(islower(*cp2))
	  *cp2 = toupper(*cp2);
	
	/* Continue with the remaining string */
	while(*cp2) cp2++;
	cp1++;
      }
    }
  
  /* Return their name without any trailing stuff */
  strncpy(buffer,fullname,len);
  return(buffer);
}

tm_char*
ip2a(long ip) {
  static tm_char str[32];
  struct hostent *h;

  h=gethostbyaddr((char*)&ip,sizeof(ip),AF_INET);

  sprintf(str,"%s (%d.%d.%d.%d)",h->h_name,(htonl(ip)>>24)&0xff,(htonl(ip)>>16)&0xff,
	  (htonl(ip)>>8)&0xff,htonl(ip)&0xff);
  return(str);
}

tm_void 
send_ack(tm_bool ok) {
  tm_char ack[]={NET_ACK,0};

  ack[1]=ok;
  
  if (sendto(sock, ack, 2, 0, (struct sockaddr *)&addr, sizeof addr)!= 2) {
    fprintf(stderr, "cant sendto %s", ip2a(addr.sin_addr.s_addr));
    perror("");
  }
#ifdef NET_DEBUG
  printf("sent acknowledge to %s\n", ip2a(addr.sin_addr.s_addr));
#endif
}

tm_bool
open_udp(tm_void) {
  int i;

  if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("cant create socket");
    return(ERROR);
  }

  i=1;
  if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i))!=0) {
    perror("cant set socket options");
    return(ERROR);
  }

  /* Port an Adresse binden */
  addr.sin_family = AF_INET;
  addr.sin_port = g_port;
  addr.sin_addr.s_addr = INADDR_ANY;
  if(bind(sock, (struct sockaddr *)&addr, sizeof addr)) {
    perror("cant bind socket");
    return(ERROR);
  }

  return(OK);
}

tm_bool
close_udp(tm_void) {
  if(sock>=0) close(sock);
  return(OK);
}

tm_bool 
send_udp(tm_char *data, tm_word len) {
#ifdef NET_DEBUG
  printf("io server: send %d bytes to %s\n",len,ip2a(addr.sin_addr.s_addr));
#endif
  if (sendto(sock, data, len, 0, (struct sockaddr *)&addr, sizeof addr)!= len) {
    fprintf(stderr, "cant sendto %s", ip2a(addr.sin_addr.s_addr));
    perror("");
  }
}

tm_bool
stat_udp(tm_word to) {
  struct timeval timeout;
  tm_fd_set fdset;

  timeout.tv_usec=0; timeout.tv_sec=to;
  TM_FD_ZERO(&fdset);
  TM_FD_SET(sock,&fdset);
  if(select(TM_LIMIT,&fdset,NULL,NULL,&timeout)>0)
    return(OK);
  return(ERROR);
}

tm_word
recv_udp(char *data, tm_word len) {
  int i,ret;

  i = sizeof addr;
  if ((ret = recvfrom(sock, data, len, 0,
			  (struct sockaddr *)&addr, &i)) <= 0) {
    perror("recvfrom");
    return(ERROR);
  }
  return(ret);
}

tm_bool 
init_net(void) 
{
  tm_char buf[128],str[MAX_PKT_LEN];
  int bufused,i,iter,c;
  struct hostent *h;
  struct tm_packet_string;
  tm_fd_set fdset;
  struct timeval timeout;

  for(i=0;i<MAX_PLAYERS;i++) addr_assigned[i]=FALSE;

  /* get my machines name */
  if (gethostname(buf, sizeof buf - 1)) {
    fprintf(stderr, "unable get host name\n");
    return(ERROR);
  }

  /* get my machines ip address */
  h = gethostbyname(buf);
  if (!h) {
    fprintf(stderr, "cant get address for %s\n",buf);
    return(ERROR);
  }

  my_ip=*(int*)h->h_addr_list[0];

  /* access the host */
  h = gethostbyname(g_name);
  if (!h) {
    fprintf(stderr, "cant get address for %s\n",g_name);
    return(ERROR);
  }
  next_ip = addr.sin_addr.s_addr = *(int *)h->h_addr_list[0];

  fprintf(stderr,"    %s\n",ip2a(my_ip));
  fprintf(stderr," -> %s\nentering %s player mode\n",ip2a(next_ip),(my_ip==next_ip)?"single":"multi");

  /* ask server to open socket */
  buf[0]=IO_OPEN;
  memcpy(&buf[1],&next_ip,sizeof(next_ip));
  io_write(buf,sizeof(next_ip)+1);

  if(next_ip==my_ip) {
    str[0]=IO_MYIP;
    memcpy(&str[1],&my_ip,sizeof(my_ip));
    io_write(str,sizeof(my_ip)+1);
  } else {
    i_am_master=FALSE;
    /* ping next machine */
    c=1;
    do {
#ifdef NET_DEBUG
      printf("sending net connect\n");
#endif

      /* ask io server to create connection */
      str[0]=IO_CONNECT;
      memcpy(&str[1],&next_ip,sizeof(next_ip));
      memcpy(&str[1+sizeof(next_ip)],&my_ip,sizeof(my_ip));
      io_write(str,2*sizeof(next_ip)+1);

      /* wait for answer */
      timeout.tv_usec=0; timeout.tv_sec=1;
      TM_FD_ZERO(&fdset);
      TM_FD_SET(sockets[1],&fdset);
      if(select(TM_LIMIT,&fdset,NULL,NULL,&timeout)>0) {
	read(sockets[1],&str,sizeof(long)+1);
	/* data available? */
	if(str[0]==IO_ACK) {
#ifdef NET_DEBUG
	  printf("TheMaze: connected to %s\n",ip2a(addr.sin_addr.s_addr));
	  printf("TheMaze: next_ip changes from %s ",ip2a(next_ip));
#endif
	  memcpy(&next_ip,&str[1],sizeof(next_ip));
#ifdef NET_DEBUG
	  printf("to %s\n",ip2a(next_ip));
#endif
	  c=0;
	}
      } else {
	if(c++==10) {
	  fprintf(stderr,"timeout connecting to %s\n",ip2a(addr.sin_addr.s_addr));
	  return(ERROR);
	}
      }      
    } while(c);
  }

  return(OK);
}

tm_void
exit_net(void)
{
  kill_io_server();
}



