/*
   sgiaudio.c

   30.10.97 Till   first conversion from linuxaudio
   31.10.97 Till   added do_music
*/


#include "ctypes.h"
#include "audio.h"
#include "joystick.h"
#include "net.h"
#include "misc.h"
#include "../music/music.h"
#include "../music/the_song.h"

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/signal.h>
#include </usr/include/audio.h>            /* SGI audio interface lib. header */

#include "ulaw.h"

#define INITIAL_VOL 60

extern tm_char *tm_sounds[];
extern tm_word tm_sound_len[];

int audio_vol;
tm_bool aon=ON;
tm_char audio_driver_name[]="SGI Audio V0.1";

static ALport	thePort = 0;
static ALconfig theConf = 0;

static long defConf[] = 
{
    AL_OUTPUT_RATE, 		8000,
    AL_INPUT_RATE, 		8000, 
    AL_LEFT_SPEAKER_GAIN, 	128, 
    AL_RIGHT_SPEAKER_GAIN, 	128
};

static long gainConf[] = { 
    AL_LEFT_SPEAKER_GAIN, 	128, 
    AL_RIGHT_SPEAKER_GAIN, 	128,  
};

static void audioSetGain(int volume) {
  gainConf[1] = gainConf[3] = volume;
  ALsetparams( AL_DEFAULT_DEVICE, gainConf, 4);
}


static int audioOpenPort(void) {
  ALsetparams(AL_DEFAULT_DEVICE, defConf, 8);
  theConf = ALnewconfig();
  ALsetchannels(theConf, AL_MONO);
  ALsetwidth(theConf, AL_SAMPLE_16);
  
  if (theConf) {
    thePort = ALopenport ("TheMaze audio", "w", theConf);
    if (thePort)
      return TRUE; 
  }
  
  return FALSE;   
}

static void audioClosePort(void) {
    ALfreeconfig(theConf);
    ALcloseport(thePort);
    theConf = 0;
    thePort = 0;
}

int flush_audio(void) {
  int ok;

  /* Big hack but I don't know how to do it otherwise */
  audioClosePort();
  ok = audioOpenPort();
  audioSetGain(audio_vol);  

  if (ok)
    return TRUE;
  
  return FALSE;
}

tm_void show_audio_options(tm_void) {
  console_printf("%s audio driver options:\n",audio_driver_name);
  console_puts("as=n\tswitch audio output on (n=1) or off (n=0), default is on");
  console_printf("av=n\tset audio volume to n (0..100), default is %d\n", INITIAL_VOL);
}

tm_bool  g_on_f,g_on,g_vol_f;
tm_uword g_vol;

tm_void init_audio_options(tm_void) {
  g_on_f  = FALSE;
  g_vol_f = FALSE;
}

tm_bool set_audio_option(tm_char *cmd) {
  tm_uword val=atoi(&cmd[2]);

  if(cmd[1]=='=') {
    if((cmd[0]=='s')&&((val==0)||(val==1))) {
      g_on_f=TRUE;
      g_on=(tm_bool)val;
      return(OK);
    }
    if((cmd[0]=='v')&&(val>=0)&&(val<=100)) {
      g_vol_f=TRUE;
      g_vol=val;
      return(OK);
    }
  }
  return(FALSE);
}

tm_bool init_audio(tm_void) {
  tm_bool ok = audioOpenPort();
  audioSetGain(INITIAL_VOL);
  audio_vol=INITIAL_VOL;

  return ok;
}

tm_bool exit_audio(tm_void) {
  while (ALgetfilled(thePort)) 
    sginap(1);
  
  audioClosePort();
  
  return(OK);
}

tm_bool audio_on(tm_bool *on, tm_bool get) {
  if(thePort) {
    if(get) *on = aon;
    else    aon = *on;
  } else
    *on = FALSE;

  return(OK);
}

/* volume ranges from 0..100 */
tm_bool vol_audio(tm_uword *volume, tm_bool get) {
  int vol;
  int err;

  if(thePort) {
    if(get)
      *volume = audio_vol;
    else {
      audio_vol = *volume;
      audioSetGain(audio_vol);
    }
  } else
    if(get) *volume=0;

  return(OK);
}

/* dumm-di-dumm trallala */
#define MUSIC_ENDLESS
#define MUSIC_OUTPUT_RATE     (8000)  /* this is 8.000 kHz */
#define MUSIC_BEATS_PER_MIN   (80)    /* try techno polka! */
#define MUSIC_OUTPUT_SIZE     (60*MUSIC_OUTPUT_RATE/MUSIC_BEATS_PER_MIN)
#define MUSIC_BUFFER_SIZE     (MUS_MAX_PRECALC+MUSIC_OUTPUT_SIZE)
tm_long  music_buffer[MUSIC_BUFFER_SIZE];
tm_short music_output_buffer[MUSIC_OUTPUT_SIZE];

tm_void
do_music(tm_void) {
  tm_uchar *pos=(tm_uchar*)mus_song,sample;
  tm_word  i,time,vol,len,joy,beatlen;
  tm_ulong val,factor;
  tm_long  *dest;
  tm_short *src;
  int format;

  /* play music even with wrong playback speed */
  if(aon) {
    /* clear buffer */
    memset(music_buffer,0,sizeof(music_buffer));

    /* get length of beat in n/32 */
    beatlen=*pos++;

    do {
      time=0;  /* start a new beat */
      while((time<beatlen)&&(*pos!=MUS_END)) {   /* one beat completed ? */
	    
	sample=*pos++;                      /* get sample no */
	
	if(sample==MUS_WAIT) {
	  time+=*pos++;     /* time in n/32 beats */
	} else {
	  /* get frequency factor  */
	  factor=*pos++*256*256*256; factor+=*pos++*256*256; factor+=*pos++*256; factor+=*pos++;
	  vol=*pos++; len=*pos++; /* get volume and length */
	  
	  /* compose samples in buffer */
	  dest=&music_buffer[time*(MUSIC_OUTPUT_SIZE/beatlen)]; 
	  src=(tm_short*)mus_sample[sample];
	  for(i=0;i<((tm_word)mus_sample_len[sample]<<16)/factor;i++)
	    *dest+++=(vol*src[(i*factor)>>16])>>8;
	}
#ifdef MUSIC_ENDLESS
	if(*pos==MUS_END)
	  pos=(tm_uchar*)mus_song+1;
#endif
      }
      
      /* play this beat */
      /* output buffer (4*long -> signed short ) */
      for(i=0;i<MUSIC_OUTPUT_SIZE;i++)
	music_output_buffer[i]=(music_buffer[i]/MUS_INSTRUMENTS/4)&0xffff;
      /* shift music buffer down */
      memcpy(music_buffer,&music_buffer[MUSIC_OUTPUT_SIZE],(MUSIC_BUFFER_SIZE-MUSIC_OUTPUT_SIZE)*sizeof(tm_long));
      /* clear buffer end */
      memset(&music_buffer[MUSIC_BUFFER_SIZE-MUSIC_OUTPUT_SIZE],0,MUSIC_OUTPUT_SIZE*sizeof(tm_long));
      
      /* and down to the device */
      ALwritesamps(thePort, music_output_buffer, MUSIC_OUTPUT_SIZE);
      joy=get_joystick();
      net_handle_messages();
      /* music running and nobody pressed fire and game not running? */
    } while((*pos!=MUS_END)&&!(joy&JOY_FIRE)&&!(game_running));
    
    /* wait for button release */
    if(joy&JOY_FIRE) {
      flush_audio();
      wait_msec(10);
      do
	joy=get_joystick();
      while(joy&JOY_FIRE);
    } else {
      if(!game_running)
	/* wait two beats for music to stop */
	wait_msec(2000*MUSIC_BEATS_PER_MIN/60);
      else
	flush_audio();
	}
    
    if(!aon) {
      do {
	joy=get_joystick();
	net_handle_messages();
      } while(!(joy&JOY_FIRE)&&(!game_running));
      while(joy&JOY_FIRE) joy=get_joystick();
    }
  } else {
    /* no audio: wait for key and talk to net */
    do {
      joy=get_joystick();
      net_handle_messages();
    } while(!(joy&JOY_FIRE)&&(!game_running));
    while(joy&JOY_FIRE) joy=get_joystick();
  }
}

#define AUDIO_BUFFER_LEN 32768
tm_void do_audio(tm_sound sound) {
  int i;
  short linBuff[AUDIO_BUFFER_LEN];

  if(aon) {
    /* Flush the sound port for the new sound please */
    if (ALgetfilled(thePort)) {
      if (!flush_audio()) {
	puts("Cannot flush sound system.");
	return;   
      }
    }

    if(tm_sound_len[sound]>AUDIO_BUFFER_LEN) {
      puts("sgiaudio.c warning: sound length exceeds buffer length");
      tm_sound_len[sound]=AUDIO_BUFFER_LEN;
    }

    /* Now do the ulaw stuff to the raw data */
    for(i = 0; i < tm_sound_len[sound]; i++) {
      linBuff[i] = UlToLin(tm_sounds[sound][i]);
    }
    
    /* Write the samples out to the SGI audio device */
    ALwritesamps(thePort, linBuff, tm_sound_len[sound]);
  }
}

tm_void test_audio(tm_void) {
  do_audio(SOUND_TEST);
}





