/*
   unix/x11_win.c

   video/io driver for the XWindow system

   tested with
     - SunOS 4
     - SunOS 5
     - Irix
     - Linux
     - HPUX

   1 bit support
   8 bit support
  32/24/16/15 simulates 8 bit

*/

#define WARP             /* Maus im Rechteck halten */
#define MOUSE_INVISIBLE
#define USE_CLIB         /* Grafikprimitiva duerfen string.h benutzen */
#define noCATCH_EX
#define noFADE

#define RCFILENAME  ".themazerc"

#include "../conf.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/cursorfont.h>
#ifdef XSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <sys/termios.h>
#include <ctype.h>
#include <pwd.h>
#include <sys/types.h>

#include "../ctypes.h"
#include "../gr_prim.h"
#include "../text.h"
#include "../grafik.h"
#include "../timer.h"
#include "../machine.h"
#include "../error.h"
#include "../gif.h"
#include "../misc.h"
#include "../maze.h"

extern tm_void kill_io_server(tm_void);

/* Wir definieren uns ein Window */
#define WIDTH_A (width)
#define WIDTH   (width)
#define HEIGHT  (height)
#define COL256

#define JOY_KEY     (0)
#define JOY_MOUSE   (1)
#define JOY_JOY     (2)
#define JOY_LP      (3)

#define JOY_TIMEOUT (100000)

#define STEPS 50

#define COPY_TO_SCREEN

#define JOY_LP_DEV  "/dev/lpj"

#ifdef sun
#define JOY_DEV_LIST "/dev/ttya","/dev/ttyb"
#endif

#ifdef linux
#define JOY_DEV_LIST "/dev/cua0","/dev/cua1","/dev/cua2","/dev/cua3"
#endif

#ifdef hpux
#define JOY_DEV_LIST "/dev/rs232.1","/dev/rs232.2"
#endif

#ifdef sgi
#define JOY_DEV_LIST "/dev/ttyd1","/dev/ttyd2"
#endif

#ifndef JOY_DEV_LIST
#warning Please define some serial devices for your machine!
#define JOY_DEV_LIST 
#endif

char *device_name[]={JOY_DEV_LIST,""};

tm_word joy_inp;
char *screen_base;

tm_char *offscreen_buffer;

#define RAW2
#define RAW256    /* this system supports the 256 color mode */
#include "../gr_prim.c"

char video_driver_name[]        = "X11 RAW SHM V0.91";
char video_driver_name_no_shm[] = "X11 RAW V0.91";

char joystick_driver_name[]     = "X11 Input V0.9";
char timer_driver_name[]        = "Unix Timer V1.0";

char icon_name[]="TheMaze-X11";
#include "the_maze.icn"

void 
my_usleep(long usec) {
  struct timeval timeout;
  timeout.tv_sec  = 0;
  timeout.tv_usec = usec;
  select (0, 0, 0, 0, &timeout);
}

/*
 * Check if the X Shared Memory extension is available.
 * Return:  0 = not available
 *          1 = shared XImage support available
 *          2 = shared Pixmap support available also
 */
static int 
check_for_xshm( Display *display ) {
#ifdef XSHM
   int major, minor, ignore;
   Bool pixmaps;

   if (XQueryExtension( display, "MIT-SHM", &ignore, &ignore, &ignore )) {
      if (XShmQueryVersion( display, &major, &minor, &pixmaps )==True) {
	 return (pixmaps==True) ? 2 : 1;
      } else {
	 return FALSE;
      } } else {
      return FALSE;
   }
#else
   /* Can't compile XSHM support */
   return FALSE;
#endif
}

tm_uword font_height;
extern tm_uword font_height0, font_height1;
CHARACTER charTable[128];
extern CHARACTER charTable0[128], charTable1[128];
extern tm_char dithermatrix[][8];
extern tm_picture pic_tm;
extern tm_uchar gif_grey[256];


/* ein paar Sachen fuers X11 */
Display  *display;
int      screen;
GC 	 gc;
Window   win;
XImage   *xi;
Colormap cmap;
Visual   *visual;
Pixmap   Cursorpix;
Cursor   cursor;
tm_bool  use_shm, fullscreen;
int      planes,nplanes;
int      key_auto_state;
int      rev_bits=FALSE;

/* things needed for true color support */
tm_bool  truecol;
char     *truebuf;
unsigned short    true_table_16[256];
unsigned long     true_table_32[256];
unsigned int      red_shift, red_mask;
unsigned int      blue_shift, blue_mask;
unsigned int      green_shift, green_mask;

/* things for bw support */
unsigned short    *rev_table;

#ifdef XSHM
XShmSegmentInfo shminfo;
#endif

tm_void 
show_video_options(tm_void) {
  console_printf("%s video driver options:\n",video_driver_name);
  console_printf("vs=n*m\tset window size, default is %d*%d (min %d*%d, max %d*%d)\n\tfullscreen with vs=full (only if fullscreen < maximum)\n",
		 DEF_REZ_WIDTH, DEF_REZ_HEIGHT, MIN_REZ_WIDTH, MIN_REZ_HEIGHT, MAX_REZ_WIDTH, MAX_REZ_HEIGHT);
  console_puts("vq=n\tset video quality to BW (n=2) or 256 colors (n=256), default is best");
#ifdef TEXTURE
  console_puts("vt=n\tuse texture mapping (n=on/off), default is on");
#endif
  console_puts("vp=n\tforce x server to n bits, default is best/machine dependent");
}

tm_word quality, splanes;
#ifdef TEXTURE
tm_bool texon;
#endif

tm_void 
init_video_options(tm_void) {
  quality=0;
#ifdef TEXTURE
  texon=TRUE;
#endif
  splanes=0;

  width = DEF_REZ_WIDTH;
  height = DEF_REZ_HEIGHT;
  fullscreen=FALSE;
}

tm_bool 
set_video_option(tm_char *cmd) {
  tm_word val=atoi(&cmd[2]),n;
  char str[10];

  if(cmd[1]=='=') {
    if(((cmd[0]=='q')||(cmd[0]=='Q'))&&((val==2)||(val==256))) {
      quality=val;
      return(OK);
    }
    
#ifdef TEXTURE
    if( ((cmd[0]=='t')||(cmd[0]=='T'))
       &&((cmd[2]=='O')||(cmd[2]=='o'))
       &&((cmd[3]=='F')||(cmd[3]=='f'))
       &&((cmd[4]=='F')||(cmd[4]=='f'))) {
      texon=FALSE;
      return(OK);
    }
#endif

    if((cmd[0]=='p')||(cmd[0]=='P')) {
      splanes=val;
      return(OK);
    }

    if((cmd[0]=='s')||(cmd[0]=='S')) {
      if((strncmp(&cmd[2], "full",4)==0)||
	 (strncmp(&cmd[2], "FULL",4)==0)) {
	fullscreen = TRUE;
      } else {

	n=0;
	while(isdigit(cmd[n+2])) { 
	  str[n]=cmd[n+2]; n++; 
	}
	str[n]=0;
	
	if((cmd[n+2]!='*')&&(cmd[n+2]!='x'))
	  return(FALSE);
	
	width = atoi(str);
	height = atoi(&cmd[n+3]);
	fullscreen = FALSE;
	
	if((width<MIN_REZ_WIDTH)||(height<MIN_REZ_HEIGHT)||
	   (width>MAX_REZ_WIDTH)||(height>MAX_REZ_HEIGHT)) {
	  return(FALSE);
	}
      }

      return(OK);
    }
  }
  return(FALSE);
}

/* no double buffer with unix */
tm_void
update_double_buffer(tm_void) {
}

void
setup_1(void) {
  int i;
  unsigned short *d=(short*)xi->data;
  unsigned short *s=(short*)screen_base;

  if(rev_bits) 
    for(i=0;i<(WIDTH*HEIGHT/16);i++)
      *d++=rev_table[*s++];
}

void
setup_16(void) {
  int  i;
  unsigned short *d=(short*)xi->data;
  unsigned char *s=screen_base;

  for(i=0;i<(WIDTH*HEIGHT);i++)
    *d++=true_table_16[*s++];
}

void
setup_32(void) {
  int  i;
  unsigned long *d=(long*)xi->data;
  unsigned char *s=screen_base;

  for(i=0;i<(WIDTH*HEIGHT);i++)
    *d++=true_table_32[*s++];
}

void 
flip_screen(void) {

  if(truecol) {
    if(nplanes==32) setup_32();
    if(nplanes==16) setup_16();
  } else
    if(planes==1) setup_1();

  if(use_shm)
#ifdef XSHM
    XShmPutImage(display, win, gc, xi, 0, 0, 0, 0, WIDTH, HEIGHT, False);
#else
    ;
#endif
  else
    XPutImage(display, win, gc, xi, 0, 0, 0, 0, WIDTH, HEIGHT);

  XSync(display, False);
}

void 
restore_keyboard(void) {
  XKeyboardControl key_control;

  if(joy_inp == JOY_KEY) {
    /* restore keyboard state */
    key_control.auto_repeat_mode =  key_auto_state;
    XChangeKeyboardControl(display,KBAutoRepeatMode,&key_control);
  }
}

/* new signal handler */
void 
my_handler(int sig) {
  do_error(sig,"signal caught");
  sleep(1);
  kill_io_server();
  restore_keyboard();

  exit(1);
}

tm_bool Xerr;
int 
my_xerror_handler(Display *d,  XErrorEvent *e) {
  Xerr = TRUE;
  return 0;
}

tm_void 
parse_config_file(tm_void) {
  FILE *rcfile;
  char str[128],str2[128],sid[32];
  struct passwd   *pw;
  int scanning,n,m,id,k;
  char hostname[64];

  if(gethostname(hostname,64)!=0) {
    strcpy(hostname,"unknown");
  }

  /* remove trailing stuff */
  for(n=0;n<strlen(hostname);n++)
    if(hostname[n]=='.') hostname[n]=0;
	
  /* locate user's home directory */
  pw=getpwuid(getuid());
  if(!pw) { 
    printf("Who are you? I can't find you in the passwd database!\n");
    return;
  }

  sprintf(str,"%s/%s",pw->pw_dir, RCFILENAME);

  if((rcfile = fopen(str, "r"))==NULL) {
    printf("no config file found, using global defaults\n");
    return;
  }

  while(fgets(str,128,rcfile)!=NULL) {
    k=0; m=0; n=0; scanning=FALSE; id=FALSE;
    while((str[n]!=0)&&(str[n]!=';')){
      if((!scanning)&&(!id)) {
	if(!isspace(str[n])) {
	  if(str[n]!='[') {
	    str2[m++]=str[n];
	    scanning=TRUE;
	  } else {
	    id=TRUE;
	    k=0;
	  }
	}
      } else {
	if(id) {
	  if(str[n]==']') {
	    sid[k]=0;
	    id=FALSE;
	  } else
	    sid[k++]=str[n];
	} else
	  str2[m++]=str[n];
      }
      n++;
    }

    if(id) {
      printf("syntax error in config file: '%s'\n", str);
      m=0;
    }

    if((k!=0)&&(m>0)) {
      /* if not this machines name, remove option */
      if(strcmp(sid,hostname)!=0) m=0;
    }

    if(m>0) while(str2[m-1] && isspace(str2[m-1])) m--;
    str2[m]=0;

    switch(str2[0]) {
      case 0:
	break;   /* ignore empty strings */
    case 'a':
      if(!set_audio_option(&str2[1]))
	console_printf("audio option error in config file: '%s' (ignored)\n",str2);
      break;

    case 'v':
      if(!set_video_option(&str2[1]))
	console_printf("video option error in config file: '%s' (ignored)\n",str2);
      break;

    case 'n':
      if(!set_net_option(&str2[1]))
	console_printf("net option error in config file: '%s' (ignored)\n",str2);
      break;

    case 'i':
      if(!set_joystick_option(&str2[1]))
	console_printf("input option error in config file: '%s' (ignored)\n",str2);
      break;

    default:
      console_printf("option error in config file: '%s' (ignored)\n",str2);
      break;
    }
  }
  
  fclose(rcfile);
}

tm_bool 
open_screen( tm_uword argc, tm_char **argv, tm_uword *width, tm_uword *height, tm_uword *colors ) {

  /* und hier muss erstmal ein Fenster aufgemacht werden */
  unsigned int display_width, display_height,i,j;
  Pixmap icon_pixmap;
  XColor fade_col[256],col;
  XSizeHints size_hints;
  char *display_name=NULL;
  int window_size=0;
  unsigned long valuemask=0;
  XGCValues values;
  XEvent report;
  XStandardColormap *std_map;
  XKeyboardState key_state;
  XKeyboardControl key_control;
  int (*old_xerror_handler)(Display *, XErrorEvent *);

  /* stuff for searching the right visual */
  XVisualInfo vTemplate;
  XVisualInfo *visualList;
  int visualsMatched;
  XSetWindowAttributes attributes;
  unsigned long attrvaluemask;

  if((display=XOpenDisplay(display_name))==NULL) {
    fprintf(stderr,"the_maze: cannot connect to X server %s\n",
                         XDisplayName(display_name));
    kill_io_server();
    exit(1);
  }

  if(fullscreen) {
    *width  = XScreenOfDisplay(display, screen)->width;
    *height = XScreenOfDisplay(display, screen)->height;

    if((*width<DEF_REZ_WIDTH)||(*height<DEF_REZ_HEIGHT)||
       (*width>MAX_REZ_WIDTH)||(*height>MAX_REZ_HEIGHT)) {
      *width = DEF_REZ_WIDTH;
      *height = DEF_REZ_HEIGHT;
      fullscreen = FALSE;
    }
  }

  /* copy right character info */
  if((*width>=640)&&(*height>=400)) {
    font_height = font_height1;
    memcpy(charTable, charTable1, 128* sizeof(CHARACTER));
  } else {
    font_height = font_height0;
    memcpy(charTable, charTable0, 128* sizeof(CHARACTER));
  }

  if(splanes!=0) {

    /* now search for the desired visual */
    vTemplate.screen = DefaultScreen(display);
    vTemplate.depth = splanes;
    visualList = XGetVisualInfo(display, VisualScreenMask |
				VisualDepthMask, &vTemplate, &visualsMatched);
  
    if(visualsMatched == 0) {
      fprintf(stderr,"the_maze: no visual of depth %d found\n", splanes);
      kill_io_server();
      return FALSE;
    }

    /* use first Visual matching */
    visual = visualList[0].visual;
    screen = visualList[0].screen;
    planes = nplanes = splanes;

    printf("visual found with depth %d\n",splanes);
  } else {
    /* use default visual */
    screen = DefaultScreen(display);
    visual = DefaultVisual(display,screen);
    nplanes = planes = DisplayPlanes(display,screen);
  }

  if((planes!=8)&&(planes!=24)&&(planes!=32)&&(planes!=15)&&(planes!=16))  
    planes=1;  /* no 4/16/64 color support */
  if(quality==2) planes=1;  /* the user wants bw, we give him bw */

  /* simulate 8 bit? */
  if((planes==24)||(planes==32)||(planes==15)||(planes==16)) {
    int i;

    /* Korrektur fuer nicht-ganzzahlige Modi */
    if(planes==24) nplanes=32;
    if(planes==15) nplanes=16;

    red_shift=0;   while(!((visual->  red_mask >>   red_shift)&1))   red_shift++;
    green_shift=0; while(!((visual->green_mask >> green_shift)&1)) green_shift++;
    blue_shift=0;  while(!((visual-> blue_mask >>  blue_shift)&1))  blue_shift++;
    
    red_mask   = visual->  red_mask >>   red_shift;
    green_mask = visual->green_mask >> green_shift;
    blue_mask  = visual-> blue_mask >>  blue_shift;

    truebuf=malloc(*width * *height);
    truecol=TRUE;    
  } else {
    truecol=FALSE;
    truebuf=(char*)0;
  }

  if (splanes == 0) {
    /* einfaches Window kreieren */
    win = XCreateSimpleWindow(display,RootWindow(display,screen),0,0,*width, *height,
			    0,BlackPixel(display,screen),WhitePixel(display,screen));
  } else {
    cmap   = XCreateColormap(display,  RootWindow(display,screen), visual, AllocNone);
    
    /* Spezial-Window kreieren */
    attrvaluemask = CWColormap | CWBackPixel | CWBorderPixel;
    attributes.background_pixel = WhitePixel(display,screen);
    attributes.border_pixel = BlackPixel(display,screen);
    attributes.colormap = cmap;
    win = XCreateWindow(display, RootWindow(display,screen),0,0,*width, *height,
			  0, (splanes==0)?planes:splanes, InputOutput, visual, attrvaluemask, &attributes);
  }

  if(fullscreen) {
    attributes.override_redirect = True;
    attrvaluemask = CWOverrideRedirect;

    XChangeWindowAttributes(display, win, attrvaluemask, &attributes);
  }

  /* Icon anmelden */
  icon_pixmap=XCreateBitmapFromData(display,win,icon_bitmap_bits,
				    icon_bitmap_width,icon_bitmap_height);
  
  size_hints.flags=PPosition|PSize|PMinSize;
  size_hints.x=0;
  size_hints.y=0;
  size_hints.min_width  = size_hints.width  = *width;
  size_hints.min_height = size_hints.height = *height;

  XSetStandardProperties(display,win,"TheMaze X11",icon_name,icon_pixmap,argv,
                         argc,&size_hints);

  if(joy_inp==JOY_KEY) {
    /* non repeating keyboard */
    XGetKeyboardControl(display,&key_state);
    key_auto_state = key_state.global_auto_repeat;
    key_control.auto_repeat_mode = AutoRepeatModeOff;
    XChangeKeyboardControl(display,KBAutoRepeatMode,&key_control);
  }

  /* Keyboard & Joy is simultanously possible */
  if(joy_inp!=JOY_MOUSE)
    XSelectInput(display,win,ExposureMask|KeyPressMask|KeyReleaseMask|
		 ButtonPressMask|ButtonReleaseMask|StructureNotifyMask);
  else
    XSelectInput(display,win,ExposureMask|KeyPressMask|KeyReleaseMask|
		 ButtonPressMask|ButtonReleaseMask|StructureNotifyMask|
		 PointerMotionMask);

  gc=XCreateGC(display,win,valuemask,&values);
  XSetForeground(display,gc,WhitePixel(display,screen));
  XSetBackground(display,gc,BlackPixel(display,screen));

  /* Window anzeigen */
  XMapWindow(display,win);
  
  /* warten, bis das Window da ist */
  do {
    while(!XPending(display))
      net_handle_messages();
    XNextEvent(display,&report);
    if(report.type == Expose)
      while(XCheckTypedEvent(display,Expose,&report));
  }
  while(report.type != Expose);

  if(fullscreen) {
    /* fullscreen: grab focus */
    XSetInputFocus(display, win, RevertToPointerRoot, CurrentTime);
  }

  /* unterstuetzt das X11 shared memory ? */
  use_shm = check_for_xshm(display);
  if(use_shm) {
#ifdef XSHM
    if(planes == 1)
      xi = XShmCreateImage(display, visual, 1, XYBitmap, 0, &shminfo, *width, *height);
    else
      xi = XShmCreateImage(display, visual, planes, ZPixmap, 0, &shminfo, *width, *height);

    if(xi==NULL) {
      use_shm = FALSE;
      printf("TheMaze: XShmCreateImage - error creating image\n");
    } else {
      if((shminfo.shmid = shmget( IPC_PRIVATE, *width * *height * nplanes / 8, IPC_CREAT|0777 )) == -1) {
	perror("TheMaze: shmget()");
	use_shm = FALSE;
	XDestroyImage(xi);
      } else {
	if((shminfo.shmaddr=xi->data=screen_base=(char*)shmat( shminfo.shmid,0,0))== (char*)-1) {
	  use_shm = FALSE;
	  XDestroyImage(xi);
	} else {
          /* ggf. Zwischenbuffer benutzen */
          if(truecol) screen_base=truebuf;

	  shminfo.readOnly = False;
	  Xerr = FALSE;
	  old_xerror_handler = XSetErrorHandler(my_xerror_handler);
	  XShmAttach( display, &shminfo );   /* shared memory geht nicht per fremdem Display */
	  XSync( display, False );
	  my_usleep(100000);  /* give error some time to occur */
	  XSetErrorHandler(old_xerror_handler);
	  if(Xerr) {
	    use_shm = FALSE;
	    XDestroyImage(xi);
	  } else
	    shmctl( shminfo.shmid, IPC_RMID, 0 ); /* nobody else needs it */
	}
      }
      if(!use_shm)
	fputs("TheMaze: Unable to use MIT-SHM!\n",stderr);
    }
#else
    ;
#endif
  }
  
  if(!use_shm) {
    /* ok because s2 is shorter than s1 */
    strcpy(video_driver_name,video_driver_name_no_shm);
    
    if((screen_base = malloc(*width * *height * nplanes/8))==NULL) {
      puts(no_video_buffer[language]);
      return(FALSE);
    }   
    /* Bitmap-Daten lesen und auf den Puffer schalten */
    if(planes==1)
      xi = XCreateImage(display, visual, 1, XYBitmap, 0, screen_base, *width, *height, 8, 0);
    else
      xi = XCreateImage(display, visual, planes, ZPixmap, 0, screen_base, *width, *height, 8, 0);

    if(truecol)
      screen_base=truebuf;
  }
  
#ifdef MOUSE_INVISIBLE
  /* Mauszeiger verstecken */
  Cursorpix = XCreatePixmap(display, win, 1, 1, 1); 
  col.red = col.green = col.blue = 0xffff; col.pixel = 0; col.flags = 0xff;
  cursor = XCreatePixmapCursor(display, Cursorpix, Cursorpix, &col, &col, 0, 0);
  XDefineCursor(display, win, cursor);
#endif

  *colors = (1<<planes);
  if(*colors>256)
    *colors=256;          /* max 256 colors */

#ifdef TEXTURE
  if((*colors==256)&&(texon))
    use_texture=TRUE;    /* use textures if possible */
#endif

#ifdef CATCH_EX
  /* catch some exceptions */
  signal(SIGINT,my_handler);
  signal(SIGSEGV,my_handler);
  signal(SIGBUS,my_handler);
  signal(SIGTERM,my_handler);
#endif

#ifdef WARP
  /* warp mouse into window */
  XWarpPointer(display, RootWindow(display, screen), 
	       win, 0, 0, *width, *height, (*width)/2, (*height)/2);
#endif

  /* initialize to desired color depth */
  if(planes==1)
    set_multimode(MULTI_RAW2);
  else
    set_multimode(MULTI_RAW256);

  rev_table = NULL;

  if(planes==1) {
#if 0
    if(xi->byte_order!=MSBFirst) {
       fprintf(stderr,"Sorry, the black'n white xserver has to display\nsignificant byte first\n");
       kill_io_server();
       close_screen();
       exit(1);
    }
#endif

    /* revert bit priority? */
    if(xi->bitmap_bit_order!=MSBFirst) {
      rev_bits=TRUE;
      rev_table=malloc(65536*sizeof(short));
      for(i=0;i<65536;i++) {
	rev_table[i]=0;
	for(j=0;j<16;j++)
	  if(i&(1<<j))
	    rev_table[i]|=(0x8000)>>j;
	rev_table[i]= ((rev_table[i]&0xff)<<8)|((rev_table[i]&0xff00)>>8);
      }   
      /* Zwischenpuffer benutzen */
      screen_base=truebuf=malloc(*width * (*height)/8);
    }
  }

  offscreen_buffer = (tm_uchar*)malloc(MAX(*width,pic_tm.width) * MAX(*height,pic_tm.height));
  if(offscreen_buffer == NULL) {
    console_printf("x11_win.c: out of memory allocating %d bytes\n", *width * *height);
    return(FALSE);
  }

  /* fade to black */
  if((planes>1)&&(!truecol)) {
    cmap = XCreateColormap(display, win, visual, AllocAll);
    if(fullscreen) {
      XGrabServer(display);
      if((visual->class != TrueColor ) && (visual->class != DirectColor))
	XInstallColormap(display, cmap);
    }
#ifdef FADE
    {
      XColor orig_col[256];
      
      /* get systems default colormap */
      for (j = 0; j < *colors; j++) orig_col[j].pixel = j;
      XQueryColors (display, DefaultColormap (display, screen), orig_col, *colors);
      XStoreColors (display, cmap, orig_col, *colors);
      XSetWindowColormap(display, win, cmap);

      /* fade down to black in STEPS steps */
      for (i = STEPS; i > 0; i--) {
	for (j = 0; j < *colors; j++) {
	  fade_col[j].red   = orig_col[j].red   * i / STEPS;
	  fade_col[j].green = orig_col[j].green * i / STEPS;
	  fade_col[j].blue  = orig_col[j].blue  * i / STEPS;
	  fade_col[j].pixel = j;
	  fade_col[j].flags = 0xff;
	}

	/* display colors */
	XStoreColors (display, cmap, fade_col, *colors);
	my_usleep (10);
	XSync (display, False);
      }
    }
#else
    XSetWindowColormap(display, win, cmap);

    /* don't fade, just black */
    for (j = 0; j < *colors; j++) {
      fade_col[j].red   = fade_col[j].green = fade_col[j].blue  = 0;
      fade_col[j].pixel = j;
      fade_col[j].flags = 0xff;
    }
    XStoreColors (display, cmap, fade_col, *colors);
#endif
  }

  return(OK);
}

tm_void 
fade_title_pic(tm_char *copyright) {
  XColor fade_col[256];
  int i,j,N,M,k,t;
  tm_uword col[3];
  unsigned char *src,*p,*dst;
    
#if 0
  n1=(tex_height>>1)-1; n2=tex_height>>1;
  for(M=0,cnt=0;cnt<(len>>1);cnt++) {
    while(M>real_len) { M-=real_len; n1--; n2++; }  /* This is only needed if the texture is greater than the wall */
    M+=tex_height;
    *start1=coltab[texture[n1]];  *start2=coltab[texture[n2]];
    start1-=(tm_uword)WIDTH;      start2+=(tm_uword)WIDTH;
}
#endif

  /* show title screen */
  if((width==pic_tm.width)&&(height==pic_tm.height))
    show_gif("tm",screen_base);
  else {

    /* this is extreme naughty, but it works */
    k=colors;
    colors=256;
    show_gif("tm",offscreen_buffer);
    colors=k;

    if(colors==256) {
      /* now scale title pic to screen size */
      for(dst=screen_base,src=offscreen_buffer,M=0,i=0;i<height;i++)  {  /* do this for all lines */
	while(M>height) { M-=height; src+=pic_tm.width; }
	M+=pic_tm.height;
	
	for(p=src,N=0,j=0;j<width;j++) {
	  while(N>width) { N-=width; p++; }
	  N+=pic_tm.width;
	  *dst++=*p;
	}
      }
    } else {
      /* now scale title pic to screen size */
      for(dst=screen_base,src=offscreen_buffer,M=0,i=0;i<height;i++)  {  /* do this for all lines */
	while(M>height) { M-=height; src+=pic_tm.width; }
	M+=pic_tm.height;
	
	for(t=gif_grey[src[0]],k=0,N=0,j=0;j<width;j++) {
	  while(N>width) { N-=width; k++; t=gif_grey[src[k]]; }
	  N+=pic_tm.width;

	  if( t > dithermatrix[i&7][j&7] )
	    dst[j/8]|= 0x80>>(j&7);
	  else
	    dst[j/8]&=~( 0x80>>(j&7));
	}
	dst += width/8;
      }
    }
  }

  /* display copyright message */
  draw_text((WIDTH-get_text_width(copyright))/2,HEIGHT/6,copyright,COL_RED);

#ifdef FADE
  if((planes>1)&&(truecol)) {
    for (i = 0; i <= STEPS; i++) {
      for (j = 0; j < colors; j++) {
	col[0]=(tm_uword)(pic_tm.color[j].r*1000)/256*i/STEPS;
	col[1]=(tm_uword)(pic_tm.color[j].g*1000)/256*i/STEPS;
	col[2]=(tm_uword)(pic_tm.color[j].b*1000)/256*i/STEPS;
	set_color(j,col);
      }
      wait_msec(14);
      flip_screen();
      
      if(i==0)
	wait_msec(1000);
    }
  }
#else
  if(truecol) {
    for (j = 0; j < colors; j++) {
      col[0]=(tm_uword)(pic_tm.color[j].r*1000)/256;
      col[1]=(tm_uword)(pic_tm.color[j].g*1000)/256;
      col[2]=(tm_uword)(pic_tm.color[j].b*1000)/256;
      set_color(j,col);
    }
  }
#endif

  flip_screen();

  if((planes>1)&&(!truecol)) {
#ifdef FADE
    for (i = 0; i <= STEPS; i++) {
      for (j = 0; j < colors; j++) {
	fade_col[j].red   = pic_tm.color[j].r * 256 * i / STEPS;
	fade_col[j].green = pic_tm.color[j].g * 256 * i / STEPS;
	fade_col[j].blue  = pic_tm.color[j].b * 256 * i / STEPS;
	fade_col[j].pixel = j;
	fade_col[j].flags = 0xff;
      }
      XStoreColors (display, cmap, fade_col, colors);
      wait_msec(14);
      XSync (display, False);
    }
#else
    for (j = 0; j < colors; j++) {
      fade_col[j].red   = pic_tm.color[j].r * 256;
      fade_col[j].green = pic_tm.color[j].g * 256;
      fade_col[j].blue  = pic_tm.color[j].b * 256;
      fade_col[j].pixel = j;
      fade_col[j].flags = 0xff;
    }
    XStoreColors (display, cmap, fade_col, colors);
#endif
  }
}

void 
close_screen(void) {

  /* set keyboard to original state */
  restore_keyboard();

  /* free simulation buffer for true color */
  if(truebuf!=(char*)0) 
    free(truebuf);

#ifdef MOUSE_INVISIBLE
  XUndefineCursor(display, win);
  XFreePixmap(display, Cursorpix);
#endif

#if 0
  /* linux core dumps here */
  if(!use_shm)
    free(screen_base);
#endif

  XDestroyImage(xi);

  if(fullscreen) {
    if((visual->class != TrueColor ) && (visual->class != DirectColor))
      XUninstallColormap(display, cmap);
    XUngrabServer(display);
  }

  if(((planes!=1)&&(!truecol))||(splanes!=0))
    XFreeColormap(display, cmap);

  XFreeGC(display,gc);
  XCloseDisplay(display);

  free(offscreen_buffer);

  if(rev_table!=NULL)
    free(rev_table);
}

void 
set_color(tm_uword col, tm_uword prom[3]) {
  XColor color;

  color.red   = prom[0]*65;
  color.green = prom[1]*65;
  color.blue  = prom[2]*65;

  if(!truecol) {
    color.pixel = col;
    color.flags = 0xff;
    XStoreColor(display, cmap, &color);
  } else {
    true_table_16[col]=true_table_32[col]=
      ((((int)prom[0]*  red_mask)/1000)<<  red_shift)+
      ((((int)prom[1]*green_mask)/1000)<<green_shift)+
      ((((int)prom[2]* blue_mask)/1000)<< blue_shift);
  }
}

void  
save_colors(void) {
}

void 
restore_colors(void) {
}

tm_void 
copy_back_screen(tm_void) {
}

/* 
    Und hier kommt der Joystick-Treiber fuer X11

*/

#include "../joystick.h"
#include "joy_par/lpj.h"
#include <X11/keysym.h>

tm_char ***joy_list;
tm_char joy_device[128],*joy_xlist[]={joy_device,""};
tm_word fkeys;

tm_void 
show_joystick_options(tm_void) {
  tm_word i;
  console_printf("%s input driver options:\n",joystick_driver_name);
  console_puts("id=n\tset input device to mouse (n=m), keyboard (n=k), \n\tserial joystick (n=j) or parallel joystick (n=p), default is mouse");
  console_printf("ij=n\tset serial joystick device to n, default is to search\n\t");
  for(i=0;device_name[i][0]!=0;i++) 
    console_printf("%s'%s'",(i==0)?"":(device_name[i+1][0]==0)?" and ":", ",device_name[i]);
  console_puts("");
}

tm_void 
init_joystick_options(tm_void) {
  joy_inp=JOY_MOUSE;  
  joy_list=(tm_char***)&device_name;
}

tm_bool 
set_joystick_option(tm_char *cmd) {
  tm_word val=atoi(&cmd[2]);

  if(cmd[1]=='=') {
    if((cmd[0]=='d')&&((cmd[2]=='k')||(cmd[2]=='m')||(cmd[2]=='j')||(cmd[2]=='p'))) {
      if(cmd[2]=='k') joy_inp=JOY_KEY;
      if(cmd[2]=='m') joy_inp=JOY_MOUSE;
      if(cmd[2]=='j') joy_inp=JOY_JOY;
      if(cmd[2]=='p') joy_inp=JOY_LP;
      return(OK);
    }
    if(cmd[0]=='j') {
      joy_list=(tm_char***)&joy_xlist;
      strcpy(joy_device,&cmd[2]);
      return(OK);
    }
  }
  return(FALSE);
}

#ifdef WARP
tm_uword warp_counter=0;
#endif
tm_uword joy;
tm_word  joy_x,joy_y;

int sio_fd, lp_fd;
struct termios sio_term;

tm_bool 
init_ser_joy(tm_void) {
  tm_char in_buf;
  tm_word i,dev;

  for(dev=0;strlen((char*)joy_list[dev])!=0;dev++) {
    fprintf(stderr,"%s ",joy_list[dev]);

    if((sio_fd = open((char*)joy_list[dev], O_RDWR | O_NONBLOCK))<0)
      perror("error opening");
    else {
      /* joystick uses 1200 baud (possible 19200 is untested) */ 
      if(tcgetattr(sio_fd, &sio_term)!=0)
	perror("error getting termio infop");
      else {
	/* We don't want any special function (NO ECHO!!!) but 8 Bit */
	sio_term.c_iflag=0;
	sio_term.c_oflag=0;
	sio_term.c_cflag=CS8|CREAD|HUPCL;
	sio_term.c_lflag=0;
  
	if(cfsetispeed(&sio_term,B1200)!=0)
	  perror("error setting input bit rate");
	else {
	  if(cfsetospeed(&sio_term,B1200)!=0)
	    perror("error setting output bit rate");
	  else {
	    if(tcsetattr(sio_fd, TCSANOW, &sio_term)!=0)
       	      perror("error setting parameters");
	    else {
	      /* hardware setup is done, now detect the joystick */
	      for(i=0;i<10;i++) {              /* try ten times */
		write(sio_fd,"",1);            /* send one byte 0x00 */
		my_usleep(JOY_TIMEOUT);           /* give sio some time to work */
		while(read(sio_fd,&in_buf,1)>0) {
		  if( ((in_buf&0xe0)==0xc0)&&  /* upper three bits must be 110  */
		      ((in_buf&0x03)!=0x03)&&  /* up & down the same time is impossible (second button?) */ 
		      ((in_buf&0x0a)!=0x0a)) { /* left & right the same time is impossible (third button?) */
		    fputs("used for joystick input\n",stderr);
		    return OK;                  /* this seems to be a joystick */
	      } } }
	      fputs("error detecting joystick\n",stderr);
      } } } }
      close(sio_fd);
  } }
  kill_io_server();
  return ERROR;
}

tm_bool 
init_joystick(void) {
  if(joy_inp==JOY_MOUSE)
     return(TRUE);

  if(joy_inp==JOY_JOY)
    return(init_ser_joy());

  if(joy_inp==JOY_LP) {
    lp_fd = open(JOY_LP_DEV, O_RDONLY);
    if(lp_fd <0) {
      perror("open /dev/lpj");
      kill_io_server();
      return(FALSE);
    }
  }

  fkeys = 0;
  return(TRUE);
}

tm_void
exit_joystick(void) {

  if(joy_inp==JOY_JOY) {
    tcflush(sio_fd,TCIOFLUSH);
    if(sio_fd>=0)
      close(sio_fd);
  }

  if(joy_inp==JOY_LP)
    close(lp_fd);
}

tm_uword 
get_joystick(void) {
  tm_uword  this_joy=0;
  XEvent report;
  char buffer[20];
  int  bufsize=20;
  static int ret=-1;
  KeySym key;
  XComposeStatus compose;
  char in_buf;

  /* Key-Releases duerfen nur im ersten Durchlauf kommen,   */
  /* sonst koennte ein Press sofort wieder anulliert werden */

  if(XPending(display))
  {
    XNextEvent(display,&report);

    switch(report.type)
    {
    case Expose:
      while(XCheckTypedEvent(display,Expose,&report));
      flip_screen();
      break;

    case ButtonPress:
      if(((XButtonPressedEvent*)&report)->button == 1) 
	{ joy |= JOY_FIRE0; this_joy |= JOY_FIRE0; }
      if(((XButtonPressedEvent*)&report)->button == 2) 
	{ joy |= JOY_FIRE2; this_joy |= JOY_FIRE2; }
      if(((XButtonPressedEvent*)&report)->button == 3) 
	{ joy |= JOY_FIRE1; this_joy |= JOY_FIRE1; }
      break;
      
    case ButtonRelease:
      if(((XButtonReleasedEvent*)&report)->button == 1)  joy &=~ JOY_FIRE0;
      if(((XButtonReleasedEvent*)&report)->button == 2)  joy &=~ JOY_FIRE2;
      if(((XButtonReleasedEvent*)&report)->button == 3)  joy &=~ JOY_FIRE1;
      break;
	
    case KeyPress:
      XLookupString((XKeyEvent*)&report, buffer, bufsize, &key, &compose);
      switch(key&0xffff) {
      case XK_Help:    /* my (german, sun clone) sun keyboard returns XK_Help on F1 ?!?! */
      case XK_F1:      fkeys |= 1;
      case XK_space:   joy |=  JOY_FIRE0; this_joy |= JOY_FIRE0; break;
      case XK_F2:      fkeys |= 2;
      case XK_Return:  joy |=  JOY_FIRE1; this_joy |= JOY_FIRE1; break;
      case XK_F3:      fkeys |= 4;
	               joy |=  JOY_FIRE2; this_joy |= JOY_FIRE2; break;
      case XK_Up:      joy |=  JOY_UP;    this_joy |= JOY_UP;    break;
      case XK_Down:    joy |=  JOY_DOWN;  this_joy |= JOY_DOWN;  break;
      case XK_Left:    joy |=  JOY_LEFT;  this_joy |= JOY_LEFT;  break;
      case XK_Right:   joy |=  JOY_RIGHT; this_joy |= JOY_RIGHT; break;

	/* we mirror the ten function keys (use 1-0 on machines /wo f-keys */
      case XK_F4:      fkeys |= 8;   break;
      case XK_F5:      fkeys |= 16;  break;
      case XK_F6:      fkeys |= 32;  break;
      case XK_F7:      fkeys |= 64;  break;
      case XK_F8:      fkeys |= 128; break;
      case XK_F9:      fkeys |= 256; break;
      case XK_F10:     fkeys |= 512; break;
      }
      break;

    case KeyRelease:
      XLookupString((XKeyEvent*)&report, buffer, bufsize, &key, &compose);
      switch(key&0xffff) {
      case XK_Help:
      case XK_F1:       fkeys &= ~1;
      case XK_space:    joy &=~ JOY_FIRE0; break;
      case XK_F2:       fkeys &= ~2;
      case XK_Return:   joy &=~ JOY_FIRE1; break;
      case XK_F3:       fkeys &= ~4;
                        joy &=~ JOY_FIRE2; break;
      case XK_Up:       joy &=~ JOY_UP;    break;
      case XK_Down:     joy &=~ JOY_DOWN;  break;
      case XK_Left:     joy &=~ JOY_LEFT;  break;
      case XK_Right:    joy &=~ JOY_RIGHT; break;

	/* we mirror the ten function keys (use 1-0 on machines /wo f-keys */
      case XK_F4:      fkeys &= ~8;   break;
      case XK_F5:      fkeys &= ~16;  break;
      case XK_F6:      fkeys &= ~32;  break;
      case XK_F7:      fkeys &= ~64;  break;
      case XK_F8:      fkeys &= ~128; break;
      case XK_F9:      fkeys &= ~256; break;
      case XK_F10:     fkeys &= ~512; break;
      }
      break;

    case MotionNotify:
      while(XCheckTypedEvent(display,MotionNotify,&report));
      /* Maus im Bereich -100/100 festhalten */
      joy_x = ((XPointerMovedEvent*)&report)->x - (WIDTH/2);
      joy_y = ((XPointerMovedEvent*)&report)->y - (HEIGHT/2);

#ifdef WARP
      if(joy_x<-100) XWarpPointer(display, win, win, 0, 0, WIDTH, HEIGHT, WIDTH/2-100, HEIGHT/2+joy_y);
      if(joy_x>100)  XWarpPointer(display, win, win, 0, 0, WIDTH, HEIGHT, WIDTH/2+100, HEIGHT/2+joy_y);
      if(joy_y<-100) XWarpPointer(display, win, win, 0, 0, WIDTH, HEIGHT, WIDTH/2+joy_x, HEIGHT/2-100);
      if(joy_y>100)  XWarpPointer(display, win, win, 0, 0, WIDTH, HEIGHT, WIDTH/2+joy_x, HEIGHT/2+100);
#endif
      if(joy_x<-100) joy_x=-100;
      if(joy_x>100)  joy_x=100;
      if(joy_y<-100) joy_y=-100;
      if(joy_y>100)  joy_y=100;
      break;

    case MappingNotify:
      XRefreshKeyboardMapping((XMappingEvent*)&report);
      break;
    }
  }

  if(joy_inp==JOY_JOY) {
    while(read(sio_fd,&in_buf,1)>0)
      ret=in_buf&0x1f;

    write(sio_fd,"",1);

    /* keyboard has higher priority */
    if((this_joy==0)&&(ret>0)) {
      if((ret&0x01) && !(ret&0x02)) this_joy|=JOY_UP;       /* UP * /DOWN    */
      if((ret&0x02) && !(ret&0x01)) this_joy|=JOY_DOWN;     /* DOWN * /UP    */
      if((ret&0x04) && !(ret&0x08)) this_joy|=JOY_LEFT;     /* LEFT * /RIGHT */
      if((ret&0x08) && !(ret&0x04)) this_joy|=JOY_RIGHT;    /* RIGHT * /LEFT */
      if((ret&0x01) &&  (ret&0x02)) this_joy|=JOY_FIRE2;    /* UP * DOWN     */
      if((ret&0x04) &&  (ret&0x08)) this_joy|=JOY_FIRE1;    /* LEFT * RIGHT  */
      if(ret&0x10)                  this_joy|=JOY_FIRE0;    /* FIRE          */
    }
  }

  if(joy_inp==JOY_LP) {
    read(lp_fd, &in_buf,1);

    /* keyboard has higher priority */
    if(this_joy==0) {
      if((in_buf & LPJ_JOY_UP)    && !(in_buf & LPJ_JOY_DOWN))  this_joy|=JOY_UP;       /* UP * /DOWN    */
      if((in_buf & LPJ_JOY_DOWN)  && !(in_buf & LPJ_JOY_UP))    this_joy|=JOY_DOWN;     /* DOWN * /UP    */
      if((in_buf & LPJ_JOY_LEFT)  && !(in_buf & LPJ_JOY_RIGHT)) this_joy|=JOY_LEFT;     /* LEFT * /RIGHT */
      if((in_buf & LPJ_JOY_RIGHT) && !(in_buf & LPJ_JOY_LEFT))  this_joy|=JOY_RIGHT;    /* RIGHT * /LEFT */
      if((in_buf & LPJ_JOY_UP)    &&  (in_buf & LPJ_JOY_DOWN))  this_joy|=JOY_FIRE2;    /* UP * DOWN     */
      if((in_buf & LPJ_JOY_LEFT)  &&  (in_buf & LPJ_JOY_RIGHT)) this_joy|=JOY_FIRE1;    /* LEFT * RIGHT  */
      if(in_buf & LPJ_JOY_FIRE)                                 this_joy|=JOY_FIRE0;    /* FIRE          */
    }
  }

  return(joy | this_joy);
}

tm_uword 
get_joystick_analog(tm_word *x, tm_word *y) {
  tm_uword joy;

  /* only mouse is realy analog */
  if(joy_inp!=JOY_MOUSE) {
    *x=*y=0;

    joy=get_joystick();
    if(joy&JOY_UP)    *y=-100;
    if(joy&JOY_DOWN)  *y=+100;
    if(joy&JOY_LEFT)  *x=-100;
    if(joy&JOY_RIGHT) *x=+100;
  } else {
    joy=(get_joystick()&(JOY_FIRE0|JOY_FIRE1|JOY_FIRE2));
    *x = joy_x;
    *y = joy_y;
  }

  return(joy);
}

tm_uword 
kind_of_joystick(void) {
  if(joy_inp==JOY_MOUSE)
     return(ABSOLUTE_DEVICE);

  return(RELATIVE_DEVICE);
}

/*
  Timer-Treiber fuer X11
  Weihnachten '95  Till   erste Version
*/

#include "../timer.h"
#include <sys/time.h>

tm_uword 
init_timer(void) { 
  struct timeval  tp;
  struct timezone tzp;

  gettimeofday(&tp,&tzp);
  srnd(tp.tv_usec);

  return(OK); 
}

tm_void 
exit_timer(void) { 
}

unsigned long 
get_ticks(void) {
  struct timeval  tp;
  struct timezone tzp;

  gettimeofday(&tp,&tzp);
  return((tm_ulong)(tp.tv_usec/1000)+(tm_ulong)(1000*tp.tv_sec));
}

tm_ulong 
ticks2msec(tm_ulong ticks) {
  return(ticks);
}

tm_ulong 
msec2ticks(tm_ulong msec) {
  return(msec);
}









